---- MODULE MemoirLL1TypeSafety ----

(***************************************************************************)
(* This module proves the type safety of the `^\basicname^' spec.          *)
(***************************************************************************)

EXTENDS MemoirLL1TypeLemmas

THEOREM LL1TypeSafe == LL1Spec => []LL1TypeInvariant \*`^\label{THEOREM-LL1TypeSafe}^'
  (*************************************************************************)
  (* The top level of the proof is boilerplate TLA+ for an Inv1-style      *)
  (* proof.  First, we prove that the initial state satisfies              *)
  (* LL1TypeInvariant.  Second, we prove that the LL1Next predicate        *)
  (* inductively preserves LL1TypeInvariant.  Third, we use temporal       *)
  (* induction to prove that these two conditions satisfy type safety over *)
  (* all behaviors.                                                        *)
  (*************************************************************************)
  <1>1. LL1Init => LL1TypeInvariant
    (***********************************************************************)
    (* The base case follows directly from the definition of LL1Init.      *)
    (* There are a bunch of steps, but they are simple expansions of       *)
    (* definitions and appeals to the type safety of the initial           *)
    (* definitions.                                                        *)
    (***********************************************************************)
    <2>1. HAVE LL1Init
    <2>2. PICK symmetricKey \in SymmetricKeyType : LL1Init!(symmetricKey)!1
      BY <2>1 DEF LL1Init
    <2> initialPrivateStateEnc == SymmetricEncrypt(symmetricKey, InitialPrivateState)
    <2> initialStateHash == Hash(InitialPublicState, initialPrivateStateEnc)
    <2> initialHistoryStateBinding == Hash(BaseHashValue, initialStateHash)
    <2> initialAuthenticator == GenerateMAC(symmetricKey, initialHistoryStateBinding)
    <2> initialUntrustedStorage == [
            publicState |-> InitialPublicState,
            privateStateEnc |-> initialPrivateStateEnc,
            historySummary |-> BaseHashValue,
            authenticator |-> initialAuthenticator ]
    <2> initialTrustedStorage == [
            historySummary |-> BaseHashValue,
            symmetricKey |-> symmetricKey ]
    <2>3. /\  initialPrivateStateEnc \in PrivateStateEncType
          /\  initialStateHash \in HashType
          /\  initialHistoryStateBinding \in HashType
          /\  initialAuthenticator \in MACType
          /\  initialUntrustedStorage \in LL1UntrustedStorageType
          /\  initialTrustedStorage \in LL1TrustedStorageType
      <3>1. symmetricKey \in SymmetricKeyType
        BY <2>2
      <3>2. QED
        BY <3>1, LL1InitDefsTypeSafeLemma
    <2> HIDE DEF initialPrivateStateEnc, initialStateHash, initialAuthenticator,
             initialUntrustedStorage, initialTrustedStorage
    <2>4. LL1AvailableInputs \subseteq InputType
      <3>1. LL1AvailableInputs = InitialAvailableInputs
        BY <2>2
      <3>2. InitialAvailableInputs \subseteq InputType
        BY ConstantsTypeSafe DEF ConstantsTypeSafe
      <3>3. QED
        BY <3>1, <3>2
    <2>5. LL1ObservedOutputs \subseteq OutputType
      <3>1. LL1ObservedOutputs = {}
        BY <2>2
      <3>2. QED
        BY <3>1
    <2>6. LL1ObservedAuthenticators \subseteq MACType
      <3>1. LL1ObservedAuthenticators = {initialAuthenticator}
        BY <2>2
        DEF initialAuthenticator, initialHistoryStateBinding,
            initialStateHash, initialPrivateStateEnc
      <3>2. initialAuthenticator \in MACType
        BY <2>3
      <3>3. QED
        BY <3>1, <3>2
    <2>7. LL1Disk \in LL1UntrustedStorageType
      <3>1. LL1Disk = initialUntrustedStorage
        BY <2>2
        DEF initialUntrustedStorage, initialAuthenticator,
            initialHistoryStateBinding, initialStateHash, initialPrivateStateEnc
      <3>2. initialUntrustedStorage \in LL1UntrustedStorageType
        BY <2>3
      <3>3. QED
        BY <3>1, <3>2
    <2>8. LL1RAM \in LL1UntrustedStorageType
      <3>1. LL1RAM = initialUntrustedStorage
        BY <2>2
        DEF initialUntrustedStorage, initialAuthenticator,
            initialHistoryStateBinding, initialStateHash, initialPrivateStateEnc
      <3>2. initialUntrustedStorage \in LL1UntrustedStorageType
        BY <2>3
      <3>3. QED
        BY <3>1, <3>2
    <2>9. LL1NVRAM \in LL1TrustedStorageType
      <3>1. LL1NVRAM = initialTrustedStorage
        BY <2>2 DEF initialTrustedStorage
      <3>2. initialTrustedStorage \in LL1TrustedStorageType
        BY <2>3
      <3>3. QED
        BY <3>1, <3>2
    <2>10. QED
      BY <2>4, <2>5, <2>6, <2>7, <2>8, <2>9 DEF LL1TypeInvariant
  <1>2. LL1TypeInvariant /\ [LL1Next]_LL1Vars => LL1TypeInvariant'
    (***********************************************************************)
    (* The induction step is also straightforward.  We assume the          *)
    (* antecedents of the implication, then show that the consequent holds *)
    (* for all eight LL1Next actions plus stuttering.                      *)
    (***********************************************************************)
    <2>1. HAVE LL1TypeInvariant /\ [LL1Next]_LL1Vars
    <2>2. CASE UNCHANGED LL1Vars
      (*********************************************************************)
      (* Type safety is inductively trivial for a stuttering step.         *)
      (*********************************************************************)
      <3>1. LL1AvailableInputs' \subseteq InputType
        <4>1. LL1AvailableInputs \subseteq InputType
          BY <2>1 DEF LL1TypeInvariant
        <4>2. UNCHANGED LL1AvailableInputs
          BY <2>2 DEF LL1Vars
        <4>3. QED
          BY <4>1, <4>2
      <3>2. LL1ObservedOutputs' \subseteq OutputType
        <4>1. LL1ObservedOutputs \subseteq OutputType
          BY <2>1 DEF LL1TypeInvariant
        <4>2. UNCHANGED LL1ObservedOutputs
          BY <2>2 DEF LL1Vars
        <4>3. QED
          BY <4>1, <4>2
      <3>3. LL1ObservedAuthenticators' \subseteq MACType
        <4>1. LL1ObservedAuthenticators \subseteq MACType
          BY <2>1 DEF LL1TypeInvariant
        <4>2. UNCHANGED LL1ObservedAuthenticators
          BY <2>2 DEF LL1Vars
        <4>3. QED
          BY <4>1, <4>2
      <3>4. LL1Disk' \in LL1UntrustedStorageType
        <4>1. LL1Disk \in LL1UntrustedStorageType
          BY <2>1 DEF LL1TypeInvariant
        <4>2. UNCHANGED LL1Disk
          BY <2>2 DEF LL1Vars
        <4>3. QED
          BY <4>1, <4>2
      <3>5. LL1RAM' \in LL1UntrustedStorageType
        <4>1. LL1RAM \in LL1UntrustedStorageType
          BY <2>1 DEF LL1TypeInvariant
        <4>2. UNCHANGED LL1RAM
          BY <2>2 DEF LL1Vars
        <4>3. QED
          BY <4>1, <4>2
      <3>6. LL1NVRAM' \in LL1TrustedStorageType
        <4>1. LL1NVRAM \in LL1TrustedStorageType
          BY <2>1 DEF LL1TypeInvariant
        <4>2. UNCHANGED LL1NVRAM
          BY <2>2 DEF LL1Vars
        <4>3. QED
          BY <4>1, <4>2
      <3>7. QED
        BY <3>1, <3>2, <3>3, <3>4, <3>5, <3>6 DEF LL1TypeInvariant
    <2>3. CASE LL1Next
      <3>1. CASE LL1MakeInputAvailable
        (*******************************************************************)
        (* Type safety is straightforward for a LL1MakeInputAvailable      *)
        (* action.                                                         *)
        (*******************************************************************)
        <4>1. PICK input \in InputType : LL1MakeInputAvailable!(input)
          BY <3>1 DEF LL1MakeInputAvailable
        <4>2. LL1AvailableInputs' \subseteq InputType
          <5>1. LL1AvailableInputs \subseteq InputType
            BY <2>1 DEF LL1TypeInvariant
          <5>2. LL1AvailableInputs' = LL1AvailableInputs \union {input}
            BY <4>1
          <5>3. input \in InputType
            BY <4>1
          <5>4. QED
            BY <5>1, <5>2, <5>3
        <4>3. LL1ObservedOutputs' \subseteq OutputType
          <5>1. LL1ObservedOutputs \subseteq OutputType
            BY <2>1 DEF LL1TypeInvariant
          <5>2. UNCHANGED LL1ObservedOutputs
            BY <4>1
          <5>3. QED
            BY <5>1, <5>2
        <4>4. LL1ObservedAuthenticators' \subseteq MACType
          <5>1. LL1ObservedAuthenticators \subseteq MACType
            BY <2>1 DEF LL1TypeInvariant
          <5>2. UNCHANGED LL1ObservedAuthenticators
            BY <4>1
          <5>3. QED
            BY <5>1, <5>2
        <4>5. LL1Disk' \in LL1UntrustedStorageType
          <5>1. LL1Disk \in LL1UntrustedStorageType
            BY <2>1 DEF LL1TypeInvariant
          <5>2. UNCHANGED LL1Disk
            BY <4>1
          <5>3. QED
            BY <5>1, <5>2
        <4>6. LL1RAM' \in LL1UntrustedStorageType
          <5>1. LL1RAM \in LL1UntrustedStorageType
            BY <2>1 DEF LL1TypeInvariant
          <5>2. UNCHANGED LL1RAM
            BY <4>1
          <5>3. QED
            BY <5>1, <5>2
        <4>7. LL1NVRAM' \in LL1TrustedStorageType
          <5>1. LL1NVRAM \in LL1TrustedStorageType
            BY <2>1 DEF LL1TypeInvariant
          <5>2. UNCHANGED LL1NVRAM
            BY <4>1
          <5>3. QED
            BY <5>1, <5>2
        <4>8. QED
          BY <4>2, <4>3, <4>4, <4>5, <4>6, <4>7 DEF LL1TypeInvariant
      <3>2. CASE LL1PerformOperation
        (*******************************************************************)
        (* For a LL1PerformOperation action, we just walk through the      *)
        (* definitions.  Type safety follows directly.                     *)
        (*******************************************************************)
        <4>1. PICK input \in LL1AvailableInputs : LL1PerformOperation!(input)!1
          BY <3>2 DEF LL1PerformOperation
        <4> stateHash == Hash(LL1RAM.publicState, LL1RAM.privateStateEnc)
        <4> historyStateBinding == Hash(LL1RAM.historySummary, stateHash)
        <4> privateState == SymmetricDecrypt(LL1NVRAM.symmetricKey, LL1RAM.privateStateEnc)
        <4> sResult == Service(LL1RAM.publicState, privateState, input)
        <4> newPrivateStateEnc ==
                SymmetricEncrypt(LL1NVRAM.symmetricKey, sResult.newPrivateState)
        <4> newHistorySummary == Hash(LL1NVRAM.historySummary, input)
        <4> newStateHash == Hash(sResult.newPublicState, newPrivateStateEnc)
        <4> newHistoryStateBinding == Hash(newHistorySummary, newStateHash)
        <4> newAuthenticator == GenerateMAC(LL1NVRAM.symmetricKey, newHistoryStateBinding)
        <4>2. /\  stateHash \in HashType
              /\  historyStateBinding \in HashType
              /\  privateState \in PrivateStateType
              /\  sResult \in ServiceResultType
              /\  sResult.newPublicState \in PublicStateType
              /\  sResult.newPrivateState \in PrivateStateType
              /\  sResult.output \in OutputType
              /\  newPrivateStateEnc \in PrivateStateEncType
              /\  newHistorySummary \in HashType
              /\  newStateHash \in HashType
              /\  newHistoryStateBinding \in HashType
              /\  newAuthenticator \in MACType
          <5>1. input \in LL1AvailableInputs
            BY <4>1
          <5>2. LL1TypeInvariant
            BY <2>1
          <5>3. QED
            BY <5>1, <5>2, LL1PerformOperationDefsTypeSafeLemma
        <4> HIDE DEF stateHash, historyStateBinding, privateState, sResult, newPrivateStateEnc,
                newHistorySummary, newStateHash, newHistoryStateBinding, newAuthenticator
        <4>3. LL1AvailableInputs' \subseteq InputType
          <5>1. LL1AvailableInputs \subseteq InputType
            BY <2>1 DEF LL1TypeInvariant
          <5>2. UNCHANGED LL1AvailableInputs
            BY <4>1
          <5>3. QED
            BY <5>1, <5>2
        <4>4. LL1ObservedOutputs' \subseteq OutputType
          <5>1. LL1ObservedOutputs \subseteq OutputType
            BY <2>1 DEF LL1TypeInvariant
          <5>2. LL1ObservedOutputs' = LL1ObservedOutputs \union {sResult.output}
            BY <4>1 DEF sResult, privateState
          <5>3. sResult.output \in OutputType
            BY <4>2
          <5>4. QED
            BY <5>1, <5>2, <5>3
        <4>5. LL1ObservedAuthenticators' \subseteq MACType
          <5>1. LL1ObservedAuthenticators \subseteq MACType
            BY <2>1 DEF LL1TypeInvariant
          <5>2. LL1ObservedAuthenticators' =
                    LL1ObservedAuthenticators \union {newAuthenticator}
            BY <4>1 DEF newAuthenticator, newHistoryStateBinding, newStateHash,
                        newHistorySummary, newPrivateStateEnc, sResult, privateState
          <5>3. newAuthenticator \in MACType
            BY <4>2
          <5>4. QED
            BY <5>1, <5>2, <5>3
        <4>6. LL1Disk' \in LL1UntrustedStorageType
          <5>1. LL1Disk \in LL1UntrustedStorageType
            BY <2>1 DEF LL1TypeInvariant
          <5>2. UNCHANGED LL1Disk
            BY <4>1
          <5>3. QED
            BY <5>1, <5>2
        <4>7. LL1RAM' \in LL1UntrustedStorageType
          <5>1. LL1RAM' = [ publicState |-> sResult.newPublicState,
                         privateStateEnc |-> newPrivateStateEnc,
                         historySummary |-> newHistorySummary,
                         authenticator |-> newAuthenticator ]
            BY <4>1 DEF newAuthenticator, newHistoryStateBinding, newStateHash,
                        newHistorySummary, newPrivateStateEnc, sResult, privateState
          <5>2. sResult.newPublicState \in PublicStateType
            BY <4>2
          <5>3. newPrivateStateEnc \in PrivateStateEncType
            BY <4>2
          <5>4. newHistorySummary \in HashType
            BY <4>2
          <5>5. newAuthenticator \in MACType
            BY <4>2
          <5>6. QED
            BY <5>1, <5>2, <5>3, <5>4, <5>5 DEF LL1UntrustedStorageType
        <4>8. LL1NVRAM' \in LL1TrustedStorageType
          <5>1. LL1NVRAM' = [ historySummary |-> newHistorySummary,
                           symmetricKey |-> LL1NVRAM.symmetricKey ]
            BY <4>1 DEF LL1TypeInvariant, newHistorySummary
          <5>2. newHistorySummary \in HashType
            BY <4>2
          <5>3. LL1NVRAM.symmetricKey \in SymmetricKeyType
            BY <2>1, LL1SubtypeImplicationLemma DEF LL1SubtypeImplication
          <5>4. QED
            BY <5>1, <5>2, <5>3 DEF LL1TrustedStorageType
        <4>9. QED
          BY <4>3, <4>4, <4>5, <4>6, <4>7, <4>8 DEF LL1TypeInvariant
      <3>3. CASE LL1RepeatOperation
        (*******************************************************************)
        (* For a LL1RepeatOperation action, we just walk through the       *)
        (* definitions.  Type safety follows directly.                     *)
        (*******************************************************************)
        <4>1. PICK input \in LL1AvailableInputs : LL1RepeatOperation!(input)!1
          BY <3>3 DEF LL1RepeatOperation
        <4> stateHash == Hash(LL1RAM.publicState, LL1RAM.privateStateEnc)
        <4> historyStateBinding == Hash(LL1RAM.historySummary, stateHash)
        <4> privateState == SymmetricDecrypt(LL1NVRAM.symmetricKey, LL1RAM.privateStateEnc)
        <4> sResult == Service(LL1RAM.publicState, privateState, input)
        <4> newPrivateStateEnc ==
                SymmetricEncrypt(LL1NVRAM.symmetricKey, sResult.newPrivateState)
        <4> newStateHash == Hash(sResult.newPublicState, newPrivateStateEnc)
        <4> newHistoryStateBinding == Hash(LL1NVRAM.historySummary, newStateHash)
        <4> newAuthenticator == GenerateMAC(LL1NVRAM.symmetricKey, newHistoryStateBinding)
        <4>2. /\  stateHash \in HashType
              /\  historyStateBinding \in HashType
              /\  privateState \in PrivateStateType
              /\  sResult \in ServiceResultType
              /\  sResult.newPublicState \in PublicStateType
              /\  sResult.newPrivateState \in PrivateStateType
              /\  sResult.output \in OutputType
              /\  newPrivateStateEnc \in PrivateStateEncType
              /\  newStateHash \in HashType
              /\  newHistoryStateBinding \in HashType
              /\  newAuthenticator \in MACType
          <5>1. input \in LL1AvailableInputs
            BY <4>1
          <5>2. LL1TypeInvariant
            BY <2>1
          <5>3. QED
            BY <5>1, <5>2, LL1RepeatOperationDefsTypeSafeLemma
        <4> HIDE DEF stateHash, historyStateBinding, privateState, sResult, newPrivateStateEnc,
                newStateHash, newHistoryStateBinding, newAuthenticator
        <4>3. LL1AvailableInputs' \subseteq InputType
          <5>1. LL1AvailableInputs \subseteq InputType
            BY <2>1 DEF LL1TypeInvariant
          <5>2. UNCHANGED LL1AvailableInputs
            BY <4>1
          <5>3. QED
            BY <5>1, <5>2
        <4>4. LL1ObservedOutputs' \subseteq OutputType
          <5>1. LL1ObservedOutputs \subseteq OutputType
            BY <2>1 DEF LL1TypeInvariant
          <5>2. LL1ObservedOutputs' = LL1ObservedOutputs \union {sResult.output}
            BY <4>1 DEF sResult, privateState
          <5>3. sResult.output \in OutputType
            BY <4>2
          <5>4. QED
            BY <5>1, <5>2, <5>3
        <4>5. LL1ObservedAuthenticators' \subseteq MACType
          <5>1. LL1ObservedAuthenticators \subseteq MACType
            BY <2>1 DEF LL1TypeInvariant
          <5>2. LL1ObservedAuthenticators' =
                    LL1ObservedAuthenticators \union {newAuthenticator}
            BY <4>1 DEF newAuthenticator, newHistoryStateBinding, newStateHash,
                        newPrivateStateEnc, sResult, privateState
          <5>3. newAuthenticator \in MACType
            BY <4>2
          <5>4. QED
            BY <5>1, <5>2, <5>3
        <4>6. LL1Disk' \in LL1UntrustedStorageType
          <5>1. LL1Disk \in LL1UntrustedStorageType
            BY <2>1 DEF LL1TypeInvariant
          <5>2. UNCHANGED LL1Disk
            BY <4>1
          <5>3. QED
            BY <5>1, <5>2
        <4>7. LL1RAM' \in LL1UntrustedStorageType
          <5>1. LL1RAM' = [ publicState |-> sResult.newPublicState,
                         privateStateEnc |-> newPrivateStateEnc,
                         historySummary |-> LL1NVRAM.historySummary,
                         authenticator |-> newAuthenticator ]
            BY <4>1 DEF newAuthenticator, newHistoryStateBinding, newStateHash,
                        newPrivateStateEnc, sResult, privateState
          <5>2. sResult.newPublicState \in PublicStateType
            BY <4>2
          <5>3. newPrivateStateEnc \in PrivateStateEncType
            BY <4>2
          <5>4. LL1NVRAM.historySummary \in HashType
            BY <2>1, LL1SubtypeImplicationLemma DEF LL1SubtypeImplication
          <5>5. newAuthenticator \in MACType
            BY <4>2
          <5>6. QED
            BY <5>1, <5>2, <5>3, <5>4, <5>5 DEF LL1UntrustedStorageType
        <4>8. LL1NVRAM' \in LL1TrustedStorageType
          <5>1. LL1NVRAM \in LL1TrustedStorageType
            BY <2>1 DEF LL1TypeInvariant
          <5>2. UNCHANGED LL1NVRAM
            BY <4>1
          <5>3. QED
            BY <5>1, <5>2
        <4>9. QED
          BY <4>3, <4>4, <4>5, <4>6, <4>7, <4>8 DEF LL1TypeInvariant
      <3>4. CASE LL1Restart
        (*******************************************************************)
        (* Type safety is straightforward for a LL1Restart action.         *)
        (*******************************************************************)
        <4>1. PICK  untrustedStorage \in LL1UntrustedStorageType,
                    randomSymmetricKey \in SymmetricKeyType \ {LL1NVRAM.symmetricKey},
                    hash \in HashType :
                    LL1Restart!(untrustedStorage, randomSymmetricKey, hash)
          BY <3>4 DEF LL1Restart
        <4>2. LL1AvailableInputs' \subseteq InputType
          <5>1. LL1AvailableInputs \subseteq InputType
            BY <2>1 DEF LL1TypeInvariant
          <5>2. UNCHANGED LL1AvailableInputs
            BY <4>1
          <5>3. QED
            BY <5>1, <5>2
        <4>3. LL1ObservedOutputs' \subseteq OutputType
          <5>1. LL1ObservedOutputs \subseteq OutputType
            BY <2>1 DEF LL1TypeInvariant
          <5>2. UNCHANGED LL1ObservedOutputs
            BY <4>1
          <5>3. QED
            BY <5>1, <5>2
        <4>4. LL1ObservedAuthenticators' \subseteq MACType
          <5>1. LL1ObservedAuthenticators \subseteq MACType
            BY <2>1 DEF LL1TypeInvariant
          <5>2. UNCHANGED LL1ObservedAuthenticators
            BY <4>1
          <5>3. QED
            BY <5>1, <5>2
        <4>5. LL1Disk' \in LL1UntrustedStorageType
          <5>1. LL1Disk \in LL1UntrustedStorageType
            BY <2>1 DEF LL1TypeInvariant
          <5>2. UNCHANGED LL1Disk
            BY <4>1
          <5>3. QED
            BY <5>1, <5>2
        <4>6. LL1RAM' \in LL1UntrustedStorageType
          <5>1. untrustedStorage \in LL1UntrustedStorageType
            BY <4>1
          <5>2. LL1RAM' = untrustedStorage
            BY <4>1
          <5>3. QED
            BY <5>1, <5>2
        <4>7. LL1NVRAM' \in LL1TrustedStorageType
          <5>1. LL1NVRAM \in LL1TrustedStorageType
            BY <2>1 DEF LL1TypeInvariant
          <5>2. UNCHANGED LL1NVRAM
            BY <4>1
          <5>3. QED
            BY <5>1, <5>2
        <4>8. QED
          BY <4>2, <4>3, <4>4, <4>5, <4>6, <4>7 DEF LL1TypeInvariant
      <3>5. CASE LL1ReadDisk
        (*******************************************************************)
        (* Type safety is straightforward for a LL1ReadDisk action.        *)
        (*******************************************************************)
        <4>1. LL1AvailableInputs' \subseteq InputType
          <5>1. LL1AvailableInputs \subseteq InputType
            BY <2>1 DEF LL1TypeInvariant
          <5>2. UNCHANGED LL1AvailableInputs
            BY <3>5 DEF LL1ReadDisk
          <5>3. QED
            BY <5>1, <5>2
        <4>2. LL1ObservedOutputs' \subseteq OutputType
          <5>1. LL1ObservedOutputs \subseteq OutputType
            BY <2>1 DEF LL1TypeInvariant
          <5>2. UNCHANGED LL1ObservedOutputs
            BY <3>5 DEF LL1ReadDisk
          <5>3. QED
            BY <5>1, <5>2
        <4>3. LL1ObservedAuthenticators' \subseteq MACType
          <5>1. LL1ObservedAuthenticators \subseteq MACType
            BY <2>1 DEF LL1TypeInvariant
          <5>2. UNCHANGED LL1ObservedAuthenticators
            BY <3>5 DEF LL1ReadDisk
          <5>3. QED
            BY <5>1, <5>2
        <4>4. LL1Disk' \in LL1UntrustedStorageType
          <5>1. LL1Disk \in LL1UntrustedStorageType
            BY <2>1 DEF LL1TypeInvariant
          <5>2. UNCHANGED LL1Disk
            BY <3>5 DEF LL1ReadDisk
          <5>3. QED
            BY <5>1, <5>2
        <4>5. LL1RAM' \in LL1UntrustedStorageType
          <5>1. LL1Disk \in LL1UntrustedStorageType
            BY <2>1 DEF LL1TypeInvariant
          <5>2. LL1RAM' = LL1Disk
            BY <3>5 DEF LL1ReadDisk
          <5>3. QED
            BY <5>1, <5>2
        <4>6. LL1NVRAM' \in LL1TrustedStorageType
          <5>1. LL1NVRAM \in LL1TrustedStorageType
            BY <2>1 DEF LL1TypeInvariant
          <5>2. UNCHANGED LL1NVRAM
            BY <3>5 DEF LL1ReadDisk
          <5>3. QED
            BY <5>1, <5>2
        <4>7. QED
          BY <4>1, <4>2, <4>3, <4>4, <4>5, <4>6 DEF LL1TypeInvariant
      <3>6. CASE LL1WriteDisk
        (*******************************************************************)
        (* Type safety is straightforward for a LL1WriteDisk action.       *)
        (*******************************************************************)
        <4>1. LL1AvailableInputs' \subseteq InputType
          <5>1. LL1AvailableInputs \subseteq InputType
            BY <2>1 DEF LL1TypeInvariant
          <5>2. UNCHANGED LL1AvailableInputs
            BY <3>6 DEF LL1WriteDisk
          <5>3. QED
            BY <5>1, <5>2
        <4>2. LL1ObservedOutputs' \subseteq OutputType
          <5>1. LL1ObservedOutputs \subseteq OutputType
            BY <2>1 DEF LL1TypeInvariant
          <5>2. UNCHANGED LL1ObservedOutputs
            BY <3>6 DEF LL1WriteDisk
          <5>3. QED
            BY <5>1, <5>2
        <4>3. LL1ObservedAuthenticators' \subseteq MACType
          <5>1. LL1ObservedAuthenticators \subseteq MACType
            BY <2>1 DEF LL1TypeInvariant
          <5>2. UNCHANGED LL1ObservedAuthenticators
            BY <3>6 DEF LL1WriteDisk
          <5>3. QED
            BY <5>1, <5>2
        <4>4. LL1Disk' \in LL1UntrustedStorageType
          <5>1. LL1RAM \in LL1UntrustedStorageType
            BY <2>1 DEF LL1TypeInvariant
          <5>2. LL1Disk' = LL1RAM
            BY <3>6 DEF LL1WriteDisk
          <5>3. QED
            BY <5>1, <5>2
        <4>5. LL1RAM' \in LL1UntrustedStorageType
          <5>1. LL1RAM \in LL1UntrustedStorageType
            BY <2>1 DEF LL1TypeInvariant
          <5>2. UNCHANGED LL1RAM
            BY <3>6 DEF LL1WriteDisk
          <5>3. QED
            BY <5>1, <5>2
        <4>6. LL1NVRAM' \in LL1TrustedStorageType
          <5>1. LL1NVRAM \in LL1TrustedStorageType
            BY <2>1 DEF LL1TypeInvariant
          <5>2. UNCHANGED LL1NVRAM
            BY <3>6 DEF LL1WriteDisk
          <5>3. QED
            BY <5>1, <5>2
        <4>7. QED
          BY <4>1, <4>2, <4>3, <4>4, <4>5, <4>6 DEF LL1TypeInvariant
      <3>7. CASE LL1CorruptRAM
        (*******************************************************************)
        (* Type safety is straightforward for a LL1CorruptRAM action.      *)
        (*******************************************************************)
        <4>1. PICK  untrustedStorage \in LL1UntrustedStorageType,
                    fakeSymmetricKey \in SymmetricKeyType \ {LL1NVRAM.symmetricKey},
                    hash \in HashType :
                    LL1CorruptRAM!(untrustedStorage, fakeSymmetricKey, hash)
          BY <3>7 DEF LL1CorruptRAM
        <4>2. LL1AvailableInputs' \subseteq InputType
          <5>1. LL1AvailableInputs \subseteq InputType
            BY <2>1 DEF LL1TypeInvariant
          <5>2. UNCHANGED LL1AvailableInputs
            BY <4>1
          <5>3. QED
            BY <5>1, <5>2
        <4>3. LL1ObservedOutputs' \subseteq OutputType
          <5>1. LL1ObservedOutputs \subseteq OutputType
            BY <2>1 DEF LL1TypeInvariant
          <5>2. UNCHANGED LL1ObservedOutputs
            BY <4>1
          <5>3. QED
            BY <5>1, <5>2
        <4>4. LL1ObservedAuthenticators' \subseteq MACType
          <5>1. LL1ObservedAuthenticators \subseteq MACType
            BY <2>1 DEF LL1TypeInvariant
          <5>2. UNCHANGED LL1ObservedAuthenticators
            BY <4>1
          <5>3. QED
            BY <5>1, <5>2
        <4>5. LL1Disk' \in LL1UntrustedStorageType
          <5>1. LL1Disk \in LL1UntrustedStorageType
            BY <2>1 DEF LL1TypeInvariant
          <5>2. UNCHANGED LL1Disk
            BY <4>1
          <5>3. QED
            BY <5>1, <5>2
        <4>6. LL1RAM' \in LL1UntrustedStorageType
          <5>1. untrustedStorage \in LL1UntrustedStorageType
            BY <4>1
          <5>2. LL1RAM' = untrustedStorage
            BY <4>1
          <5>3. QED
            BY <5>1, <5>2
        <4>7. LL1NVRAM' \in LL1TrustedStorageType
          <5>1. LL1NVRAM \in LL1TrustedStorageType
            BY <2>1 DEF LL1TypeInvariant
          <5>2. UNCHANGED LL1NVRAM
            BY <4>1
          <5>3. QED
            BY <5>1, <5>2
        <4>8. QED
          BY <4>2, <4>3, <4>4, <4>5, <4>6, <4>7 DEF LL1TypeInvariant
      <3>8. CASE LL1RestrictedCorruption
        (*******************************************************************)
        (* Type safety is straightforward for a LL1RestrictedCorruption    *)
        (* action.                                                         *)
        (*******************************************************************)
        <4>2. LL1AvailableInputs' \subseteq InputType
          <5>1. LL1AvailableInputs \subseteq InputType
            BY <2>1 DEF LL1TypeInvariant
          <5>2. UNCHANGED LL1AvailableInputs
            BY <3>8 DEF LL1RestrictedCorruption
          <5>3. QED
            BY <5>1, <5>2
        <4>3. LL1ObservedOutputs' \subseteq OutputType
          <5>1. LL1ObservedOutputs \subseteq OutputType
            BY <2>1 DEF LL1TypeInvariant
          <5>2. UNCHANGED LL1ObservedOutputs
            BY <3>8 DEF LL1RestrictedCorruption
          <5>3. QED
            BY <5>1, <5>2
        <4>4. LL1ObservedAuthenticators' \subseteq MACType
          <5>1. LL1ObservedAuthenticators \subseteq MACType
            BY <2>1 DEF LL1TypeInvariant
          <5>2. UNCHANGED LL1ObservedAuthenticators
            BY <3>8 DEF LL1RestrictedCorruption
          <5>3. QED
            BY <5>1, <5>2
        <4>5. LL1Disk' \in LL1UntrustedStorageType
          <5>1. LL1Disk \in LL1UntrustedStorageType
            BY <2>1 DEF LL1TypeInvariant
          <5>2. UNCHANGED LL1Disk
            BY <3>8 DEF LL1RestrictedCorruption
          <5>3. QED
            BY <5>1, <5>2
        <4>6. LL1RAM' \in LL1UntrustedStorageType
          <5>1. CASE LL1RestrictedCorruption!ram!unchanged
            <6>1. LL1RAM \in LL1UntrustedStorageType
              BY <2>1 DEF LL1TypeInvariant
            <6>2. UNCHANGED LL1RAM
              BY <5>1
            <6>3. QED
              BY <6>1, <6>2
          <5>2. CASE LL1RestrictedCorruption!ram!trashed
            <6>1. PICK untrustedStorage \in LL1UntrustedStorageType,
                      randomSymmetricKey \in SymmetricKeyType \ {LL1NVRAM.symmetricKey},
                      hash \in HashType :
                      LL1RestrictedCorruption!ram!trashed!(
                          untrustedStorage, randomSymmetricKey, hash)
              BY <5>2
            <6>2. untrustedStorage \in LL1UntrustedStorageType
              BY <6>1
            <6>3. LL1RAM' = untrustedStorage
              BY <6>1
            <6>4. QED
              BY <6>2, <6>3
          <5>3. QED
            BY <3>8, <5>1, <5>2 DEF LL1RestrictedCorruption
        <4>7. LL1NVRAM' \in LL1TrustedStorageType
          <5>1. PICK garbageHistorySummary \in HashType :
                    LL1RestrictedCorruption!nvram!(garbageHistorySummary)
            BY <3>8 DEF LL1RestrictedCorruption
          <5>2. garbageHistorySummary \in HashType
            BY <5>1
          <5>3. LL1NVRAM.symmetricKey \in SymmetricKeyType
            BY <2>1, LL1SubtypeImplicationLemma DEF LL1SubtypeImplication
          <5>4. LL1NVRAM' = [ historySummary |-> garbageHistorySummary,
                           symmetricKey |-> LL1NVRAM.symmetricKey ]
            BY <5>1
          <5>5. QED
            BY <5>2, <5>3, <5>4 DEF LL1TrustedStorageType
        <4>8. QED
          BY <4>2, <4>3, <4>4, <4>5, <4>6, <4>7 DEF LL1TypeInvariant
      <3>9. QED
        BY <2>3, <3>1, <3>2, <3>3, <3>4, <3>5, <3>6, <3>7, <3>8 DEF LL1Next
    <2>4. QED
      BY <2>1, <2>2, <2>3
  <1>3. QED
    (***********************************************************************)
    (* Using the Inv1 proof rule, the base case and the induction step     *)
    (* together imply that the invariant always holds.                     *)
    (***********************************************************************)
    <2>1. LL1TypeInvariant /\ [][LL1Next]_LL1Vars => []LL1TypeInvariant
      BY <1>2, Inv1
    <2>2. QED
      BY <1>1, <2>1 DEF LL1Spec

====
