---- MODULE MemoirLL2RefinementLemmas ----

(***************************************************************************)
(* This module states and proves several lemmas about the operators        *)
(* defined in the LL2Refinement module.                                    *)
(*                                                                         *)
(* This module includes the following theorems:                            *)
(*     HistorySummaryRecordCompositionLemma                                *)
(*     LL1DiskRecordCompositionLemma                                       *)
(*     LL1RAMRecordCompositionLemma                                        *)
(*     LL1NVRAMRecordCompositionLemma                                      *)
(*     CheckpointHasBaseExtensionLemma                                     *)
(*     SuccessorHasNonBaseExtensionLemma                                   *)
(*     HistorySummariesMatchUniqueLemma                                    *)
(*     AuthenticatorsMatchUniqueLemma                                      *)
(*     AuthenticatorSetsMatchUniqueLemma                                   *)
(*     LL2NVRAMLogicalHistorySummaryTypeSafe                               *)
(*     AuthenticatorInSetLemma                                             *)
(*     AuthenticatorGeneratedLemma                                         *)
(*     AuthenticatorValidatedLemma                                         *)
(*     HistorySummariesMatchAcrossCheckpointLemma                          *)
(***************************************************************************)

EXTENDS MemoirLL2TypeSafety

(***************************************************************************)
(* The HistorySummaryRecordCompositionLemma is a formality needed for the  *)
(* prover to compose a HistorySummary record from fields of the            *)
(* appropriate type.                                                       *)
(***************************************************************************)

THEOREM HistorySummaryRecordCompositionLemma == \*`^\label{THEOREM-HistorySummaryRecordCompositionLemma}^'
    \A historySummary \in HistorySummaryType :
        historySummary = [
            anchor |-> historySummary.anchor,
            extension |-> historySummary.extension ]
  <1>1. TAKE historySummary \in HistorySummaryType
  <1> historysummary == [
          anchor |-> historySummary.anchor,
          extension |-> historySummary.extension ]
  <1>2. historysummary = [i \in {"anchor", "extension"} |-> historysummary[i]]
    OBVIOUS
  <1>3. historySummary = [i \in {"anchor", "extension"} |-> historySummary[i]]
    BY <1>1 DEF HistorySummaryType
  <1>4. QED
    BY <1>2, <1>3

----

(***************************************************************************)
(* The LL1DiskRecordCompositionLemma is a formality needed for the prover  *)
(* to compose an LL1Disk record from fields of the appropriate type.       *)
(***************************************************************************)

THEOREM LL1DiskRecordCompositionLemma == \*`^\label{THEOREM-LL1DiskRecordCompositionLemma}^'
    /\  LL1Disk \in LL1UntrustedStorageType =>
            LL1Disk = [
                publicState |-> LL1Disk.publicState,
                privateStateEnc |-> LL1Disk.privateStateEnc,
                historySummary |-> LL1Disk.historySummary,
                authenticator |-> LL1Disk.authenticator ]
    /\  LL1Disk' \in LL1UntrustedStorageType =>
            LL1Disk' = [
                publicState |-> LL1Disk.publicState',
                privateStateEnc |-> LL1Disk.privateStateEnc',
                historySummary |-> LL1Disk.historySummary',
                authenticator |-> LL1Disk.authenticator' ]
  <1>1. LL1Disk \in LL1UntrustedStorageType =>
            LL1Disk = [
                publicState |-> LL1Disk.publicState,
                privateStateEnc |-> LL1Disk.privateStateEnc,
                historySummary |-> LL1Disk.historySummary,
                authenticator |-> LL1Disk.authenticator ]
    <2>1. HAVE LL1Disk \in LL1UntrustedStorageType
    <2> ll1disk == [
            publicState |-> LL1Disk.publicState,
            privateStateEnc |-> LL1Disk.privateStateEnc,
            historySummary |-> LL1Disk.historySummary,
            authenticator |-> LL1Disk.authenticator ]
    <2>2. ll1disk =
              [i \in { "publicState",
                       "privateStateEnc",
                       "historySummary",
                       "authenticator" }
                  |-> ll1disk[i]]
      OBVIOUS
    <2>3. LL1Disk =
              [i \in { "publicState",
                       "privateStateEnc",
                       "historySummary",
                       "authenticator" }
                  |-> LL1Disk[i]]
      BY <2>1 DEF LL1UntrustedStorageType
    <2>4. QED
      BY <2>2, <2>3
  <1>2. LL1Disk' \in LL1UntrustedStorageType =>
            LL1Disk' = [
                publicState |-> LL1Disk.publicState',
                privateStateEnc |-> LL1Disk.privateStateEnc',
                historySummary |-> LL1Disk.historySummary',
                authenticator |-> LL1Disk.authenticator' ]
    <2>1. HAVE LL1Disk' \in LL1UntrustedStorageType
    <2> ll1diskprime == [
            publicState |-> LL1Disk.publicState',
            privateStateEnc |-> LL1Disk.privateStateEnc',
            historySummary |-> LL1Disk.historySummary',
            authenticator |-> LL1Disk.authenticator' ]
    <2>2. ll1diskprime =
              [i \in { "publicState",
                       "privateStateEnc",
                       "historySummary",
                       "authenticator" }
                  |-> ll1diskprime[i]]
      OBVIOUS
    <2>3. LL1Disk' =
              [i \in { "publicState",
                       "privateStateEnc",
                       "historySummary",
                       "authenticator" }
                  |-> LL1Disk[i]']
      BY <2>1 DEF LL1UntrustedStorageType
    <2>4. QED
      BY <2>2, <2>3
  <1>3. QED
    BY <1>1, <1>2

----

(***************************************************************************)
(* The LL1RAMRecordCompositionLemma is a formality needed for the prover   *)
(* to compose an LL1RAM record from fields of the appropriate type.        *)
(***************************************************************************)

THEOREM LL1RAMRecordCompositionLemma == \*`^\label{THEOREM-LL1RAMRecordCompositionLemma}^'
    /\  LL1RAM \in LL1UntrustedStorageType =>
            LL1RAM = [
                publicState |-> LL1RAM.publicState,
                privateStateEnc |-> LL1RAM.privateStateEnc,
                historySummary |-> LL1RAM.historySummary,
                authenticator |-> LL1RAM.authenticator ]
    /\  LL1RAM' \in LL1UntrustedStorageType =>
            LL1RAM' = [
                publicState |-> LL1RAM.publicState',
                privateStateEnc |-> LL1RAM.privateStateEnc',
                historySummary |-> LL1RAM.historySummary',
                authenticator |-> LL1RAM.authenticator' ]
  <1>1. LL1RAM \in LL1UntrustedStorageType =>
            LL1RAM = [
                publicState |-> LL1RAM.publicState,
                privateStateEnc |-> LL1RAM.privateStateEnc,
                historySummary |-> LL1RAM.historySummary,
                authenticator |-> LL1RAM.authenticator ]
    <2>1. HAVE LL1RAM \in LL1UntrustedStorageType
    <2> ll1ram == [
            publicState |-> LL1RAM.publicState,
            privateStateEnc |-> LL1RAM.privateStateEnc,
            historySummary |-> LL1RAM.historySummary,
            authenticator |-> LL1RAM.authenticator ]
    <2>2. ll1ram =
              [i \in { "publicState",
                       "privateStateEnc",
                       "historySummary",
                       "authenticator" }
                  |-> ll1ram[i]]
      OBVIOUS
    <2>3. LL1RAM =
              [i \in { "publicState",
                       "privateStateEnc",
                       "historySummary",
                       "authenticator" }
                  |-> LL1RAM[i]]
      BY <2>1 DEF LL1UntrustedStorageType
    <2>4. QED
      BY <2>2, <2>3
  <1>2. LL1RAM' \in LL1UntrustedStorageType =>
            LL1RAM' = [
                publicState |-> LL1RAM.publicState',
                privateStateEnc |-> LL1RAM.privateStateEnc',
                historySummary |-> LL1RAM.historySummary',
                authenticator |-> LL1RAM.authenticator' ]
    <2>1. HAVE LL1RAM' \in LL1UntrustedStorageType
    <2> ll1ramprime == [
            publicState |-> LL1RAM.publicState',
            privateStateEnc |-> LL1RAM.privateStateEnc',
            historySummary |-> LL1RAM.historySummary',
            authenticator |-> LL1RAM.authenticator' ]
    <2>2. ll1ramprime =
              [i \in { "publicState",
                       "privateStateEnc",
                       "historySummary",
                       "authenticator" }
                  |-> ll1ramprime[i]]
      OBVIOUS
    <2>3. LL1RAM' =
              [i \in { "publicState",
                       "privateStateEnc",
                       "historySummary",
                       "authenticator" }
                  |-> LL1RAM[i]']
      BY <2>1 DEF LL1UntrustedStorageType
    <2>4. QED
      BY <2>2, <2>3
  <1>3. QED
    BY <1>1, <1>2

----

(***************************************************************************)
(* The LL1NVRAMRecordCompositionLemma is a formality needed for the prover *)
(* to compose an LL1NVRAM record from fields of the appropriate type.      *)
(***************************************************************************)

THEOREM LL1NVRAMRecordCompositionLemma == \*`^\label{THEOREM-LL1NVRAMRecordCompositionLemma}^'
    /\  LL1NVRAM \in LL1TrustedStorageType =>
            LL1NVRAM = [
                historySummary |-> LL1NVRAM.historySummary,
                symmetricKey |-> LL1NVRAM.symmetricKey ]
    /\  LL1NVRAM' \in LL1TrustedStorageType =>
            LL1NVRAM' = [
                historySummary |-> LL1NVRAM.historySummary',
                symmetricKey |-> LL1NVRAM.symmetricKey' ]
  <1>1. LL1NVRAM \in LL1TrustedStorageType =>
            LL1NVRAM = [
                historySummary |-> LL1NVRAM.historySummary,
                symmetricKey |-> LL1NVRAM.symmetricKey ]
    <2>1. HAVE LL1NVRAM \in LL1TrustedStorageType
    <2> ll1nvram == [
            historySummary |-> LL1NVRAM.historySummary,
            symmetricKey |-> LL1NVRAM.symmetricKey ]
    <2>2. ll1nvram = [i \in {"historySummary", "symmetricKey"} |-> ll1nvram[i]]
      OBVIOUS
    <2>3. LL1NVRAM = [i \in {"historySummary", "symmetricKey"} |-> LL1NVRAM[i]]
      BY <2>1 DEF LL1TrustedStorageType
    <2>4. QED
      BY <2>2, <2>3
  <1>2. LL1NVRAM' \in LL1TrustedStorageType =>
            LL1NVRAM' = [
                historySummary |-> LL1NVRAM.historySummary',
                symmetricKey |-> LL1NVRAM.symmetricKey' ]
    <2>1. HAVE LL1NVRAM' \in LL1TrustedStorageType
    <2> ll1nvramprime == [
            historySummary |-> LL1NVRAM.historySummary',
            symmetricKey |-> LL1NVRAM.symmetricKey' ]
    <2>2. ll1nvramprime = [i \in {"historySummary", "symmetricKey"} |-> ll1nvramprime[i]]
      OBVIOUS
    <2>3. LL1NVRAM' = [i \in {"historySummary", "symmetricKey"} |-> LL1NVRAM[i]']
      BY <2>1 DEF LL1TrustedStorageType
    <2>4. QED
      BY <2>2, <2>3
  <1>3. QED
    BY <1>1, <1>2

----

(***************************************************************************)
(* The CheckpointHasBaseExtensionLemma proves that a history summary       *)
(* produced by the Checkpoint function has an extension field that equals  *)
(* the base hash value.                                                    *)
(***************************************************************************)

THEOREM CheckpointHasBaseExtensionLemma == \*`^\label{THEOREM-CheckpointHasBaseExtensionLemma}^'
    \A historySummary \in HistorySummaryType :
        Checkpoint(historySummary).extension = BaseHashValue
  <1>1. TAKE historySummary \in HistorySummaryType
  <1> checkpointedAnchor == Hash(historySummary.anchor, historySummary.extension)
  <1> checkpointedHistorySummary == [
          anchor |-> checkpointedAnchor,
          extension |-> BaseHashValue ]
  <1> HIDE DEF checkpointedAnchor, checkpointedHistorySummary
  <1>2. CASE historySummary.extension = BaseHashValue
    <2>1. Checkpoint(historySummary) = historySummary
      BY <1>2 DEF Checkpoint
    <2>2. QED
      BY <1>2, <2>1
  <1>3. CASE historySummary.extension # BaseHashValue
    <2>1. Checkpoint(historySummary) = checkpointedHistorySummary
      BY <1>3 DEF Checkpoint, checkpointedHistorySummary, checkpointedAnchor
    <2>2. checkpointedHistorySummary.extension = BaseHashValue
      BY DEF checkpointedHistorySummary
    <2>3. QED
      BY <2>1, <2>2
  <1>4. QED
    BY <1>2, <1>3

----

(***************************************************************************)
(* The SuccessorHasNonBaseExtensionLemma proves that a history summary     *)
(* produced by the Successor function has an extension field that does not *)
(* equal the base hash value.                                              *)
(***************************************************************************)

THEOREM SuccessorHasNonBaseExtensionLemma == \*`^\label{THEOREM-SuccessorHasNonBaseExtensionLemma}^'
    \A historySummary \in HistorySummaryType, input \in InputType, hashBarrier \in HashType :
        Successor(historySummary, input, hashBarrier).extension # BaseHashValue
  <1>1. TAKE historySummary \in HistorySummaryType, input \in InputType, hashBarrier \in HashType
  <1> securedInput == Hash(hashBarrier, input)
  <1> newAnchor == historySummary.anchor
  <1> newExtension == Hash(historySummary.extension, securedInput)
  <1> newHistorySummary == [
          anchor |-> newAnchor,
          extension |-> newExtension ]
  <1> HIDE DEF securedInput, newAnchor, newExtension, newHistorySummary
  <1>2. newExtension # BaseHashValue
    <2>1. historySummary.extension \in HashDomain
      <3>1. historySummary.extension \in HashType
        <4>1. historySummary \in HistorySummaryType
          BY <1>1
        <4>2. QED
          BY <4>1 DEF HistorySummaryType
      <3>2. QED
        BY <3>1 DEF HashDomain
    <2>2. securedInput \in HashDomain
      <3>1. securedInput \in HashType
        <4>1. hashBarrier \in HashDomain
          <5>1. hashBarrier \in HashType
            BY <1>1
          <5>2. QED
            BY <5>1 DEF HashDomain
        <4>2. input \in HashDomain
          <5>1. input \in InputType
            BY <1>1
          <5>2. QED
            BY <5>1 DEF HashDomain
        <4>3. QED
          BY <4>1, <4>2, HashTypeSafe DEF securedInput
      <3>2. QED
        BY <3>1 DEF HashDomain
    <2>3. QED
      BY <2>1, <2>2, BaseHashValueUnique DEF newExtension
  <1>3. Successor(historySummary, input, hashBarrier).extension = newExtension
    BY DEF Successor, newExtension, securedInput
  <1>4. QED
    BY <1>2, <1>3

----

(***************************************************************************)
(* The HistorySummariesMatchUniqueLemma asserts that there is a unique      *)
(* `^\basicname^' history summary that matches a particular `^\fastname^'  *)
(* history summary.                                                        *)
(***************************************************************************)

THEOREM HistorySummariesMatchUniqueLemma == \*`^\label{THEOREM-HistorySummariesMatchUniqueLemma}^'
    \A  ll1HistorySummary1, ll1HistorySummary2 \in HashType,
        ll2HistorySummary \in HistorySummaryType,
        hashBarrier \in HashType :
          ( /\  HistorySummariesMatch(ll1HistorySummary1, ll2HistorySummary, hashBarrier)
            /\  HistorySummariesMatch(ll1HistorySummary2, ll2HistorySummary, hashBarrier) )
        =>
            ll1HistorySummary1 = ll1HistorySummary2
  (*************************************************************************)
  (* It is convenient to define a predicate that captures the quantified   *)
  (* expression of the theorem.                                            *)
  (*************************************************************************)
  <1> HistorySummariesMatchUnique(
          ll1HistorySummary1, ll1HistorySummary2, ll2HistorySummary, hashBarrier) ==
            ( /\  HistorySummariesMatch(ll1HistorySummary1, ll2HistorySummary, hashBarrier)
              /\  HistorySummariesMatch(ll1HistorySummary2, ll2HistorySummary, hashBarrier) )
          =>
              ll1HistorySummary1 = ll1HistorySummary2
  (*************************************************************************)
  (* To prove the universally quantified expressoin, we take a set of      *)
  (* variables of the appropriate types.                                   *)
  (*************************************************************************)
  <1>1. TAKE ll1HistorySummary1, ll1HistorySummary2 \in HashType,
             ll2HistorySummary \in HistorySummaryType,
             hashBarrier \in HashType
  <1> ll2InitialHistorySummary == [ anchor |-> BaseHashValue, extension |-> BaseHashValue ]
  (*************************************************************************)
  (* Our proof is inductive.  First, we prove the base case, which is when *)
  (* the `^\fastname^' history summary equals the initial history summary. *)
  (*************************************************************************)
  <1>2. ll2HistorySummary = ll2InitialHistorySummary =>
            HistorySummariesMatchUnique(
                ll1HistorySummary1, ll1HistorySummary2, ll2HistorySummary, hashBarrier)
    (***********************************************************************)
    (* We assume the antecedent of the base step.                          *)
    (***********************************************************************)
    <2>1. HAVE ll2HistorySummary = ll2InitialHistorySummary
    (***********************************************************************)
    (* The HistorySummariesMatchUnique predicate states an implication.    *)
    (* It suffices to assume the antecedent and prove the consequent.      *)
    (***********************************************************************)
    <2>2. SUFFICES
          ASSUME
              /\  HistorySummariesMatch(ll1HistorySummary1, ll2HistorySummary, hashBarrier)
              /\  HistorySummariesMatch(ll1HistorySummary2, ll2HistorySummary, hashBarrier)
          PROVE
              ll1HistorySummary1 = ll1HistorySummary2
      BY DEF HistorySummariesMatchUnique
    (***********************************************************************)
    (* `^\basicname^' history summary 1 equals the base hash value, by the *)
    (* definition of HistorySummariesMatch for a `^\fastname^' history     *)
    (* summary that equals the initial history summary.                    *)
    (***********************************************************************)
    <2>3. ll1HistorySummary1 = BaseHashValue
      <3>1. HistorySummariesMatch(ll1HistorySummary1, ll2InitialHistorySummary, hashBarrier) =
                ( ll1HistorySummary1 = BaseHashValue )
        <4>1. ll2InitialHistorySummary \in HistorySummaryType
          BY <1>1, <2>1
        <4>2. QED
          BY <1>1, <4>1, HistorySummariesMatchDefinition
      <3>2. HistorySummariesMatch(ll1HistorySummary1, ll2InitialHistorySummary, hashBarrier)
        BY <2>1, <2>2
      <3>3. QED
        BY <3>1, <3>2
    (***********************************************************************)
    (* `^\basicname^' history summary 2 equals the base hash value, by the *)
    (* definition of HistorySummariesMatch for a `^\fastname^' history     *)
    (* summary that equals the initial history summary.                    *)
    (***********************************************************************)
    <2>4. ll1HistorySummary2 = BaseHashValue
      <3>1. HistorySummariesMatch(ll1HistorySummary2, ll2InitialHistorySummary, hashBarrier) =
                ( ll1HistorySummary2 = BaseHashValue )
        <4>1. ll2InitialHistorySummary \in HistorySummaryType
          BY <1>1, <2>1
        <4>2. QED
          BY <1>1, <4>1, HistorySummariesMatchDefinition
      <3>2. HistorySummariesMatch(ll1HistorySummary2, ll2InitialHistorySummary, hashBarrier)
        BY <2>1, <2>2
      <3>3. QED
        BY <3>1, <3>2
    (***********************************************************************)
    (* Since the two `^\basicname^' history summaries each equal the base  *)
    (* hash value, they equal each other.                                  *)
    (***********************************************************************)
    <2>5. QED
      BY <2>3, <2>4
  (*************************************************************************)
  (* Second, we prove the inductive case: If the previous history          *)
  (* summaries defined in the HistorySummariesMatchRecursion predicate are *)
  (* unique, then their successor history summaries are unique.            *)
  (*************************************************************************)
  <1>3. \A  input1, input2 \in InputType,
            previousLL1HistorySummary1, previousLL1HistorySummary2 \in HashType,
            previousLL2HistorySummary \in HistorySummaryType :
              ( /\  HistorySummariesMatchRecursion(
                        ll1HistorySummary1, ll2HistorySummary, hashBarrier)!(
                        input1, previousLL1HistorySummary1, previousLL2HistorySummary)
                /\  HistorySummariesMatchRecursion(
                        ll1HistorySummary2, ll2HistorySummary, hashBarrier)!(
                        input2, previousLL1HistorySummary2, previousLL2HistorySummary)
                /\  HistorySummariesMatchUnique(
                        previousLL1HistorySummary1,
                        previousLL1HistorySummary2,
                        previousLL2HistorySummary,
                        hashBarrier) )
            =>
                HistorySummariesMatchUnique(
                    ll1HistorySummary1, ll1HistorySummary2, ll2HistorySummary, hashBarrier)
    (***********************************************************************)
    (* To prove the universally quantified expressoin, we take a set of    *)
    (* variables of the appropriate types.                                 *)
    (***********************************************************************)
    <2>1. TAKE input1, input2 \in InputType,
               previousLL1HistorySummary1, previousLL1HistorySummary2 \in HashType,
               previousLL2HistorySummary \in HistorySummaryType
    (***********************************************************************)
    (* We assume the antecedent of the inductive step.                     *)
    (***********************************************************************)
    <2>2. HAVE /\  HistorySummariesMatchRecursion(
                       ll1HistorySummary1, ll2HistorySummary, hashBarrier)!(
                       input1, previousLL1HistorySummary1, previousLL2HistorySummary)
               /\  HistorySummariesMatchRecursion(
                       ll1HistorySummary2, ll2HistorySummary, hashBarrier)!(
                       input2, previousLL1HistorySummary2, previousLL2HistorySummary)
               /\  HistorySummariesMatchUnique(
                       previousLL1HistorySummary1,
                       previousLL1HistorySummary2,
                       previousLL2HistorySummary,
                       hashBarrier)
    (***********************************************************************)
    (* The HistorySummariesMatchUnique predicate states an implication.    *)
    (* It suffices to assume the antecedent and prove the consequent.      *)
    (***********************************************************************)
    <2>3. SUFFICES
          ASSUME
              /\  HistorySummariesMatch(ll1HistorySummary1, ll2HistorySummary, hashBarrier)
              /\  HistorySummariesMatch(ll1HistorySummary2, ll2HistorySummary, hashBarrier)
          PROVE
              ll1HistorySummary1 = ll1HistorySummary2
      BY DEF HistorySummariesMatchUnique
    (***********************************************************************)
    (* We prove that the two inputs from the separate instances of the     *)
    (* HistorySummariesMatchRecursion predicate are equal to each other.   *)
    (***********************************************************************)
    <2>4. input1 = input2
      <3>1. LL2HistorySummaryIsSuccessor(
                ll2HistorySummary, previousLL2HistorySummary, input1, hashBarrier)
        BY <2>2
      (*********************************************************************)
      (* We re-state the definitions from LL2HistorySummaryIsSuccessor for *)
      (* input 1.                                                          *)
      (*********************************************************************)
      <3> ll2SuccessorHistorySummary1 == Successor(previousLL2HistorySummary, input1, hashBarrier)
      <3> ll2CheckpointedSuccessorHistorySummary1 == Checkpoint(ll2SuccessorHistorySummary1)
      (*********************************************************************)
      (* We hide the definitions.                                          *)
      (*********************************************************************)
      <3> HIDE DEF ll2SuccessorHistorySummary1, ll2CheckpointedSuccessorHistorySummary1
      (*********************************************************************)
      (* We re-state the definitions from Successor for input 1.           *)
      (*********************************************************************)
      <3> securedInput1 == Hash(hashBarrier, input1)
      <3> newAnchor == previousLL2HistorySummary.anchor
      <3> newExtension1 == Hash(previousLL2HistorySummary.extension, securedInput1)
      <3> ll2NewHistorySummary1 == [
                anchor |-> newAnchor,
                extension |-> newExtension1 ]
      (*********************************************************************)
      (* We prove the types of the definitions from Successor and the      *)
      (* definition of ll2SuccessorHistorySummary1, with help from the     *)
      (* SuccessorDefsTypeSafeLemma.                                       *)
      (*********************************************************************)
      <3>2. /\  securedInput1 \in HashType
            /\  newAnchor \in HashType
            /\  newExtension1 \in HashType
            /\  ll2NewHistorySummary1 \in HistorySummaryType
            /\  ll2NewHistorySummary1.anchor \in HashType
            /\  ll2NewHistorySummary1.extension \in HashType
        BY <1>1, <2>1, SuccessorDefsTypeSafeLemma
      <3>3. ll2SuccessorHistorySummary1 \in HistorySummaryType
        BY <3>2 DEF ll2SuccessorHistorySummary1, Successor
      (*********************************************************************)
      (* We hide the definitions.                                          *)
      (*********************************************************************)
      <3> HIDE DEF securedInput1, newAnchor, newExtension1, ll2NewHistorySummary1
      (*********************************************************************)
      (* We re-state the definitions from Checkpoint for input 1.          *)
      (*********************************************************************)
      <3> checkpointedAnchor1 ==
                Hash(ll2SuccessorHistorySummary1.anchor, ll2SuccessorHistorySummary1.extension)
      <3> ll2CheckpointedHistorySummary1 == [
                anchor |-> checkpointedAnchor1,
                extension |-> BaseHashValue ]
      (*********************************************************************)
      (* We prove the types of the definitions from Checkpoint and the     *)
      (* definition of ll2CheckpointedSuccessorHistorySummary1, with help  *)
      (* from the CheckpointDefsTypeSafeLemma.                             *)
      (*********************************************************************)
      <3>4. /\  checkpointedAnchor1 \in HashType
            /\  ll2CheckpointedHistorySummary1 \in HistorySummaryType
            /\  ll2CheckpointedHistorySummary1.anchor \in HashType
            /\  ll2CheckpointedHistorySummary1.extension \in HashType
        BY <3>3, CheckpointDefsTypeSafeLemma
      <3>5. ll2CheckpointedSuccessorHistorySummary1 \in HistorySummaryType
        BY <3>3, <3>4 DEF ll2CheckpointedSuccessorHistorySummary1, Checkpoint
      (*********************************************************************)
      (* We hide the definitions.                                          *)
      (*********************************************************************)
      <3> HIDE DEF checkpointedAnchor1, ll2CheckpointedHistorySummary1
      <3>6. LL2HistorySummaryIsSuccessor(
                ll2HistorySummary, previousLL2HistorySummary, input2, hashBarrier)
        BY <2>2
      (*********************************************************************)
      (* We re-state the definitions from LL2HistorySummaryIsSuccessor for *)
      (* input 2.                                                          *)
      (*********************************************************************)
      <3> ll2SuccessorHistorySummary2 == Successor(previousLL2HistorySummary, input2, hashBarrier)
      <3> ll2CheckpointedSuccessorHistorySummary2 == Checkpoint(ll2SuccessorHistorySummary2)
      (*********************************************************************)
      (* We hide the definitions.                                          *)
      (*********************************************************************)
      <3> HIDE DEF ll2SuccessorHistorySummary2, ll2CheckpointedSuccessorHistorySummary2
      (*********************************************************************)
      (* We re-state the definitions from Successor for input 2.           *)
      (*********************************************************************)
      <3> securedInput2 == Hash(hashBarrier, input2)
      <3> newExtension2 == Hash(previousLL2HistorySummary.extension, securedInput2)
      <3> ll2NewHistorySummary2 == [
                anchor |-> newAnchor,
                extension |-> newExtension2 ]
      (*********************************************************************)
      (* We prove the types of the definitions from Successor and the      *)
      (* definition of ll2SuccessorHistorySummary2, with help from the     *)
      (* SuccessorDefsTypeSafeLemma.                                       *)
      (*********************************************************************)
      <3>7. /\  securedInput2 \in HashType
            /\  newExtension2 \in HashType
            /\  ll2NewHistorySummary2 \in HistorySummaryType
            /\  ll2NewHistorySummary2.anchor \in HashType
            /\  ll2NewHistorySummary2.extension \in HashType
        BY <1>1, <2>1, SuccessorDefsTypeSafeLemma DEF newAnchor
      <3>8. ll2SuccessorHistorySummary2 \in HistorySummaryType
        BY <3>7 DEF ll2SuccessorHistorySummary2, Successor, newAnchor
      (*********************************************************************)
      (* We hide the definitions.                                          *)
      (*********************************************************************)
      <3> HIDE DEF securedInput2, newExtension2, ll2NewHistorySummary2
      (*********************************************************************)
      (* We re-state the definitions from Checkpoint for input 2.          *)
      (*********************************************************************)
      <3> checkpointedAnchor2 ==
                Hash(ll2SuccessorHistorySummary2.anchor, ll2SuccessorHistorySummary2.extension)
      <3> ll2CheckpointedHistorySummary2 == [
                anchor |-> checkpointedAnchor2,
                extension |-> BaseHashValue ]
      (*********************************************************************)
      (* We prove the types of the definitions from Checkpoint and the     *)
      (* definition of ll2CheckpointedSuccessorHistorySummary2, with help  *)
      (* from the CheckpointDefsTypeSafeLemma.                             *)
      (*********************************************************************)
      <3>9. /\  checkpointedAnchor2 \in HashType
            /\  ll2CheckpointedHistorySummary2 \in HistorySummaryType
            /\  ll2CheckpointedHistorySummary2.anchor \in HashType
            /\  ll2CheckpointedHistorySummary2.extension \in HashType
        BY <3>8, CheckpointDefsTypeSafeLemma
      <3>10. ll2CheckpointedSuccessorHistorySummary2 \in HistorySummaryType
        BY <3>8, <3>9 DEF ll2CheckpointedSuccessorHistorySummary2, Checkpoint
      (*********************************************************************)
      (* We hide the definitions.                                          *)
      (*********************************************************************)
      <3> HIDE DEF checkpointedAnchor2, ll2CheckpointedHistorySummary2
      (*********************************************************************)
      (* We prove that the successor history summaries, as defined in the  *)
      (* LET of the LL2HistorySummaryIsSuccessor predicate, are equal to   *)
      (* each other.                                                       *)
      (*********************************************************************)
      <3>11. ll2SuccessorHistorySummary1 = ll2SuccessorHistorySummary2
        (*******************************************************************)
        (* We consider two cases.  In the first case, the `^\fastname^'    *)
        (* history summary has an extension field that equals the base     *)
        (* hash value.                                                     *)
        (*******************************************************************)
        <4>1. CASE ll2HistorySummary.extension = BaseHashValue
          (*****************************************************************)
          (* The checkpointed successor history summaries, as defined in   *)
          (* the LET within the LL2HistorySummaryIsSuccessor predicate,    *)
          (* are equal.                                                    *)
          (*****************************************************************)
          <5>1. ll2CheckpointedSuccessorHistorySummary1 = ll2CheckpointedSuccessorHistorySummary2
            (***************************************************************)
            (* The LL2HistorySummaryIsSuccessor predicate expresses a      *)
            (* disjunction.  An history summary can be a successor either  *)
            (* by being a direct successor or by being a checkpoint of a   *)
            (* successor.  We prove that, in this case, the history        *)
            (* summary is a checkpoint of a successor formed with input 1. *)
            (* It cannot be a direct successor because its extension field *)
            (* is the base hash value, and any direct successor will have  *)
            (* a non-base extension field.                                 *)
            (***************************************************************)
            <6>1. ll2HistorySummary = ll2CheckpointedSuccessorHistorySummary1
              <7>1. \/  ll2HistorySummary = ll2SuccessorHistorySummary1
                    \/  ll2HistorySummary = ll2CheckpointedSuccessorHistorySummary1
                <8>1. LL2HistorySummaryIsSuccessor(
                          ll2HistorySummary, previousLL2HistorySummary, input1, hashBarrier)
                  BY <2>2
                <8>2. QED
                  BY <8>1
                  DEF LL2HistorySummaryIsSuccessor, ll2SuccessorHistorySummary1,
                      ll2CheckpointedSuccessorHistorySummary1
              <7>2. ll2HistorySummary # ll2SuccessorHistorySummary1
                <8>1. ll2SuccessorHistorySummary1.extension # BaseHashValue
                  BY <1>1, <2>1, SuccessorHasNonBaseExtensionLemma DEF ll2SuccessorHistorySummary1
                <8>2. QED
                  BY <4>1, <8>1
              <7>3. QED
                BY <7>1, <7>2
            (***************************************************************)
            (* The same logic as above applies to the the checkpointed     *)
            (* successor formed with input 2.                              *)
            (***************************************************************)
            <6>2. ll2HistorySummary = ll2CheckpointedSuccessorHistorySummary2
              <7>1. \/  ll2HistorySummary = ll2SuccessorHistorySummary2
                    \/  ll2HistorySummary = ll2CheckpointedSuccessorHistorySummary2
                <8>1. LL2HistorySummaryIsSuccessor(
                          ll2HistorySummary, previousLL2HistorySummary, input2, hashBarrier)
                  BY <2>2
                <8>2. QED
                  BY <8>1
                  DEF LL2HistorySummaryIsSuccessor, ll2SuccessorHistorySummary2,
                      ll2CheckpointedSuccessorHistorySummary2
              <7>2. ll2HistorySummary # ll2SuccessorHistorySummary2
                <8>1. ll2SuccessorHistorySummary2.extension # BaseHashValue
                  BY <1>1, <2>1, SuccessorHasNonBaseExtensionLemma DEF ll2SuccessorHistorySummary2
                <8>2. QED
                  BY <4>1, <8>1
              <7>3. QED
                BY <7>1, <7>2
            (***************************************************************)
            (* Because the two checkpointed successor history summaries    *)
            (* are each equal to the same value, they are equal to each    *)
            (* other.                                                      *)
            (***************************************************************)
            <6>3. QED
              BY <6>1, <6>2
          (*****************************************************************)
          (* We prove the conclusion: The successor history summaries, as  *)
          (* defined in the LET of the LL2HistorySummaryIsSuccessor        *)
          (* predicate, are equal to each other.                           *)
          (*****************************************************************)
          <5>2. QED
            (***************************************************************)
            (* The checkpointed history summaries, as defined in the LET   *)
            (* of the Checkpoint operator, are equal to each other.        *)
            (***************************************************************)
            <6>1. ll2CheckpointedHistorySummary1 = ll2CheckpointedHistorySummary2
              (*************************************************************)
              (* As proven above, the checkpointed history summaries, as   *)
              (* defined in the LET of the LL2HistorySummaryIsSuccessor    *)
              (* predicate, are equal.                                     *)
              (*************************************************************)
              <7>1. ll2CheckpointedSuccessorHistorySummary1 = ll2CheckpointedSuccessorHistorySummary2
                BY <5>1
              (*************************************************************)
              (* The extension field of each successor history summary is  *)
              (* not equal to the base hash value, because they are direct *)
              (* successors.                                               *)
              (*************************************************************)
              <7>2. ll2SuccessorHistorySummary1.extension # BaseHashValue
                BY <1>1, <2>1, SuccessorHasNonBaseExtensionLemma DEF ll2SuccessorHistorySummary1
              <7>3. ll2SuccessorHistorySummary2.extension # BaseHashValue
                BY <1>1, <2>2, SuccessorHasNonBaseExtensionLemma DEF ll2SuccessorHistorySummary2
              (*************************************************************)
              (* The conclusion follows from the definitions, because the  *)
              (* Checkpoint operator is injective under the preimage       *)
              (* constraint of an extension field that is unequal to the   *)
              (* base hash value.                                          *)
              (*************************************************************)
              <7>4. QED
                BY <7>1, <7>2, <7>3
                DEF ll2CheckpointedSuccessorHistorySummary1, ll2CheckpointedSuccessorHistorySummary2,
                      Checkpoint, ll2CheckpointedHistorySummary1, ll2CheckpointedHistorySummary2,
                      checkpointedAnchor1, checkpointedAnchor2
            (***************************************************************)
            (* The individual fields of the successor history summaries    *)
            (* are equal to each other, thanks to the collision resistance *)
            (* of the hash function.                                       *)
            (***************************************************************)
            <6>2. /\  ll2SuccessorHistorySummary1.anchor = ll2SuccessorHistorySummary2.anchor
                  /\  ll2SuccessorHistorySummary1.extension = ll2SuccessorHistorySummary2.extension
              <7>1. checkpointedAnchor1 = checkpointedAnchor2
                BY <6>1 DEF ll2CheckpointedHistorySummary1, ll2CheckpointedHistorySummary2
              <7>2. ll2SuccessorHistorySummary1.anchor \in HashDomain
                BY <3>3 DEF HistorySummaryType, HashDomain
              <7>3. ll2SuccessorHistorySummary1.extension \in HashDomain
                BY <3>3 DEF HistorySummaryType, HashDomain
              <7>4. ll2SuccessorHistorySummary2.anchor \in HashDomain
                BY <3>8 DEF HistorySummaryType, HashDomain
              <7>5. ll2SuccessorHistorySummary2.extension \in HashDomain
                BY <3>8 DEF HistorySummaryType, HashDomain
              <7>6. QED
                (***********************************************************)
                (* Ideally, this QED step should just read:                *)
                (*                                                         *)
                (*  BY <<7>>1, <<7>>2, <<7>>3, <<7>>4, <<7>>5, HashCollisionResistant *)
                (*  DEF checkpointedAnchor1, checkpointedAnchor2           *)
                (*                                                         *)
                (* However, the prover seems to get a little confused in   *)
                (* this instance.  We make life easier for the prover by   *)
                (* definining some local variables and hiding their        *)
                (* definitions before appealing to the                     *)
                (* HashCollisionResistant assumption.                      *)
                (***********************************************************)
                <8> h1a == ll2SuccessorHistorySummary1.anchor
                <8> h2a == ll2SuccessorHistorySummary1.extension
                <8> h1b == ll2SuccessorHistorySummary2.anchor
                <8> h2b == ll2SuccessorHistorySummary2.extension
                <8>1. h1a \in HashDomain
                  BY <7>2
                <8>2. h2a \in HashDomain
                  BY <7>3
                <8>3. h1b \in HashDomain
                  BY <7>4
                <8>4. h2b \in HashDomain
                  BY <7>5
                <8>5. Hash(h1a, h2a) = Hash(h1b, h2b)
                  BY <7>1 DEF checkpointedAnchor1, checkpointedAnchor2
                <8>6. h1a = h1b /\ h2a = h2b
                  <9> HIDE DEF h1a, h2a, h1b, h2b
                  <9>1. QED
                    BY <8>1, <8>2, <8>3, <8>4, <8>5, HashCollisionResistant
                <8>7. QED
                  BY <8>6
            (***************************************************************)
            (* Because the fields are equal, the records are equal, but    *)
            (* proving this requires that we prove the types of the        *)
            (* records and invoke the                                      *)
            (* HistorySummaryRecordCompositionLemma.                       *)
            (***************************************************************)
            <6>3. ll2SuccessorHistorySummary1 \in HistorySummaryType
              BY <3>3
            <6>4. ll2SuccessorHistorySummary2 \in HistorySummaryType
              BY <3>8
            <6>5. QED
              BY <6>2, <6>3, <6>4, HistorySummaryRecordCompositionLemma DEF HistorySummaryType
        (*******************************************************************)
        (* In the second case, the `^\fastname^' history summary has an    *)
        (* extension field that does not equal the base hash value.        *)
        (*******************************************************************)
        <4>2. CASE ll2HistorySummary.extension # BaseHashValue
          (*****************************************************************)
          (* The LL2HistorySummaryIsSuccessor predicate expresses a        *)
          (* disjunction.  An history summary can be a successor either by *)
          (* being a direct successor or by being a checkpoint of a        *)
          (* successor.  We prove that, in this case, history summary 1 is *)
          (* a direct successor formed with input 1.  It cannot be a       *)
          (* checkpoint because its extension field is not the base hash   *)
          (* value, and any checkpoint will have a base extension field.   *)
          (*****************************************************************)
          <5>1. ll2HistorySummary = ll2SuccessorHistorySummary1
            <6>1. \/  ll2HistorySummary = ll2SuccessorHistorySummary1
                  \/  ll2HistorySummary = ll2CheckpointedSuccessorHistorySummary1
              <7>1. LL2HistorySummaryIsSuccessor(
                        ll2HistorySummary, previousLL2HistorySummary, input1, hashBarrier)
                BY <2>2
              <7>2. QED
                BY <7>1
                DEF LL2HistorySummaryIsSuccessor, ll2SuccessorHistorySummary1,
                    ll2CheckpointedSuccessorHistorySummary1
            <6>2. ll2HistorySummary # ll2CheckpointedSuccessorHistorySummary1
              <7>1. ll2CheckpointedSuccessorHistorySummary1.extension = BaseHashValue
                BY <3>3, CheckpointHasBaseExtensionLemma
                DEF ll2CheckpointedSuccessorHistorySummary1
              <7>2. QED
                BY <4>2, <7>1
            <6>3. QED
              BY <6>1, <6>2
          (*****************************************************************)
          (* The same logic as above applies to the the successor formed   *)
          (* with input 2.                                                 *)
          (*****************************************************************)
          <5>2. ll2HistorySummary = ll2SuccessorHistorySummary2
            <6>1. \/  ll2HistorySummary = ll2SuccessorHistorySummary2
                  \/  ll2HistorySummary = ll2CheckpointedSuccessorHistorySummary2
              <7>1. LL2HistorySummaryIsSuccessor(
                        ll2HistorySummary, previousLL2HistorySummary, input2, hashBarrier)
                BY <2>2
              <7>2. QED
                BY <7>1
                DEF LL2HistorySummaryIsSuccessor, ll2SuccessorHistorySummary2,
                    ll2CheckpointedSuccessorHistorySummary2
            <6>2. ll2HistorySummary # ll2CheckpointedSuccessorHistorySummary2
              <7>1. ll2CheckpointedSuccessorHistorySummary2.extension = BaseHashValue
                BY <3>8, CheckpointHasBaseExtensionLemma
                DEF ll2CheckpointedSuccessorHistorySummary2
              <7>2. QED
                BY <4>2, <7>1
            <6>3. QED
              BY <6>1, <6>2
          (*****************************************************************)
          (* Because the two successor history summaries are each equal to *)
          (* the same value, they are equal to each other.                 *)
          (*****************************************************************)
          <5>3. QED
            BY <5>1, <5>2
        (*******************************************************************)
        (* The two cases are exhaustive.                                   *)
        (*******************************************************************)
        <4>3. QED
          BY <4>1, <4>2
      (*********************************************************************)
      (* The secured inputss are equal to each other, because the new      *)
      (* extensions are equal to each other, and the hash is               *)
      (* collision-resistant.                                              *)
      (*********************************************************************)
      <3>12. securedInput1 = securedInput2
        <4>1. newExtension1 = newExtension2
          <5>1. ll2NewHistorySummary1 = ll2NewHistorySummary2
            BY <3>11
            DEF ll2SuccessorHistorySummary1, ll2SuccessorHistorySummary2, Successor,
                  ll2NewHistorySummary1, ll2NewHistorySummary2, newExtension1, newExtension2,
                  newAnchor, securedInput1, securedInput2
          <5>2. QED
            BY <5>1 DEF ll2NewHistorySummary1, ll2NewHistorySummary2
        <4>2. previousLL2HistorySummary.extension \in HashDomain
          BY <2>1 DEF HistorySummaryType, HashDomain
        <4>3. securedInput1 \in HashDomain
          BY <3>2 DEF HashDomain
        <4>4. securedInput2 \in HashDomain
          BY <3>7 DEF HashDomain
        <4>5. QED
          (*****************************************************************)
          (* Ideally, this QED step should just read:                      *)
          (*                                                               *)
          (*  BY <<4>>1, <<4>>2, <<4>>3, <<4>>4, HashCollisionResistant    *)
          (*                                                               *)
          (* However, the prover seems to get a little confused in this    *)
          (* instance.  We make life easier for the prover by definining   *)
          (* some local variables and hiding their definitions before      *)
          (* appealing to the HashCollisionResistant assumption.           *)
          (*****************************************************************)
          <5> h1a == previousLL2HistorySummary.extension
          <5> h2a == securedInput1
          <5> h2b == securedInput2
          <5>1. h1a \in HashDomain
            BY <4>2
          <5>2. h2a \in HashDomain
            BY <4>3
          <5>3. h2b \in HashDomain
            BY <4>4
          <5>4. Hash(h1a, h2a) = Hash(h1a, h2b)
            BY <4>1 DEF newExtension1, newExtension2
          <5>5. h2a = h2b
            <6> HIDE DEF h1a, h2a, h2b
            <6>1. QED
              BY <5>1, <5>2, <5>3, <5>4, HashCollisionResistant
          <5>6. QED
            BY <5>5
      (*********************************************************************)
      (* The input values are equal to each other, because the secured     *)
      (* inputs are equal, and the hash function is collision-resistant.   *)
      (*********************************************************************)
      <3>16. QED
        <4>1. securedInput1 = securedInput2
          BY <3>12
        <4>2. hashBarrier \in HashDomain
          BY <1>1 DEF HashDomain
        <4>3. input1 \in HashDomain
          BY <2>1 DEF HashDomain
        <4>4. input2 \in HashDomain
          BY <2>2 DEF HashDomain
        <4>5. QED
          BY <4>1, <4>2, <4>3, <4>4, HashCollisionResistant
          DEF securedInput1, securedInput2
    (***********************************************************************)
    (* We prove that the two previous `^\basicname^' history summaries     *)
    (* from the separate instances of the HistorySummariesMatchRecursion   *)
    (* predicate are equal to each other.  This follows from the recursive *)
    (* use of the HistorySummariesMatchUnique predicate on the previous    *)
    (* history summaries.                                                  *)
    (***********************************************************************)
    <2>5. previousLL1HistorySummary1 = previousLL1HistorySummary2
      <3>1. HistorySummariesMatch(
                previousLL1HistorySummary1, previousLL2HistorySummary, hashBarrier)
        BY <2>2 DEF HistorySummariesMatchRecursion
      <3>2. HistorySummariesMatch(
                previousLL1HistorySummary2, previousLL2HistorySummary, hashBarrier)
        BY <2>2 DEF HistorySummariesMatchRecursion
      <3>3. HistorySummariesMatchUnique(
                       previousLL1HistorySummary1,
                       previousLL1HistorySummary2,
                       previousLL2HistorySummary,
                       hashBarrier)
        BY <2>2
      <3>4. QED
        BY <3>1, <3>2, <3>3 DEF HistorySummariesMatchUnique
    (***********************************************************************)
    (* The conclusion follows directly from the equality of the previous   *)
    (* history summaries and the inputs, since the current history         *)
    (* summaries are generated as hashes of those values.                  *)
    (***********************************************************************)
    <2>6. QED
      <3>1. ll1HistorySummary1 = Hash(previousLL1HistorySummary1, input1)
        BY <2>2 DEF HistorySummariesMatchRecursion
      <3>2. ll1HistorySummary2 = Hash(previousLL1HistorySummary2, input2)
        BY <2>2 DEF HistorySummariesMatchRecursion
      <3>3. QED
        BY <2>4, <2>5, <3>1, <3>2
  (*************************************************************************)
  (* Due to language limitations, we do not state or prove the step that   *)
  (* ties together the base case and the inductive case into a proof for   *)
  (* all cases.                                                            *)
  (*************************************************************************)
  <1>4. QED
    OMITTED

----

(***************************************************************************)
(* The AuthenticatorsMatchUniqueLemma asserts that there is a unique       *)
(* `^\basicname^' authenticator that matches a particular `^\fastname^'    *)
(* authenticator.                                                          *)
(***************************************************************************)

THEOREM AuthenticatorsMatchUniqueLemma == \*`^\label{THEOREM-AuthenticatorsMatchUniqueLemma}^'
    \A  ll1Authenticator1, ll1Authenticator2 \in MACType,
        ll2Authenticator \in MACType,
        symmetricKey1, symmetricKey2 \in SymmetricKeyType,
        hashBarrier \in HashType :
          ( /\  AuthenticatorsMatch(
                    ll1Authenticator1, ll2Authenticator, symmetricKey1, hashBarrier)
            /\  AuthenticatorsMatch(
                    ll1Authenticator2, ll2Authenticator, symmetricKey2, hashBarrier) )
        =>
            ll1Authenticator1 = ll1Authenticator2
  (*************************************************************************)
  (* To prove the universally quantified expression, we take a set of      *)
  (* variables in the appropriate types.                                   *)
  (*************************************************************************)
  <1>1. TAKE ll1Authenticator1, ll1Authenticator2 \in MACType,
             ll2Authenticator \in MACType,
             symmetricKey1, symmetricKey2 \in SymmetricKeyType,
             hashBarrier \in HashType
  (*************************************************************************)
  (* We assume the antecedent of the implication.                          *)
  (*************************************************************************)
  <1>2. HAVE /\  AuthenticatorsMatch(
                     ll1Authenticator1, ll2Authenticator, symmetricKey1, hashBarrier)
             /\  AuthenticatorsMatch(
                     ll1Authenticator2, ll2Authenticator, symmetricKey2, hashBarrier)
  (*************************************************************************)
  (* We pick a set of variables that witness the existentials inside the   *)
  (* AuthenticatorsMatch predicate for authenticator 1.                    *)
  (*************************************************************************)
  <1>3. PICK stateHash1 \in HashType,
             ll1HistorySummary1 \in HashType,
             ll2HistorySummary1 \in HistorySummaryType :
             AuthenticatorsMatch(
                 ll1Authenticator1, ll2Authenticator, symmetricKey1, hashBarrier)!(
                 stateHash1, ll1HistorySummary1, ll2HistorySummary1)
    BY <1>2 DEF AuthenticatorsMatch
  (*************************************************************************)
  (* We re-state the definitions from the LET in AuthenticatorsMatch for   *)
  (* authenticator 1.                                                      *)
  (*************************************************************************)
  <1> ll1HistoryStateBinding1 == Hash(ll1HistorySummary1, stateHash1)
  <1> ll2HistorySummaryHash1 == Hash(ll2HistorySummary1.anchor, ll2HistorySummary1.extension)
  <1> ll2HistoryStateBinding1 == Hash(ll2HistorySummaryHash1, stateHash1)
  (*************************************************************************)
  (* We prove the types of the definitions, with the help of the           *)
  (* AuthenticatorsMatchDefsTypeSafeLemma.                                 *)
  (*************************************************************************)
  <1>4. /\  ll1HistoryStateBinding1 \in HashType
        /\  ll2HistorySummaryHash1 \in HashType
        /\  ll2HistoryStateBinding1 \in HashType
    BY <1>3, AuthenticatorsMatchDefsTypeSafeLemma
  (*************************************************************************)
  (* We hide the definitions.                                              *)
  (*************************************************************************)
  <1> HIDE DEF ll1HistoryStateBinding1, ll2HistorySummaryHash1, ll2HistoryStateBinding1
  (*************************************************************************)
  (* We pick a set of variables that witness the existentials inside the   *)
  (* AuthenticatorsMatch predicate for authenticator 2.                    *)
  (*************************************************************************)
  <1>5. PICK stateHash2 \in HashType,
             ll1HistorySummary2 \in HashType,
             ll2HistorySummary2 \in HistorySummaryType :
             AuthenticatorsMatch(
                 ll1Authenticator2, ll2Authenticator, symmetricKey2, hashBarrier)!(
                 stateHash2, ll1HistorySummary2, ll2HistorySummary2)
    BY <1>2 DEF AuthenticatorsMatch
  (*************************************************************************)
  (* We re-state the definitions from the LET in AuthenticatorsMatch for   *)
  (* authenticator 2.                                                      *)
  (*************************************************************************)
  <1> ll1HistoryStateBinding2 == Hash(ll1HistorySummary2, stateHash2)
  <1> ll2HistorySummaryHash2 == Hash(ll2HistorySummary2.anchor, ll2HistorySummary2.extension)
  <1> ll2HistoryStateBinding2 == Hash(ll2HistorySummaryHash2, stateHash2)
  (*************************************************************************)
  (* We prove the types of the definitions, with the help of the           *)
  (* AuthenticatorsMatchDefsTypeSafeLemma.                                 *)
  (*************************************************************************)
  <1>6. /\  ll1HistoryStateBinding2 \in HashType
        /\  ll2HistorySummaryHash2 \in HashType
        /\  ll2HistoryStateBinding2 \in HashType
    BY <1>3, AuthenticatorsMatchDefsTypeSafeLemma
   (************************************************************************)
   (* We hide the definitions.                                             *)
   (************************************************************************)
  <1> HIDE DEF ll1HistoryStateBinding2, ll2HistorySummaryHash2, ll2HistoryStateBinding2
  (*************************************************************************)
  (* Beginning the proof proper, we first prove that symmetric key 1 and   *)
  (* history state binding 1 validate the MAC generated with symmetric key *)
  (* 2 and history state binding 2.                                        *)
  (*************************************************************************)
  <1>7. ValidateMAC(
            symmetricKey1,
            ll2HistoryStateBinding1,
            GenerateMAC(symmetricKey2, ll2HistoryStateBinding2))
    <2>1. ValidateMAC(symmetricKey1, ll2HistoryStateBinding1, ll2Authenticator)
      BY <1>3 DEF ll2HistoryStateBinding1, ll2HistorySummaryHash1
    <2>2. ll2Authenticator = GenerateMAC(symmetricKey2, ll2HistoryStateBinding2)
      <3>1. ValidateMAC(symmetricKey2, ll2HistoryStateBinding2, ll2Authenticator)
        BY <1>5 DEF ll2HistoryStateBinding2, ll2HistorySummaryHash2
      <3>2. QED
        BY <1>1, <1>6, <3>1, MACConsistent
    <2>3. QED
      BY <2>1, <2>2
  (*************************************************************************)
  (* Next, we prove that the two `^\fastname^' history summary hashes are  *)
  (* equal and the two state hashes are equal.  This follows from the      *)
  (* collision resistance of the MAC functions applied to the previous     *)
  (* step, followed by the collision resistance of the hash function.      *)
  (*************************************************************************)
  <1>8. /\ ll2HistorySummaryHash1 = ll2HistorySummaryHash2
        /\ stateHash1 = stateHash2
    <2>1. ll2HistoryStateBinding1 = ll2HistoryStateBinding2
      BY <1>1, <1>4, <1>6, <1>7, MACCollisionResistant
    <2>2. ll2HistorySummaryHash1 \in HashDomain
      BY <1>4 DEF HashDomain
    <2>3. stateHash1 \in HashDomain
      BY <1>3 DEF HashDomain
    <2>4. ll2HistorySummaryHash2 \in HashDomain
      BY <1>6 DEF HashDomain
    <2>5. stateHash2 \in HashDomain
      BY <1>5 DEF HashDomain
    <2>6. QED
      BY <2>1, <2>2, <2>3, <2>4, <2>5, HashCollisionResistant
      DEF ll2HistoryStateBinding1, ll2HistoryStateBinding2
  (*************************************************************************)
  (* Because the `^\fastname^' history summary hashes are equal, the       *)
  (* `^\fastname^' history summaries are equal.                            *)
  (*************************************************************************)
  <1>9. ll2HistorySummary1 = ll2HistorySummary2
    <2>1. /\  ll2HistorySummary1.anchor = ll2HistorySummary2.anchor
          /\  ll2HistorySummary1.extension = ll2HistorySummary2.extension
      <3>1. ll2HistorySummary1.anchor \in HashDomain
        BY <1>3 DEF HistorySummaryType, HashDomain
      <3>2. ll2HistorySummary1.extension \in HashDomain
        BY <1>3 DEF HistorySummaryType, HashDomain
      <3>3. ll2HistorySummary2.anchor \in HashDomain
        BY <1>5 DEF HistorySummaryType, HashDomain
      <3>4. ll2HistorySummary2.extension \in HashDomain
        BY <1>5 DEF HistorySummaryType, HashDomain
      <3>5. QED
        (*******************************************************************)
        (* Ideally, this QED step should just read:                        *)
        (*                                                                 *)
        (*  BY <<3>>1, <<3>>2, <<3>>3, <<3>>4, <<1>>8, HashCollisionResistant *)
        (*  DEF ll2HistorySummaryHash1, ll2HistorySummaryHash2             *)
        (*                                                                 *)
        (* However, the prover seems to get a little confused in this      *)
        (* instance.  We make life easier for the prover by definining     *)
        (* some local variables and hiding their definitions before        *)
        (* appealing to the HashCollisionResistant assumption.             *)
        (*******************************************************************)
        <4> h1a == ll2HistorySummary1.anchor
        <4> h2a == ll2HistorySummary1.extension
        <4> h1b == ll2HistorySummary2.anchor
        <4> h2b == ll2HistorySummary2.extension
        <4>1. h1a \in HashDomain
          BY <3>1
        <4>2. h2a \in HashDomain
          BY <3>2
        <4>3. h1b \in HashDomain
          BY <3>3
        <4>4. h2b \in HashDomain
          BY <3>4
        <4>5. Hash(h1a, h2a) = Hash(h1b, h2b)
          BY <1>8 DEF ll2HistorySummaryHash1, ll2HistorySummaryHash2
        <4>6. h1a = h1b /\ h2a = h2b
          <5> HIDE DEF h1a, h2a, h1b, h2b
          <5>1. QED
            BY <4>1, <4>2, <4>3, <4>4, <4>5, HashCollisionResistant
        <4>7. QED
          BY <4>6
    <2>2. QED
      BY <1>3, <1>5, <2>1, HistorySummaryRecordCompositionLemma
  (*************************************************************************)
  (* Finally, we prove the equality of the `^\basicname^' state            *)
  (* authenticators.  This equality follows from the fact that they are    *)
  (* generated with identical symmetric keys and identical history state   *)
  (* bindings.  The equality of the symmetric keys follows from the        *)
  (* unforgeability of the MAC.  The equality of the `^\basicname^'        *)
  (* history state bindings follows from the equality of the               *)
  (* `^\basicname^' history summaries and the equality of the state        *)
  (* hashes.  And the equality of the `^\basicname^' history summaries     *)
  (* follows from the equality of the `^\fastname^' history summaries, by  *)
  (* employing the HistorySummariesMatchUniqueLemma.                       *)
  (*************************************************************************)
  <1>10. QED
    <2>1. ll1HistoryStateBinding1 = ll1HistoryStateBinding2
      <3>1. ll1HistorySummary1 = ll1HistorySummary2
        <4>1. HistorySummariesMatch(ll1HistorySummary1, ll2HistorySummary1, hashBarrier)
          BY <1>3
        <4>2. HistorySummariesMatch(ll1HistorySummary2, ll2HistorySummary2, hashBarrier)
          BY <1>5
        <4>3. QED
          BY <1>3, <1>5, <1>9, <4>1, <4>2, HistorySummariesMatchUniqueLemma
      <3>2. QED
        BY <1>8, <3>1 DEF ll1HistoryStateBinding1, ll1HistoryStateBinding2
    <2>2. symmetricKey1 = symmetricKey2
      BY <1>1, <1>4, <1>6, <1>7, MACUnforgeable
    <2>3. ll1Authenticator1 = GenerateMAC(symmetricKey1, ll1HistoryStateBinding1)
      BY <1>3 DEF ll1HistoryStateBinding1
    <2>4. ll1Authenticator2 = GenerateMAC(symmetricKey2, ll1HistoryStateBinding2)
      BY <1>5 DEF ll1HistoryStateBinding2
    <2>5. QED
      BY <2>1, <2>2, <2>3, <2>4

----

(***************************************************************************)
(* The AuthenticatorSetsMatchUniqueLemma asserts that there is a unique    *)
(* `^\basicname^' set of authenticators that matches a particular          *)
(* `^\fastname^' set of authenticators.                                    *)
(***************************************************************************)

THEOREM AuthenticatorSetsMatchUniqueLemma == \*`^\label{THEOREM-AuthenticatorSetsMatchUniqueLemma}^'
    \A  ll1AuthenticatorSet1, ll1AuthenticatorSet2 \in SUBSET MACType,
        ll2AuthenticatorSet \in SUBSET MACType,
        symmetricKey \in SymmetricKeyType,
        hashBarrier \in HashType :
          ( /\  AuthenticatorSetsMatch(
                    ll1AuthenticatorSet1, ll2AuthenticatorSet, symmetricKey, hashBarrier)
            /\  AuthenticatorSetsMatch(
                    ll1AuthenticatorSet2, ll2AuthenticatorSet, symmetricKey, hashBarrier) )
        =>
            ll1AuthenticatorSet1 = ll1AuthenticatorSet2
  (*************************************************************************)
  (* To prove the universally quantified expression, we take a set of      *)
  (* variables in the appropriate types.                                   *)
  (*************************************************************************)
  <1>1. TAKE ll1AuthenticatorSet1, ll1AuthenticatorSet2 \in SUBSET MACType,
             ll2AuthenticatorSet \in SUBSET MACType,
             symmetricKey \in SymmetricKeyType,
             hashBarrier \in HashType
  (*************************************************************************)
  (* We assume the antecedent of the implication.                          *)
  (*************************************************************************)
  <1>2. HAVE /\  AuthenticatorSetsMatch(
                     ll1AuthenticatorSet1, ll2AuthenticatorSet, symmetricKey, hashBarrier)
             /\  AuthenticatorSetsMatch(
                     ll1AuthenticatorSet2, ll2AuthenticatorSet, symmetricKey, hashBarrier)
  (*************************************************************************)
  (* The proof has two main steps, which are mirror images of each other.  *)
  (* First, we prove that every `^\basicname^' authenticator in set 1 is   *)
  (* also in set 2.                                                        *)
  (*************************************************************************)
  <1>3. \A ll1Authenticator \in ll1AuthenticatorSet1 :
            ll1Authenticator \in ll1AuthenticatorSet2
    (***********************************************************************)
    (* To prove the universally quantified expression, we take an          *)
    (* arbitrary authenticator in `^\basicname^' set 1.                    *)
    (***********************************************************************)
    <2>1. TAKE ll1Authenticator1 \in ll1AuthenticatorSet1
    (***********************************************************************)
    (* Then we pick a matching authenticator in the `^\fastname^' set.     *)
    (***********************************************************************)
    <2>2. PICK ll2Authenticator \in ll2AuthenticatorSet :
              AuthenticatorsMatch(
                  ll1Authenticator1, ll2Authenticator, symmetricKey, hashBarrier)
      BY <1>2 DEF AuthenticatorSetsMatch
    (***********************************************************************)
    (* We then pick a matching authenticator in set 2.                     *)
    (***********************************************************************)
    <2>3. PICK ll1Authenticator2 \in ll1AuthenticatorSet2 :
              AuthenticatorsMatch(
                  ll1Authenticator2, ll2Authenticator, symmetricKey, hashBarrier)
      BY <1>2 DEF AuthenticatorSetsMatch
    (***********************************************************************)
    (* The two `^\basicname^' authenticators match, by the                 *)
    (* AuthenticatorsMatchUniqueLemma.                                     *)
    (***********************************************************************)
    <2>4. ll1Authenticator1 = ll1Authenticator2
      <3>1. ll1Authenticator1 \in MACType
        BY <1>1, <2>1
      <3>2. ll1Authenticator2 \in MACType
        BY <1>1, <2>3
      <3>3. ll2Authenticator \in MACType
        BY <1>1, <2>2
      <3>4. QED
        BY <2>2, <2>3, <3>1, <3>2, <3>3, AuthenticatorsMatchUniqueLemma
    <2>5. QED
      BY <2>3, <2>4
  (*************************************************************************)
  (* Second, we prove that every `^\basicname^' authenticator in set 2 is  *)
  (* also in set 1.                                                        *)
  (*************************************************************************)
  <1>4. \A ll1Authenticator \in ll1AuthenticatorSet2 :
            ll1Authenticator \in ll1AuthenticatorSet1
    (***********************************************************************)
    (* To prove the universally quantified expression, we take an          *)
    (* arbitrary authenticator in `^\basicname^' set 2.                    *)
    (***********************************************************************)
    <2>1. TAKE ll1Authenticator2 \in ll1AuthenticatorSet2
    (***********************************************************************)
    (* Then we pick a matching authenticator in the `^\fastname^' set.     *)
    (***********************************************************************)
    <2>2. PICK ll2Authenticator \in ll2AuthenticatorSet :
              AuthenticatorsMatch(
                  ll1Authenticator2, ll2Authenticator, symmetricKey, hashBarrier)
      BY <1>2 DEF AuthenticatorSetsMatch
    (***********************************************************************)
    (* We then pick a matching authenticator in set 1.                     *)
    (***********************************************************************)
    <2>3. PICK ll1Authenticator1 \in ll1AuthenticatorSet1 :
              AuthenticatorsMatch(
                  ll1Authenticator1, ll2Authenticator, symmetricKey, hashBarrier)
      BY <1>2 DEF AuthenticatorSetsMatch
    (***********************************************************************)
    (* The two `^\basicname^' authenticators match, by the                 *)
    (* AuthenticatorsMatchUniqueLemma.                                     *)
    (***********************************************************************)
    <2>4. ll1Authenticator1 = ll1Authenticator2
      <3>1. ll1Authenticator1 \in MACType
        BY <1>1, <2>3
      <3>2. ll1Authenticator2 \in MACType
        BY <1>1, <2>1
      <3>3. ll2Authenticator \in MACType
        BY <1>1, <2>2
      <3>4. QED
        BY <2>2, <2>3, <3>1, <3>2, <3>3, AuthenticatorsMatchUniqueLemma
    <2>5. QED
      BY <2>3, <2>4
  <1>5. QED
    BY <1>3, <1>4

----

(***************************************************************************)
(* The LL2NVRAMLogicalHistorySummaryTypeSafe lemma asserts that the        *)
(* LL2NVRAMLogicalHistorySummary is of HistorySummaryType as long as the   *)
(* `^\fastname^' type invariant is satisfied.                              *)
(***************************************************************************)

THEOREM LL2NVRAMLogicalHistorySummaryTypeSafe == \*`^\label{THEOREM-LL2NVRAMLogicalHistorySummaryTypeSafe}^'
    /\  LL2TypeInvariant => LL2NVRAMLogicalHistorySummary \in HistorySummaryType
    /\  LL2TypeInvariant' => LL2NVRAMLogicalHistorySummary' \in HistorySummaryType
  <1>1. LL2TypeInvariant => LL2NVRAMLogicalHistorySummary \in HistorySummaryType
    <2>1. HAVE LL2TypeInvariant
    <2>2. LL2NVRAM.historySummaryAnchor \in HashType
      BY <2>1, LL2SubtypeImplicationLemma DEF LL2SubtypeImplication
    <2>3. LL2SPCR \in HashType
      BY <2>1 DEF LL2TypeInvariant
    <2>4. BaseHashValue \in HashType
      BY ConstantsTypeSafe
    <2>5. CrazyHashValue \in HashType
      BY CrazyHashValueTypeSafe
    <2>6. QED
      BY <2>2, <2>3, <2>4, <2>5 DEF LL2NVRAMLogicalHistorySummary, HistorySummaryType
  <1>2. LL2TypeInvariant' => LL2NVRAMLogicalHistorySummary' \in HistorySummaryType
    <2>1. HAVE LL2TypeInvariant'
    <2>2. LL2NVRAM.historySummaryAnchor' \in HashType
      BY <2>1, LL2SubtypeImplicationLemma DEF LL2SubtypeImplication
    <2>3. LL2SPCR' \in HashType
      BY <2>1 DEF LL2TypeInvariant
    <2>4. BaseHashValue \in HashType
      BY ConstantsTypeSafe
    <2>5. CrazyHashValue \in HashType
      BY CrazyHashValueTypeSafe
    <2>6. QED
      BY <2>2, <2>3, <2>4, <2>5 DEF LL2NVRAMLogicalHistorySummary, HistorySummaryType
  <1>3. QED
    BY <1>1, <1>2

----

(***************************************************************************)
(* The AuthenticatorInSetLemma states that if (1) two state authenticators *)
(* match across the two specs, (2) two authenticator sets match across the *)
(* two specs, and (3) in the `^\fastname^' spec, the authenticator is an   *)
(* element of the authenticator set, then in the `^\basicname^' spec, the  *)
(* authenticator is an element of the authenticator set.                   *)
(***************************************************************************)

THEOREM AuthenticatorInSetLemma == \*`^\label{THEOREM-AuthenticatorInSetLemma}^'
    \A  ll1Authenticator \in MACType,
        ll2Authenticator \in MACType,
        ll1Authenticators \in SUBSET MACType,
        ll2Authenticators \in SUBSET MACType,
        symmetricKey1 \in SymmetricKeyType,
        symmetricKey2 \in SymmetricKeyType,
        hashBarrier \in HashType :
          ( /\  AuthenticatorsMatch(
                    ll1Authenticator, ll2Authenticator, symmetricKey1, hashBarrier)
            /\  AuthenticatorSetsMatch(
                    ll1Authenticators, ll2Authenticators, symmetricKey2, hashBarrier)
            /\  ll2Authenticator \in ll2Authenticators )
        =>
            ll1Authenticator \in ll1Authenticators
  (*************************************************************************)
  (* To prove the universally quantified expression, we take a set of      *)
  (* variables of the appropriate types.                                   *)
  (*************************************************************************)
  <1>1. TAKE ll1Authenticator1 \in MACType,
             ll2Authenticator1 \in MACType,
             ll1Authenticators \in SUBSET MACType,
             ll2Authenticators \in SUBSET MACType,
             symmetricKey1 \in SymmetricKeyType,
             symmetricKey2 \in SymmetricKeyType,
             hashBarrier \in HashType
  (*************************************************************************)
  (* To prove the implication, we assume the antecedent.                   *)
  (*************************************************************************)
  <1>2. HAVE /\  AuthenticatorsMatch(
                     ll1Authenticator1, ll2Authenticator1, symmetricKey1, hashBarrier)
             /\  AuthenticatorSetsMatch(
                     ll1Authenticators, ll2Authenticators, symmetricKey2, hashBarrier)
             /\  ll2Authenticator1 \in ll2Authenticators
  (*************************************************************************)
  (* We pick a new `^\basicname^' authenticator that (1) is in the given   *)
  (* `^\basicname^' authenticator set and (2) matches the given            *)
  (* `^\fastname^' authenticator.                                          *)
  (*************************************************************************)
  <1>3. PICK ll1Authenticator2 \in ll1Authenticators :
            AuthenticatorsMatch(
                ll1Authenticator2, ll2Authenticator1, symmetricKey2, hashBarrier)
    (***********************************************************************)
    (* We first prove that such an element exists.                         *)
    (***********************************************************************)
    <2>1. \E ll1Authenticator2 \in ll1Authenticators :
              AuthenticatorsMatch(
                  ll1Authenticator2, ll2Authenticator1, symmetricKey2, hashBarrier)
      (*********************************************************************)
      (* By hypothesis of the lemma, the given `^\fastname^' authenticator *)
      (* is an element of the given `^\fastname^' authenticator set.       *)
      (*********************************************************************)
      <3>1. ll2Authenticator1 \in ll2Authenticators
        BY <1>2
      (*********************************************************************)
      (* Because the two authenticator sets match, there exists a matching *)
      (* `^\basicname^' authenticator for every `^\fastname^'              *)
      (* authenticator.  This follows from the definition of               *)
      (* AuthenticatorSetsMatch.                                           *)
      (*********************************************************************)
      <3>2. \A ll2Authenticator2 \in ll2Authenticators :
                \E ll1Authenticator2 \in ll1Authenticators :
                    AuthenticatorsMatch(
                        ll1Authenticator2, ll2Authenticator2, symmetricKey2, hashBarrier)
        <4>1. AuthenticatorSetsMatch(
                  ll1Authenticators, ll2Authenticators, symmetricKey2, hashBarrier)
          BY <1>2
        <4>2. QED
         BY <4>1 DEF AuthenticatorSetsMatch
      (*********************************************************************)
      (* Therefore, there exists a matching `^\basicname^' authenticator   *)
      (* for the given `^\fastname^' authenticator.                        *)
      (*********************************************************************)
      <3>3. QED
        BY <3>1, <3>2
    (***********************************************************************)
    (* Because such an element exists, we can pick it.                     *)
    (***********************************************************************)
    <2>2. QED
      BY <2>1
  (*************************************************************************)
  (* We prove that the given `^\basicname^' authenticator equals the newly *)
  (* picked `^\basicname^' authenticator.                                  *)
  (*************************************************************************)
  <1>4. ll1Authenticator1 = ll1Authenticator2
    (***********************************************************************)
    (* The given `^\basicname^' authenticator matches the given            *)
    (* `^\fastname^' authenticator, by hypothesis of the lemma.            *)
    (***********************************************************************)
    <2>1. AuthenticatorsMatch(
              ll1Authenticator1, ll2Authenticator1, symmetricKey1, hashBarrier)
      BY <1>2
    (***********************************************************************)
    (* The newly picked `^\basicname^' authenticator matches the given     *)
    (* `^\fastname^' authenticator, by property of the pick.               *)
    (***********************************************************************)
    <2>2. AuthenticatorsMatch(
              ll1Authenticator2, ll2Authenticator1, symmetricKey2, hashBarrier)
      BY <1>3
    (***********************************************************************)
    (* Since both `^\basicname^' authenticators match the given            *)
    (* `^\fastname^' authenticator, the AuthenticatorsMatchUniqueLemma     *)
    (* tells us that they must be equal.                                   *)
    (***********************************************************************)
    <2>3. QED
      (*********************************************************************)
      (* We first have to prove some types for the                         *)
      (* AuthenticatorsMatchUniqueLemma.                                   *)
      (*********************************************************************)
      <3>1. ll1Authenticator1 \in MACType
        BY <1>1
      <3>2. ll1Authenticator2 \in MACType
        <4>1. ll1Authenticator2 \in ll1Authenticators
          BY <1>3
        <4>2. ll1Authenticators \in SUBSET MACType
          BY <1>1
        <4>3. QED
          BY <4>1, <4>2
      <3>3. ll2Authenticator1 \in MACType
        BY <1>1
      <3>4. symmetricKey1 \in SymmetricKeyType
        BY <1>1
      <3>5. symmetricKey2 \in SymmetricKeyType
        BY <1>1
      <3>6. hashBarrier \in HashType
        BY <1>1
      (*********************************************************************)
      (* Then we can apply the AuthenticatorsMatchUniqueLemma directly.    *)
      (*********************************************************************)
      <3>7. QED
        BY <2>1, <2>2, <3>1, <3>2, <3>3, <3>4, <3>5, <3>6,
            AuthenticatorsMatchUniqueLemma
    (***********************************************************************)
    (* The newly picked `^\basicname^' authenticator is an element of the  *)
    (* given `^\fastname^' authenticator set, by property of the pick.     *)
    (***********************************************************************)
  <1>5. ll1Authenticator2 \in ll1Authenticators
    BY <1>3
  (*************************************************************************)
  (* Since the two `^\basicname^' authenticators are equal, and the newly  *)
  (* picked authenticator is an element of the given `^\fastname^'         *)
  (* authenticator set, it follows that the given authenticator is an      *)
  (* element of the given `^\fastname^' authenticator set.                 *)
  (*************************************************************************)
  <1>6. QED
    BY <1>4, <1>5

----

(***************************************************************************)
(* The AuthenticatorGeneratedLemma states that if (1) two inputs summaries *)
(* match across the two specs, (2) two authenticators match across the two *)
(* specs, and (3) in the `^\fastname^' spec, the authenticator is          *)
(* generated as a MAC for the history state binding formed from the        *)
(* history summary and any state hash, then in the `^\basicname^' spec,    *)
(* the authenticator is generated as a MAC for the history state binding   *)
(* formed from the history summary and the same state hash.                *)
(***************************************************************************)

THEOREM AuthenticatorGeneratedLemma == \*`^\label{THEOREM-AuthenticatorGeneratedLemma}^'
    \A  stateHash \in HashType,
        ll1HistorySummary \in HashType,
        ll2HistorySummary \in HistorySummaryType,
        ll1Authenticator \in MACType,
        ll2Authenticator \in MACType,
        symmetricKey1 \in SymmetricKeyType,
        symmetricKey2 \in SymmetricKeyType,
        hashBarrier \in HashType :
        LET
            ll1HistoryStateBinding == Hash(ll1HistorySummary, stateHash)
            ll2HistorySummaryHash == Hash(ll2HistorySummary.anchor, ll2HistorySummary.extension)
            ll2HistoryStateBinding == Hash(ll2HistorySummaryHash, stateHash)
        IN
          ( /\  HistorySummariesMatch(ll1HistorySummary, ll2HistorySummary, hashBarrier)
            /\  AuthenticatorsMatch(
                    ll1Authenticator, ll2Authenticator, symmetricKey1, hashBarrier)
            /\  ll2Authenticator = GenerateMAC(symmetricKey2, ll2HistoryStateBinding) )
        =>
            ll1Authenticator = GenerateMAC(symmetricKey2, ll1HistoryStateBinding)
  (*************************************************************************)
  (* To prove the universally quantified expression, we take a set of      *)
  (* variables of the appropriate types.                                   *)
  (*************************************************************************)
  <1>1. TAKE stateHash1 \in HashType,
             ll1HistorySummary1 \in HashType,
             ll2HistorySummary1 \in HistorySummaryType,
             ll1Authenticator \in MACType,
             ll2Authenticator \in MACType,
             symmetricKey1 \in SymmetricKeyType,
             symmetricKey2 \in SymmetricKeyType,
             hashBarrier \in HashType
  (*************************************************************************)
  (* We re-state the definitions from the LET in the lemma.                *)
  (*************************************************************************)
  <1> ll1HistoryStateBinding1 == Hash(ll1HistorySummary1, stateHash1)
  <1> ll2HistorySummaryHash1 == Hash(ll2HistorySummary1.anchor, ll2HistorySummary1.extension)
  <1> ll2HistoryStateBinding1 == Hash(ll2HistorySummaryHash1, stateHash1)
  (*************************************************************************)
  (* We prove the types of the definitions, with help from the             *)
  (* AuthenticatorsMatchDefsTypeSafeLemma.                                 *)
  (*************************************************************************)
  <1>2. /\  ll1HistoryStateBinding1 \in HashType
        /\  ll2HistorySummaryHash1 \in HashType
        /\  ll2HistoryStateBinding1 \in HashType
    BY <1>1, AuthenticatorsMatchDefsTypeSafeLemma
  (*************************************************************************)
  (* We hide the definitions.                                              *)
  (*************************************************************************)
  <1> HIDE DEF ll1HistoryStateBinding1, ll2HistorySummaryHash1, ll2HistoryStateBinding1
  (*************************************************************************)
  (* To prove the implication, it suffices to assume the antecedent and    *)
  (* prove the consequent.                                                 *)
  (*************************************************************************)
  <1>3. SUFFICES
        ASSUME
            /\  HistorySummariesMatch(ll1HistorySummary1, ll2HistorySummary1, hashBarrier)
            /\  AuthenticatorsMatch(
                    ll1Authenticator, ll2Authenticator, symmetricKey1, hashBarrier)
            /\  ll2Authenticator = GenerateMAC(symmetricKey2, ll2HistoryStateBinding1)
        PROVE
            ll1Authenticator = GenerateMAC(symmetricKey2, ll1HistoryStateBinding1)
    BY DEF ll1HistoryStateBinding1, ll2HistorySummaryHash1, ll2HistoryStateBinding1
  (*************************************************************************)
  (* We pick a set of variables that satisfy the extentially quantified    *)
  (* variables in AuthenticatorsMatch.                                     *)
  (*************************************************************************)
  <1>4. PICK stateHash2 \in HashType,
             ll1HistorySummary2 \in HashType,
             ll2HistorySummary2 \in HistorySummaryType :
             AuthenticatorsMatch(
                 ll1Authenticator, ll2Authenticator, symmetricKey1, hashBarrier)!(
                 stateHash2, ll1HistorySummary2, ll2HistorySummary2)!1
    (***********************************************************************)
    (* We prove that such a set of variables exists.  This follows because *)
    (* AuthenticatorsMatch is assumed by the lemma.                        *)
    (***********************************************************************)
    <2>1. \E stateHash2 \in HashType,
              ll1HistorySummary2 \in HashType,
              ll2HistorySummary2 \in HistorySummaryType :
              AuthenticatorsMatch(
                  ll1Authenticator, ll2Authenticator, symmetricKey1, hashBarrier)!(
                  stateHash2, ll1HistorySummary2, ll2HistorySummary2)!1
      <3>1. AuthenticatorsMatch(
                ll1Authenticator, ll2Authenticator, symmetricKey1, hashBarrier)
        BY <1>3
      <3>2. QED
        BY <3>1 DEF AuthenticatorsMatch
    <2>2. QED
      BY <2>1
  (*************************************************************************)
  (* We re-state the definitions from the LET in AuthenticatorsMatch.      *)
  (*************************************************************************)
  <1> ll1HistoryStateBinding2 == Hash(ll1HistorySummary2, stateHash2)
  <1> ll2HistorySummaryHash2 == Hash(ll2HistorySummary2.anchor, ll2HistorySummary2.extension)
  <1> ll2HistoryStateBinding2 == Hash(ll2HistorySummaryHash2, stateHash2)
  (*************************************************************************)
  (* We prove the types of the definitions, with help from the             *)
  (* AuthenticatorsMatchDefsTypeSafeLemma.                                 *)
  (*************************************************************************)
  <1>5. /\  ll1HistoryStateBinding2 \in HashType
        /\  ll2HistorySummaryHash2 \in HashType
        /\  ll2HistoryStateBinding2 \in HashType
    BY <1>4, AuthenticatorsMatchDefsTypeSafeLemma
  (*************************************************************************)
  (* We hide the definitions.                                              *)
  (*************************************************************************)
  <1> HIDE DEF ll1HistoryStateBinding2, ll2HistorySummaryHash2, ll2HistoryStateBinding2
  (*************************************************************************)
  (* The biggest outer step of the proof is showing that in the            *)
  (* `^\fastname^' spec, the given values of history summary and state     *)
  (* hash equal the newly picked values of history summary and state hash. *)
  (*************************************************************************)
  <1>6. /\  ll2HistorySummary1 = ll2HistorySummary2
        /\  stateHash1 = stateHash2
    (***********************************************************************)
    (* In the `^\fastname^' spec, the given values of history summary hash *)
    (* and state hash equal the newly picked values of history summary     *)
    (* hash and state hash.                                                *)
    (***********************************************************************)
    <2>1. /\  ll2HistorySummaryHash1 = ll2HistorySummaryHash2
          /\  stateHash1 = stateHash2
      (*********************************************************************)
      (* In the `^\fastname^' spec, the given history state binding equals *)
      (* the newly picked history state binding.  This follows from the    *)
      (* collision resistance of the MAC.                                  *)
      (*********************************************************************)
      <3>1. ll2HistoryStateBinding1 = ll2HistoryStateBinding2
        <4>1. ll2Authenticator = GenerateMAC(symmetricKey2, ll2HistoryStateBinding1)
          BY <1>3
        <4>2. ValidateMAC(symmetricKey1, ll2HistoryStateBinding2, ll2Authenticator)
          BY <1>4 DEF ll2HistoryStateBinding2, ll2HistorySummaryHash2
        <4>3. QED
          <5>1. symmetricKey1 \in SymmetricKeyType
            BY <1>1
          <5>2. symmetricKey2 \in SymmetricKeyType
            BY <1>1
          <5>3. ll2HistoryStateBinding1 \in HashType
            BY <1>2
          <5>4. ll2HistoryStateBinding2 \in HashType
            BY <1>5
          <5>5. QED
            BY <4>1, <4>2, <5>1, <5>2, <5>3, <5>4, MACCollisionResistant
      (*********************************************************************)
      (* The history summary hashes and state hashes are each respectively *)
      (* equal, because the hash is collision resistant.                   *)
      (*********************************************************************)
      <3>2. QED
        <4>1. ll2HistorySummaryHash1 \in HashDomain
          <5>1. ll2HistorySummaryHash1 \in HashType
            BY <1>2
          <5>2. QED
            BY <5>1 DEF HashDomain
        <4>2. ll2HistorySummaryHash2 \in HashDomain
          <5>1. ll2HistorySummaryHash2 \in HashType
            BY <1>5
          <5>2. QED
            BY <5>1 DEF HashDomain
        <4>3. stateHash1 \in HashDomain
          <5>1. stateHash1 \in HashType
            BY <1>2
          <5>2. QED
            BY <5>1 DEF HashDomain
        <4>4. stateHash2 \in HashDomain
          <5>1. stateHash2 \in HashType
            BY <1>5
          <5>2. QED
            BY <5>1 DEF HashDomain
        <4>5. QED
          BY <3>1, <4>1, <4>2, <4>3, <4>4, HashCollisionResistant
          DEF ll2HistoryStateBinding1, ll2HistoryStateBinding2
    (***********************************************************************)
    (* In the `^\fastname^' spec, the given value of history summary       *)
    (* equals the newly picked value of history summary.                   *)
    (***********************************************************************)
    <2>2. ll2HistorySummary1 = ll2HistorySummary2
      (*********************************************************************)
      (* The corresponding fields of the two history summaries (the given  *)
      (* one and the newly picked one) are equal, because the hash is      *)
      (* collision resistant.                                              *)
      (*********************************************************************)
      <3>1. /\ ll2HistorySummary1.anchor = ll2HistorySummary2.anchor
            /\ ll2HistorySummary1.extension = ll2HistorySummary2.extension
        <4>1. ll2HistorySummaryHash1 = ll2HistorySummaryHash2
          BY <2>1
        <4>2. ll2HistorySummary1.anchor \in HashDomain
          <5>1. ll2HistorySummary1.anchor \in HashType
            BY <1>1 DEF HistorySummaryType
          <5>2. QED
            BY <5>1 DEF HashDomain
        <4>3. ll2HistorySummary1.extension \in HashDomain
          <5>1. ll2HistorySummary1.extension \in HashType
            BY <1>1 DEF HistorySummaryType
          <5>2. QED
            BY <5>1 DEF HashDomain
        <4>4. ll2HistorySummary2.anchor \in HashDomain
          <5>1. ll2HistorySummary2.anchor \in HashType
            BY <1>4 DEF HistorySummaryType
          <5>2. QED
            BY <5>1 DEF HashDomain
        <4>5. ll2HistorySummary2.extension \in HashDomain
          <5>1. ll2HistorySummary2.extension \in HashType
            BY <1>4 DEF HistorySummaryType
          <5>2. QED
            BY <5>1 DEF HashDomain
        <4>6. QED
          (*****************************************************************)
          (* Ideally, this QED step should just read:                      *)
          (*                                                               *)
          (*  BY <<4>>1, <<4>>2, <<4>>3, <<4>>4, <<4>>5, HashCollisionResistant *)
          (*  DEF ll2HistorySummaryHash1, ll2HistorySummaryHash2           *)
          (*                                                               *)
          (* However, the prover seems to get a little confused in this    *)
          (* instance.  We make life easier for the prover by definining   *)
          (* some local variables and hiding their definitions before      *)
          (* appealing to the HashCollisionResistant assumption.           *)
          (*****************************************************************)
          <5> h1a == ll2HistorySummary1.anchor
          <5> h2a == ll2HistorySummary1.extension
          <5> h1b == ll2HistorySummary2.anchor
          <5> h2b == ll2HistorySummary2.extension
          <5>1. Hash(h1a, h2a) = Hash(h1b, h2b)
            BY <4>1 DEF ll2HistorySummaryHash1, ll2HistorySummaryHash2
          <5>2. h1a \in HashDomain
            BY <4>2
          <5>3. h2a \in HashDomain
            BY <4>3
          <5>4. h1b \in HashDomain
            BY <4>4
          <5>5. h2b \in HashDomain
            BY <4>5
          <5>6. h1a = h1b /\ h2a = h2b
            <6> HIDE DEF h1a, h2a, h1b, h2b
            <6>1. QED
              BY <5>1, <5>2, <5>3, <5>4, <5>5, HashCollisionResistant
          <5>7. QED
            BY <5>6
      (*********************************************************************)
      (* Because the fields are equal, the records are equal, but proving  *)
      (* this requires that we prove the types of the records and invoke   *)
      (* the HistorySummaryRecordCompositionLemma.                         *)
      (*********************************************************************)
      <3>2. ll2HistorySummary1 \in HistorySummaryType
        BY <1>1
      <3>3. ll2HistorySummary2 \in HistorySummaryType
        BY <1>4
      <3>4. QED
        BY <3>1, <3>2, <3>3, HistorySummaryRecordCompositionLemma DEF HistorySummaryType
    (***********************************************************************)
    (* For clarity, we restate one of the conjuncts above.                 *)
    (***********************************************************************)
    <2>3. stateHash1 = stateHash2
      BY <2>1
    <2>4. QED
      BY <2>2, <2>3
  (*************************************************************************)
  (* We then prove that in the `^\basicname^' spec, the given value of     *)
  (* history summary equals the newly picked value of history summary.     *)
  (* This follows from the HistorySummariesMatchUniqueLemma.               *)
  (*************************************************************************)
  <1>7. ll1HistorySummary1 = ll1HistorySummary2
    <2>1. HistorySummariesMatch(ll1HistorySummary1, ll2HistorySummary1, hashBarrier)
      BY <1>3
    <2>2. HistorySummariesMatch(ll1HistorySummary2, ll2HistorySummary2, hashBarrier)
      BY <1>4
    <2>3. ll1HistorySummary1 \in HashType
      BY <1>1
    <2>4. ll1HistorySummary2 \in HashType
      BY <1>4
    <2>5. ll2HistorySummary1 \in HistorySummaryType
      BY <1>1
    <2>6. hashBarrier \in HashType
      BY <1>1
    <2>7. QED
      BY <1>6, <2>1, <2>2, <2>3, <2>4, <2>5, <2>6,
      HistorySummariesMatchUniqueLemma
  (*************************************************************************)
  (* In the final step, we show in the `^\basicname^' spec that (1) the    *)
  (* given value of the history state binding equals the newly picked      *)
  (* value of the history state binding, (2) the two symmetric keys are    *)
  (* equal, and (3) the given authenticator is generated as a MAC of the   *)
  (* newly picked history state binding.  This directly implies that the   *)
  (* given authenticator is generated as a MAC of the given history state  *)
  (* binding, which is the goal of the lemma.                              *)
  (*************************************************************************)
  <1>8. QED
    <2>1. ll1HistoryStateBinding1 = ll1HistoryStateBinding2
      <3>1. ll1HistorySummary1 = ll1HistorySummary2
        BY <1>7
      <3>2. stateHash1 = stateHash2
        BY <1>6
      <3>3. QED
        BY <3>1, <3>2 DEF ll1HistoryStateBinding1, ll1HistoryStateBinding2
    <2>2. symmetricKey1 = symmetricKey2
      <3>1. ll2Authenticator = GenerateMAC(symmetricKey2, ll2HistoryStateBinding1)
        BY <1>3
      <3>2. ValidateMAC(symmetricKey1, ll2HistoryStateBinding2, ll2Authenticator)
        BY <1>4 DEF ll2HistoryStateBinding2, ll2HistorySummaryHash2
      <3>3. QED
        <4>1. symmetricKey1 \in SymmetricKeyType
          BY <1>1
        <4>2. symmetricKey2 \in SymmetricKeyType
          BY <1>1
        <4>3. ll2HistoryStateBinding1 \in HashType
          BY <1>2
        <4>4. ll2HistoryStateBinding2 \in HashType
          BY <1>5
        <4>5. QED
          BY <3>1, <3>2, <4>1, <4>2, <4>3, <4>4, MACUnforgeable
    <2>3. ll1Authenticator = GenerateMAC(symmetricKey1, ll1HistoryStateBinding2)
      BY <1>4, <1>6 DEF ll1HistoryStateBinding2
    <2>4. QED
      BY <2>1, <2>2, <2>3 DEF ll1HistoryStateBinding2

----

(***************************************************************************)
(* The AuthenticatorValidatedLemma states that if (1) two inputs summaries *)
(* match across the two specs, (2) two authenticators match across the two *)
(* specs, and (3) in the `^\fastname^' spec, the authenticator is a valid  *)
(* MAC for the history state binding formed from the history summary and   *)
(* any state hash, then in the `^\basicname^' spec, the authenticator is a *)
(* valid MAC for the history state binding formed from the history summary *)
(* and the same state hash.                                                *)
(***************************************************************************)

THEOREM AuthenticatorValidatedLemma == \*`^\label{THEOREM-AuthenticatorValidatedLemma}^'
    \A  stateHash \in HashType,
        ll1HistorySummary \in HashType,
        ll2HistorySummary \in HistorySummaryType,
        ll1Authenticator \in MACType,
        ll2Authenticator \in MACType,
        symmetricKey1 \in SymmetricKeyType,
        symmetricKey2 \in SymmetricKeyType,
        hashBarrier \in HashType :
        LET
            ll1HistoryStateBinding == Hash(ll1HistorySummary, stateHash)
            ll2HistorySummaryHash == Hash(ll2HistorySummary.anchor, ll2HistorySummary.extension)
            ll2HistoryStateBinding == Hash(ll2HistorySummaryHash, stateHash)
        IN
          ( /\  HistorySummariesMatch(ll1HistorySummary, ll2HistorySummary, hashBarrier)
            /\  AuthenticatorsMatch(
                    ll1Authenticator, ll2Authenticator, symmetricKey1, hashBarrier)
            /\  ValidateMAC(symmetricKey2, ll2HistoryStateBinding, ll2Authenticator) )
        =>
            ValidateMAC(symmetricKey2, ll1HistoryStateBinding, ll1Authenticator)
  (*************************************************************************)
  (* To prove the universally quantified expression, we take a set of      *)
  (* variables of the appropriate types.                                   *)
  (*************************************************************************)
  <1>1. TAKE stateHash \in HashType,
             ll1HistorySummary \in HashType,
             ll2HistorySummary \in HistorySummaryType,
             ll1Authenticator \in MACType,
             ll2Authenticator \in MACType,
             symmetricKey1 \in SymmetricKeyType,
             symmetricKey2 \in SymmetricKeyType,
             hashBarrier \in HashType
  (*************************************************************************)
  (* We re-state the definitions from the LET in the lemma.                *)
  (*************************************************************************)
  <1> ll1HistoryStateBinding == Hash(ll1HistorySummary, stateHash)
  <1> ll2HistorySummaryHash == Hash(ll2HistorySummary.anchor, ll2HistorySummary.extension)
  <1> ll2HistoryStateBinding == Hash(ll2HistorySummaryHash, stateHash)
  (*************************************************************************)
  (* We prove the types of the definitions, with help from the             *)
  (* AuthenticatorsMatchDefsTypeSafeLemma.                                 *)
  (*************************************************************************)
  <1>2. /\  ll1HistoryStateBinding \in HashType
        /\  ll2HistorySummaryHash \in HashType
        /\  ll2HistoryStateBinding \in HashType
    BY <1>1, AuthenticatorsMatchDefsTypeSafeLemma
  (*************************************************************************)
  (* We hide the definitions.                                              *)
  (*************************************************************************)
  <1> HIDE DEF ll1HistoryStateBinding, ll2HistorySummaryHash, ll2HistoryStateBinding
  (*************************************************************************)
  (* To prove the implication, it suffices to assume the antecedent and    *)
  (* prove the consequent.                                                 *)
  (*************************************************************************)
  <1>3. SUFFICES
        ASSUME
            /\  HistorySummariesMatch(ll1HistorySummary, ll2HistorySummary, hashBarrier)
            /\  AuthenticatorsMatch(
                    ll1Authenticator, ll2Authenticator, symmetricKey1, hashBarrier)
            /\  ValidateMAC(symmetricKey2, ll2HistoryStateBinding, ll2Authenticator)
        PROVE
            ValidateMAC(symmetricKey2, ll1HistoryStateBinding, ll1Authenticator)
    BY DEF ll1HistoryStateBinding, ll2HistorySummaryHash, ll2HistoryStateBinding
  (*************************************************************************)
  (* Using the MACConsistent property, we show that, since the             *)
  (* `^\fastname^' authenticator is a valid MAC of the `^\fastname^'       *)
  (* history state binding, it must have been generated as a MAC of this   *)
  (* history state binding.                                                *)
  (*************************************************************************)
  <1>4. ll2Authenticator = GenerateMAC(symmetricKey2, ll2HistoryStateBinding)
    <2>1. ValidateMAC(symmetricKey2, ll2HistoryStateBinding, ll2Authenticator)
      BY <1>3
    <2>2. symmetricKey1 \in SymmetricKeyType
      BY <1>1
    <2>3. symmetricKey2 \in SymmetricKeyType
      BY <1>1
    <2>4. ll2HistoryStateBinding \in HashType
      BY <1>2
    <2>5. ll2Authenticator \in MACType
      BY <1>1
    <2>6. QED
      BY <2>1, <2>2, <2>3, <2>4, <2>5, MACConsistent
  (*************************************************************************)
  (* Then, using the AuthenticatorGeneratedLemma, we show that the         *)
  (* `^\basicname^' authenticator is generated as a MAC of the             *)
  (* `^\basicname^' history state binding.                                 *)
  (*************************************************************************)
  <1>5. ll1Authenticator = GenerateMAC(symmetricKey2, ll1HistoryStateBinding)
    <2>1. HistorySummariesMatch(ll1HistorySummary, ll2HistorySummary, hashBarrier)
      BY <1>3
    <2>2. AuthenticatorsMatch(
              ll1Authenticator, ll2Authenticator, symmetricKey1, hashBarrier)
      BY <1>3
    <2>3. QED
      (*********************************************************************)
      (* Ideally, this QED step should just read:                          *)
      (*                                                                   *)
      (*  BY <<1>>1, <<1>>4, <<2>>1, <<2>>2, AuthenticatorGeneratedLemma   *)
      (*  DEF ll1HistoryStateBinding, ll2HistorySummaryHash, ll2HistoryStateBinding *)
      (*                                                                   *)
      (* However, the prover seems to get a little confused in this        *)
      (* instance.  We make life easier for the prover by explicitly       *)
      (* staging the instantiation of the quantified variables within the  *)
      (* definition of AuthenticatorGeneratedLemma.                        *)
      (*********************************************************************)
      <3>1. \A  samLL2Authenticator \in MACType,
                samSymmetricKey2 \in SymmetricKeyType,
                samHashBarrier \in HashType :
                AuthenticatorGeneratedLemma!(stateHash, ll1HistorySummary,
                    ll2HistorySummary, ll1Authenticator, samLL2Authenticator,
                    symmetricKey1, samSymmetricKey2, samHashBarrier)!1
        BY <1>1, AuthenticatorGeneratedLemma
      <3>2. AuthenticatorGeneratedLemma!(stateHash, ll1HistorySummary,
                ll2HistorySummary, ll1Authenticator, ll2Authenticator,
                symmetricKey1, symmetricKey2, hashBarrier)!1
        BY <1>1, <3>1
      <3>3. QED
        BY <1>1, <1>4, <2>1, <2>2, <3>2
        DEF ll1HistoryStateBinding, ll2HistorySummaryHash, ll2HistoryStateBinding
  (*************************************************************************)
  (* Finally, using the MACComplete property, we show that the             *)
  (* `^\basicname^' authenticator is a valid MAC of the `^\basicname^'     *)
  (* history state binding.                                                *)
  (*************************************************************************)
  <1>6. QED
    <2>1. symmetricKey2 \in SymmetricKeyType
      BY <1>1
    <2>2. ll1HistoryStateBinding \in HashType
      BY <1>2
    <2>3. QED
      BY <1>5, <2>1, <2>2, MACComplete DEF ll1HistoryStateBinding

----

(***************************************************************************)
(* The HistorySummariesMatchAcrossCheckpointLemma asserts that for any     *)
(* pair of `^\fastname^' history summaries for which one is a checkpoint   *)
(* of the other, there is a unique `^\basicname^' history summary that     *)
(* matches either `^\fastname^' history summary.                           *)
(***************************************************************************)

THEOREM HistorySummariesMatchAcrossCheckpointLemma == \*`^\label{THEOREM-HistorySummariesMatchAcrossCheckpointLemma}^'
    \A  ll1HistorySummary1, ll1HistorySummary2 \in HashType,
        ll2HistorySummary1, ll2HistorySummary2 \in HistorySummaryType,
        hashBarrier \in HashType :
          ( /\  HistorySummariesMatch(ll1HistorySummary1, ll2HistorySummary1, hashBarrier)
            /\  HistorySummariesMatch(ll1HistorySummary2, ll2HistorySummary2, hashBarrier)
            /\  ll2HistorySummary2 = Checkpoint(ll2HistorySummary1) )
        =>
            ll1HistorySummary1 = ll1HistorySummary2
  (*************************************************************************)
  (* To prove the universally quantified expression, we take a set of      *)
  (* variables of the appropriate types.                                   *)
  (*************************************************************************)
  <1>1. TAKE ll1HistorySummary1, ll1HistorySummary2 \in HashType,
             ll2HistorySummary1, ll2HistorySummary2 \in HistorySummaryType,
             hashBarrier \in HashType
  (*************************************************************************)
  (* We assume the antecedent.                                             *)
  (*************************************************************************)
  <1>2. HAVE /\  HistorySummariesMatch(ll1HistorySummary1, ll2HistorySummary1, hashBarrier)
             /\  HistorySummariesMatch(ll1HistorySummary2, ll2HistorySummary2, hashBarrier)
             /\  ll2HistorySummary2 = Checkpoint(ll2HistorySummary1)
  (*************************************************************************)
  (* There are two separate cases, one for the base case of the            *)
  (* HistorySummariesMatch predicate and one for the recursion.  The base  *)
  (* case is trivial.                                                      *)
  (*************************************************************************)
  <1> ll2InitialHistorySummary == [ anchor |-> BaseHashValue, extension |-> BaseHashValue ]
  <1>3. CASE ll2HistorySummary1 = ll2InitialHistorySummary
    <2>1. ll2HistorySummary2 = Checkpoint(ll2HistorySummary1)
      BY <1>2
    <2>2. ll2HistorySummary1.extension = BaseHashValue
      BY <1>3
    <2>3. ll2HistorySummary1 = ll2HistorySummary2
      BY <2>1, <2>2 DEF Checkpoint
    <2>4. QED
      BY <1>1, <1>2, <2>3, HistorySummariesMatchUniqueLemma
  (*************************************************************************)
  (* The recursive case is fairly involved.                                *)
  (*************************************************************************)
  <1>4. CASE ll2HistorySummary1 # ll2InitialHistorySummary
    (***********************************************************************)
    (* We'll first pick values for the existential variables inside the    *)
    (* HistorySummariesMatchRecursion predicate that satisfy the predicate *)
    (* for the `1' variables in the antecedent of the theorem.  We know    *)
    (* the HistorySummariesMatchRecursion predicate holds, because the     *)
    (* `^\fastname^' history summary does not equal the initial history    *)
    (* summary, according to this case.                                    *)
    (***********************************************************************)
    <2>1. PICK input1 \in InputType,
               ll1PreviousHistorySummary1 \in HashType,
               ll2PreviousHistorySummary1 \in HistorySummaryType :
               HistorySummariesMatchRecursion(
                   ll1HistorySummary1, ll2HistorySummary1, hashBarrier)!(
                   input1, ll1PreviousHistorySummary1, ll2PreviousHistorySummary1)
      <3>1. /\  ll1HistorySummary1 \in HashType
            /\  ll2HistorySummary1 \in HistorySummaryType
            /\  hashBarrier \in HashType
        BY <1>1
      <3>2. HistorySummariesMatch(ll1HistorySummary1, ll2HistorySummary1, hashBarrier)
        BY <1>2
      <3>3. ll2HistorySummary1 # ll2InitialHistorySummary
        BY <1>4
      <3>4. QED
        BY <3>1, <3>2, <3>3, HistorySummariesMatchDefinition
        DEF HistorySummariesMatchRecursion
    (***********************************************************************)
    (* We'll next pick values for the existential variables inside the     *)
    (* HistorySummariesMatchRecursion predicate that satisfy the predicate *)
    (* for the `2' variables in the antecedent of the theorem.  We know    *)
    (* the HistorySummariesMatchRecursion predicate holds, because the     *)
    (* `^\fastname^' history summary does not equal the initial history    *)
    (* summary, but proving this latter point is slightly involved.        *)
    (***********************************************************************)
    <2>2. PICK input2 \in InputType,
               ll1PreviousHistorySummary2 \in HashType,
               ll2PreviousHistorySummary2 \in HistorySummaryType :
               HistorySummariesMatchRecursion(
                   ll1HistorySummary2, ll2HistorySummary2, hashBarrier)!(
                   input2, ll1PreviousHistorySummary2, ll2PreviousHistorySummary2)
      <3>1. /\  ll1HistorySummary2 \in HashType
            /\  ll2HistorySummary2 \in HistorySummaryType
            /\  hashBarrier \in HashType
        BY <1>1
      <3>2. HistorySummariesMatch(ll1HistorySummary2, ll2HistorySummary2, hashBarrier)
        BY <1>2
      <3>3. ll2HistorySummary2 # ll2InitialHistorySummary
        (*******************************************************************)
        (* Inputs summary 2 does not equal the initial history summary     *)
        (* because its anchor field does not equal the base hash value.    *)
        (*******************************************************************)
        <4>1. ll2HistorySummary2.anchor # BaseHashValue
      (*********************************************************************)
      (* We use two more levels of case analysis.  For the first level, we *)
      (* show that at least one of the two fields in history summary 1 is  *)
      (* not equal to the base hash value.                                 *)
      (*********************************************************************)
          <5>1. \/  ll2HistorySummary1.anchor # BaseHashValue
                \/  ll2HistorySummary1.extension # BaseHashValue
            <6>1. ll2HistorySummary1 \in HistorySummaryType
              BY <3>1
            <6>2. QED
              BY <1>4, <6>1, HistorySummaryRecordCompositionLemma DEF HistorySummaryType
        (*******************************************************************)
        (* We consider the case in which the extension field of history    *)
        (* summary 1 is not equal to the base hash value.  In this case,   *)
        (* the anchor field of history summary 2 is formed by the hash     *)
        (* function, so it cannot be equal to the base hash value.         *)
        (*******************************************************************)
          <5>2. CASE ll2HistorySummary1.extension # BaseHashValue
            <6>1. ll2HistorySummary2.anchor =
                      Hash(ll2HistorySummary1.anchor, ll2HistorySummary1.extension)
              <7>1. ll2HistorySummary2 = Checkpoint(ll2HistorySummary1)
                BY <1>2
              <7>2. QED
                BY <5>2, <7>1 DEF Checkpoint
            <6>2. ll2HistorySummary1.anchor \in HashDomain
              BY <1>1 DEF HistorySummaryType, HashDomain
            <6>3. ll2HistorySummary1.extension \in HashDomain
              BY <1>1 DEF HistorySummaryType, HashDomain
            <6>4. QED
              BY <6>1, <6>2, <6>3, BaseHashValueUnique
        (*******************************************************************)
        (* We consider the case in which the anchor field of history       *)
        (* summary 1 is not equal to the base hash value.  We consider two *)
        (* sub-cases.                                                      *)
        (*******************************************************************)
          <5>3. CASE ll2HistorySummary1.anchor # BaseHashValue
            (***************************************************************)
            (* In the first sub-case, the extension field of history       *)
            (* summary 1 is also not equal to the base hash value.  We     *)
            (* already proved this case above.                             *)
            (***************************************************************)
            <6>1. CASE ll2HistorySummary1.extension # BaseHashValue
              BY <5>2, <6>1
            (***************************************************************)
            (* In the second sub-case, the extension field of history      *)
            (* summary 1 equals the base hash value, so the Checkpoint     *)
            (* operator acts as an identity operator.  Thus, the anchor    *)
            (* field of history summary 2 equals the anchor field of       *)
            (* history summmary 1, which is not equal to the base hash     *)
            (* value.                                                      *)
            (***************************************************************)
            <6>2. CASE ll2HistorySummary1.extension = BaseHashValue
              <7>1. ll2HistorySummary2.anchor = ll2HistorySummary1.anchor
                <8>1. ll2HistorySummary2 = Checkpoint(ll2HistorySummary1)
                  BY <1>2
                <8>2. QED
                  BY <6>2, <8>1 DEF Checkpoint
              <7>2. QED
               BY <5>3, <7>1
            <6>3. QED
              BY <6>1, <6>2
          <5>4. QED
            BY <5>1, <5>2, <5>3 
        <4>2. ll2InitialHistorySummary.anchor = BaseHashValue
          OBVIOUS
        <4>3. QED
          BY <4>1, <4>2
      <3>4. QED
        BY <3>1, <3>2, <3>3, HistorySummariesMatchDefinition
        DEF HistorySummariesMatchRecursion
    (***********************************************************************)
    (* We prove that the `^\fastname^' previous history summaries are      *)
    (* equal, and the inputs are equal.                                    *)
    (***********************************************************************)
    <2>3. /\  ll2PreviousHistorySummary1 = ll2PreviousHistorySummary2
          /\  input1 = input2
      (*********************************************************************)
      (* Several useful facts follow directly from the theorem's           *)
      (* assumptions and the above picks.                                  *)
      (*********************************************************************)
      <3>1. LL2HistorySummaryIsSuccessor(
                ll2HistorySummary1, ll2PreviousHistorySummary1, input1, hashBarrier)
        BY <2>1
      <3>2. LL2HistorySummaryIsSuccessor(
                ll2HistorySummary2, ll2PreviousHistorySummary2, input2, hashBarrier)
        BY <2>2
      <3>3. ll2HistorySummary2 = Checkpoint(ll2HistorySummary1)
        BY <1>2
      (*********************************************************************)
      (* We re-state the definitions from LL2HistorySummaryIsSuccessor for *)
      (* history summary 1.                                                *)
      (*********************************************************************)
      <3> ll2SuccessorHistorySummary1 == Successor(ll2PreviousHistorySummary1, input1, hashBarrier)
      <3> ll2CheckpointedSuccessorHistorySummary1 == Checkpoint(ll2SuccessorHistorySummary1)
      (*********************************************************************)
      (* We hide the definitions.                                          *)
      (*********************************************************************)
      <3> HIDE DEF ll2SuccessorHistorySummary1, ll2CheckpointedSuccessorHistorySummary1
      (*********************************************************************)
      (* We re-state the definitions from Successor for history summary 1. *)
      (*********************************************************************)
      <3> securedInput1 == Hash(hashBarrier, input1)
      <3> newAnchor1 == ll2PreviousHistorySummary1.anchor
      <3> newExtension1 == Hash(ll2PreviousHistorySummary1.extension, securedInput1)
      <3> ll2NewHistorySummary1 == [
                anchor |-> newAnchor1,
                extension |-> newExtension1 ]
      (*********************************************************************)
      (* We prove the types of the definitions from Successor and the      *)
      (* definition of ll2SuccessorHistorySummary1, with help from the     *)
      (* SuccessorDefsTypeSafeLemma.                                       *)
      (*********************************************************************)
      <3>4. /\  securedInput1 \in HashType
            /\  newAnchor1 \in HashType
            /\  newExtension1 \in HashType
            /\  ll2NewHistorySummary1 \in HistorySummaryType
            /\  ll2NewHistorySummary1.anchor \in HashType
            /\  ll2NewHistorySummary1.extension \in HashType
          BY <1>1, <2>1, SuccessorDefsTypeSafeLemma
      <3>5. ll2SuccessorHistorySummary1 \in HistorySummaryType
          BY <3>4 DEF ll2SuccessorHistorySummary1, Successor
      (*********************************************************************)
      (* We hide the definitions.                                          *)
      (*********************************************************************)
      <3> HIDE DEF securedInput1, newAnchor1, newExtension1, ll2NewHistorySummary1
      (*********************************************************************)
      (* We re-state the definitions from Checkpoint for history summary   *)
      (* 1.                                                                *)
      (*********************************************************************)
      <3> checkpointedAnchor1 ==
                Hash(ll2SuccessorHistorySummary1.anchor, ll2SuccessorHistorySummary1.extension)
      <3> ll2CheckpointedHistorySummary1 == [
                anchor |-> checkpointedAnchor1,
                extension |-> BaseHashValue ]
      (*********************************************************************)
      (* We prove the types of the definitions from Checkpoint and the     *)
      (* definition of ll2CheckpointedSuccessorHistorySummary1, with help  *)
      (* from the CheckpointDefsTypeSafeLemma.                             *)
      (*********************************************************************)
      <3>6. /\  checkpointedAnchor1 \in HashType
            /\  ll2CheckpointedHistorySummary1 \in HistorySummaryType
            /\  ll2CheckpointedHistorySummary1.anchor \in HashType
            /\  ll2CheckpointedHistorySummary1.extension \in HashType
          BY <3>5, CheckpointDefsTypeSafeLemma
      <3>7. ll2CheckpointedSuccessorHistorySummary1 \in HistorySummaryType
          BY <3>5, <3>6 DEF ll2CheckpointedSuccessorHistorySummary1, Checkpoint
      (*********************************************************************)
      (* We hide the definitions.                                          *)
      (*********************************************************************)
      <3> HIDE DEF checkpointedAnchor1, ll2CheckpointedHistorySummary1
      (*********************************************************************)
      (* We re-state the definitions from LL2HistorySummaryIsSuccessor for *)
      (* history summary 2.                                                *)
      (*********************************************************************)
      <3> ll2SuccessorHistorySummary2 == Successor(ll2PreviousHistorySummary2, input2, hashBarrier)
      <3> ll2CheckpointedSuccessorHistorySummary2 == Checkpoint(ll2SuccessorHistorySummary2)
      (*********************************************************************)
      (* We hide the definitions.                                          *)
      (*********************************************************************)
      <3> HIDE DEF ll2SuccessorHistorySummary2, ll2CheckpointedSuccessorHistorySummary2
      (*********************************************************************)
      (* We re-state the definitions from Successor for history summary 2. *)
      (*********************************************************************)
      <3> securedInput2 == Hash(hashBarrier, input2)
      <3> newAnchor2 == ll2PreviousHistorySummary2.anchor
      <3> newExtension2 == Hash(ll2PreviousHistorySummary2.extension, securedInput2)
      <3> ll2NewHistorySummary2 == [
                anchor |-> newAnchor2,
                extension |-> newExtension2 ]
      (*********************************************************************)
      (* We prove the types of the definitions from Successor and the      *)
      (* definition of ll2SuccessorHistorySummary2, with help from the     *)
      (* SuccessorDefsTypeSafeLemma.                                       *)
      (*********************************************************************)
      <3>8. /\  securedInput2 \in HashType
            /\  newAnchor2 \in HashType
            /\  newExtension2 \in HashType
            /\  ll2NewHistorySummary2 \in HistorySummaryType
            /\  ll2NewHistorySummary2.anchor \in HashType
            /\  ll2NewHistorySummary2.extension \in HashType
          BY <1>1, <2>1, SuccessorDefsTypeSafeLemma
      <3>9. ll2SuccessorHistorySummary2 \in HistorySummaryType
          BY <3>8 DEF ll2SuccessorHistorySummary2, Successor
      (*********************************************************************)
      (* We hide the definitions.                                          *)
      (*********************************************************************)
      <3> HIDE DEF securedInput2, newAnchor2, newExtension2, ll2NewHistorySummary2
      (*********************************************************************)
      (* We re-state the definitions from Checkpoint for history summary   *)
      (* 2.                                                                *)
      (*********************************************************************)
      <3> checkpointedAnchor2 ==
                Hash(ll2SuccessorHistorySummary2.anchor, ll2SuccessorHistorySummary2.extension)
      <3> ll2CheckpointedHistorySummary2 == [
                anchor |-> checkpointedAnchor2,
                extension |-> BaseHashValue ]
      (*********************************************************************)
      (* We prove the types of the definitions from Checkpoint and the     *)
      (* definition of ll2CheckpointedSuccessorHistorySummary2, with help  *)
      (* from the CheckpointDefsTypeSafeLemma.                             *)
      (*********************************************************************)
      <3>10. /\  checkpointedAnchor2 \in HashType
             /\  ll2CheckpointedHistorySummary2 \in HistorySummaryType
             /\  ll2CheckpointedHistorySummary2.anchor \in HashType
             /\  ll2CheckpointedHistorySummary2.extension \in HashType
          BY <3>9, CheckpointDefsTypeSafeLemma
      <3>11. ll2CheckpointedSuccessorHistorySummary2 \in HistorySummaryType
          BY <3>9, <3>10 DEF ll2CheckpointedSuccessorHistorySummary2, Checkpoint
      (*********************************************************************)
      (* We hide the definitions.                                          *)
      (*********************************************************************)
      <3> HIDE DEF checkpointedAnchor2, ll2CheckpointedHistorySummary2
      (*********************************************************************)
      (* The most critical part of the proof is showing that the           *)
      (* checkpointed successor history summaries, as defined in the LET   *)
      (* within the LL2HistorySummaryIsSuccessor predicate, are equal.     *)
      (* This requires separately considering the cases of whether the     *)
      (* extension field of ll2HistorySummary equals the base hash value.  *)
      (*********************************************************************)
      <3>12. ll2CheckpointedSuccessorHistorySummary1 = ll2CheckpointedSuccessorHistorySummary2
        (*******************************************************************)
        (* In the first case, the extension field of ll2HistorySummary     *)
        (* equals the base hash value.                                     *)
        (*******************************************************************)
        <4>1. CASE ll2HistorySummary1.extension = BaseHashValue
          (*****************************************************************)
          (* The LL2HistorySummaryIsSuccessor predicate expresses a        *)
          (* disjunction.  An history summary can be a successor either by *)
          (* being a direct successor or by being a checkpoint of a        *)
          (* successor.  We prove that, in this case, history summary 1 is *)
          (* a checkpoint of a successor.  It cannot be a direct successor *)
          (* because its extension field is the base hash value, and any   *)
          (* direct successor will have a non-base extension field.        *)
          (*****************************************************************)
          <5>1. ll2HistorySummary1 = ll2CheckpointedSuccessorHistorySummary1
            <6>1. \/  ll2HistorySummary1 = ll2SuccessorHistorySummary1
                  \/  ll2HistorySummary1 = ll2CheckpointedSuccessorHistorySummary1
              BY <3>1
                DEF LL2HistorySummaryIsSuccessor, ll2SuccessorHistorySummary1,
                    ll2CheckpointedSuccessorHistorySummary1
            <6>2. ll2HistorySummary1 # ll2SuccessorHistorySummary1
              <7>1. ll2SuccessorHistorySummary1.extension # BaseHashValue
                BY <1>1, <2>1, SuccessorHasNonBaseExtensionLemma DEF ll2SuccessorHistorySummary1
              <7>2. QED
                BY <4>1, <7>1
            <6>3. QED
              BY <6>1, <6>2
          (*****************************************************************)
          (* Inputs summary 2 is also a checkpoint of a successor, by the  *)
          (* same argument as above for history summary 1.  The one twist  *)
          (* is that we need to prove that the its extension field is the  *)
          (* base hash value, which follows from the fact that - when      *)
          (* history summary 1 has a base extension field, history         *)
          (* summaries 1 and 2 are equal.                                  *)
          (*****************************************************************)
          <5>2. ll2HistorySummary2 = ll2CheckpointedSuccessorHistorySummary2
            <6>1. \/  ll2HistorySummary2 = ll2SuccessorHistorySummary2
                  \/  ll2HistorySummary2 = ll2CheckpointedSuccessorHistorySummary2
              BY <3>2
              DEF LL2HistorySummaryIsSuccessor, ll2SuccessorHistorySummary2,
                  ll2CheckpointedSuccessorHistorySummary2
            <6>2. ll2HistorySummary2 # ll2SuccessorHistorySummary2
              <7>1. ll2SuccessorHistorySummary2.extension # BaseHashValue
                BY <1>1, <2>2, SuccessorHasNonBaseExtensionLemma DEF ll2SuccessorHistorySummary2
              <7>2. QED
                <8>1. ll2HistorySummary2.extension = BaseHashValue
                  <9>1. ll2HistorySummary1 = ll2HistorySummary2
                    BY <3>3, <4>1 DEF Checkpoint
                  <9>2. QED
                    BY <4>1, <9>1
                <8>2. QED
                  BY <7>1, <8>1
            <6>3. QED
              BY <6>1, <6>2
          (*****************************************************************)
          (* The conclusion follows from the above-proven equalities,      *)
          (* given the fact that, when history summary 1 has a base        *)
          (* extension field, the Checkpoint operator acts as an identity  *)
          (* operator, so history summaries 1 and 2 are equal.             *)
          (*****************************************************************)
          <5>3. QED
            <6>1. ll2HistorySummary1 = ll2HistorySummary2
              BY <3>3, <4>1 DEF Checkpoint
            <6>2. QED
              BY <5>1, <5>2, <6>1
        (*******************************************************************)
        (* In the second case, the extension field of ll2HistorySummary    *)
        (* does not equal the base hash value.                             *)
        (*******************************************************************)
        <4>2. CASE ll2HistorySummary1.extension # BaseHashValue
          (*****************************************************************)
          (* The LL2HistorySummaryIsSuccessor predicate expresses a        *)
          (* disjunction.  An history summary can be a successor either by *)
          (* being a direct successor or by being a checkpoint of a        *)
          (* successor.  We prove that, in this case, history summary 1 is *)
          (* a direct successor.  It cannot be a checkpoint because its    *)
          (* extension field is not the base hash value, and any           *)
          (* checkpoint will have a base extension field.                  *)
          (*****************************************************************)
          <5>1. ll2HistorySummary1 = ll2SuccessorHistorySummary1
            <6>1. \/  ll2HistorySummary1 = ll2SuccessorHistorySummary1
                  \/  ll2HistorySummary1 = ll2CheckpointedSuccessorHistorySummary1
              BY <3>1
              DEF LL2HistorySummaryIsSuccessor, ll2SuccessorHistorySummary1,
                  ll2CheckpointedSuccessorHistorySummary1
            <6>2. ll2HistorySummary1 # ll2CheckpointedSuccessorHistorySummary1
              <7>1. ll2CheckpointedSuccessorHistorySummary1.extension = BaseHashValue
                BY <3>5, CheckpointHasBaseExtensionLemma
                  DEF ll2CheckpointedSuccessorHistorySummary1
              <7>2. QED
                BY <4>2, <7>1
            <6>3. QED
              BY <6>1, <6>2
          (*****************************************************************)
          (* Inputs summary 2 is a checkpoint of a successor, by the       *)
          (* reverse of the above argument for history summary 1.  We know *)
          (* that the extension field of history summary 2 must be the     *)
          (* base hash value, because it is a checkpoint of history        *)
          (* summary 1.                                                    *)
          (*****************************************************************)
          <5>2. ll2HistorySummary2 = ll2CheckpointedSuccessorHistorySummary2
            <6>1. \/  ll2HistorySummary2 = ll2SuccessorHistorySummary2
                  \/  ll2HistorySummary2 = ll2CheckpointedSuccessorHistorySummary2
              BY <3>2
              DEF LL2HistorySummaryIsSuccessor, ll2SuccessorHistorySummary2,
                  ll2CheckpointedSuccessorHistorySummary2
            <6>2. ll2HistorySummary2 # ll2SuccessorHistorySummary2
              <7>1. ll2SuccessorHistorySummary2.extension # BaseHashValue
                BY <1>1, <2>2, SuccessorHasNonBaseExtensionLemma DEF ll2SuccessorHistorySummary2
              <7>2. ll2HistorySummary2.extension = BaseHashValue
                BY <1>1, <1>2, CheckpointHasBaseExtensionLemma
              <7>3. QED
                BY <7>1, <7>2
            <6>3. QED
              BY <6>1, <6>2
          (*****************************************************************)
          (* The conclusion follows from the above-proven equalities,      *)
          (* given the parallel construction of checkpoint from history    *)
          (* summary in both the history summaries in the antecedent of    *)
          (* the theorem and the history summaries defined within the LET  *)
          (* of the LL2HistorySummaryIsSuccessor predicate.                *)
          (*****************************************************************)
          <5>3. QED
            <6>1. ll2CheckpointedSuccessorHistorySummary1 = Checkpoint(ll2HistorySummary1)
              BY <5>1 DEF ll2CheckpointedSuccessorHistorySummary1
            <6>2. ll2HistorySummary2 = Checkpoint(ll2HistorySummary1)
              BY <1>2
            <6>3. ll2CheckpointedSuccessorHistorySummary2 = ll2HistorySummary2
              BY <5>2
            <6>4. QED
              BY <6>1, <6>2, <6>3
        (*******************************************************************)
        (* The two cases are exhaustive.                                   *)
        (*******************************************************************)
        <4>3. QED
          BY <4>1, <4>2
      (*********************************************************************)
      (* We prove that the successor history summaries, as defined in the  *)
      (* LET of the LL2HistorySummaryIsSuccessor predicate, are equal to   *)
      (* each other.                                                       *)
      (*********************************************************************)
      <3>13. ll2SuccessorHistorySummary1 = ll2SuccessorHistorySummary2
        (*******************************************************************)
        (* The checkpointed history summaries, as defined in the LET of    *)
        (* the Checkpoint operator, are equal to each other.               *)
        (*******************************************************************)
        <4>1. ll2CheckpointedHistorySummary1 = ll2CheckpointedHistorySummary2
          (*****************************************************************)
          (* As proven above by case analysis, the checkpointed inputs     *)
          (* summaries, as defined in the LET of the                       *)
          (* LL2HistorySummaryIsSuccessor predicate, are equal.            *)
          (*****************************************************************)
          <5>1. ll2CheckpointedSuccessorHistorySummary1 = ll2CheckpointedSuccessorHistorySummary2
            BY <3>12
          (*****************************************************************)
          (* The extension field of each successor history summary is not  *)
          (* equal to the base hash value, because they are direct         *)
          (* successors.                                                   *)
          (*****************************************************************)
          <5>2. ll2SuccessorHistorySummary1.extension # BaseHashValue
            BY <1>1, <2>1, SuccessorHasNonBaseExtensionLemma DEF ll2SuccessorHistorySummary1
          <5>3. ll2SuccessorHistorySummary2.extension # BaseHashValue
            BY <1>1, <2>2, SuccessorHasNonBaseExtensionLemma DEF ll2SuccessorHistorySummary2
          (*****************************************************************)
          (* The conclusion follows from the definitions, because the      *)
          (* Checkpoint operator is injective under the preimage           *)
          (* constraint of an extension field that is unequal to the base  *)
          (* hash value.                                                   *)
          (*****************************************************************)
          <5>4. QED
            BY <5>1, <5>2, <5>3
            DEF ll2CheckpointedSuccessorHistorySummary1, ll2CheckpointedSuccessorHistorySummary2,
                  Checkpoint, ll2CheckpointedHistorySummary1, ll2CheckpointedHistorySummary2,
                  checkpointedAnchor1, checkpointedAnchor2
        (*******************************************************************)
        (* The individual fields of the successor history summaries are    *)
        (* equal to each other, thanks to the collision resistance of the  *)
        (* hash function.                                                  *)
        (*******************************************************************)
        <4>2. /\  ll2SuccessorHistorySummary1.anchor = ll2SuccessorHistorySummary2.anchor
              /\  ll2SuccessorHistorySummary1.extension = ll2SuccessorHistorySummary2.extension
          <5>1. checkpointedAnchor1 = checkpointedAnchor2
            BY <4>1 DEF ll2CheckpointedHistorySummary1, ll2CheckpointedHistorySummary2
          <5>2. ll2SuccessorHistorySummary1.anchor \in HashDomain
            BY <3>5 DEF HistorySummaryType, HashDomain
          <5>3. ll2SuccessorHistorySummary1.extension \in HashDomain
            BY <3>5 DEF HistorySummaryType, HashDomain
          <5>4. ll2SuccessorHistorySummary2.anchor \in HashDomain
            BY <3>9 DEF HistorySummaryType, HashDomain
          <5>5. ll2SuccessorHistorySummary2.extension \in HashDomain
            BY <3>9 DEF HistorySummaryType, HashDomain
          <5>6. QED
            (***************************************************************)
            (* Ideally, this QED step should just read:                    *)
            (*                                                             *)
            (*  BY <<5>>1, <<5>>2, <<5>>3, <<5>>4, <<5>>5, HashCollisionResistant *)
            (*  DEF checkpointedAnchor1, checkpointedAnchor2               *)
            (*                                                             *)
            (* However, the prover seems to get a little confused in this  *)
            (* instance.  We make life easier for the prover by definining *)
            (* some local variables and hiding their definitions before    *)
            (* appealing to the HashCollisionResistant assumption.         *)
            (***************************************************************)
            <6> h1a == ll2SuccessorHistorySummary1.anchor
            <6> h2a == ll2SuccessorHistorySummary1.extension
            <6> h1b == ll2SuccessorHistorySummary2.anchor
            <6> h2b == ll2SuccessorHistorySummary2.extension
            <6>1. h1a \in HashDomain
              BY <5>2
            <6>2. h2a \in HashDomain
              BY <5>3
            <6>3. h1b \in HashDomain
              BY <5>4
            <6>4. h2b \in HashDomain
              BY <5>5
            <6>5. Hash(h1a, h2a) = Hash(h1b, h2b)
              BY <5>1 DEF checkpointedAnchor1, checkpointedAnchor2
            <6>6. h1a = h1b /\ h2a = h2b
              <7> HIDE DEF h1a, h2a, h1b, h2b
              <7>1. QED
                BY <6>1, <6>2, <6>3, <6>4, <6>5, HashCollisionResistant
            <6>7. QED
              BY <6>6
        (*******************************************************************)
        (* Because the fields are equal, the records are equal, but        *)
        (* proving this requires that we prove the types of the records    *)
        (* and invoke the HistorySummaryRecordCompositionLemma.            *)
        (*******************************************************************)
        <4>3. ll2SuccessorHistorySummary1 \in HistorySummaryType
          BY <3>5
        <4>4. ll2SuccessorHistorySummary2 \in HistorySummaryType
          BY <3>9
        <4>5. QED
          BY <4>2, <4>3, <4>4, HistorySummaryRecordCompositionLemma DEF HistorySummaryType
      (*********************************************************************)
      (* The values that are hashed to produce the new extension value     *)
      (* within the Successor operator are equal to each other, because    *)
      (* the new extensions are equal to each other, and the hash is       *)
      (* collision-resistant.                                              *)
      (*********************************************************************)
      <3>14. /\  ll2PreviousHistorySummary1.extension = ll2PreviousHistorySummary2.extension
             /\  securedInput1 = securedInput2
        <4>1. newExtension1 = newExtension2
          <5>1. ll2NewHistorySummary1 = ll2NewHistorySummary2
            BY <3>13
            DEF ll2SuccessorHistorySummary1, ll2SuccessorHistorySummary2, Successor,
                  ll2NewHistorySummary1, ll2NewHistorySummary2, newExtension1, newExtension2,
                  newAnchor1, newAnchor2, securedInput1, securedInput2
          <5>2. QED
            BY <5>1 DEF ll2NewHistorySummary1, ll2NewHistorySummary2
        <4>2. ll2PreviousHistorySummary1.extension \in HashDomain
          BY <2>1 DEF HistorySummaryType, HashDomain
        <4>3. securedInput1 \in HashDomain
          BY <3>4 DEF HashDomain
        <4>4. ll2PreviousHistorySummary2.extension \in HashDomain
          BY <2>2 DEF HistorySummaryType, HashDomain
        <4>5. securedInput2 \in HashDomain
          BY <3>8 DEF HashDomain
        <4>6. QED
          (*****************************************************************)
          (* Ideally, this QED step should just read:                      *)
          (*                                                               *)
          (*  BY <<4>>1, <<4>>2, <<4>>3, <<4>>4, <<4>>5, HashCollisionResistant *)
          (*                                                               *)
          (* However, the prover seems to get a little confused in this    *)
          (* instance.  We make life easier for the prover by definining   *)
          (* some local variables and hiding their definitions before      *)
          (* appealing to the HashCollisionResistant assumption.           *)
          (*****************************************************************)
          <5> h1a == ll2PreviousHistorySummary1.extension
          <5> h2a == securedInput1
          <5> h1b == ll2PreviousHistorySummary2.extension
          <5> h2b == securedInput2
          <5>1. h1a \in HashDomain
            BY <4>2
          <5>2. h2a \in HashDomain
            BY <4>3
          <5>3. h1b \in HashDomain
            BY <4>4
          <5>4. h2b \in HashDomain
            BY <4>5
          <5>5. Hash(h1a, h2a) = Hash(h1b, h2b)
            BY <4>1 DEF newExtension1, newExtension2
          <5>6. h1a = h1b /\ h2a = h2b
            <6> HIDE DEF h1a, h2a, h1b, h2b
            <6>1. QED
              BY <5>1, <5>2, <5>3, <5>4, <5>5, HashCollisionResistant
          <5>7. QED
            BY <5>6
      (*********************************************************************)
      (* The previous history summaries are equal to each other, because   *)
      (* the fields of each record are equal.                              *)
      (*********************************************************************)
      <3>15. ll2PreviousHistorySummary1 = ll2PreviousHistorySummary2
        <4>1. ll2PreviousHistorySummary1.anchor = ll2PreviousHistorySummary2.anchor
          <5>1. ll2NewHistorySummary1 = ll2NewHistorySummary2
            BY <3>13
            DEF ll2SuccessorHistorySummary1, ll2SuccessorHistorySummary2, Successor,
                  ll2NewHistorySummary1, ll2NewHistorySummary2, newExtension1, newExtension2,
                  newAnchor1, newAnchor2, securedInput1, securedInput2
          <5>2. QED
            BY <5>1 DEF ll2NewHistorySummary1, ll2NewHistorySummary2, newAnchor1, newAnchor2
        <4>2. ll2PreviousHistorySummary1.extension = ll2PreviousHistorySummary2.extension
          BY <3>14
        (*******************************************************************)
        (* Because the fields are equal, the records are equal, but        *)
        (* proving this requires that we prove the types of the records    *)
        (* and invoke the HistorySummaryRecordCompositionLemma.            *)
        (*******************************************************************)
        <4>3. ll2PreviousHistorySummary1 \in HistorySummaryType
          BY <2>1
        <4>4. ll2PreviousHistorySummary1 \in HistorySummaryType
          BY <2>2
        <4>5. QED
          BY <4>1, <4>2, <4>3, <4>4, HistorySummaryRecordCompositionLemma
      (*********************************************************************)
      (* The input values are equal to each other, because the secured     *)
      (* inputs are equal, and the hash function is collision-resistant.   *)
      (*********************************************************************)
      <3>16. input1 = input2
        <4>1. securedInput1 = securedInput2
          BY <3>14
        <4>2. hashBarrier \in HashDomain
          BY <1>1 DEF HashDomain
        <4>3. input1 \in HashDomain
          BY <2>1 DEF HashDomain
        <4>4. input2 \in HashDomain
          BY <2>2 DEF HashDomain
        <4>5. QED
          BY <4>1, <4>2, <4>3, <4>4, HashCollisionResistant
          DEF securedInput1, securedInput2
      (*********************************************************************)
      (* Each of the conjuncts is true, so the conjunction is true.        *)
      (*********************************************************************)
      <3>17. QED
        BY <3>15, <3>16
    (***********************************************************************)
    (* The `^\basicname^' previous history summaries are equal, because    *)
    (* the `^\fastname^' previous history summaries are equal, and the     *)
    (* matches are unique.                                                 *)
    (***********************************************************************)
    <2>4. ll1PreviousHistorySummary1 = ll1PreviousHistorySummary2
      <3>1. HistorySummariesMatch(ll1PreviousHistorySummary1, ll2PreviousHistorySummary1, hashBarrier)
        BY <2>1
      <3>2. HistorySummariesMatch(ll1PreviousHistorySummary2, ll2PreviousHistorySummary2, hashBarrier)
        BY <2>2
      <3>3. ll2PreviousHistorySummary1 = ll2PreviousHistorySummary2
        BY <2>3
      <3>4. QED
        <4>1. ll1PreviousHistorySummary1 \in HashType
          BY <2>1
        <4>2. ll1PreviousHistorySummary2 \in HashType
          BY <2>2
        <4>3. ll2PreviousHistorySummary1 \in HistorySummaryType
          BY <2>1
        <4>4. ll2PreviousHistorySummary2 \in HistorySummaryType
          BY <2>2
        <4>5. hashBarrier \in HashType
          BY <1>1
        <4>6. QED
          BY <3>1, <3>2, <3>3, <4>1, <4>2, <4>3, <4>4, <4>5, HistorySummariesMatchUniqueLemma
    (***********************************************************************)
    (* The `^\basicname^' history summaries are equal by construction,     *)
    (* given that the inputs to the hash functions that produce these      *)
    (* inputs are equal.                                                   *)
    (***********************************************************************)
    <2>5. QED
      <3>1. ll1HistorySummary1 = Hash(ll1PreviousHistorySummary1, input1)
        BY <2>1
      <3>2. ll1HistorySummary2 = Hash(ll1PreviousHistorySummary2, input2)
        BY <2>2
      <3>3. input1 = input2
        BY <2>3
      <3>4. QED
        BY <2>4, <3>1, <3>2, <3>3
  (*************************************************************************)
  (* The base case and the recursive case are exhaustive.                  *)
  (*************************************************************************)
  <1>5. QED
    BY <1>3, <1>4

====
