---- MODULE MemoirLL2Implementation ----

(***************************************************************************)
(* This module proves that the `^\fastname^' spec implements the           *)
(* `^\basicname^' spec, under the defined refinement.                      *)
(***************************************************************************)

EXTENDS MemoirLL2ImplementationLemmas

(***************************************************************************)
(* The LL2Implementation theorem is where the rubber meets the road.  This *)
(* is the ultimate proof that the `^\fastname^' spec implements the        *)
(* `^\basicname^' spec, under the defined refinement.                      *)
(***************************************************************************)

THEOREM LL2Implementation == LL2Spec /\ []LL2Refinement => LL1Spec \*`^\label{THEOREM-LL2Implementation}^'
  (*************************************************************************)
  (* This proof will require the LL2TypeInvariant.  Fortunately, the       *)
  (* LL2TypeSafe theorem has already proven that the `^\fastname^' spec    *)
  (* satisfies its type invariant.                                         *)
  (*************************************************************************)
  <1>1. LL2Spec => []LL2TypeInvariant
    BY LL2TypeSafe
  (*************************************************************************)
  (* The top level of the proof is boilerplate TLA+ for a StepSimulation   *)
  (* proof.  First, we prove that the initial predicate of the             *)
  (* `^\fastname^' spec, conjoined with the LL2Refinement and type         *)
  (* invariant, implies the initial predicate of the `^\basicname^' spec.  *)
  (* Second, we prove that the LL2Next predicate, conjoined with the       *)
  (* LL2Refinement and type invariant in both primed and unprimed states,  *)
  (* implies the LL1Next predicate.  Third, we use temporal induction to   *)
  (* prove that these two conditions imply that, if the LL2Refinement and  *)
  (* the type invariant always hold, the LL2Spec implies the LL1Spec.      *)
  (*************************************************************************)
  <1>2. LL2Init /\ LL2Refinement /\ LL2TypeInvariant => LL1Init
    (***********************************************************************)
    (* We begin the base case by assuming the antecedent.                  *)
    (***********************************************************************)
    <2>1. HAVE LL2Init /\ LL2Refinement /\ LL2TypeInvariant
    (***********************************************************************)
    (* We pick a symmetric key and a hash barrier that satisfy the LL2Init *)
    (* predicate.                                                          *)
    (***********************************************************************)
    <2>2. PICK symmetricKey \in SymmetricKeyType, hashBarrier \in HashType :
              LL2Init!(symmetricKey, hashBarrier)!1
      BY <2>1 DEF LL2Init
    (***********************************************************************)
    (* We re-state the definitions from LL2Init.                           *)
    (***********************************************************************)
    <2> initialPrivateStateEnc == SymmetricEncrypt(symmetricKey, InitialPrivateState)
    <2> initialStateHash == Hash(InitialPublicState, initialPrivateStateEnc)
    <2> ll2InitialHistorySummary == [
            anchor |-> BaseHashValue,
            extension |-> BaseHashValue ]
    <2> ll2InitialHistorySummaryHash == Hash(BaseHashValue, BaseHashValue)
    <2> ll2InitialHistoryStateBinding == Hash(ll2InitialHistorySummaryHash, initialStateHash)
    <2> ll2InitialAuthenticator == GenerateMAC(symmetricKey, ll2InitialHistoryStateBinding)
    <2> ll2InitialUntrustedStorage == [
            publicState |-> InitialPublicState,
            privateStateEnc |-> initialPrivateStateEnc,
            historySummary |-> ll2InitialHistorySummary,
            authenticator |-> ll2InitialAuthenticator ]
    <2> ll2InitialTrustedStorage == [
            historySummaryAnchor |-> BaseHashValue,
            symmetricKey |-> symmetricKey,
            hashBarrier |-> hashBarrier,
            extensionInProgress |-> FALSE ]
    (***********************************************************************)
    (* We prove that the definitions from LL2Init satisfy their types,     *)
    (* using the LL2InitDefsTypeSafeLemma.                                 *)
    (***********************************************************************)
    <2>3.   /\  initialPrivateStateEnc \in PrivateStateEncType
            /\  initialStateHash \in HashType
            /\  ll2InitialHistorySummary \in HistorySummaryType
            /\  ll2InitialHistorySummaryHash \in HashType
            /\  ll2InitialHistoryStateBinding \in HashType
            /\  ll2InitialAuthenticator \in MACType
            /\  ll2InitialUntrustedStorage \in LL2UntrustedStorageType
            /\  ll2InitialTrustedStorage \in LL2TrustedStorageType
      <3>1. symmetricKey \in SymmetricKeyType
        BY <2>2
      <3>2. QED
        BY <3>1, LL2InitDefsTypeSafeLemma
    (***********************************************************************)
    (* We hide the definitions from LL2Init.                               *)
    (***********************************************************************)
    <2> HIDE DEF initialPrivateStateEnc, initialStateHash, ll2InitialHistorySummary,
            ll2InitialHistorySummaryHash, ll2InitialHistoryStateBinding, ll2InitialAuthenticator,
            ll2InitialUntrustedStorage, ll2InitialTrustedStorage
    (***********************************************************************)
    (* We re-state the definitions from LL1Init, except for                *)
    (* initialPrivateStateEnc and initialStateHash, which are the same as  *)
    (* the definitions from LL2Init.                                       *)
    (***********************************************************************)
    <2> ll1InitialHistoryStateBinding == Hash(BaseHashValue, initialStateHash)
    <2> ll1InitialAuthenticator == GenerateMAC(symmetricKey, ll1InitialHistoryStateBinding)
    <2> ll1InitialUntrustedStorage == [
            publicState |-> InitialPublicState,
            privateStateEnc |-> initialPrivateStateEnc,
            historySummary |-> BaseHashValue,
            authenticator |-> ll1InitialAuthenticator ]
    <2> ll1InitialTrustedStorage == [
            historySummary |-> BaseHashValue,
            symmetricKey |-> symmetricKey ]
    (***********************************************************************)
    (* We prove that the definitions from LL1Init satisfy their types,     *)
    (* using the LL1InitDefsTypeSafeLemma.                                 *)
    (***********************************************************************)
    <2>4. /\  ll1InitialHistoryStateBinding \in HashType
          /\  ll1InitialAuthenticator \in MACType
          /\  ll1InitialUntrustedStorage \in LL1UntrustedStorageType
          /\  ll1InitialTrustedStorage \in LL1TrustedStorageType
      <3>1. symmetricKey \in SymmetricKeyType
        BY <2>2
      <3>2. QED
        BY <3>1, LL1InitDefsTypeSafeLemma DEF initialPrivateStateEnc, initialStateHash
    (***********************************************************************)
    (* We hide the definitions from LL1Init.                               *)
    (***********************************************************************)
    <2> HIDE DEF ll1InitialHistoryStateBinding, ll1InitialAuthenticator,
            ll1InitialUntrustedStorage, ll1InitialTrustedStorage
    (***********************************************************************)
    (* One fact that will be useful in several places is that the initial  *)
    (* history summaries match across the two specs.  This follows fairly  *)
    (* simply from the definition of HistorySummariesMatch.                *)
    (***********************************************************************)
    <2>5. HistorySummariesMatch(BaseHashValue, ll2InitialHistorySummary, LL2NVRAM.hashBarrier)
      <3>1. BaseHashValue \in HashType
        BY ConstantsTypeSafe
      <3>2. ll2InitialHistorySummary \in HistorySummaryType
        BY ConstantsTypeSafe, BaseHashValueTypeSafe DEF ll2InitialHistorySummary, HistorySummaryType
      <3>3. LL2NVRAM.hashBarrier \in HashType
        BY <2>1, LL2SubtypeImplicationLemma DEF LL2SubtypeImplication
      <3>4. QED
        BY <3>1, <3>2, <3>3, HistorySummariesMatchDefinition DEF ll2InitialHistorySummary
    (***********************************************************************)
    (* The most interesting part of the proof for the initial state is     *)
    (* proving that the corresponding initial authenticator values match   *)
    (* across the two specs.  We will use this in three places below: for  *)
    (* LL1Disk.authenticator, LL1RAM.authenticator, and                    *)
    (* LL1ObservedAuthenticators.  We prove the match using the definition *)
    (* of the AuthenticatorsMatch predicate.                               *)
    (***********************************************************************)
    <2>6. AuthenticatorsMatch(
              ll1InitialAuthenticator,
              ll2InitialAuthenticator,
              LL2NVRAM.symmetricKey,
              LL2NVRAM.hashBarrier)
      (*********************************************************************)
      (* First, we prove some types needed by the definition of the        *)
      (* AuthenticatorsMatch predicate.                                    *)
      (*********************************************************************)
      <3>1. initialStateHash \in HashType
        BY <2>3
      <3>2. BaseHashValue \in HashType
        BY ConstantsTypeSafe
      <3>3. ll2InitialHistorySummary \in HistorySummaryType
        BY <2>3
      (*********************************************************************)
      (* We then prove that, in the `^\fastname^' spec, the initial        *)
      (* authenticator is a valid MAC for the initial history state        *)
      (* binding.  We will use the MACComplete property.                   *)
      (*********************************************************************)
      <3>4. ValidateMAC(
                LL2NVRAM.symmetricKey,
                ll2InitialHistoryStateBinding,
                ll2InitialAuthenticator)
        (*******************************************************************)
        (* In the `^\fastname^' spec, the initial authenticator is         *)
        (* generated as a MAC of the initial history state binding.        *)
        (*******************************************************************)
        <4>1. ll2InitialAuthenticator =
                  GenerateMAC(LL2NVRAM.symmetricKey, ll2InitialHistoryStateBinding)
          <5>1. LL2NVRAM.symmetricKey = symmetricKey
            BY <2>2
          <5>2. QED
            BY <5>1
            DEF ll2InitialAuthenticator, ll2InitialHistoryStateBinding,
                ll2InitialHistorySummaryHash, initialStateHash, initialPrivateStateEnc
        (*******************************************************************)
        (* We can thus use the MACComplete property to show that the       *)
        (* generated MAC validates appropriately.  To do this, we first    *)
        (* need to prove some types.                                       *)
        (*******************************************************************)
        <4>2. LL2NVRAM.symmetricKey \in SymmetricKeyType
          BY <2>1, LL2SubtypeImplicationLemma DEF LL2SubtypeImplication
        <4>3. ll2InitialHistoryStateBinding \in HashType
          BY <2>3
        (*******************************************************************)
        (* Then, we appeal to the MACComplete property in a                *)
        (* straightforward way.                                            *)
        (*******************************************************************)
        <4>4. QED
          BY <4>1, <4>2, <4>3, MACComplete
      (*********************************************************************)
      (* We then prove that, in the `^\basicname^' spec, the initial       *)
      (* authenticator is generated as a MAC of the initial history state  *)
      (* binding.                                                          *)
      (*********************************************************************)
      <3>5. ll1InitialAuthenticator = GenerateMAC(
                LL2NVRAM.symmetricKey, ll1InitialHistoryStateBinding)
        <4>1. LL2NVRAM.symmetricKey = symmetricKey
          BY <2>2
        <4>2. QED
          BY <4>1
          DEF ll1InitialAuthenticator, ll1InitialHistoryStateBinding,
              initialStateHash, initialPrivateStateEnc
      (*********************************************************************)
      (* The initial history summaries match across the two specs, as we   *)
      (* proved above.                                                     *)
      (*********************************************************************)
      <3>6. HistorySummariesMatch(BaseHashValue, ll2InitialHistorySummary, LL2NVRAM.hashBarrier)
        BY <2>5
      (*********************************************************************)
      (* We then invoke the definition of the AuthenticatorsMatch          *)
      (* predicate.                                                        *)
      (*********************************************************************)
      <3>7. QED
        BY <3>1, <3>2, <3>3, <3>4, <3>5, <3>6
        DEF AuthenticatorsMatch, ll2InitialAuthenticator,
            ll1InitialHistoryStateBinding, ll2InitialHistoryStateBinding,
            ll2InitialHistorySummaryHash, ll2InitialHistorySummary
    (***********************************************************************)
    (* The next six steps assert each conjunct in the LL1Init predicate.   *)
    (* For the LL1Disk variable, we prove each field within the record     *)
    (* separately.                                                         *)
    (***********************************************************************)
    <2>7. LL1Disk = ll1InitialUntrustedStorage
      (*********************************************************************)
      (* The LL1Disk's public state equals the initial public state,       *)
      (* because the refinement asserts a direct equality between this     *)
      (* field for the two specs.                                          *)
      (*********************************************************************)
      <3>1. LL1Disk.publicState = InitialPublicState
        <4>1. LL2Disk.publicState = InitialPublicState
          BY <2>2
        <4>2. LL1Disk.publicState = LL2Disk.publicState
          BY <2>1 DEF LL2Refinement
        <4>3. QED
          BY <4>1, <4>2
      (*********************************************************************)
      (* The LL1Disk's private encrypted state equals the initial private  *)
      (* encrypted state, because the refinement asserts a direct equality *)
      (* between this field for the two specs.                             *)
      (*********************************************************************)
      <3>2. LL1Disk.privateStateEnc = initialPrivateStateEnc
        <4>1. LL2Disk.privateStateEnc = initialPrivateStateEnc
          BY <2>2 DEF initialPrivateStateEnc
        <4>2. LL1Disk.privateStateEnc = LL2Disk.privateStateEnc
          BY <2>1 DEF LL2Refinement
        <4>3. QED
          BY <4>1, <4>2
      (*********************************************************************)
      (* The LL1Disk's history summary equals the base hash value.  There  *)
      (* are four steps: (1) The refinement asserts that the corresponding *)
      (* fields from the two specs match according to the                  *)
      (* HistorySummariesMatch predicate.  (2) We prove that the initial   *)
      (* history summary values match across the two specs.  (3) We prove  *)
      (* that the `^\fastname^' disk's history summary equals its initial  *)
      (* history summary value.  (4) We use the                            *)
      (* HistorySummariesMatchUniqueLemma to prove that the field in the   *)
      (* `^\basicname^' spec must equal the base hash value.               *)
      (*********************************************************************)
      <3>3. LL1Disk.historySummary = BaseHashValue
        (*******************************************************************)
        (* The corresponding fields from the two specs match, thanks to    *)
        (* the refinement.                                                 *)
        (*******************************************************************)
        <4>1. HistorySummariesMatch(
                    LL1Disk.historySummary,
                    LL2Disk.historySummary,
                    LL2NVRAM.hashBarrier)
          BY <2>1 DEF LL2Refinement
        (*******************************************************************)
        (* The initial history summaries match across the two specs, as we *)
        (* proved above.                                                   *)
        (*******************************************************************)
        <4>2. HistorySummariesMatch(BaseHashValue, ll2InitialHistorySummary, LL2NVRAM.hashBarrier)
          BY <2>5
        (*******************************************************************)
        (* The history summary in the `^\fastname^' spec's disk equals the *)
        (* initial history summary, by the definition of LL2Init.          *)
        (*******************************************************************)
        <4>3. LL2Disk.historySummary = ll2InitialHistorySummary
          BY <2>2 DEF ll2InitialHistorySummary
        (*******************************************************************)
        (* We use the HistorySummariesMatchUniqueLemma to prove that the   *)
        (* field in the `^\basicname^' spec equals the base hash value.    *)
        (* This requires proving some types.                               *)
        (*******************************************************************)
        <4>4. QED
          <5>1. LL1Disk.historySummary \in HashType
            BY <2>1 DEF LL2Refinement, LL1UntrustedStorageType
          <5>2. BaseHashValue \in HashType
            BY ConstantsTypeSafe
          <5>3. LL2Disk.historySummary \in HistorySummaryType
            BY <2>1, LL2SubtypeImplicationLemma DEF LL2SubtypeImplication
          <5>4. LL2NVRAM.hashBarrier \in HashType
            BY <2>1, LL2SubtypeImplicationLemma DEF LL2SubtypeImplication
          <5>5. QED
            BY <4>1, <4>2, <4>3, <5>1, <5>2, <5>3, <5>4, HistorySummariesMatchUniqueLemma
      (*********************************************************************)
      (* The LL1Disk's authenticator equals the initial authenticator.     *)
      (* There are four steps: (1) The refinement asserts that the         *)
      (* corresponding fields from the two specs match according to the    *)
      (* HistorySummariesMatch predicate.  (2) We prove that the initial   *)
      (* authenticator values match across the two specs.  (3) We prove    *)
      (* that the `^\fastname^' disk's authenticator equals its initial    *)
      (* authenticator value.  (4) We use the                              *)
      (* AuthenticatorsMatchUniqueLemma to prove that the field in the     *)
      (* `^\basicname^' spec must equal the initial authenticator value.   *)
      (*********************************************************************)
      <3>4. LL1Disk.authenticator = ll1InitialAuthenticator
        (*******************************************************************)
        (* The corresponding fields from the two specs match, thanks to    *)
        (* the refinement.                                                 *)
        (*******************************************************************)
        <4>1. \E symmetricKey1 \in SymmetricKeyType :
                  AuthenticatorsMatch(
                      LL1Disk.authenticator,
                      LL2Disk.authenticator,
                      symmetricKey1,
                      LL2NVRAM.hashBarrier)
          BY <2>1 DEF LL2Refinement
        (*******************************************************************)
        (* The corresponding initial authenticator values match across the *)
        (* two specs.  We proved this above.                               *)
        (*******************************************************************)
        <4>2. AuthenticatorsMatch(
                  ll1InitialAuthenticator,
                  ll2InitialAuthenticator,
                  LL2NVRAM.symmetricKey,
                  LL2NVRAM.hashBarrier)
          BY <2>6
        (*******************************************************************)
        (* The `^\fastname^' disk's authenticator equals its initial       *)
        (* authenticator value, as asserted by the LL2Init predicate.      *)
        (*******************************************************************)
        <4>3. LL2Disk.authenticator = ll2InitialAuthenticator
          BY <2>2
          DEF ll2InitialAuthenticator, ll2InitialHistoryStateBinding,
              ll2InitialHistorySummaryHash, initialStateHash, initialPrivateStateEnc
        (*******************************************************************)
        (* We use the AuthenticatorsMatchUniqueLemma to prove that the     *)
        (* field in the `^\basicname^' spec equals the initial             *)
        (* authenticator value.  This requires proving some types.         *)
        (*******************************************************************)
        <4>4. QED
          <5>1. LL1Disk.authenticator \in MACType
            BY <2>1 DEF LL2Refinement, LL1UntrustedStorageType
          <5>2. ll1InitialAuthenticator \in MACType
            BY <2>4
          <5>3. LL2Disk.authenticator \in MACType
            BY <2>1, LL2SubtypeImplicationLemma DEF LL2SubtypeImplication
          <5>4. LL2NVRAM.symmetricKey \in SymmetricKeyType
            BY <2>1, LL2SubtypeImplicationLemma DEF LL2SubtypeImplication
          <5>5. LL2NVRAM.hashBarrier \in HashType
            BY <2>1, LL2SubtypeImplicationLemma DEF LL2SubtypeImplication
          <5>6. QED
            BY <4>1, <4>2, <4>3, <5>1, <5>2, <5>3, <5>4, <5>5,
                AuthenticatorsMatchUniqueLemma
      (*********************************************************************)
      (* The refinement asserts that the Disk record has the appropriate   *)
      (* type.                                                             *)
      (*********************************************************************)
      <3>5. LL1Disk \in LL1UntrustedStorageType
        BY <2>1 DEF LL2Refinement
      (*********************************************************************)
      (* We use the LL1DiskRecordCompositionLemma to unify the field       *)
      (* equalities into a record equality.                                *)
      (*********************************************************************)
      <3>6. QED
        BY <3>1, <3>2, <3>3, <3>4, <3>5, LL1DiskRecordCompositionLemma
        DEF ll1InitialUntrustedStorage
    (***********************************************************************)
    (* The second conjunct in the LL1Init predicate relates to the LL1RAM  *)
    (* variable.  We prove each field within the record separately.        *)
    (***********************************************************************)
    <2>8. LL1RAM = ll1InitialUntrustedStorage
      (*********************************************************************)
      (* The LL1RAM's public state equals the initial public state,        *)
      (* because the refinement asserts a direct equality between this     *)
      (* field for the two specs.                                          *)
      (*********************************************************************)
      <3>1. LL1RAM.publicState = InitialPublicState
        <4>1. LL2RAM.publicState = InitialPublicState
          BY <2>2
        <4>2. LL1RAM.publicState = LL2RAM.publicState
          BY <2>1 DEF LL2Refinement
        <4>3. QED
          BY <4>1, <4>2
      (*********************************************************************)
      (* The LL1RAM's private encrypted state equals the initial private   *)
      (* encrypted state, because the refinement asserts a direct equality *)
      (* between this field for the two specs.                             *)
      (*********************************************************************)
      <3>2. LL1RAM.privateStateEnc = initialPrivateStateEnc
        <4>1. LL2RAM.privateStateEnc = initialPrivateStateEnc
          BY <2>2 DEF initialPrivateStateEnc
        <4>2. LL1RAM.privateStateEnc = LL2RAM.privateStateEnc
          BY <2>1 DEF LL2Refinement
        <4>3. QED
          BY <4>1, <4>2
      (*********************************************************************)
      (* The LL1RAM's history summary equals the base hash value.  There   *)
      (* are four steps: (1) The refinement asserts that the corresponding *)
      (* fields from the two specs match according to the                  *)
      (* HistorySummariesMatch predicate.  (2) We prove that the initial   *)
      (* history summary values match across the two specs.  (3) We prove  *)
      (* that the `^\fastname^' RAM's history summary equals its initial   *)
      (* history summary value.  (4) We use the                            *)
      (* HistorySummariesMatchUniqueLemma to prove that the field in the   *)
      (* `^\basicname^' spec must equal the base hash value.               *)
      (*********************************************************************)
      <3>3. LL1RAM.historySummary = BaseHashValue
        (*******************************************************************)
        (* The corresponding fields from the two specs match, thanks to    *)
        (* the refinement.                                                 *)
        (*******************************************************************)
        <4>1. HistorySummariesMatch(
                    LL1RAM.historySummary, LL2RAM.historySummary, LL2NVRAM.hashBarrier)
          BY <2>1 DEF LL2Refinement
        (*******************************************************************)
        (* The initial history summaries match across the two specs, as we *)
        (* proved above.                                                   *)
        (*******************************************************************)
        <4>2. HistorySummariesMatch(BaseHashValue, ll2InitialHistorySummary, LL2NVRAM.hashBarrier)
          BY <2>5
        (*******************************************************************)
        (* The history summary in the `^\fastname^' spec's RAM equals the  *)
        (* initial history summary, by the definition of LL2Init.          *)
        (*******************************************************************)
        <4>3. LL2RAM.historySummary = ll2InitialHistorySummary
          BY <2>2 DEF ll2InitialHistorySummary
        (*******************************************************************)
        (* We use the HistorySummariesMatchUniqueLemma to prove that the   *)
        (* field in the `^\basicname^' spec equals the base hash value.    *)
        (* This requires proving some types.                               *)
        (*******************************************************************)
        <4>4. QED
          <5>1. LL1RAM.historySummary \in HashType
            BY <2>1 DEF LL2Refinement, LL1UntrustedStorageType
          <5>2. BaseHashValue \in HashType
            BY ConstantsTypeSafe
          <5>3. LL2RAM.historySummary \in HistorySummaryType
            BY <2>1, LL2SubtypeImplicationLemma DEF LL2SubtypeImplication
          <5>4. LL2NVRAM.hashBarrier \in HashType
            BY <2>1, LL2SubtypeImplicationLemma DEF LL2SubtypeImplication
          <5>5. QED
            BY <4>1, <4>2, <4>3, <5>1, <5>2, <5>3, <5>4, HistorySummariesMatchUniqueLemma
      (*********************************************************************)
      (* The LL1RAM's authenticator equals the initial authenticator.      *)
      (* There are four steps: (1) The refinement asserts that the         *)
      (* corresponding fields from the two specs match according to the    *)
      (* HistorySummariesMatch predicate.  (2) We prove that the initial   *)
      (* authenticator values match across the two specs.  (3) We prove    *)
      (* that the `^\fastname^' RAM's authenticator equals its initial     *)
      (* authenticator value.  (4) We use the                              *)
      (* AuthenticatorsMatchUniqueLemma to prove that the field in the     *)
      (* `^\basicname^' spec must equal the initial authenticator value.   *)
      (*********************************************************************)
      <3>4. LL1RAM.authenticator = ll1InitialAuthenticator
        (*******************************************************************)
        (* The corresponding fields from the two specs match, thanks to    *)
        (* the refinement.                                                 *)
        (*******************************************************************)
        <4>1. \E symmetricKey1 \in SymmetricKeyType :
                  AuthenticatorsMatch(
                      LL1RAM.authenticator,
                      LL2RAM.authenticator,
                      symmetricKey1,
                      LL2NVRAM.hashBarrier)
          BY <2>1 DEF LL2Refinement
        (*******************************************************************)
        (* The corresponding initial authenticator values match across the *)
        (* two specs.  We proved this above.                               *)
        (*******************************************************************)
        <4>2. AuthenticatorsMatch(
                  ll1InitialAuthenticator,
                  ll2InitialAuthenticator,
                  LL2NVRAM.symmetricKey,
                  LL2NVRAM.hashBarrier)
          BY <2>6
        (*******************************************************************)
        (* The `^\fastname^' RAM's authenticator equals its initial        *)
        (* authenticator value, as asserted by the LL2Init predicate.      *)
        (*******************************************************************)
        <4>3. LL2RAM.authenticator = ll2InitialAuthenticator
          BY <2>2
          DEF ll2InitialAuthenticator, ll2InitialHistoryStateBinding,
              ll2InitialHistorySummaryHash, initialStateHash, initialPrivateStateEnc
        (*******************************************************************)
        (* We use the AuthenticatorsMatchUniqueLemma to prove that the     *)
        (* field in the `^\basicname^' spec equals the initial             *)
        (* authenticator value.  This requires proving some types.         *)
        (*******************************************************************)
        <4>4. QED
          <5>1. LL1RAM.authenticator \in MACType
            BY <2>1 DEF LL2Refinement, LL1UntrustedStorageType
          <5>2. ll1InitialAuthenticator \in MACType
            BY <2>4
          <5>3. LL2RAM.authenticator \in MACType
            BY <2>1, LL2SubtypeImplicationLemma DEF LL2SubtypeImplication
          <5>4. LL2NVRAM.symmetricKey \in SymmetricKeyType
            BY <2>1, LL2SubtypeImplicationLemma DEF LL2SubtypeImplication
          <5>5. LL2NVRAM.hashBarrier \in HashType
            BY <2>1, LL2SubtypeImplicationLemma DEF LL2SubtypeImplication
          <5>6. QED
            BY <4>1, <4>2, <4>3, <5>1, <5>2, <5>3, <5>4, <5>5,
                AuthenticatorsMatchUniqueLemma
      (*********************************************************************)
      (* The refinement asserts that the RAM record has the appropriate    *)
      (* type.                                                             *)
      (*********************************************************************)
      <3>5. LL1RAM \in LL1UntrustedStorageType
        BY <2>1 DEF LL2Refinement
      (*********************************************************************)
      (* We use the LL1RAMRecordCompositionLemma to unify the field        *)
      (* equalities into a record equality.                                *)
      (*********************************************************************)
      <3>6. QED
        BY <3>1, <3>2, <3>3, <3>4, <3>5, LL1RAMRecordCompositionLemma
        DEF ll1InitialUntrustedStorage
    (***********************************************************************)
    (* The third conjunct in the LL1Init predicate relates to the LL1NVRAM *)
    (* variable.                                                           *)
    (***********************************************************************)
    <2>9. LL1NVRAM = ll1InitialTrustedStorage
      (*********************************************************************)
      (* The LL1NVRAM's history summary equals the base hash value.  There *)
      (* are three steps: (1) The refinement asserts that the              *)
      (* corresponding fields from the two specs match according to the    *)
      (* HistorySummariesMatch predicate.  (2) We prove that the initial   *)
      (* history summary values match across the two specs.  (3) We prove  *)
      (* that the `^\fastname^' NVRAM's logical history summary equals its *)
      (* initial history summary value.  (4) We use the                    *)
      (* HistorySummariesMatchUniqueLemma to prove that the field in the   *)
      (* `^\basicname^' spec must equal the base hash value.               *)
      (*********************************************************************)
      <3>1. LL1NVRAM.historySummary = BaseHashValue
        (*******************************************************************)
        (* The corresponding fields from the two specs match, thanks to    *)
        (* the refinement.                                                 *)
        (*******************************************************************)
        <4>1. HistorySummariesMatch(
                  LL1NVRAM.historySummary, LL2NVRAMLogicalHistorySummary, LL2NVRAM.hashBarrier)
          BY <2>1 DEF LL2Refinement
        (*******************************************************************)
        (* The initial history summaries match across the two specs, as we *)
        (* proved above.                                                   *)
        (*******************************************************************)
        <4>2. HistorySummariesMatch(BaseHashValue, ll2InitialHistorySummary, LL2NVRAM.hashBarrier)
          BY <2>5
        (*******************************************************************)
        (* The logical value of the history summary in the `^\fastname^'   *)
        (* spec's NVRAM equals the initial history summary.  This follows  *)
        (* from the definition of LL2NVRAMLogicalHistorySummary.           *)
        (*******************************************************************)
        <4>3. LL2NVRAMLogicalHistorySummary = ll2InitialHistorySummary
          <5>1. LL2NVRAM.extensionInProgress <=> FALSE    (** changed = by <=> *)
            BY <2>2
          <5>2. LL2NVRAM.historySummaryAnchor = BaseHashValue
            BY <2>2
          <5>3. LL2SPCR = BaseHashValue
            BY <2>2
          <5>4. QED
            BY <5>1, <5>2, <5>3 DEF LL2NVRAMLogicalHistorySummary, ll2InitialHistorySummary
        (*******************************************************************)
        (* We use the HistorySummariesMatchUniqueLemma to prove that the   *)
        (* field in the `^\basicname^' spec equals the base hash value.    *)
        (* This requires proving some types.                               *)
        (*******************************************************************)
        <4>4. QED
          <5>1. LL1NVRAM.historySummary \in HashType
            BY <2>1 DEF LL2Refinement, LL1TrustedStorageType
          <5>2. BaseHashValue \in HashType
            BY ConstantsTypeSafe
          <5>3. LL2NVRAMLogicalHistorySummary \in HistorySummaryType
            BY <2>1, LL2NVRAMLogicalHistorySummaryTypeSafe
          <5>4. LL2NVRAM.hashBarrier \in HashType
            BY <2>1, LL2SubtypeImplicationLemma DEF LL2SubtypeImplication
          <5>5. QED
            BY <4>1, <4>2, <4>3, <5>1, <5>2, <5>3, <5>4, HistorySummariesMatchUniqueLemma
      (*********************************************************************)
      (* The symmetric key in the LL1NVRAM matches its initial value by    *)
      (* direct equality through the refinement.                           *)
      (*********************************************************************)
      <3>2. LL1NVRAM.symmetricKey = symmetricKey
        <4>1. LL1NVRAM.symmetricKey = LL2NVRAM.symmetricKey
          BY <2>1 DEF LL2Refinement
        <4>2. LL2NVRAM.symmetricKey = symmetricKey
          BY <2>2
        <4>3. QED
          BY <4>1, <4>2
      (*********************************************************************)
      (* The refinement asserts that the NVRAM record has the appropriate  *)
      (* type.                                                             *)
      (*********************************************************************)
      <3>3. LL1NVRAM \in LL1TrustedStorageType
        BY <2>1 DEF LL2Refinement
      (*********************************************************************)
      (* We use the LL1NVRAMRecordCompositionLemma to unify the field      *)
      (* equalities into a record equality.                                *)
      (*********************************************************************)
      <3>4. QED
        BY <3>1, <3>2, <3>3, LL1NVRAMRecordCompositionLemma
        DEF ll1InitialTrustedStorage
    (***********************************************************************)
    (* The fourth conjunct in the LL1Init predicate relates to the         *)
    (* LL1AvailableInputs variable.  The proof is straightforward, because *)
    (* the equality is direct.                                             *)
    (***********************************************************************)
    <2>10. LL1AvailableInputs = InitialAvailableInputs
      <3>1. LL2AvailableInputs = InitialAvailableInputs
        BY <2>2
      <3>2. LL1AvailableInputs = LL2AvailableInputs
        BY <2>1 DEF LL2Refinement
      <3>3. QED
        BY <3>1, <3>2
    (***********************************************************************)
    (* The fifth conjunct in the LL1Init predicate relates to the          *)
    (* LL1ObservedOutputs variable.  The proof is straightforward, because *)
    (* the equality is direct.                                             *)
    (***********************************************************************)
    <2>11. LL1ObservedOutputs = {}
      <3>1. LL2ObservedOutputs = {}
        BY <2>2
      <3>2. LL1ObservedOutputs = LL2ObservedOutputs
        BY <2>1 DEF LL2Refinement
      <3>3. QED
        BY <3>1, <3>2
    (***********************************************************************)
    (* The sixth conjunct in the LL1Init predicate relates to the          *)
    (* LL1ObservedAuthenticators variable.  There are four steps: (1) The  *)
    (* refinement asserts that the corresponding fields from the two specs *)
    (* match according to the HistorySummariesMatch predicate.  (2) We     *)
    (* prove that the initial authenticator values match across the two    *)
    (* specs.  (3) We prove that the `^\fastname^' set of observed         *)
    (* authenticators equals its initial value.  (4) We use the            *)
    (* AuthenticatorSetsMatchUniqueLemma to prove that the variable in the *)
    (* `^\basicname^' spec must equal the initial value.                   *)
    (***********************************************************************)
    <2>12. LL1ObservedAuthenticators = {ll1InitialAuthenticator}
      (*********************************************************************)
      (* The corresponding variables from the two specs match, thanks to   *)
      (* the refinement.                                                   *)
      (*********************************************************************)
      <3>1. AuthenticatorSetsMatch(
                LL1ObservedAuthenticators,
                LL2ObservedAuthenticators,
                LL2NVRAM.symmetricKey,
                LL2NVRAM.hashBarrier)
        BY <2>1 DEF LL2Refinement
        (*******************************************************************)
        (* The corresponding sets of initial authenticator values match    *)
        (* across the two specs.  Since the sets are singletons, this      *)
        (* follows trivially from proving that one element in each set     *)
        (* matches the other.  And we proved that these elements match     *)
        (* above.                                                          *)
        (*******************************************************************)
      <3>2. AuthenticatorSetsMatch(
                {ll1InitialAuthenticator},
                {ll2InitialAuthenticator},
                LL2NVRAM.symmetricKey,
                LL2NVRAM.hashBarrier)
        <4>1. AuthenticatorsMatch(
                ll1InitialAuthenticator,
                ll2InitialAuthenticator,
                LL2NVRAM.symmetricKey,
                LL2NVRAM.hashBarrier)
          BY <2>6
        <4>2. QED
          BY <4>1 DEF AuthenticatorSetsMatch
      (*********************************************************************)
      (* The `^\fastname^' spec's set of observed authenticators equals    *)
      (* its initial authenticator value, as asserted by the LL2Init       *)
      (* predicate.                                                        *)
      (*********************************************************************)
      <3>3. LL2ObservedAuthenticators = {ll2InitialAuthenticator}
        BY <2>2
        DEF ll2InitialAuthenticator, ll2InitialHistoryStateBinding,
            ll2InitialHistorySummaryHash, initialStateHash, initialPrivateStateEnc
      (*********************************************************************)
      (* We use the AuthenticatorSetsMatchUniqueLemma to prove that the    *)
      (* variable in the `^\basicname^' spec equals its initial value.     *)
      (* This requires proving some types.                                 *)
      (*********************************************************************)
      <3>4. QED
        <4>1. LL1ObservedAuthenticators \in SUBSET MACType
          BY <2>1 DEF LL2Refinement
        <4>2. {ll1InitialAuthenticator} \in SUBSET MACType
          <5>1. ll1InitialAuthenticator \in MACType
            BY <2>4
          <5>2. QED
            BY <5>1
        <4>3. LL2ObservedAuthenticators \in SUBSET MACType
          BY <2>1 DEF LL2TypeInvariant
        <4>4. LL2NVRAM.symmetricKey \in SymmetricKeyType
          BY <2>1, LL2SubtypeImplicationLemma DEF LL2SubtypeImplication
        <4>5. LL2NVRAM.hashBarrier \in HashType
          BY <2>1, LL2SubtypeImplicationLemma DEF LL2SubtypeImplication
        <4>6. QED
          BY <3>1, <3>2, <3>3, <4>1, <4>2, <4>3, <4>4, <4>5,
              AuthenticatorSetsMatchUniqueLemma
    (***********************************************************************)
    (* The conjunction of the above six conjuncts.                         *)
    (***********************************************************************)
    <2>13. QED
      BY <2>7, <2>8, <2>9, <2>10, <2>11, <2>12
      DEF LL1Init, initialPrivateStateEnc, initialStateHash, ll1InitialHistoryStateBinding,
          ll1InitialAuthenticator, ll1InitialUntrustedStorage, ll1InitialTrustedStorage
  (*************************************************************************)
  (* For the induction step, we will need the refinement and the type      *)
  (* invariant to be true in both the unprimed and primed states.          *)
  (*************************************************************************)
  <1>3. ( /\  [LL2Next]_LL2Vars
          /\  LL2Refinement
          /\  LL2Refinement'
          /\  LL2TypeInvariant
          /\  LL2TypeInvariant' )
      =>
          [LL1Next]_LL1Vars
    (***********************************************************************)
    (* We assume the antecedents.                                          *)
    (***********************************************************************)
    <2>1. HAVE /\  [LL2Next]_LL2Vars
               /\  LL2Refinement
               /\  LL2Refinement'
               /\  LL2TypeInvariant
               /\  LL2TypeInvariant'
    (***********************************************************************)
    (* We then prove that each step in the `^\fastname^' spec refines to a *)
    (* step in the `^\basicname^' spec.  First, a `^\fastname^' stuttering *)
    (* step refines to a `^\basicname^' stuttering step.  We prove the     *)
    (* UNCHANGED status for each `^\basicname^' variable in turn, using    *)
    (* the lemmas we have proven for this purpose.                         *)
    (***********************************************************************)
    <2>2. UNCHANGED LL2Vars => UNCHANGED LL1Vars
      <3>1. HAVE UNCHANGED LL2Vars
      <3>2. UNCHANGED LL1AvailableInputs
        BY <2>1, <3>1, UnchangedAvailableInputsLemma DEF LL2Vars
      <3>3. UNCHANGED LL1ObservedOutputs
        BY <2>1, <3>1, UnchangedObservedOutputsLemma DEF LL2Vars
      <3>4. UNCHANGED LL1ObservedAuthenticators
        BY <2>1, <3>1, UnchangedObservedAuthenticatorsLemma DEF LL2Vars
      <3>5. UNCHANGED LL1Disk
        BY <2>1, <3>1, UnchangedDiskLemma DEF LL2Vars
      <3>6. UNCHANGED LL1RAM
        BY <2>1, <3>1, UnchangedRAMLemma DEF LL2Vars
      <3>7. UNCHANGED LL1NVRAM
        BY <2>1, <3>1, UnchangedNVRAMLemma DEF LL2Vars
      <3>8. QED
        BY <3>2, <3>3, <3>4, <3>5, <3>6, <3>7 DEF LL1Vars
    (***********************************************************************)
    (* A `^\fastname^' LL2MakeInputAvailable action refines to a           *)
    (* `^\basicname^' LL1MakeInputAvailable action.                        *)
    (***********************************************************************)
    <2>3. LL2MakeInputAvailable => LL1MakeInputAvailable
      <3>1. HAVE LL2MakeInputAvailable
      <3>2. PICK input \in InputType : LL2MakeInputAvailable!(input)
        BY <3>1 DEF LL2MakeInputAvailable
      (*********************************************************************)
      (* We prove each conjunct in the LL1MakeInputAvailable action        *)
      (* separately.                                                       *)
      (*********************************************************************)
      <3>3. input \notin LL1AvailableInputs
        <4>1. input \notin LL2AvailableInputs
          BY <3>2
        <4>2. LL1AvailableInputs = LL2AvailableInputs
          BY <2>1 DEF LL2Refinement
        <4>3. LL1AvailableInputs' = LL2AvailableInputs'
          BY <2>1 DEF LL2Refinement
        <4>4. QED
          BY <4>1, <4>2, <4>3
      <3>4. LL1AvailableInputs' = LL1AvailableInputs \union {input}
        <4>1. LL2AvailableInputs' = LL2AvailableInputs \union {input}
          BY <3>2
        <4>2. LL1AvailableInputs = LL2AvailableInputs
          BY <2>1 DEF LL2Refinement
        <4>3. LL1AvailableInputs' = LL2AvailableInputs'
          BY <2>1 DEF LL2Refinement
        <4>4. QED
          BY <4>1, <4>2, <4>3
      <3>5. UNCHANGED LL1Disk
        BY <2>1, <3>2, UnchangedDiskLemma
      <3>6. UNCHANGED LL1RAM
        BY <2>1, <3>2, UnchangedRAMLemma
      <3>7. UNCHANGED LL1NVRAM
        BY <2>1, <3>2, UnchangedNVRAMLemma
      <3>8. UNCHANGED LL1ObservedOutputs
        BY <2>1, <3>2, UnchangedObservedOutputsLemma
      <3>9. UNCHANGED LL1ObservedAuthenticators
        BY <2>1, <3>2, UnchangedObservedAuthenticatorsLemma
      <3>. QED
        BY <3>3, <3>4, <3>5, <3>6, <3>7, <3>8, <3>9 DEF LL1MakeInputAvailable
    (***********************************************************************)
    (* A `^\fastname^' LL2PerformOperation action refines to a             *)
    (* `^\basicname^' LL1PerformOperation action.                          *)
    (***********************************************************************)
    <2>4. LL2PerformOperation => LL1PerformOperation
      (*********************************************************************)
      (* We assume the antecedent.                                         *)
      (*********************************************************************)
      <3>1. HAVE LL2PerformOperation
      (*********************************************************************)
      (* We pick an input that satisfies the LL2PerformOperation           *)
      (* predicate.                                                        *)
      (*********************************************************************)
      <3>2. PICK input \in LL2AvailableInputs : LL2PerformOperation!(input)!1
        BY <3>1 DEF LL2PerformOperation
      (*********************************************************************)
      (* We re-state the definitions from LL2PerformOperation.             *)
      (*********************************************************************)
      <3> ll2HistorySummaryHash ==
              Hash(LL2RAM.historySummary.anchor, LL2RAM.historySummary.extension)
      <3> ll2StateHash == Hash(LL2RAM.publicState, LL2RAM.privateStateEnc)
      <3> ll2HistoryStateBinding == Hash(ll2HistorySummaryHash, ll2StateHash)
      <3> ll2PrivateState == SymmetricDecrypt(LL2NVRAM.symmetricKey, LL2RAM.privateStateEnc)
      <3> ll2SResult == Service(LL2RAM.publicState, ll2PrivateState, input)
      <3> ll2NewPrivateStateEnc ==
              SymmetricEncrypt(LL2NVRAM.symmetricKey, ll2SResult.newPrivateState)
      <3> ll2CurrentHistorySummary == [
              anchor |-> LL2NVRAM.historySummaryAnchor,
              extension |-> LL2SPCR ]
      <3> ll2NewHistorySummary ==
              Successor(ll2CurrentHistorySummary, input, LL2NVRAM.hashBarrier)
      <3> ll2NewHistorySummaryHash ==
              Hash(ll2NewHistorySummary.anchor, ll2NewHistorySummary.extension)
      <3> ll2NewStateHash == Hash(ll2SResult.newPublicState, ll2NewPrivateStateEnc)
      <3> ll2NewHistoryStateBinding == Hash(ll2NewHistorySummaryHash, ll2NewStateHash)
      <3> ll2NewAuthenticator == GenerateMAC(LL2NVRAM.symmetricKey, ll2NewHistoryStateBinding)
      (*********************************************************************)
      (* We prove that the definitions from LL2PerformOperation satisfy    *)
      (* their types, using the LL2PerformOperationDefsTypeSafeLemma.      *)
      (*********************************************************************)
      <3>3. /\  ll2HistorySummaryHash \in HashType
            /\  ll2StateHash \in HashType
            /\  ll2HistoryStateBinding \in HashType
            /\  ll2PrivateState \in PrivateStateType
            /\  ll2SResult \in ServiceResultType
            /\  ll2SResult.newPublicState \in PublicStateType
            /\  ll2SResult.newPrivateState \in PrivateStateType
            /\  ll2SResult.output \in OutputType
            /\  ll2NewPrivateStateEnc \in PrivateStateEncType
            /\  ll2CurrentHistorySummary \in HistorySummaryType
            /\  ll2CurrentHistorySummary.anchor \in HashType
            /\  ll2CurrentHistorySummary.extension \in HashType
            /\  ll2NewHistorySummary \in HistorySummaryType
            /\  ll2NewHistorySummary.anchor \in HashType
            /\  ll2NewHistorySummary.extension \in HashType
            /\  ll2NewHistorySummaryHash \in HashType
            /\  ll2NewStateHash \in HashType
            /\  ll2NewHistoryStateBinding \in HashType
            /\  ll2NewAuthenticator \in MACType
        <4>1. input \in LL2AvailableInputs
          BY <3>2
        <4>2. LL2TypeInvariant
          BY <2>1
        <4>3. QED
          BY <4>1, <4>2, LL2PerformOperationDefsTypeSafeLemma
      (*********************************************************************)
      (* We hide the definitions.                                          *)
      (*********************************************************************)
      <3> HIDE DEF ll2HistorySummaryHash, ll2StateHash, ll2HistoryStateBinding, ll2PrivateState,
              ll2SResult, ll2NewPrivateStateEnc, ll2CurrentHistorySummary, ll2NewHistorySummary,
              ll2NewHistorySummaryHash, ll2NewStateHash, ll2NewHistoryStateBinding, ll2NewAuthenticator
      (*********************************************************************)
      (* One fact that will be needed many places is that the input is in  *)
      (* the `^\basicname^' set of available inputs.                       *)
      (*********************************************************************)
      <3>4. input \in LL1AvailableInputs
        <4>1. input \in LL2AvailableInputs
          BY <3>2
        <4>2. LL1AvailableInputs = LL2AvailableInputs
          BY <2>1 DEF LL2Refinement
        <4>3. QED
          BY <4>1, <4>2
      (*********************************************************************)
      (* We re-state the definitions from LL1PerformOperation.             *)
      (*********************************************************************)
      <3> ll1StateHash == Hash(LL1RAM.publicState, LL1RAM.privateStateEnc)
      <3> ll1HistoryStateBinding == Hash(LL1RAM.historySummary, ll1StateHash)
      <3> ll1PrivateState == SymmetricDecrypt(LL1NVRAM.symmetricKey, LL1RAM.privateStateEnc)
      <3> ll1sResult == Service(LL1RAM.publicState, ll1PrivateState, input)
      <3> ll1NewPrivateStateEnc ==
              SymmetricEncrypt(LL1NVRAM.symmetricKey, ll1sResult.newPrivateState)
      <3> ll1NewHistorySummary == Hash(LL1NVRAM.historySummary, input)
      <3> ll1NewStateHash == Hash(ll1sResult.newPublicState, ll1NewPrivateStateEnc)
      <3> ll1NewHistoryStateBinding == Hash(ll1NewHistorySummary, ll1NewStateHash)
      <3> ll1NewAuthenticator == GenerateMAC(LL1NVRAM.symmetricKey, ll1NewHistoryStateBinding)
      (*********************************************************************)
      (* We prove that the definitions from LL1PerformOperation satisfy    *)
      (* their types, using the LL1PerformOperationDefsTypeSafeLemma and   *)
      (* the TypeSafetyRefinementLemma.                                    *)
      (*********************************************************************)
      <3>5. /\  ll1StateHash \in HashType
            /\  ll1HistoryStateBinding \in HashType
            /\  ll1PrivateState \in PrivateStateType
            /\  ll1sResult \in ServiceResultType
            /\  ll1sResult.newPublicState \in PublicStateType
            /\  ll1sResult.newPrivateState \in PrivateStateType
            /\  ll1sResult.output \in OutputType
            /\  ll1NewPrivateStateEnc \in PrivateStateEncType
            /\  ll1NewHistorySummary \in HashType
            /\  ll1NewStateHash \in HashType
            /\  ll1NewHistoryStateBinding \in HashType
            /\  ll1NewAuthenticator \in MACType
        <4>1. input \in LL1AvailableInputs
          BY <3>4
        <4>2. LL1TypeInvariant
          BY <2>1, TypeSafetyRefinementLemma
        <4>3. QED
          BY <4>1, <4>2, LL1PerformOperationDefsTypeSafeLemma
      (*********************************************************************)
      (* We hide the definitions from LL1PerformOperation.                 *)
      (*********************************************************************)
      <3> HIDE DEF ll1StateHash, ll1HistoryStateBinding, ll1PrivateState, ll1sResult, ll1NewPrivateStateEnc,
              ll1NewHistorySummary, ll1NewStateHash, ll1NewHistoryStateBinding, ll1NewAuthenticator
      (*********************************************************************)
      (* We prove the correspondences between the definitions in           *)
      (* LL1PerformOperation and LL2PerformOperation.  The state hashes    *)
      (* are directly equal.                                               *)
      (*********************************************************************)
      <3>6. ll1StateHash = ll2StateHash
        <4>1. LL1RAM.publicState = LL2RAM.publicState
          BY <2>1 DEF LL2Refinement
        <4>2. LL1RAM.privateStateEnc = LL2RAM.privateStateEnc
          BY <2>1 DEF LL2Refinement
        <4>3. QED
          BY <4>1, <4>2 DEF ll1StateHash, ll2StateHash
      (*********************************************************************)
      (* The private states are directly equal across the two specs.       *)
      (*********************************************************************)
      <3>7. ll1PrivateState = ll2PrivateState
        <4>1. LL1NVRAM.symmetricKey = LL2NVRAM.symmetricKey
          BY <2>1 DEF LL2Refinement
        <4>2. LL1RAM.privateStateEnc = LL2RAM.privateStateEnc
          BY <2>1 DEF LL2Refinement
        <4>3. QED
          BY <4>1, <4>2 DEF ll1PrivateState, ll2PrivateState
      (*********************************************************************)
      (* The service results are directly equal across the two specs.      *)
      (*********************************************************************)
      <3>8. ll1sResult = ll2SResult
        <4>1. LL1RAM.publicState = LL2RAM.publicState
          BY <2>1 DEF LL2Refinement
        <4>2. ll1PrivateState = ll2PrivateState
          BY <3>7
        <4>3. QED
          BY <4>1, <4>2 DEF ll1sResult, ll2SResult
      (*********************************************************************)
      (* The new encrypted private states are directly equal across the    *)
      (* two specs.                                                        *)
      (*********************************************************************)
      <3>9. ll1NewPrivateStateEnc = ll2NewPrivateStateEnc
        <4>1. LL1NVRAM.symmetricKey = LL2NVRAM.symmetricKey
          BY <2>1 DEF LL2Refinement
        <4>2. ll1sResult.newPrivateState = ll2SResult.newPrivateState
          BY <3>8
        <4>3. QED
          BY <4>1, <4>2 DEF ll1NewPrivateStateEnc, ll2NewPrivateStateEnc
      (*********************************************************************)
      (* The new history summaries match across the two specs, as          *)
      (* specified by the HistorySummariesMatch predicate.                 *)
      (*********************************************************************)
      <3>10. HistorySummariesMatch(
                 ll1NewHistorySummary, ll2NewHistorySummary, LL2NVRAM.hashBarrier')
        (*******************************************************************)
        (* First, we prove that the HistorySummariesMatch predicate equals *)
        (* the HistorySummariesMatchRecursion predicate in this case.  We  *)
        (* assert each condition required by the definition of the         *)
        (* predicate.                                                      *)
        (*******************************************************************)
        <4>1. HistorySummariesMatch(
                  ll1NewHistorySummary, ll2NewHistorySummary, LL2NVRAM.hashBarrier') <=>  (** Changed = by <=> . Zenon also fails because of this. *)
              HistorySummariesMatchRecursion(
                  ll1NewHistorySummary, ll2NewHistorySummary, LL2NVRAM.hashBarrier')
          (*****************************************************************)
          (* We begin by proving the types for the                         *)
          (* HistorySummariesMatchRecursion predicate.                     *)
          (*****************************************************************)
          <5>1. ll1NewHistorySummary \in HashType
            BY <3>5
          <5>2. ll2NewHistorySummary \in HistorySummaryType
            BY <3>3
          <5>3. LL2NVRAM.hashBarrier' \in HashType
              BY <2>1, LL2SubtypeImplicationLemma DEF LL2SubtypeImplication, LL2TypeInvariant, LL2TrustedStorageType
          <5> ll2InitialHistorySummary == [ anchor |-> BaseHashValue, extension |-> BaseHashValue ]
          (*****************************************************************)
          (* We then prove that this is not the base case for the          *)
          (* HistorySummariesMatch predicate.                              *)
          (*****************************************************************)
          <5>4. ll2NewHistorySummary # ll2InitialHistorySummary
            (***************************************************************)
            (* This proof has a lot of sub-steps, but it is pretty simple. *)
            (* We just use the BaseHashValueUnique property to show that   *)
            (* the extension field in the ll2NewHistorySummary record      *)
            (* cannot match the base hash value, which is the value of the *)
            (* extension field in the initial history summary.             *)
            (***************************************************************)
            <6>1. ll2NewHistorySummary =
                      Successor(ll2CurrentHistorySummary, input, LL2NVRAM.hashBarrier)
              BY DEF ll2NewHistorySummary
            <6>2. ll2NewHistorySummary.extension =
                      Hash(ll2CurrentHistorySummary.extension, Hash(LL2NVRAM.hashBarrier, input))
              BY <6>1 DEF Successor
            <6>3. ll2NewHistorySummary.extension # BaseHashValue
              <7>1. ll2CurrentHistorySummary.extension \in HashDomain
                <8>1. ll2CurrentHistorySummary.extension \in HashType
                  <9>1. ll2CurrentHistorySummary.extension = LL2SPCR
                    BY DEF ll2CurrentHistorySummary
                  <9>2. LL2SPCR \in HashType
                    BY <2>1 DEF LL2TypeInvariant
                  <9>3. QED
                    BY <9>1, <9>2
                <8>2. QED
                  BY <8>1 DEF HashDomain
              <7>2. Hash(LL2NVRAM.hashBarrier, input) \in HashDomain
                <8>1. Hash(LL2NVRAM.hashBarrier, input) \in HashType
                  <9>1. LL2NVRAM.hashBarrier \in HashDomain
                    <10>1. LL2NVRAM.hashBarrier \in HashType
                      BY <2>1, LL2SubtypeImplicationLemma DEF LL2SubtypeImplication
                    <10>2. QED
                      BY <10>1 DEF HashDomain
                  <9>2. input \in HashDomain
                    <10>1. input \in InputType
                      <11>1. input \in LL2AvailableInputs
                        BY <3>2
                      <11>2. LL2AvailableInputs \subseteq InputType
                        BY <2>1 DEF LL2TypeInvariant
                      <11>3. QED
                        BY <11>1, <11>2
                    <10>2. QED
                      BY <10>1 DEF HashDomain
                  <9>3. QED
                    BY <9>1, <9>2, HashTypeSafe
                <8>2. QED
                  BY <8>1 DEF HashDomain
              <7>3. QED
                BY <6>2, <7>1, <7>2, BaseHashValueUnique
            <6>4. QED
              BY <6>3
          (*****************************************************************)
          (* Since this is not the base case, the HistorySummariesMatch    *)
          (* predicate equals the HistorySummariesMatchRecursion           *)
          (* predicate.                                                    *)
          (*****************************************************************)
          <5>5. QED
            BY <5>1, <5>2, <5>3, <5>4, HistorySummariesMatchDefinition
        (*******************************************************************)
        (* Then, we prove that the HistorySummariesMatchRecursion          *)
        (* predicate is satisfied.  We assert each condition required by   *)
        (* the definition of the predicate.                                *)
        (*******************************************************************)
        <4>2. HistorySummariesMatchRecursion(
                  ll1NewHistorySummary, ll2NewHistorySummary, LL2NVRAM.hashBarrier')
          (*****************************************************************)
          (* We begin by proving the types for the existentially           *)
          (* quantified variables in the HistorySummariesMatchRecursion    *)
          (* predicate.                                                    *)
          (*****************************************************************)
          <5>1. input \in InputType
            <6>1. input \in LL2AvailableInputs
              BY <3>2
            <6>2. LL2AvailableInputs \subseteq InputType
              BY <2>1 DEF LL2TypeInvariant
            <6>3. QED
              BY <6>1, <6>2
          <5>2. LL1NVRAM.historySummary \in HashType
            BY <2>1 DEF LL2Refinement, LL1TrustedStorageType
          <5>3. LL2NVRAMLogicalHistorySummary \in HistorySummaryType
            BY <2>1, LL2NVRAMLogicalHistorySummaryTypeSafe
          (*****************************************************************)
          (* We then prove the three conjuncts in the                      *)
          (* HistorySummariesMatchRecursion predicate.  The first conjunct *)
          (* follows directly from the refinement.                         *)
          (*****************************************************************)
          <5>4. HistorySummariesMatch(
                    LL1NVRAM.historySummary,
                    LL2NVRAMLogicalHistorySummary,
                    LL2NVRAM.hashBarrier')
            <6>1. UNCHANGED LL2NVRAM.hashBarrier
              BY <3>2
            <6>2. QED
              BY <2>1, <6>1 DEF LL2Refinement
          (*****************************************************************)
          (* The second conjunct in the HistorySummariesMatchRecursion     *)
          (* predicate is true by definition.                              *)
          (*****************************************************************)
          <5>5. ll1NewHistorySummary = Hash(LL1NVRAM.historySummary, input)
            BY DEF ll1NewHistorySummary
          (*****************************************************************)
          (* The third conjunct in the HistorySummariesMatchRecursion      *)
          (* predicate is true because the ll2NewHistorySummary definition *)
          (* in the LL2PerformOperation action yields a value that is the  *)
          (* successor of the logical history summary in the NVRAM.        *)
          (*****************************************************************)
          <5>6. LL2HistorySummaryIsSuccessor(
                    ll2NewHistorySummary,
                    LL2NVRAMLogicalHistorySummary,
                    input,
                    LL2NVRAM.hashBarrier')
            <6>1. LL2NVRAMLogicalHistorySummary = ll2CurrentHistorySummary
              <7>1. CASE LL2NVRAM.extensionInProgress <=> TRUE    (** changed = by <=> *)
                <8>2. LL2NVRAMLogicalHistorySummary = [
                          anchor |-> LL2NVRAM.historySummaryAnchor,
                          extension |-> LL2SPCR ]
                  <9>1. LL2SPCR # BaseHashValue
                    BY <3>2, <7>1
                  <9>2. QED
                    BY <7>1, <9>1 DEF LL2NVRAMLogicalHistorySummary
                <8>3. ll2CurrentHistorySummary = [
                          anchor |-> LL2NVRAM.historySummaryAnchor,
                          extension |-> LL2SPCR ]
                  BY DEF ll2CurrentHistorySummary
                <8>4. QED
                  BY <8>2, <8>3
              <7>2. CASE LL2NVRAM.extensionInProgress <=> FALSE
                <8>1. LL2NVRAMLogicalHistorySummary = [
                          anchor |-> LL2NVRAM.historySummaryAnchor,
                          extension |-> BaseHashValue ]
                  BY <7>2 DEF LL2NVRAMLogicalHistorySummary
                <8>2. ll2CurrentHistorySummary = [
                          anchor |-> LL2NVRAM.historySummaryAnchor,
                          extension |-> LL2SPCR ]
                  BY DEF ll2CurrentHistorySummary
                <8>3. LL2SPCR = BaseHashValue
                  BY <3>2, <7>2
                <8>4. QED
                  BY <8>1, <8>2, <8>3
              <7>3. LL2NVRAM.extensionInProgress \in BOOLEAN
                BY <2>1, LL2SubtypeImplicationLemma DEF LL2SubtypeImplication
              <7>4. QED
                BY <7>1, <7>2, <7>3
            <6>2. ll2NewHistorySummary =
                      Successor(LL2NVRAMLogicalHistorySummary, input, LL2NVRAM.hashBarrier)
              BY <6>1 DEF ll2NewHistorySummary
            <6>3. UNCHANGED LL2NVRAM.hashBarrier
              BY <3>2
            <6>4. QED
              BY <6>2, <6>3 DEF LL2HistorySummaryIsSuccessor
          (*****************************************************************)
          (* All conjuncts in the HistorySummariesMatchRecursion predicate *)
          (* are satisfied.                                                *)
          (*****************************************************************)
          <5>7. QED
            BY <5>1, <5>2, <5>3, <5>4, <5>5, <5>6 DEF HistorySummariesMatchRecursion
          (*****************************************************************)
          (* Since the HistorySummariesMatch predicate equals the          *)
          (* HistorySummariesMatchRecursion predicate, and the latter      *)
          (* predicate is satisfied, the former predicate is satisfied.    *)
          (*****************************************************************)
        <4>3. QED
          BY <4>1, <4>2
      (*********************************************************************)
      (* The new state hashes are directly equal across the two specs.     *)
      (*********************************************************************)
      <3>11. ll1NewStateHash = ll2NewStateHash
        <4>1. ll1sResult.newPublicState = ll2SResult.newPublicState
          BY <3>8
        <4>2. ll1NewPrivateStateEnc = ll2NewPrivateStateEnc
          BY <3>9
        <4>3. QED
          BY <4>1, <4>2 DEF ll1NewStateHash, ll2NewStateHash
      (*********************************************************************)
      (* The new authenticators match across the two specs, as specified   *)
      (* by the AuthenticatorsMatch predicate.                             *)
      (*********************************************************************)
      <3>12. AuthenticatorsMatch(
                ll1NewAuthenticator,
                ll2NewAuthenticator,
                LL2NVRAM.symmetricKey',
                LL2NVRAM.hashBarrier')
        (*******************************************************************)
        (* First, we prove some types needed by the definition of the      *)
        (* AuthenticatorsMatch predicate.                                  *)
        (*******************************************************************)
        <4>1. ll2NewStateHash \in HashType
          BY <3>3
        <4>2. ll1NewHistorySummary \in HashType
          BY <3>5
        <4>3. ll2NewHistorySummary \in HistorySummaryType
          BY <3>3
        (*******************************************************************)
        (* We then prove that, in the `^\fastname^' spec, the new          *)
        (* authenticator is a valid MAC for the new history state binding. *)
        (* We will use the MACComplete property.                           *)
        (*******************************************************************)
        <4>4. ValidateMAC(LL2NVRAM.symmetricKey', ll2NewHistoryStateBinding, ll2NewAuthenticator)
          (*****************************************************************)
          (* In the `^\fastname^' spec, the new authenticator is generated *)
          (* as a MAC of the new history state binding.                    *)
          (*****************************************************************)
          <5>1. ll2NewAuthenticator =
                    GenerateMAC(LL2NVRAM.symmetricKey', ll2NewHistoryStateBinding)
            <6>1. UNCHANGED LL2NVRAM.symmetricKey
              BY <3>2
            <6>2. QED
              BY <6>1 DEF ll2NewAuthenticator
          (*****************************************************************)
          (* We can thus use the MACComplete property to show that the     *)
          (* generated MAC validates appropriately.  To do this, we first  *)
          (* need to prove some types.                                     *)
          (*****************************************************************)
          <5>2. LL2NVRAM.symmetricKey' \in SymmetricKeyType
            BY <2>1, LL2SubtypeImplicationLemma DEF LL2SubtypeImplication
          <5>3. ll2NewHistoryStateBinding \in HashType
            BY <3>3
          (*****************************************************************)
          (* Then, we appeal to the MACComplete property in a              *)
          (* straightforward way.                                          *)
          (*****************************************************************)
          <5>4. QED
            BY <5>1, <5>2, <5>3, MACComplete
        (*******************************************************************)
        (* We then prove that, in the `^\basicname^' spec, the new         *)
        (* authenticator is generated as a MAC of the new history state    *)
        (* binding.                                                        *)
        (*******************************************************************)
        <4>5. ll1NewAuthenticator = GenerateMAC(LL2NVRAM.symmetricKey', ll1NewHistoryStateBinding)
          <5>1. LL1NVRAM.symmetricKey = LL2NVRAM.symmetricKey'
            <6>1. LL1NVRAM.symmetricKey = LL2NVRAM.symmetricKey
              BY <2>1 DEF LL2Refinement
            <6>2. UNCHANGED LL2NVRAM.symmetricKey
              BY <3>2
            <6>3. QED
              BY <6>1, <6>2
          <5>2. QED
            BY <5>1 DEF ll1NewAuthenticator
        (*******************************************************************)
        (* The new history summaries match across the two specs, as we     *)
        (* proved above.                                                   *)
        (*******************************************************************)
        <4>6. HistorySummariesMatch(
                  ll1NewHistorySummary, ll2NewHistorySummary, LL2NVRAM.hashBarrier')
          BY <3>10
        (*******************************************************************)
        (* We then invoke the definition of the AuthenticatorsMatch        *)
        (* predicate.                                                      *)
        (*******************************************************************)
        <4>7. QED
          BY <3>11, <4>1, <4>2, <4>3, <4>4, <4>5, <4>6
          DEF AuthenticatorsMatch, ll1NewHistoryStateBinding,
              ll2NewHistorySummaryHash, ll2NewHistoryStateBinding
      (*********************************************************************)
      (* The remainder of the proof for LL2PerformOperation is a series of *)
      (* assertions, one for each conjunct in the definition of the        *)
      (* LL1PerformOperation action.                                       *)
      (*                                                                   *)
      (* The first conjunct in LL1PerformOperation.  This is basically     *)
      (* just an application of AuthenticatorValidatedLemma                *)
      (*********************************************************************)
      <3>13. ValidateMAC(LL1NVRAM.symmetricKey, ll1HistoryStateBinding, LL1RAM.authenticator)
        (*******************************************************************)
        (* We need the fact that the symmetric keys in the NVRAM are equal *)
        (* across the two specs.                                           *)
        (*******************************************************************)
        <4>1. LL1NVRAM.symmetricKey = LL2NVRAM.symmetricKey
          BY <2>1 DEF LL2Refinement
        (*******************************************************************)
        (* We prove the types that are needed for                          *)
        (* AuthenticatorValidatedLemma.                                    *)
        (*******************************************************************)
        <4>2. ll2StateHash \in HashType
          BY <3>3
        <4>3. LL1RAM.historySummary \in HashType
          BY <2>1 DEF LL2Refinement, LL1UntrustedStorageType
        <4>4. LL2RAM.historySummary \in HistorySummaryType
          BY <2>1, LL2SubtypeImplicationLemma DEF LL2SubtypeImplication
        <4>5. LL1RAM.authenticator \in MACType
          BY <2>1 DEF LL2Refinement, LL1UntrustedStorageType
        <4>6. LL2RAM.authenticator \in MACType
          BY <2>1, LL2SubtypeImplicationLemma DEF LL2SubtypeImplication
        <4>7. LL2NVRAM.symmetricKey \in SymmetricKeyType
          BY <2>1, LL2SubtypeImplicationLemma DEF LL2SubtypeImplication
        <4>8. LL2NVRAM.hashBarrier \in HashType
          BY <2>1, LL2SubtypeImplicationLemma DEF LL2SubtypeImplication
        (*******************************************************************)
        (* There are three preconditions for AuthenticatorValidatedLemma.  *)
        (* The first precondition follows from the refinement.             *)
        (*******************************************************************)
        <4>9. HistorySummariesMatch(
                  LL1RAM.historySummary, LL2RAM.historySummary, LL2NVRAM.hashBarrier)
          BY <2>1 DEF LL2Refinement
        (*******************************************************************)
        (* The second precondition also follows from the refinement.       *)
        (*******************************************************************)
        <4>10. PICK symmetricKey \in SymmetricKeyType :
                  AuthenticatorsMatch(
                      LL1RAM.authenticator,
                      LL2RAM.authenticator,
                      symmetricKey,
                      LL2NVRAM.hashBarrier)
          BY <2>1 DEF LL2Refinement
        (*******************************************************************)
        (* The third precondition follows from the definition of the       *)
        (* LL2PerformOperation action.                                     *)
        (*******************************************************************)
        <4>11. ValidateMAC(LL2NVRAM.symmetricKey, ll2HistoryStateBinding, LL2RAM.authenticator)
          BY <3>2 DEF ll2HistoryStateBinding, ll2StateHash, ll2HistorySummaryHash
        <4>12. QED
          (*****************************************************************)
          (* Ideally, this QED step should just read:                      *)
          (*                                                               *)
          (*  BY <<3>>6, <<4>>1, <<4>>2, <<4>>3, <<4>>4, <<4>>5, <<4>>6, <<4>>7, <<4>>8, <<4>>9, <<4>>10, <<4>>11, *)
          (*      AuthenticatorValidatedLemma                              *)
          (*  DEF ll1HistoryStateBinding, ll2HistoryStateBinding, ll2HistorySummaryHash *)
          (*                                                               *)
          (* However, the prover seems to get a little confused in this    *)
          (* instance.  We make life easier for the prover by explicitly   *)
          (* staging the instantiation of the quantified variables within  *)
          (* the definition of AuthenticatorValidatedLemma.                *)
          (*****************************************************************)
          <5>1. \A  samLL2Authenticator \in MACType,
                    samSymmetricKey2 \in SymmetricKeyType,
                    samHashBarrier \in HashType :
                    AuthenticatorValidatedLemma!(ll2StateHash, LL1RAM.historySummary,
                        LL2RAM.historySummary, LL1RAM.authenticator, samLL2Authenticator,
                        symmetricKey, samSymmetricKey2, samHashBarrier)!1
            BY <4>2, <4>3, <4>4, <4>5, AuthenticatorValidatedLemma
          <5>2. AuthenticatorValidatedLemma!(ll2StateHash, LL1RAM.historySummary,
                    LL2RAM.historySummary, LL1RAM.authenticator, LL2RAM.authenticator,
                    symmetricKey, LL2NVRAM.symmetricKey, LL2NVRAM.hashBarrier)!1
            BY <4>2, <4>3, <4>4, <4>5, <4>6, <4>7, <4>8, <4>10, <5>1
          <5>3. QED
            BY <3>6, <4>1, <4>2, <4>3, <4>4, <4>5, <4>6, <4>7, <4>8, <4>9, <4>10, <4>11, <5>2
            DEF ll1HistoryStateBinding, ll2HistoryStateBinding, ll2HistorySummaryHash
      (*********************************************************************)
      (* The second conjunct in LL1PerformOperation: In the `^\basicname^' *)
      (* spec, the history summary in the NVRAM equals the history summary *)
      (* in the RAM.                                                       *)
      (*********************************************************************)
      <3>14. LL1NVRAM.historySummary = LL1RAM.historySummary
        (*******************************************************************)
        (* The history summaries in the NVRAM match across the two specs.  *)
        (*******************************************************************)
        <4>1. HistorySummariesMatch(
                  LL1NVRAM.historySummary,
                  LL2NVRAMLogicalHistorySummary,
                  LL2NVRAM.hashBarrier)
          BY <2>1 DEF LL2Refinement
        (*******************************************************************)
        (* The history summaries in the RAM match across the two specs.    *)
        (*******************************************************************)
        <4>2. HistorySummariesMatch(
                  LL1RAM.historySummary, LL2RAM.historySummary, LL2NVRAM.hashBarrier)
          BY <2>1 DEF LL2Refinement
        (*******************************************************************)
        (* In the `^\fastname^' spec, we separately handle the two cases   *)
        (* of whether an extension is in progress.  If an extension is in  *)
        (* progress, then the logical history summary in the NVRAM equals  *)
        (* the history summary in the RAM, so we can use the               *)
        (* HistorySummariesMatchUniqueLemma.                               *)
        (*******************************************************************)
        <4>3. CASE LL2NVRAM.extensionInProgress
          <5>1. LL2NVRAMLogicalHistorySummary = LL2RAM.historySummary
            <6>1. LL2NVRAMLogicalHistorySummary = ll2CurrentHistorySummary
              <7>1. LL2NVRAMLogicalHistorySummary = [
                        anchor |-> LL2NVRAM.historySummaryAnchor,
                        extension |-> LL2SPCR ]
                <8>1. LL2SPCR # BaseHashValue
                  BY <3>2, <4>3
                <8>2. QED
                  BY <4>3, <8>1 DEF LL2NVRAMLogicalHistorySummary
              <7>2. ll2CurrentHistorySummary = [
                        anchor |-> LL2NVRAM.historySummaryAnchor,
                        extension |-> LL2SPCR ]
                BY DEF ll2CurrentHistorySummary
              <7>3. QED
                BY <7>1, <7>2
            <6>2. ll2CurrentHistorySummary = LL2RAM.historySummary
              BY <3>2, <4>3 DEF ll2CurrentHistorySummary
            <6>3. QED
              BY <6>1, <6>2
          (*****************************************************************)
          (* We use the HistorySummariesMatchUniqueLemma to prove that the *)
          (* fields are equal.  This requires proving some types.          *)
          (*****************************************************************)
          <5>2. QED
            <6>1. LL1NVRAM.historySummary \in HashType
              BY <2>1 DEF LL2Refinement, LL1TrustedStorageType
            <6>2. LL1RAM.historySummary \in HashType
              BY <2>1 DEF LL2Refinement, LL1UntrustedStorageType
            <6>3. LL2RAM.historySummary \in HistorySummaryType
              BY <2>1, LL2SubtypeImplicationLemma DEF LL2SubtypeImplication
            <6>4. LL2NVRAM.hashBarrier \in HashType
              BY <2>1, LL2SubtypeImplicationLemma DEF LL2SubtypeImplication
            <6>5. QED
              BY <4>1, <4>2, <5>1, <6>1, <6>2, <6>3, <6>4, HistorySummariesMatchUniqueLemma
        (*******************************************************************)
        (* If an extension is not in progress, then the logical history    *)
        (* summary in the NVRAM is a checkpoint of the history summary in  *)
        (* the RAM, so we can use the                                      *)
        (* HistorySummariesMatchAcrossCheckpointLemma.                     *)
        (*******************************************************************)
        <4>4. CASE ~LL2NVRAM.extensionInProgress
          <5>1. LL2NVRAMLogicalHistorySummary = Checkpoint(LL2RAM.historySummary)
            <6>1. LL2NVRAMLogicalHistorySummary = ll2CurrentHistorySummary
              <7>1. LL2NVRAMLogicalHistorySummary = [
                        anchor |-> LL2NVRAM.historySummaryAnchor,
                        extension |-> BaseHashValue ]
                BY <4>4 DEF LL2NVRAMLogicalHistorySummary
              <7>2. ll2CurrentHistorySummary = [
                        anchor |-> LL2NVRAM.historySummaryAnchor,
                        extension |-> LL2SPCR ]
                BY DEF ll2CurrentHistorySummary
              <7>3. LL2SPCR = BaseHashValue
                BY <3>2, <4>4
              <7>4. QED
                BY <7>1, <7>2, <7>3
            <6>2. ll2CurrentHistorySummary = Checkpoint(LL2RAM.historySummary)
              BY <3>2, <4>4 DEF ll2CurrentHistorySummary
            <6>3. QED
              BY <6>1, <6>2
          (*****************************************************************)
          (* We use the HistorySummariesMatchAcrossCheckpointLemma to      *)
          (* prove that the fields are equal.  This requires proving some  *)
          (* types.                                                        *)
          (*****************************************************************)
          <5>2. QED
            <6>1. LL1NVRAM.historySummary \in HashType
              BY <2>1 DEF LL2Refinement, LL1TrustedStorageType
            <6>2. LL1RAM.historySummary \in HashType
              BY <2>1 DEF LL2Refinement, LL1UntrustedStorageType
            <6>3. LL2NVRAMLogicalHistorySummary \in HistorySummaryType
              BY <2>1, LL2NVRAMLogicalHistorySummaryTypeSafe
            <6>4. LL2RAM.historySummary \in HistorySummaryType
              BY <2>1, LL2SubtypeImplicationLemma DEF LL2SubtypeImplication
            <6>5. LL2NVRAM.hashBarrier \in HashType
              BY <2>1, LL2SubtypeImplicationLemma DEF LL2SubtypeImplication
            <6>6. QED
              BY <4>1, <4>2, <5>1, <6>1, <6>2, <6>3, <6>4, <6>5, HistorySummariesMatchAcrossCheckpointLemma
        <4>5. QED
          BY <4>3, <4>4
      (*********************************************************************)
      (* The third conjunct in LL1PerformOperation.  This conjunct asserts *)
      (* a record equality, so we prove each field in the record           *)
      (* separately.                                                       *)
      (*********************************************************************)
      <3>15. LL1RAM' = [
                publicState |-> ll1sResult.newPublicState,
                privateStateEnc |-> ll1NewPrivateStateEnc,
                historySummary |-> ll1NewHistorySummary,
                authenticator |-> ll1NewAuthenticator ]
        (*******************************************************************)
        (* The public state field in the primed `^\basicname^' RAM equals  *)
        (* the public state in the result of the service, because the      *)
        (* primed RAM variables match across the two specs.                *)
        (*******************************************************************)
        <4>1. LL1RAM.publicState' = ll1sResult.newPublicState
          <5>1. LL1RAM.publicState' = LL2RAM.publicState'
            BY <2>1 DEF LL2Refinement
          <5>2. LL2RAM.publicState' = ll2SResult.newPublicState
            BY <3>2 DEF ll2SResult, ll2PrivateState
          <5>3. ll1sResult.newPublicState = ll2SResult.newPublicState
            BY <3>8
          <5>4. QED
            BY <5>1, <5>2, <5>3
        (*******************************************************************)
        (* The encrypted private state field in the primed `^\basicname^'  *)
        (* RAM equals the encrypted private state in the result of the     *)
        (* service, because the primed RAM variables match across the two  *)
        (* specs.                                                          *)
        (*******************************************************************)
        <4>2. LL1RAM.privateStateEnc' = ll1NewPrivateStateEnc
          <5>1. LL1RAM.privateStateEnc' = LL2RAM.privateStateEnc'
            BY <2>1 DEF LL2Refinement
          <5>2. LL2RAM.privateStateEnc' = ll2NewPrivateStateEnc
            BY <3>2 DEF ll2NewPrivateStateEnc, ll2SResult, ll2PrivateState
          <5>3. ll1NewPrivateStateEnc = ll2NewPrivateStateEnc
            BY <3>9
          <5>4. QED
            BY <5>1, <5>2, <5>3
        (*******************************************************************)
        (* The history summary field in the primed `^\basicname^' RAM      *)
        (* equals the new history summary defined in the                   *)
        (* LL1PerformOperation action, because the history summaries in    *)
        (* the RAM variables match across the specs by refinement.  We use *)
        (* the HistorySummariesMatchUniqueLemma to prove the equality.     *)
        (*******************************************************************)
        <4>3. LL1RAM.historySummary' = ll1NewHistorySummary
          <5>1. HistorySummariesMatch(
                    LL1RAM.historySummary', LL2RAM.historySummary', LL2NVRAM.hashBarrier')
            BY <2>1 DEF LL2Refinement
          <5>2. LL2RAM.historySummary' = ll2NewHistorySummary
            BY <3>2 DEF ll2NewHistorySummary, ll2CurrentHistorySummary
          <5>3. HistorySummariesMatch(
                    ll1NewHistorySummary, ll2NewHistorySummary, LL2NVRAM.hashBarrier')
            BY <3>10
          (*****************************************************************)
          (* We use the HistorySummariesMatchUniqueLemma to prove that the *)
          (* fields are equal.  This requires proving some types.          *)
          (*****************************************************************)
          <5>4. QED
            <6>1. LL1RAM.historySummary' \in HashType
              BY <2>1 DEF LL2Refinement, LL1UntrustedStorageType
            <6>2. ll1NewHistorySummary \in HashType
              BY <3>5
            <6>3. LL2RAM.historySummary' \in HistorySummaryType
              BY <2>1, LL2SubtypeImplicationLemma DEF LL2SubtypeImplication
            <6>4. LL2NVRAM.hashBarrier' \in HashType
              BY <2>1, LL2SubtypeImplicationLemma DEF LL2SubtypeImplication, LL2TypeInvariant, LL2TrustedStorageType
            <6>5. QED
              BY <5>1, <5>2, <5>3, <6>1, <6>2, <6>3, <6>4, HistorySummariesMatchUniqueLemma
        (*******************************************************************)
        (* The authenticator field in the primed `^\basicname^' RAM equals *)
        (* the new authenticator defined in the LL1PerformOperation        *)
        (* action, because the authenticators in the RAM variables match   *)
        (* across the specs by refinement.  We use the                     *)
        (* AuthenticatorsMatchUniqueLemma to prove the equality.           *)
        (*******************************************************************)
        <4>4. LL1RAM.authenticator' = ll1NewAuthenticator
          <5>1. PICK symmetricKey \in SymmetricKeyType :
                    AuthenticatorsMatch(
                        LL1RAM.authenticator',
                        LL2RAM.authenticator',
                        symmetricKey,
                        LL2NVRAM.hashBarrier')
            BY <2>1 DEF LL2Refinement
          <5>2. LL2RAM.authenticator' = ll2NewAuthenticator
            BY <3>2
            DEF ll2NewAuthenticator, ll2NewHistoryStateBinding, ll2NewHistorySummaryHash,
                ll2NewStateHash, ll2NewPrivateStateEnc, ll2SResult, ll2PrivateState,
                ll2NewHistorySummary, ll2CurrentHistorySummary
          <5>3. AuthenticatorsMatch(
                    ll1NewAuthenticator,
                    ll2NewAuthenticator,
                    LL2NVRAM.symmetricKey',
                    LL2NVRAM.hashBarrier')
            BY <3>12
          (*****************************************************************)
          (* We use the AuthenticatorsMatchUniqueLemma to prove that the   *)
          (* fields are equal.  This requires proving some types.          *)
          (*****************************************************************)
          <5>4. QED
            <6>1. LL1RAM.authenticator' \in MACType
              BY <2>1 DEF LL2Refinement, LL1UntrustedStorageType
            <6>2. ll1NewAuthenticator \in MACType
              BY <3>5
            <6>3. LL2RAM.authenticator' \in MACType
              BY <2>1, LL2SubtypeImplicationLemma DEF LL2SubtypeImplication
            <6>4. symmetricKey \in SymmetricKeyType
              BY <5>1
            <6>5. LL2NVRAM.symmetricKey' \in SymmetricKeyType
              BY <2>1, LL2SubtypeImplicationLemma DEF LL2SubtypeImplication
            <6>6. LL2NVRAM.hashBarrier' \in HashType
              BY <2>1, LL2SubtypeImplicationLemma DEF LL2SubtypeImplication, LL2TypeInvariant, LL2TrustedStorageType
            <6>7. QED
              BY <5>1, <5>2, <5>3, <6>1, <6>2, <6>3, <6>4, <6>5, <6>6,
                  AuthenticatorsMatchUniqueLemma
        (*******************************************************************)
        (* The refinement asserts that the RAM record has the appropriate  *)
        (* type.                                                           *)
        (*******************************************************************)
        <4>5. LL1RAM' \in LL1UntrustedStorageType
          BY <2>1 DEF LL2Refinement
        (*******************************************************************)
        (* We use the LL1RAMRecordCompositionLemma to unify the field      *)
        (* equalities into a record equality.                              *)
        (*******************************************************************)
        <4>6. QED
          BY <4>1, <4>2, <4>3, <4>4, <4>5, LL1RAMRecordCompositionLemma
      (*********************************************************************)
      (* The fourth conjunct in LL1PerformOperation.  This conjunct        *)
      (* asserts a record equality, so we prove each field in the record   *)
      (* separately.                                                       *)
      (*********************************************************************)
      <3>16. LL1NVRAM' = [
                historySummary |-> ll1NewHistorySummary,
                symmetricKey |-> LL1NVRAM.symmetricKey ]
        (*******************************************************************)
        (* The history summary field in the primed `^\basicname^' NVRAM    *)
        (* equals the new history summary defined in the                   *)
        (* LL1PerformOperation action, because the logical history summary *)
        (* in the `^\fastname^' NVRAM and SPCR matches the history summary *)
        (* in the `^\basicname^' NVRAM by refinement.  We use the          *)
        (* HistorySummariesMatchUniqueLemma to prove the equality.         *)
        (*******************************************************************)
        <4>1. LL1NVRAM.historySummary' = ll1NewHistorySummary
          <5>1. HistorySummariesMatch(
                    LL1NVRAM.historySummary',
                    LL2NVRAMLogicalHistorySummary',
                    LL2NVRAM.hashBarrier')
            BY <2>1 DEF LL2Refinement
          (*****************************************************************)
          (* In the `^\fastname^' spec, the logical history summary equals *)
          (* the new history summary defined in the LL2PerformOperation    *)
          (* action.  Proving this is slightly involved, because we have   *)
          (* to show that the updates performed by LL2PerformOperation are *)
          (* logically equivalent to the operation of the Successor        *)
          (* operator.                                                     *)
          (*****************************************************************)
          <5>2. LL2NVRAMLogicalHistorySummary' = ll2NewHistorySummary
            <6>1. LL2NVRAMLogicalHistorySummary' = [
                      anchor |-> LL2NVRAM.historySummaryAnchor',
                      extension |-> LL2SPCR' ]
              <7>1. LL2NVRAM.extensionInProgress' <=> TRUE (** changed = by <=> *)
                BY <3>2
              <7>2. LL2SPCR' # BaseHashValue
                <8>1. LL2SPCR' = ll2NewHistorySummary.extension
                  BY <3>2 DEF ll2NewHistorySummary, ll2CurrentHistorySummary
                <8>2. ll2NewHistorySummary.extension # BaseHashValue
                  <9>1. ll2CurrentHistorySummary \in HistorySummaryType
                    BY <3>3
                  <9>2. input \in InputType
                    <10>1. input \in LL2AvailableInputs
                      BY <3>2
                    <10>2. LL2AvailableInputs \subseteq InputType
                      BY <2>1 DEF LL2TypeInvariant
                    <10>3. QED
                      BY <10>1, <10>2
                  <9>3. LL2NVRAM.hashBarrier \in HashType
                    BY <2>1, LL2SubtypeImplicationLemma DEF LL2SubtypeImplication
                  <9>4. QED
                    BY <9>1, <9>2, <9>3, SuccessorHasNonBaseExtensionLemma
                    DEF ll2NewHistorySummary
                <8>3. QED
                  BY <8>1, <8>2
              <7>3. QED
                BY <7>1, <7>2 DEF LL2NVRAMLogicalHistorySummary
            <6>2. ll2NewHistorySummary = [
                      anchor |-> LL2NVRAM.historySummaryAnchor',
                      extension |-> LL2SPCR' ]
              <7>1. ll2NewHistorySummary.anchor = LL2NVRAM.historySummaryAnchor'
                <8>1. ll2NewHistorySummary =
                          Successor(ll2CurrentHistorySummary, input, LL2NVRAM.hashBarrier)
                  BY DEF ll2NewHistorySummary
                <8>2. ll2NewHistorySummary.anchor = ll2CurrentHistorySummary.anchor
                  BY <8>1 DEF Successor
                <8>3. ll2NewHistorySummary.anchor = LL2NVRAM.historySummaryAnchor
                  BY <8>2 DEF ll2CurrentHistorySummary
                <8>4. UNCHANGED LL2NVRAM.historySummaryAnchor
                  BY <3>2
                <8>5. QED
                  BY <8>3, <8>4
              <7>2. ll2NewHistorySummary.extension = LL2SPCR'
                BY <3>2 DEF ll2NewHistorySummary, ll2CurrentHistorySummary
              <7>3. ll2NewHistorySummary \in HistorySummaryType
                BY <3>3
              <7>4. QED
                BY <7>1, <7>2, <7>3, HistorySummaryRecordCompositionLemma
            <6>3. QED
              BY <6>1, <6>2
          <5>3. HistorySummariesMatch(
                    ll1NewHistorySummary, ll2NewHistorySummary, LL2NVRAM.hashBarrier')
            BY <3>10
          (*****************************************************************)
          (* We use the HistorySummariesMatchUniqueLemma to prove that the *)
          (* fields are equal.  This requires proving some types.          *)
          (*****************************************************************)
          <5>4. QED
            <6>1. LL1NVRAM.historySummary' \in HashType
              BY <2>1 DEF LL2Refinement, LL1TrustedStorageType
            <6>2. ll1NewHistorySummary \in HashType
              BY <3>5
            <6>3. LL2NVRAMLogicalHistorySummary' \in HistorySummaryType
              BY <2>1, LL2NVRAMLogicalHistorySummaryTypeSafe
            <6>4. LL2NVRAM.hashBarrier' \in HashType
              BY <2>1, LL2SubtypeImplicationLemma DEF LL2SubtypeImplication, LL2TypeInvariant, LL2TrustedStorageType
            <6>5. QED
              BY <5>1, <5>2, <5>3, <6>1, <6>2, <6>3, <6>4, HistorySummariesMatchUniqueLemma
        (*******************************************************************)
        (* The symmetric key in the NVRAM is unchanged by the              *)
        (* UnchangedNVRAMSymmetricKeyLemma.                                *)
        (*******************************************************************)
        <4>2. UNCHANGED LL1NVRAM.symmetricKey
          BY <2>1, <3>2, UnchangedNVRAMSymmetricKeyLemma
        (*******************************************************************)
        (* The refinement asserts that the NVRAM record has the            *)
        (* appropriate type.                                               *)
        (*******************************************************************)
        <4>3. LL1NVRAM' \in LL1TrustedStorageType
          BY <2>1 DEF LL2Refinement
        (*******************************************************************)
        (* We use the LL1NVRAMRecordCompositionLemma to unify the field    *)
        (* equalities into a record equality.                              *)
        (*******************************************************************)
        <4>4. QED
          BY <4>1, <4>2, <4>3, LL1NVRAMRecordCompositionLemma
      (*********************************************************************)
      (* The fifth conjunct in LL1PerformOperation.  The set of observed   *)
      (* outputs is equal across the two specs.                            *)
      (*********************************************************************)
      <3>17. LL1ObservedOutputs' = LL1ObservedOutputs \union {ll1sResult.output}
        <4>1. LL1ObservedOutputs = LL2ObservedOutputs
          BY <2>1 DEF LL2Refinement
        <4>2. LL1ObservedOutputs' = LL2ObservedOutputs'
          BY <2>1 DEF LL2Refinement
        <4>3. ll1sResult.output = ll2SResult.output
          BY <3>8
        <4>4. LL2ObservedOutputs' = LL2ObservedOutputs \union {ll2SResult.output}
          BY <3>2 DEF ll2SResult, ll2PrivateState
        <4>5. QED
          BY <4>1, <4>2, <4>3, <4>4
      (*********************************************************************)
      (* The sixth conjunct in LL1PerformOperation.  The disk is unchanged *)
      (* by the UnchangedDiskLemma.                                        *)
      (*********************************************************************)
      <3>18. UNCHANGED LL1Disk
        BY <2>1, <3>2, UnchangedDiskLemma
      (*********************************************************************)
      (* The seventh conjunct in LL1PerformOperation.  The set of          *)
      (* available inputs is unchanged by the                              *)
      (* UnchangedAvailableInputsLemma.                                    *)
      (*********************************************************************)
      <3>19. UNCHANGED LL1AvailableInputs
        BY <2>1, <3>2, UnchangedAvailableInputsLemma
      (*********************************************************************)
      (* The eighth conjunct in LL1PerformOperation.  We prove that the    *)
      (* primed set of observed authenticators matches across the specs,   *)
      (* that the unprimed set of observed authenticators matches across   *)
      (* the specs, and that the new authenticator matches across the      *)
      (* specs.  The AuthenticatorSetsMatchUniqueLemma then proves the     *)
      (* equality.                                                         *)
      (*********************************************************************)
      <3>20. LL1ObservedAuthenticators' =
                 LL1ObservedAuthenticators \union {ll1NewAuthenticator}
        (*******************************************************************)
        (* The primed set of observed authenticators matches across the    *)
        (* specs.  This follows directly from the refinement.              *)
        (*******************************************************************)
        <4>1. AuthenticatorSetsMatch(
            LL1ObservedAuthenticators',
            LL2ObservedAuthenticators',
            LL2NVRAM.symmetricKey',
            LL2NVRAM.hashBarrier')
          BY <2>1 DEF LL2Refinement
        (*******************************************************************)
        (* The union matches across the specs.  We prove this by proving   *)
        (* the matching of each constituent set.                           *)
        (*******************************************************************)
        <4>2. AuthenticatorSetsMatch(
            LL1ObservedAuthenticators \union {ll1NewAuthenticator},
            LL2ObservedAuthenticators \union {ll2NewAuthenticator},
            LL2NVRAM.symmetricKey',
            LL2NVRAM.hashBarrier')
          <5>1. UNCHANGED <<LL2NVRAM.symmetricKey, LL2NVRAM.hashBarrier>>
            BY <3>2
          (*****************************************************************)
          (* The unprimed set of observed authenticators matches across    *)
          (* the specs.  This follows directly from the refinement.        *)
          (*****************************************************************)
          <5>2. AuthenticatorSetsMatch(
                    LL1ObservedAuthenticators,
                    LL2ObservedAuthenticators,
                    LL2NVRAM.symmetricKey',
                    LL2NVRAM.hashBarrier')
              BY <2>1, <5>1 DEF LL2Refinement
          (*****************************************************************)
          (* The new authenticator matches across the specs.  This follows *)
          (* directly from the refinement.                                 *)
          (*****************************************************************)
          <5>3. AuthenticatorsMatch(
                    ll1NewAuthenticator,
                    ll2NewAuthenticator,
                    LL2NVRAM.symmetricKey',
                    LL2NVRAM.hashBarrier')
              BY <2>1, <5>1 DEF LL2Refinement
          <5>4. QED
            BY <5>2, <5>3 DEF AuthenticatorSetsMatch
        (*******************************************************************)
        (* In the `^\fastname^' spec, the primed set of observed           *)
        (* authenticators is formed from the union of the unprimed set of  *)
        (* observed authenticators and the new authenticator.              *)
        (*******************************************************************)
        <4>3. LL2ObservedAuthenticators' =
                    LL2ObservedAuthenticators \union {ll2NewAuthenticator}
          BY <3>2
          DEF ll2NewAuthenticator, ll2NewHistoryStateBinding, ll2NewHistorySummaryHash,
              ll2NewStateHash, ll2NewPrivateStateEnc, ll2SResult, ll2PrivateState,
              ll2NewHistorySummary, ll2CurrentHistorySummary
        <4>4. QED
          (*****************************************************************)
          (* We use the AuthenticatorSetsMatchUniqueLemma to prove that    *)
          (* the sets are equal.  This requires proving some types.        *)
          (*****************************************************************)
          <5>1. LL1ObservedAuthenticators' \in SUBSET MACType
            BY <2>1 DEF LL2Refinement, LL1UntrustedStorageType
          <5>2. LL1ObservedAuthenticators \union {ll1NewAuthenticator} \in
                    SUBSET MACType
            <6>1. LL1ObservedAuthenticators \in SUBSET MACType
              BY <2>1 DEF LL2Refinement, LL1UntrustedStorageType
            <6>2. ll1NewAuthenticator \in MACType
              BY <3>5
            <6>3. QED
              BY <6>1, <6>2
          <5>3. LL2ObservedAuthenticators' \in SUBSET MACType
            BY <2>1 DEF LL2TypeInvariant
          <5>4. LL2NVRAM.symmetricKey' \in SymmetricKeyType
            BY <2>1, LL2SubtypeImplicationLemma DEF LL2SubtypeImplication
          <5>5. LL2NVRAM.hashBarrier' \in HashType
            BY <2>1, LL2SubtypeImplicationLemma DEF LL2SubtypeImplication, LL2TypeInvariant, LL2TrustedStorageType
          <5>6. QED
            BY <4>1, <4>2, <4>3, <5>1, <5>2, <5>3, <5>4, <5>5,
                AuthenticatorSetsMatchUniqueLemma
      <3>21. QED
        BY <3>4, <3>13, <3>14, <3>15, <3>16, <3>17, <3>18, <3>19, <3>20
        DEF LL1PerformOperation, ll1StateHash, ll1HistoryStateBinding, ll1PrivateState,
            ll1sResult, ll1NewPrivateStateEnc, ll1NewHistorySummary, ll1NewStateHash,
            ll1NewHistoryStateBinding, ll1NewAuthenticator
    (***********************************************************************)
    (* A `^\fastname^' LL2RepeatOperation action refines to a              *)
    (* `^\basicname^' LL1RepeatOperation action.                           *)
    (***********************************************************************)
    <2>5. LL2RepeatOperation => LL1RepeatOperation
      (*********************************************************************)
      (* We assume the antecedent.                                         *)
      (*********************************************************************)
      <3>1. HAVE LL2RepeatOperation
      (*********************************************************************)
      (* We pick an input that satisfies the LL2RepeatOperation            *)
      (* predicate.                                                        *)
      (*********************************************************************)
      <3>2. PICK input \in LL2AvailableInputs : LL2RepeatOperation!(input)!1
        BY <3>1 DEF LL2RepeatOperation
      (*********************************************************************)
      (* We re-state the definitions from LL2RepeatOperation.              *)
      (*********************************************************************)
      <3> ll2HistorySummaryHash ==
              Hash(LL2RAM.historySummary.anchor, LL2RAM.historySummary.extension)
      <3> ll2StateHash == Hash(LL2RAM.publicState, LL2RAM.privateStateEnc)
      <3> ll2HistoryStateBinding == Hash(ll2HistorySummaryHash, ll2StateHash)
      <3> ll2NewHistorySummary == Successor(LL2RAM.historySummary, input, LL2NVRAM.hashBarrier)
      <3> ll2CheckpointedHistorySummary == Checkpoint(LL2RAM.historySummary)
      <3> ll2NewCheckpointedHistorySummary ==
              Successor(ll2CheckpointedHistorySummary, input, LL2NVRAM.hashBarrier)
      <3> ll2CheckpointedNewHistorySummary == Checkpoint(ll2NewHistorySummary)
      <3> ll2CheckpointedNewCheckpointedHistorySummary ==
              Checkpoint(ll2NewCheckpointedHistorySummary)
      <3> ll2PrivateState == SymmetricDecrypt(LL2NVRAM.symmetricKey, LL2RAM.privateStateEnc)
      <3> ll2SResult == Service(LL2RAM.publicState, ll2PrivateState, input)
      <3> ll2NewPrivateStateEnc ==
              SymmetricEncrypt(LL2NVRAM.symmetricKey, ll2SResult.newPrivateState)
      <3> ll2CurrentHistorySummary == [
              anchor |-> LL2NVRAM.historySummaryAnchor,
              extension |-> LL2SPCR ]
      <3> ll2CurrentHistorySummaryHash == Hash(LL2NVRAM.historySummaryAnchor, LL2SPCR)
      <3> ll2NewStateHash == Hash(ll2SResult.newPublicState, ll2NewPrivateStateEnc)
      <3> ll2NewHistoryStateBinding == Hash(ll2CurrentHistorySummaryHash, ll2NewStateHash)
      <3> ll2NewAuthenticator == GenerateMAC(LL2NVRAM.symmetricKey, ll2NewHistoryStateBinding)
      (*********************************************************************)
      (* We prove that the definitions from LL2RepeatOperation satisfy     *)
      (* their types, using the LL2RepeatOperationDefsTypeSafeLemma.       *)
      (*********************************************************************)
      <3>3. /\  ll2HistorySummaryHash \in HashType
            /\  ll2StateHash \in HashType
            /\  ll2HistoryStateBinding \in HashType
            /\  ll2NewHistorySummary \in HistorySummaryType
            /\  ll2NewHistorySummary.anchor \in HashType
            /\  ll2NewHistorySummary.extension \in HashType
            /\  ll2CheckpointedHistorySummary \in HistorySummaryType
            /\  ll2CheckpointedHistorySummary.anchor \in HashType
            /\  ll2CheckpointedHistorySummary.extension \in HashType
            /\  ll2NewCheckpointedHistorySummary \in HistorySummaryType
            /\  ll2NewCheckpointedHistorySummary.anchor \in HashType
            /\  ll2NewCheckpointedHistorySummary.extension \in HashType
            /\  ll2CheckpointedNewHistorySummary \in HistorySummaryType
            /\  ll2CheckpointedNewHistorySummary.anchor \in HashType
            /\  ll2CheckpointedNewHistorySummary.extension \in HashType
            /\  ll2CheckpointedNewCheckpointedHistorySummary \in HistorySummaryType
            /\  ll2CheckpointedNewCheckpointedHistorySummary.anchor \in HashType
            /\  ll2CheckpointedNewCheckpointedHistorySummary.extension \in HashType
            /\  ll2PrivateState \in PrivateStateType
            /\  ll2SResult \in ServiceResultType
            /\  ll2SResult.newPublicState \in PublicStateType
            /\  ll2SResult.newPrivateState \in PrivateStateType
            /\  ll2SResult.output \in OutputType
            /\  ll2NewPrivateStateEnc \in PrivateStateEncType
            /\  ll2CurrentHistorySummary \in HistorySummaryType
            /\  ll2CurrentHistorySummary.anchor \in HashType
            /\  ll2CurrentHistorySummary.extension \in HashType
            /\  ll2CurrentHistorySummaryHash \in HashType
            /\  ll2NewStateHash \in HashType
            /\  ll2NewHistoryStateBinding \in HashType
            /\  ll2NewAuthenticator \in MACType
        <4>1. input \in LL2AvailableInputs
          BY <3>2
        <4>2. LL2TypeInvariant
          BY <2>1
        <4>3. QED
          BY <4>1, <4>2, LL2RepeatOperationDefsTypeSafeLemma
      (*********************************************************************)
      (* We hide the definitions.                                          *)
      (*********************************************************************)
      <3> HIDE DEF ll2HistorySummaryHash, ll2StateHash, ll2HistoryStateBinding, ll2NewHistorySummary,
              ll2CheckpointedHistorySummary, ll2NewCheckpointedHistorySummary,
              ll2CheckpointedNewHistorySummary, ll2CheckpointedNewCheckpointedHistorySummary,
              ll2PrivateState, ll2SResult, ll2NewPrivateStateEnc, ll2CurrentHistorySummary,
              ll2CurrentHistorySummaryHash, ll2NewStateHash, ll2NewHistoryStateBinding,
              ll2NewAuthenticator
      (*********************************************************************)
      (* One fact that will be needed many places is that the input is in  *)
      (* the `^\basicname^' set of available inputs.                       *)
      (*********************************************************************)
      <3>4. input \in LL1AvailableInputs
        <4>1. input \in LL2AvailableInputs
          BY <3>2
        <4>2. LL1AvailableInputs = LL2AvailableInputs
          BY <2>1 DEF LL2Refinement
        <4>3. QED
          BY <4>1, <4>2
      (*********************************************************************)
      (* We re-state the definitions from LL1RepeatOperation.              *)
      (*********************************************************************)
      <3> ll1StateHash == Hash(LL1RAM.publicState, LL1RAM.privateStateEnc)
      <3> ll1HistoryStateBinding == Hash(LL1RAM.historySummary, ll1StateHash)
      <3> ll1PrivateState == SymmetricDecrypt(LL1NVRAM.symmetricKey, LL1RAM.privateStateEnc)
      <3> ll1sResult == Service(LL1RAM.publicState, ll1PrivateState, input)
      <3> ll1NewPrivateStateEnc ==
              SymmetricEncrypt(LL1NVRAM.symmetricKey, ll1sResult.newPrivateState)
      <3> ll1NewStateHash == Hash(ll1sResult.newPublicState, ll1NewPrivateStateEnc)
      <3> ll1NewHistoryStateBinding == Hash(LL1NVRAM.historySummary, ll1NewStateHash)
      <3> ll1NewAuthenticator == GenerateMAC(LL1NVRAM.symmetricKey, ll1NewHistoryStateBinding)
      (*********************************************************************)
      (* We prove that the definitions from LL1RepeatOperation satisfy     *)
      (* their types, using the LL1RepeatOperationDefsTypeSafeLemma and    *)
      (* the TypeSafetyRefinementLemma.                                    *)
      (*********************************************************************)
      <3>5. /\  ll1StateHash \in HashType
            /\  ll1HistoryStateBinding \in HashType
            /\  ll1PrivateState \in PrivateStateType
            /\  ll1sResult \in ServiceResultType
            /\  ll1sResult.newPublicState \in PublicStateType
            /\  ll1sResult.newPrivateState \in PrivateStateType
            /\  ll1sResult.output \in OutputType
            /\  ll1NewPrivateStateEnc \in PrivateStateEncType
            /\  ll1NewStateHash \in HashType
            /\  ll1NewHistoryStateBinding \in HashType
            /\  ll1NewAuthenticator \in MACType
        <4>1. input \in LL1AvailableInputs
          BY <3>4
        <4>2. LL1TypeInvariant
          BY <2>1, TypeSafetyRefinementLemma
        <4>3. QED
          BY <4>1, <4>2, LL1RepeatOperationDefsTypeSafeLemma
      (*********************************************************************)
      (* We hide the definitions from LL1RepeatOperation.                  *)
      (*********************************************************************)
      <3> HIDE DEF ll1StateHash, ll1HistoryStateBinding, ll1PrivateState, ll1sResult, ll1NewPrivateStateEnc,
              ll1NewStateHash, ll1NewHistoryStateBinding, ll1NewAuthenticator
      (*********************************************************************)
      (* We prove the correspondences between the definitions in           *)
      (* LL1RepeatOperation and LL2RepeatOperation.  The state hashes are  *)
      (* directly equal.                                                   *)
      (*********************************************************************)
      <3>6. ll1StateHash = ll2StateHash
        <4>1. LL1RAM.publicState = LL2RAM.publicState
          BY <2>1 DEF LL2Refinement
        <4>2. LL1RAM.privateStateEnc = LL2RAM.privateStateEnc
          BY <2>1 DEF LL2Refinement
        <4>3. QED
          BY <4>1, <4>2 DEF ll1StateHash, ll2StateHash
      (*********************************************************************)
      (* The private states are directly equal across the two specs.       *)
      (*********************************************************************)
      <3>7. ll1PrivateState = ll2PrivateState
        <4>1. LL1NVRAM.symmetricKey = LL2NVRAM.symmetricKey
          BY <2>1 DEF LL2Refinement
        <4>2. LL1RAM.privateStateEnc = LL2RAM.privateStateEnc
          BY <2>1 DEF LL2Refinement
        <4>3. QED
          BY <4>1, <4>2 DEF ll1PrivateState, ll2PrivateState
      (*********************************************************************)
      (* The service results are directly equal across the two specs.      *)
      (*********************************************************************)
      <3>8. ll1sResult = ll2SResult
        <4>1. LL1RAM.publicState = LL2RAM.publicState
          BY <2>1 DEF LL2Refinement
        <4>2. ll1PrivateState = ll2PrivateState
          BY <3>7
        <4>3. QED
          BY <4>1, <4>2 DEF ll1sResult, ll2SResult
      (*********************************************************************)
      (* The new encrypted private states are directly equal across the    *)
      (* two specs.                                                        *)
      (*********************************************************************)
      <3>9. ll1NewPrivateStateEnc = ll2NewPrivateStateEnc
        <4>1. LL1NVRAM.symmetricKey = LL2NVRAM.symmetricKey
          BY <2>1 DEF LL2Refinement
        <4>2. ll1sResult.newPrivateState = ll2SResult.newPrivateState
          BY <3>8
        <4>3. QED
          BY <4>1, <4>2 DEF ll1NewPrivateStateEnc, ll2NewPrivateStateEnc
      (*********************************************************************)
      (* There is no definition in the `^\basicname^' spec that            *)
      (* corresponds to ll2CurrentHistorySummary in the `^\fastname^'      *)
      (* spec.  Instead, the corresponding value is the history summary    *)
      (* field of the NVRAM variable.                                      *)
      (*********************************************************************)
      <3>10. HistorySummariesMatch(
                 LL1NVRAM.historySummary,
                 ll2CurrentHistorySummary,
                 LL2NVRAM.hashBarrier')
        (*******************************************************************)
        (* The history summary in the `^\basicname^' NVRAM matches the     *)
        (* logical history summary in the `^\fastname^' NVRAM and SPCR, as *)
        (* specified by the refinement.                                    *)
        (*******************************************************************)
        <4>1. HistorySummariesMatch(
                  LL1NVRAM.historySummary,
                  LL2NVRAMLogicalHistorySummary,
                  LL2NVRAM.hashBarrier)
          BY <2>1 DEF LL2Refinement
        (*******************************************************************)
        (* In the `^\fastname^' spec, the logical history summary in the   *)
        (* NVRAM and SPCR equals the ll2CurrentHistorySummary defined by   *)
        (* the LL2RepeatOperation action.  This is a fairly                *)
        (* straightforward equivalence by definition, but it requires      *)
        (* separate consideration for the two cases of whether an          *)
        (* extension is in progress.                                       *)
        (*******************************************************************)
        <4>2. LL2NVRAMLogicalHistorySummary = ll2CurrentHistorySummary
          <5>1. CASE LL2NVRAM.extensionInProgress
            <6>1. LL2NVRAMLogicalHistorySummary = [
                      anchor |-> LL2NVRAM.historySummaryAnchor,
                      extension |-> LL2SPCR ]
              <7>1. LL2SPCR # BaseHashValue
                BY <3>2, <5>1
              <7>2. QED
                BY <5>1, <7>1 DEF LL2NVRAMLogicalHistorySummary
            <6>2. ll2CurrentHistorySummary = [
                      anchor |-> LL2NVRAM.historySummaryAnchor,
                      extension |-> LL2SPCR ]
              BY DEF ll2CurrentHistorySummary
            <6>3. QED
              BY <6>1, <6>2
          <5>2. CASE ~LL2NVRAM.extensionInProgress
            <6>1. LL2NVRAMLogicalHistorySummary = [
                      anchor |-> LL2NVRAM.historySummaryAnchor,
                      extension |-> BaseHashValue ]
              BY <5>2 DEF LL2NVRAMLogicalHistorySummary
            <6>2. ll2CurrentHistorySummary = [
                      anchor |-> LL2NVRAM.historySummaryAnchor,
                      extension |-> LL2SPCR ]
              BY DEF ll2CurrentHistorySummary
            <6>3. LL2SPCR = BaseHashValue
              BY <3>2, <5>2
            <6>4. QED
              BY <6>1, <6>2, <6>3
          <5>3. QED
            BY <5>1, <5>2
        <4>3. UNCHANGED LL2NVRAM.hashBarrier
          BY <3>2
        <4>4. QED
          BY <4>1, <4>2, <4>3
      (*********************************************************************)
      (* The new state hashes are directly equal across the two specs.     *)
      (*********************************************************************)
      <3>11. ll1NewStateHash = ll2NewStateHash
        <4>1. ll1sResult.newPublicState = ll2SResult.newPublicState
          BY <3>8
        <4>2. ll1NewPrivateStateEnc = ll2NewPrivateStateEnc
          BY <3>9
        <4>3. QED
          BY <4>1, <4>2 DEF ll1NewStateHash, ll2NewStateHash
      (*********************************************************************)
      (* The new authenticators match across the two specs, as specified   *)
      (* by the AuthenticatorsMatch predicate.                             *)
      (*********************************************************************)
      <3>12. AuthenticatorsMatch(
                ll1NewAuthenticator,
                ll2NewAuthenticator,
                LL2NVRAM.symmetricKey',
                LL2NVRAM.hashBarrier')
        (*******************************************************************)
        (* First, we prove some types needed by the definition of the      *)
        (* AuthenticatorsMatch predicate.                                  *)
        (*******************************************************************)
        <4>1. ll2NewStateHash \in HashType
          BY <3>3
        <4>2. LL1NVRAM.historySummary \in HashType
          BY <2>1 DEF LL2Refinement, LL1TrustedStorageType
        <4>3. ll2CurrentHistorySummary \in HistorySummaryType
          BY <3>3
        (*******************************************************************)
        (* We then prove that, in the `^\fastname^' spec, the new          *)
        (* authenticator is a valid MAC for the new history state binding. *)
        (* We will use the MACComplete property.                           *)
        (*******************************************************************)
        <4>4. ValidateMAC(LL2NVRAM.symmetricKey', ll2NewHistoryStateBinding, ll2NewAuthenticator)
          (*****************************************************************)
          (* In the `^\fastname^' spec, the new authenticator is generated *)
          (* as a MAC of the new history state binding.                    *)
          (*****************************************************************)
          <5>1. ll2NewAuthenticator =
                    GenerateMAC(LL2NVRAM.symmetricKey', ll2NewHistoryStateBinding)
            <6>1. UNCHANGED LL2NVRAM.symmetricKey
              BY <3>2
            <6>2. QED
              BY <6>1 DEF ll2NewAuthenticator
          (*****************************************************************)
          (* We can thus use the MACComplete property to show that the     *)
          (* generated MAC validates appropriately.  To do this, we first  *)
          (* need to prove some types.                                     *)
          (*****************************************************************)
          <5>2. LL2NVRAM.symmetricKey' \in SymmetricKeyType
            BY <2>1, LL2SubtypeImplicationLemma DEF LL2SubtypeImplication
          <5>3. ll2NewHistoryStateBinding \in HashType
            BY <3>3
          (*****************************************************************)
          (* Then, we appeal to the MACComplete property in a              *)
          (* straightforward way.                                          *)
          (*****************************************************************)
          <5>4. QED
            BY <5>1, <5>2, <5>3, MACComplete
        (*******************************************************************)
        (* We then prove that, in the `^\basicname^' spec, the new         *)
        (* authenticator is generated as a MAC of the new history state    *)
        (* binding.                                                        *)
        (*******************************************************************)
        <4>5. ll1NewAuthenticator = GenerateMAC(LL2NVRAM.symmetricKey', ll1NewHistoryStateBinding)
          <5>1. LL1NVRAM.symmetricKey = LL2NVRAM.symmetricKey'
            <6>1. LL1NVRAM.symmetricKey = LL2NVRAM.symmetricKey
              BY <2>1 DEF LL2Refinement
            <6>2. UNCHANGED LL2NVRAM.symmetricKey
              BY <3>2
            <6>3. QED
              BY <6>1, <6>2
          <5>2. QED
            BY <5>1 DEF ll1NewAuthenticator
        (*******************************************************************)
        (* The new history summaries match across the two specs, as we     *)
        (* proved above.                                                   *)
        (*******************************************************************)
        <4>6. HistorySummariesMatch(
                  LL1NVRAM.historySummary, ll2CurrentHistorySummary, LL2NVRAM.hashBarrier')
          BY <3>10
        (*******************************************************************)
        (* We then invoke the definition of the AuthenticatorsMatch        *)
        (* predicate.                                                      *)
        (*******************************************************************)
        <4>7. QED
          BY <3>11, <4>1, <4>2, <4>3, <4>4, <4>5, <4>6
          DEF AuthenticatorsMatch, ll1NewHistoryStateBinding, ll2NewHistoryStateBinding,
              ll2CurrentHistorySummary, ll2CurrentHistorySummaryHash
      (*********************************************************************)
      (* The remainder of the proof for LL2RepeatOperation is a series of  *)
      (* assertions, one for each conjunct in the definition of the        *)
      (* LL1RepeatOperation action.                                        *)
      (*                                                                   *)
      (* The first conjunct in LL1RepeatOperation.  This is basically just *)
      (* an application of AuthenticatorValidatedLemma                     *)
      (*********************************************************************)
      <3>13. ValidateMAC(LL1NVRAM.symmetricKey, ll1HistoryStateBinding, LL1RAM.authenticator)
        (*******************************************************************)
        (* We need the fact that the symmetric keys in the NVRAM are equal *)
        (* across the two specs.                                           *)
        (*******************************************************************)
        <4>1. LL1NVRAM.symmetricKey = LL2NVRAM.symmetricKey
          BY <2>1 DEF LL2Refinement
        (*******************************************************************)
        (* We prove the types that are needed for                          *)
        (* AuthenticatorValidatedLemma.                                    *)
        (*******************************************************************)
        <4>2. ll2StateHash \in HashType
          BY <3>3
        <4>3. LL1RAM.historySummary \in HashType
          BY <2>1 DEF LL2Refinement, LL1UntrustedStorageType
        <4>4. LL2RAM.historySummary \in HistorySummaryType
          BY <2>1, LL2SubtypeImplicationLemma DEF LL2SubtypeImplication
        <4>5. LL1RAM.authenticator \in MACType
          BY <2>1 DEF LL2Refinement, LL1UntrustedStorageType
        <4>6. LL2RAM.authenticator \in MACType
          BY <2>1, LL2SubtypeImplicationLemma DEF LL2SubtypeImplication
        <4>7. LL2NVRAM.symmetricKey \in SymmetricKeyType
          BY <2>1, LL2SubtypeImplicationLemma DEF LL2SubtypeImplication
        <4>8. LL2NVRAM.hashBarrier \in HashType
          BY <2>1, LL2SubtypeImplicationLemma DEF LL2SubtypeImplication
        (*******************************************************************)
        (* There are three preconditions for AuthenticatorValidatedLemma.  *)
        (* The first precondition follows from the refinement.             *)
        (*******************************************************************)
        <4>9. HistorySummariesMatch(
                  LL1RAM.historySummary, LL2RAM.historySummary, LL2NVRAM.hashBarrier)
          BY <2>1 DEF LL2Refinement
        (*******************************************************************)
        (* The second precondition also follows from the refinement.       *)
        (*******************************************************************)
        <4>10. PICK symmetricKey \in SymmetricKeyType :
                  AuthenticatorsMatch(
                      LL1RAM.authenticator,
                      LL2RAM.authenticator,
                      symmetricKey,
                      LL2NVRAM.hashBarrier)
          BY <2>1 DEF LL2Refinement
        (*******************************************************************)
        (* The third precondition follows from the definition of the       *)
        (* LL2RepeatOperation action.                                      *)
        (*******************************************************************)
        <4>11. ValidateMAC(LL2NVRAM.symmetricKey, ll2HistoryStateBinding, LL2RAM.authenticator)
          BY <3>2 DEF ll2HistoryStateBinding, ll2StateHash, ll2HistorySummaryHash
        <4>12. QED
          (*****************************************************************)
          (* Ideally, this QED step should just read:                      *)
          (*                                                               *)
          (*  BY <<3>>6, <<4>>1, <<4>>2, <<4>>3, <<4>>4, <<4>>5, <<4>>6, <<4>>7, <<4>>8, <<4>>9, <<4>>10, <<4>>11, *)
          (*      AuthenticatorValidatedLemma                              *)
          (*  DEF ll1HistoryStateBinding, ll2HistoryStateBinding, ll2HistorySummaryHash *)
          (*                                                               *)
          (* However, the prover seems to get a little confused in this    *)
          (* instance.  We make life easier for the prover by explicitly   *)
          (* staging the instantiation of the quantified variables within  *)
          (* the definition of AuthenticatorValidatedLemma.                *)
          (*****************************************************************)
          <5>1. \A  samLL2Authenticator \in MACType,
                    samSymmetricKey2 \in SymmetricKeyType,
                    samHashBarrier \in HashType :
                    AuthenticatorValidatedLemma!(ll2StateHash, LL1RAM.historySummary,
                        LL2RAM.historySummary, LL1RAM.authenticator, samLL2Authenticator,
                        symmetricKey, samSymmetricKey2, samHashBarrier)!1
            BY <4>2, <4>3, <4>4, <4>5, AuthenticatorValidatedLemma
          <5>2. AuthenticatorValidatedLemma!(ll2StateHash, LL1RAM.historySummary,
                    LL2RAM.historySummary, LL1RAM.authenticator, LL2RAM.authenticator,
                    symmetricKey, LL2NVRAM.symmetricKey, LL2NVRAM.hashBarrier)!1
            BY <4>2, <4>3, <4>4, <4>5, <4>6, <4>7, <4>8, <4>10, <5>1
          <5>3. QED
            BY <3>6, <4>1, <4>2, <4>3, <4>4, <4>5, <4>6, <4>7, <4>8, <4>9, <4>10, <4>11, <5>2
            DEF ll1HistoryStateBinding, ll2HistoryStateBinding, ll2HistorySummaryHash
      (*********************************************************************)
      (* The second conjunct in LL1RepeatOperation: In the `^\basicname^'  *)
      (* spec, the history summary in the NVRAM equals the hash of the     *)
      (* history summary in the RAM and the input.                         *)
      (*********************************************************************)
      <3>14. LL1NVRAM.historySummary = Hash(LL1RAM.historySummary, input)
        (*******************************************************************)
        (* We separately tackle the two cases of whether a checkpoint was  *)
        (* taken before the input was processed.                           *)
        (*******************************************************************)
        <4>1. \/  ll2CurrentHistorySummary = ll2CheckpointedNewHistorySummary
              \/  ll2CurrentHistorySummary = ll2CheckpointedNewCheckpointedHistorySummary
          BY <3>2
          DEF ll2CurrentHistorySummary, ll2NewHistorySummary,
              ll2CheckpointedHistorySummary, ll2NewCheckpointedHistorySummary,
              ll2CheckpointedNewHistorySummary, ll2CheckpointedNewCheckpointedHistorySummary
        (*******************************************************************)
        (* Before proceeding to the cases, we'll prove one type fact that  *)
        (* will be needed several places below.                            *)
        (*******************************************************************)
        <4>2. Hash(LL1RAM.historySummary, input) \in HashType
          <5>1. LL1RAM.historySummary \in HashDomain
            <6>1. LL1RAM.historySummary \in HashType
              BY <2>1 DEF LL2Refinement, LL1UntrustedStorageType
            <6>2. QED
              BY <6>1 DEF HashDomain
          <5>2. input \in HashDomain
            <6>1. input \in InputType
              <7>1. input \in LL2AvailableInputs
                BY <3>2
              <7>2. LL2AvailableInputs \subseteq InputType
                BY <2>1 DEF LL2TypeInvariant
              <7>3. QED
                BY <7>1, <7>2
            <6>2. QED
              BY <6>1 DEF HashDomain
          <5>3. QED
            BY <5>1, <5>2, HashTypeSafe
        (*******************************************************************)
        (* In the first case, a checkpoint was not taken before the input  *)
        (* was processed.                                                  *)
        (*******************************************************************)
        <4>3. CASE ll2CurrentHistorySummary = ll2CheckpointedNewHistorySummary
          (*****************************************************************)
          (* We will be employing the                                      *)
          (* HistorySummariesMatchAcrossCheckpointLemma, which has three   *)
          (* preconditions.  The first is that the history summaries in    *)
          (* the NVRAM match across the two specs, which is true by        *)
          (* refinement.                                                   *)
          (*****************************************************************)
          <5>1. HistorySummariesMatch(
                    LL1NVRAM.historySummary,
                    LL2NVRAMLogicalHistorySummary,
                    LL2NVRAM.hashBarrier)
            BY <2>1 DEF LL2Refinement
          (*****************************************************************)
          (* The second precondition is that the `^\basicname^' history    *)
          (* summary formed from the hash of the history summary in the    *)
          (* RAM and the input matches the new history summary in the      *)
          (* `^\fastname^' spec.                                           *)
          (*****************************************************************)
          <5>2. HistorySummariesMatch(
                    Hash(LL1RAM.historySummary, input),
                    ll2NewHistorySummary,
                    LL2NVRAM.hashBarrier)
            (***************************************************************)
            (* First, we prove that the HistorySummariesMatch predicate    *)
            (* equals the HistorySummariesMatchRecursion predicate in this *)
            (* case.  We assert each condition required by the definition  *)
            (* of the predicate.                                           *)
            (***************************************************************)
            <6>1. HistorySummariesMatch(
                      Hash(LL1RAM.historySummary, input),
                      ll2NewHistorySummary,
                      LL2NVRAM.hashBarrier) <=> (** changed = by <=> *)
                  HistorySummariesMatchRecursion(
                      Hash(LL1RAM.historySummary, input),
                      ll2NewHistorySummary,
                      LL2NVRAM.hashBarrier)
              (*************************************************************)
              (* We begin by proving the types for the                     *)
              (* HistorySummariesMatchRecursion predicate.                 *)
              (*************************************************************)
              <7>1. Hash(LL1RAM.historySummary, input) \in HashType
                BY <4>2
              <7>2. ll2NewHistorySummary \in HistorySummaryType
                BY <3>3
              <7>3. LL2NVRAM.hashBarrier \in HashType
                  BY <2>1, LL2SubtypeImplicationLemma DEF LL2SubtypeImplication
              <7> ll2InitialHistorySummary == [ anchor |-> BaseHashValue, extension |-> BaseHashValue ]
              (*************************************************************)
              (* We then prove that this is not the base case for the      *)
              (* HistorySummariesMatch predicate.                          *)
              (*************************************************************)
              <7>4. ll2NewHistorySummary # ll2InitialHistorySummary
                (***********************************************************)
                (* This proof has a lot of sub-steps, but it is pretty     *)
                (* simple.  We just use the BaseHashValueUnique property   *)
                (* to show that the extension field in the                 *)
                (* ll2NewHistorySummary record cannot match the base hash  *)
                (* value, which is the value of the extension field in the *)
                (* initial history summary.                                *)
                (***********************************************************)
                <8>1. ll2NewHistorySummary =
                          Successor(LL2RAM.historySummary, input, LL2NVRAM.hashBarrier)
                  BY DEF ll2NewHistorySummary
                <8>2. ll2NewHistorySummary.extension =
                          Hash(LL2RAM.historySummary.extension, Hash(LL2NVRAM.hashBarrier, input))
                  BY <8>1 DEF Successor
                <8>3. ll2NewHistorySummary.extension # BaseHashValue
                  <9>1. LL2RAM.historySummary.extension \in HashDomain
                    <10>1. LL2RAM.historySummary.extension \in HashType
                      <11>1. LL2RAM.historySummary \in HistorySummaryType
                        BY <2>1, LL2SubtypeImplicationLemma DEF LL2SubtypeImplication
                      <11>2. QED
                        BY <11>1 DEF HistorySummaryType
                    <10>2. QED
                      BY <10>1 DEF HashDomain
                  <9>2. Hash(LL2NVRAM.hashBarrier, input) \in HashDomain
                    <10>1. Hash(LL2NVRAM.hashBarrier, input) \in HashType
                      <11>1. LL2NVRAM.hashBarrier \in HashDomain
                        <12>1. LL2NVRAM.hashBarrier \in HashType
                          BY <2>1, LL2SubtypeImplicationLemma DEF LL2SubtypeImplication
                        <12>2. QED
                          BY <12>1 DEF HashDomain
                      <11>2. input \in HashDomain
                        <12>1. input \in InputType
                          <13>1. input \in LL2AvailableInputs
                            BY <3>2
                          <13>2. LL2AvailableInputs \subseteq InputType
                            BY <2>1 DEF LL2TypeInvariant
                          <13>3. QED
                            BY <13>1, <13>2
                        <12>2. QED
                          BY <12>1 DEF HashDomain
                      <11>3. QED
                        BY <11>1, <11>2, HashTypeSafe
                    <10>2. QED
                      BY <10>1 DEF HashDomain
                  <9>3. QED
                    BY <8>2, <9>1, <9>2, BaseHashValueUnique
                <8>4. QED
                  BY <8>3
              (*************************************************************)
              (* Since this is not the base case, the                      *)
              (* HistorySummariesMatch predicate equals the                *)
              (* HistorySummariesMatchRecursion predicate.                 *)
              (*************************************************************)
              <7>5. QED
                BY <7>1, <7>2, <7>3, <7>4, HistorySummariesMatchDefinition
            (***************************************************************)
            (* Then, we prove that the HistorySummariesMatchRecursion      *)
            (* predicate is satisfied.  We assert each condition required  *)
            (* by the definition of the predicate.                         *)
            (***************************************************************)
            <6>2. HistorySummariesMatchRecursion(
                      Hash(LL1RAM.historySummary, input),
                      ll2NewHistorySummary,
                      LL2NVRAM.hashBarrier)
              (*************************************************************)
              (* We begin by proving the types for the existentially       *)
              (* quantified variables in the                               *)
              (* HistorySummariesMatchRecursion predicate.                 *)
              (*************************************************************)
              <7>1. input \in InputType
                <8>1. input \in LL2AvailableInputs
                  BY <3>2
                <8>2. LL2AvailableInputs \subseteq InputType
                  BY <2>1 DEF LL2TypeInvariant
                <8>3. QED
                  BY <8>1, <8>2
              <7>2. LL1RAM.historySummary \in HashType
                BY <2>1 DEF LL2Refinement, LL1UntrustedStorageType
              <7>3. LL2RAM.historySummary \in HistorySummaryType
                BY <2>1, LL2SubtypeImplicationLemma DEF LL2SubtypeImplication
              (*************************************************************)
              (* We then prove the three conjuncts in the                  *)
              (* HistorySummariesMatchRecursion predicate.  The first      *)
              (* conjunct follows directly from the refinement.            *)
              (*************************************************************)
              <7>4. HistorySummariesMatch(
                        LL1RAM.historySummary,
                        LL2RAM.historySummary,
                        LL2NVRAM.hashBarrier)
                BY <2>1 DEF LL2Refinement
              (*************************************************************)
              (* The second conjunct in the HistorySummariesMatchRecursion *)
              (* predicate is that Hash(LL1RAM.historySummary, input) is   *)
              (* equal to itself.  We do not bother writing this.          *)
              (*                                                           *)
              (* The third conjunct in the HistorySummariesMatchRecursion  *)
              (* predicate is true by definition.                          *)
              (*************************************************************)
              <7>6. LL2HistorySummaryIsSuccessor(
                        ll2NewHistorySummary, LL2RAM.historySummary, input, LL2NVRAM.hashBarrier)
                BY DEF LL2HistorySummaryIsSuccessor, ll2NewHistorySummary 
              (*************************************************************)
              (* All conjuncts in the HistorySummariesMatchRecursion       *)
              (* predicate are satisfied.                                  *)
              (*************************************************************)
              <7>7. QED
                BY <7>1, <7>2, <7>3, <7>4, <7>6 DEF HistorySummariesMatchRecursion
              (*************************************************************)
              (* Since the HistorySummariesMatch predicate equals the      *)
              (* HistorySummariesMatchRecursion predicate, and the latter  *)
              (* predicate is satisfied, the former predicate is           *)
              (* satisfied.                                                *)
              (*************************************************************)
            <6>3. QED
              BY <6>1, <6>2
          (*****************************************************************)
          (* The third precondition is that in the `^\fastname^' spec, the *)
          (* logical history summary in the NVRAM and SPCR equals the      *)
          (* checkpointed new history summary.  This follows from a        *)
          (* straightforward expansion of definitions, given that an       *)
          (* extension is not in progress when an LL2RepeatOperation       *)
          (* action is executed.                                           *)
          (*****************************************************************)
          <5>3. LL2NVRAMLogicalHistorySummary = ll2CheckpointedNewHistorySummary
            <6>1. ~LL2NVRAM.extensionInProgress
              BY <3>2
            <6>2. LL2NVRAMLogicalHistorySummary = ll2CurrentHistorySummary
              <7>1. LL2NVRAMLogicalHistorySummary = [
                        anchor |-> LL2NVRAM.historySummaryAnchor,
                        extension |-> BaseHashValue ]
                BY <6>1 DEF LL2NVRAMLogicalHistorySummary
              <7>2. ll2CurrentHistorySummary = [
                        anchor |-> LL2NVRAM.historySummaryAnchor,
                        extension |-> LL2SPCR ]
                BY DEF ll2CurrentHistorySummary
              <7>3. LL2SPCR = BaseHashValue
                BY <3>2, <6>1
              <7>4. QED
                BY <7>1, <7>2, <7>3
            <6>3. ll2CurrentHistorySummary = ll2CheckpointedNewHistorySummary
              BY <4>3
            <6>4. QED
              BY <6>2, <6>3
          (*****************************************************************)
          (* We use the HistorySummariesMatchAcrossCheckpointLemma to      *)
          (* prove that the fields are equal.  This requires proving some  *)
          (* types.                                                        *)
          (*****************************************************************)
          <5>4. QED
            <6>1. LL1NVRAM.historySummary \in HashType
              BY <2>1 DEF LL2Refinement, LL1TrustedStorageType
            <6>2. Hash(LL1RAM.historySummary, input) \in HashType
              BY <4>2
            <6>3. LL2NVRAMLogicalHistorySummary \in HistorySummaryType
              BY <2>1, LL2NVRAMLogicalHistorySummaryTypeSafe
            <6>4. ll2NewHistorySummary \in HistorySummaryType
              BY <3>3
            <6>5. LL2NVRAM.hashBarrier \in HashType
              BY <2>1, LL2SubtypeImplicationLemma DEF LL2SubtypeImplication
            <6>6. QED
              BY <5>1, <5>2, <5>3, <6>1, <6>2, <6>3, <6>4, <6>5,
                  HistorySummariesMatchAcrossCheckpointLemma
              DEF ll2CheckpointedNewHistorySummary
        (*******************************************************************)
        (* In the second case, a checkpoint was taken before the input was *)
        (* processed.                                                      *)
        (*******************************************************************)
        <4>4. CASE ll2CurrentHistorySummary = ll2CheckpointedNewCheckpointedHistorySummary
          (*****************************************************************)
          (* We will be employing the                                      *)
          (* HistorySummariesMatchAcrossCheckpointLemma, which has three   *)
          (* preconditions.  The first is that the history summaries in    *)
          (* the NVRAM match across the two specs, which is true by        *)
          (* refinement.                                                   *)
          (*****************************************************************)
          <5>1. HistorySummariesMatch(
                    LL1NVRAM.historySummary,
                    LL2NVRAMLogicalHistorySummary,
                    LL2NVRAM.hashBarrier)
            BY <2>1 DEF LL2Refinement
          (*****************************************************************)
          (* The second precondition is that the `^\basicname^' history    *)
          (* summary formed from the hash of the history summary in the    *)
          (* RAM and the input matches the new checkpointed history        *)
          (* summary in the `^\fastname^' spec.                            *)
          (*****************************************************************)
          <5>2. HistorySummariesMatch(
                    Hash(LL1RAM.historySummary, input),
                    ll2NewCheckpointedHistorySummary,
                    LL2NVRAM.hashBarrier)
            (***************************************************************)
            (* First, we prove that the HistorySummariesMatch predicate    *)
            (* equals the HistorySummariesMatchRecursion predicate in this *)
            (* case.  We assert each condition required by the definition  *)
            (* of the predicate.                                           *)
            (***************************************************************)
            <6>1. HistorySummariesMatch(
                      Hash(LL1RAM.historySummary, input),
                      ll2NewCheckpointedHistorySummary,
                      LL2NVRAM.hashBarrier) <=> (** changed = by <=> *)
                  HistorySummariesMatchRecursion(
                      Hash(LL1RAM.historySummary, input),
                      ll2NewCheckpointedHistorySummary,
                      LL2NVRAM.hashBarrier)
              (*************************************************************)
              (* We begin by proving the types for the                     *)
              (* HistorySummariesMatchRecursion predicate.                 *)
              (*************************************************************)
              <7>1. Hash(LL1RAM.historySummary, input) \in HashType
                BY <4>2
              <7>2. ll2NewCheckpointedHistorySummary \in HistorySummaryType
                BY <3>3
              <7>3. LL2NVRAM.hashBarrier \in HashType
                  BY <2>1, LL2SubtypeImplicationLemma DEF LL2SubtypeImplication
              <7> ll2InitialHistorySummary == [ anchor |-> BaseHashValue, extension |-> BaseHashValue ]
              (*************************************************************)
              (* We then prove that this is not the base case for the      *)
              (* HistorySummariesMatch predicate.                          *)
              (*************************************************************)
              <7>4. ll2NewCheckpointedHistorySummary # ll2InitialHistorySummary
                (***********************************************************)
                (* This proof has a lot of sub-steps, but it is pretty     *)
                (* simple.  We just use the BaseHashValueUnique property   *)
                (* to show that the extension field in the                 *)
                (* ll2NewCheckpointedHistorySummary record cannot match    *)
                (* the base hash value, which is the value of the          *)
                (* extension field in the initial history summary.         *)
                (***********************************************************)
                <8>1. ll2NewCheckpointedHistorySummary =
                          Successor(ll2CheckpointedHistorySummary, input, LL2NVRAM.hashBarrier)
                  BY DEF ll2NewCheckpointedHistorySummary
                <8>2. ll2NewCheckpointedHistorySummary.extension =
                          Hash(ll2CheckpointedHistorySummary.extension,
                          Hash(LL2NVRAM.hashBarrier, input))
                  BY <8>1 DEF Successor
                <8>3. ll2NewCheckpointedHistorySummary.extension # BaseHashValue
                  <9>1. ll2CheckpointedHistorySummary.extension \in HashDomain
                    <10>1. ll2CheckpointedHistorySummary.extension \in HashType
                      <11>1. ll2CheckpointedHistorySummary \in HistorySummaryType
                        BY <3>3
                      <11>2. QED
                        BY <11>1 DEF HistorySummaryType
                    <10>2. QED
                      BY <10>1 DEF HashDomain
                  <9>2. Hash(LL2NVRAM.hashBarrier, input) \in HashDomain
                    <10>1. Hash(LL2NVRAM.hashBarrier, input) \in HashType
                      <11>1. LL2NVRAM.hashBarrier \in HashDomain
                        <12>1. LL2NVRAM.hashBarrier \in HashType
                          BY <2>1, LL2SubtypeImplicationLemma DEF LL2SubtypeImplication
                        <12>2. QED
                          BY <12>1 DEF HashDomain
                      <11>2. input \in HashDomain
                        <12>1. input \in InputType
                          <13>1. input \in LL2AvailableInputs
                            BY <3>2
                          <13>2. LL2AvailableInputs \subseteq InputType
                            BY <2>1 DEF LL2TypeInvariant
                          <13>3. QED
                            BY <13>1, <13>2
                        <12>2. QED
                          BY <12>1 DEF HashDomain
                      <11>3. QED
                        BY <11>1, <11>2, HashTypeSafe
                    <10>2. QED
                      BY <10>1 DEF HashDomain
                  <9>3. QED
                    BY <8>2, <9>1, <9>2, BaseHashValueUnique
                <8>4. QED
                  BY <8>3
              (*************************************************************)
              (* Since this is not the base case, the                      *)
              (* HistorySummariesMatch predicate equals the                *)
              (* HistorySummariesMatchRecursion predicate.                 *)
              (*************************************************************)
              <7>5. QED
                BY <7>1, <7>2, <7>3, <7>4, HistorySummariesMatchDefinition
            (***************************************************************)
            (* Then, we prove that the HistorySummariesMatchRecursion      *)
            (* predicate is satisfied.  We assert each condition required  *)
            (* by the definition of the predicate.                         *)
            (***************************************************************)
            <6>2. HistorySummariesMatchRecursion(
                      Hash(LL1RAM.historySummary, input),
                      ll2NewCheckpointedHistorySummary,
                      LL2NVRAM.hashBarrier)
              (*************************************************************)
              (* We begin by proving the types for the existentially       *)
              (* quantified variables in the                               *)
              (* HistorySummariesMatchRecursion predicate.                 *)
              (*************************************************************)
              <7>1. input \in InputType
                <8>1. input \in LL2AvailableInputs
                  BY <3>2
                <8>2. LL2AvailableInputs \subseteq InputType
                  BY <2>1 DEF LL2TypeInvariant
                <8>3. QED
                  BY <8>1, <8>2
              <7>2. LL1RAM.historySummary \in HashType
                BY <2>1 DEF LL2Refinement, LL1UntrustedStorageType
              <7>3. Checkpoint(LL2RAM.historySummary) \in HistorySummaryType
                <8>1. LL2RAM.historySummary \in HistorySummaryType
                  BY <2>1, LL2SubtypeImplicationLemma DEF LL2SubtypeImplication
                <8>2. QED
                  BY <8>1, CheckpointTypeSafe
              (*************************************************************)
              (* We then prove the three conjuncts in the                  *)
              (* HistorySummariesMatchRecursion predicate.                 *)
              (*                                                           *)
              (* The first conjunct is a recursive instance of the         *)
              (* HistorySummariesMatch predicate, namely that the history  *)
              (* summary in the `^\basicname^' RAM matches the             *)
              (* checkpointed history summary in the `^\fastname^' RAM.    *)
              (* We have to recursively expand the definition.             *)
              (*************************************************************)
              <7>4. HistorySummariesMatch(
                        LL1RAM.historySummary,
                        Checkpoint(LL2RAM.historySummary),
                        LL2NVRAM.hashBarrier)
                (***********************************************************)
                (* From the refinement, we know that the history summary   *)
                (* in the `^\basicname^' RAM matches the uncheckpointed    *)
                (* history summary in the `^\fastname^' RAM.               *)
                (***********************************************************)
                <8>1. HistorySummariesMatch(
                          LL1RAM.historySummary,
                          LL2RAM.historySummary,
                          LL2NVRAM.hashBarrier)
                  BY <2>1 DEF LL2Refinement
                (***********************************************************)
                (* We separately consider the base case and recursive      *)
                (* case.  We will apply the                                *)
                (* HistorySummariesMatchDefinition in both cases, so we    *)
                (* prove the necessary types once up front.                *)
                (***********************************************************)
                <8>2. LL1RAM.historySummary \in HashType
                  BY <2>1 DEF LL2Refinement, LL1UntrustedStorageType
                <8>3. Checkpoint(LL2RAM.historySummary) \in HistorySummaryType
                  BY <7>3
                <8>4. LL2NVRAM.hashBarrier \in HashType
                  BY <2>1, LL2SubtypeImplicationLemma DEF LL2SubtypeImplication
                <8> ll2InitialHistorySummary == [ anchor |-> BaseHashValue, extension |-> BaseHashValue ]
                (***********************************************************)
                (* The base case is very simple.                           *)
                (***********************************************************)
                <8>5. CASE LL2RAM.historySummary = ll2InitialHistorySummary
                  <9>1. Checkpoint(LL2RAM.historySummary) = ll2InitialHistorySummary
                    BY <8>5 DEF Checkpoint
                  <9>2. LL1RAM.historySummary = BaseHashValue
                    <10>1. LL2RAM.historySummary \in HistorySummaryType
                      BY <2>1, LL2SubtypeImplicationLemma DEF LL2SubtypeImplication
                    <10>2. QED
                      BY <8>1, <8>2, <8>4, <8>5, <10>1, HistorySummariesMatchDefinition
                  <9>3. QED
                    BY <8>2, <8>3, <8>4, <9>1, <9>2, HistorySummariesMatchDefinition
                (***********************************************************)
                (* The recursive case is pretty involved.                  *)
                (***********************************************************)
                <8>6. CASE LL2RAM.historySummary # ll2InitialHistorySummary
                  (*********************************************************)
                  (* First, we prove that the HistorySummariesMatch        *)
                  (* predicate equals the HistorySummariesMatchRecursion   *)
                  (* predicate in this case.                               *)
                  (*********************************************************)
                  <9>1. HistorySummariesMatch(
                            LL1RAM.historySummary,
                            Checkpoint(LL2RAM.historySummary),
                            LL2NVRAM.hashBarrier) <=>
                        HistorySummariesMatchRecursion(
                            LL1RAM.historySummary,
                            Checkpoint(LL2RAM.historySummary),
                            LL2NVRAM.hashBarrier)
                    (*******************************************************)
                    (* There is only one sub-step, which is showing that   *)
                    (* the `^\fastname^' checkpointed history summary is   *)
                    (* not equal to the initial history summary.  This     *)
                    (* follows naturally from the case we're in, but it    *)
                    (* takes a lot of tedious steps to get the prover to   *)
                    (* recognize this.                                     *)
                    (*******************************************************)
                    <10>1. Checkpoint(LL2RAM.historySummary) # ll2InitialHistorySummary
                      <11>1. CASE LL2RAM.historySummary.extension = BaseHashValue
                        <12>1. Checkpoint(LL2RAM.historySummary) = LL2RAM.historySummary
                          BY <11>1 DEF Checkpoint
                        <12>2. QED
                          BY <8>6, <11>1, <12>1 
                      <11>2. CASE LL2RAM.historySummary.extension # BaseHashValue
                        <12>1. Checkpoint(LL2RAM.historySummary) = [
                                   anchor |-> Hash(
                                                  LL2RAM.historySummary.anchor,
                                                  LL2RAM.historySummary.extension),
                                   extension |-> BaseHashValue ]
                          BY <11>2 DEF Checkpoint
                        <12>2. Checkpoint(LL2RAM.historySummary).anchor = Hash(
                                   LL2RAM.historySummary.anchor, LL2RAM.historySummary.extension)
                          BY <12>1
                        <12>3. Checkpoint(LL2RAM.historySummary).anchor # BaseHashValue
                          <13>1. LL2RAM.historySummary.anchor \in HashDomain
                            <14>1. LL2RAM.historySummary.anchor \in HashType
                              <15>1. LL2RAM.historySummary \in HistorySummaryType
                                BY <2>1, LL2SubtypeImplicationLemma DEF LL2SubtypeImplication
                              <15>2. QED
                                BY <15>1 DEF HistorySummaryType
                            <14>2. QED
                              BY <14>1 DEF HashDomain
                          <13>2. LL2RAM.historySummary.extension \in HashDomain
                            <14>1. LL2RAM.historySummary.extension \in HashType
                              <15>1. LL2RAM.historySummary \in HistorySummaryType
                                BY <2>1, LL2SubtypeImplicationLemma DEF LL2SubtypeImplication
                              <15>2. QED
                                BY <15>1 DEF HistorySummaryType
                            <14>2. QED
                              BY <14>1 DEF HashDomain
                          <13>3. QED
                            BY <12>2, <13>1, <13>2, BaseHashValueUnique
                        <12>4. QED
                          BY <12>2, <12>3
                      <11>3. QED
                        BY <11>1, <11>2
                    <10>2. QED
                      BY <8>2, <8>3, <8>4, <10>1, HistorySummariesMatchDefinition
                  (*********************************************************)
                  (* Then, we'll show that the                             *)
                  (* HistorySummariesMatchRecursion predicate holds.       *)
                  (*********************************************************)
                  <9>2. HistorySummariesMatchRecursion(
                            LL1RAM.historySummary,
                            Checkpoint(LL2RAM.historySummary),
                            LL2NVRAM.hashBarrier)
                    (*******************************************************)
                    (* We'll first pick values for the existential         *)
                    (* variables inside the HistorySummariesMatchRecursion *)
                    (* predicate that satisfy the predicate for the        *)
                    (* uncheckpointed history summary in the `^\fastname^' *)
                    (* RAM.  We know such variables exist, because this    *)
                    (* history summary matches the history summary in the  *)
                    (* `^\basicname^' RAM, by the refinement.              *)
                    (*******************************************************)
                    <10>1. PICK prevInput \in InputType,
                                previousLL1HistorySummary \in HashType,
                                previousLL2HistorySummary \in HistorySummaryType :
                                HistorySummariesMatchRecursion(
                                    LL1RAM.historySummary,
                                    LL2RAM.historySummary,
                                    LL2NVRAM.hashBarrier)!(
                                    prevInput,
                                    previousLL1HistorySummary,
                                    previousLL2HistorySummary)
                      <11>1. LL2RAM.historySummary \in HistorySummaryType
                        BY <2>1, LL2SubtypeImplicationLemma DEF LL2SubtypeImplication
                      <11>2. QED
                        BY <8>1, <8>2, <8>4, <8>6, <11>1, HistorySummariesMatchDefinition
                        DEF HistorySummariesMatchRecursion
                    (*******************************************************)
                    (* We then assert all of the conditions necessary to   *)
                    (* satisfy the HistorySummariesMatchRecursion for the  *)
                    (* checkpointed history summary in the `^\fastname^'   *)
                    (* RAM.  The first three of these are identical to the *)
                    (* conditions that hold for the uncheckpointed history *)
                    (* summary.                                            *)
                    (*******************************************************)
                    <10>2. prevInput \in InputType
                      BY <10>1
                    <10>3. HistorySummariesMatch(
                               previousLL1HistorySummary,
                               previousLL2HistorySummary,
                               LL2NVRAM.hashBarrier)
                      BY <10>1
                    <10>4. LL1RAM.historySummary = Hash(previousLL1HistorySummary, prevInput)
                      BY <10>1
                    (*******************************************************)
                    (* The last condition is more involved.  We show that  *)
                    (* the checkpointed history summary is a successor of  *)
                    (* the previous history summary.                       *)
                    (*******************************************************)
                    <10>5. LL2HistorySummaryIsSuccessor(
                               Checkpoint(LL2RAM.historySummary),
                               previousLL2HistorySummary,
                               prevInput,
                               LL2NVRAM.hashBarrier)
                      (*****************************************************)
                      (* We note that the uncheckpointed history summary   *)
                      (* is a successor of the previous history summary.   *)
                      (*****************************************************)
                      <11>1. LL2HistorySummaryIsSuccessor(
                                 LL2RAM.historySummary,
                                 previousLL2HistorySummary,
                                 prevInput,
                                 LL2NVRAM.hashBarrier)
                        BY <10>1
                      (*****************************************************)
                      (* The definition of LL2HistorySummaryIsSuccessor    *)
                      (* tells us that there are two ways that the         *)
                      (* uncheckpointed history summary could be a         *)
                      (* successor.  We will consider the two cases        *)
                      (* separately.                                       *)
                      (*****************************************************)
                      <11>2. \/  LL2RAM.historySummary =
                                     Successor(
                                         previousLL2HistorySummary,
                                         prevInput,
                                         LL2NVRAM.hashBarrier)
                             \/  LL2RAM.historySummary =
                                     Checkpoint(
                                         Successor(
                                             previousLL2HistorySummary,
                                             prevInput,
                                             LL2NVRAM.hashBarrier))
                        BY <11>1 DEF LL2HistorySummaryIsSuccessor
                      (*****************************************************)
                      (* The first case is trivial.  If the uncheckpointed *)
                      (* history summary is a successor by virtue of the   *)
                      (* Successor operator, then an application of the    *)
                      (* Checkpoint operater yields the checkpointed       *)
                      (* history summary.                                  *)
                      (*****************************************************)
                      <11>3. CASE LL2RAM.historySummary =
                                 Successor(previousLL2HistorySummary, prevInput, LL2NVRAM.hashBarrier)
                        <12>1. Checkpoint(LL2RAM.historySummary) =
                                   Checkpoint(
                                       Successor(
                                           previousLL2HistorySummary,
                                           prevInput,
                                           LL2NVRAM.hashBarrier))
                          BY <11>3
                        <12>2. QED
                          BY <12>1 DEF LL2HistorySummaryIsSuccessor
                      (*****************************************************)
                      (* The second case is only slightly more involved.   *)
                      (* If the uncheckpointed history summary is a        *)
                      (* successor by virtue of the Successor operator and *)
                      (* the Checkpoint operator, then a second            *)
                      (* application of the Checkpoint operater is         *)
                      (* idempotent.                                       *)
                      (*****************************************************)
                      <11>4. CASE LL2RAM.historySummary =
                                 Checkpoint(
                                     Successor(
                                         previousLL2HistorySummary,
                                         prevInput,
                                         LL2NVRAM.hashBarrier))
                        <12>1. Checkpoint(LL2RAM.historySummary) =
                                   Checkpoint(Checkpoint(
                                       Successor(
                                           previousLL2HistorySummary,
                                           prevInput,
                                           LL2NVRAM.hashBarrier)))
                          BY <11>4
                        <12>2. Checkpoint(LL2RAM.historySummary) =
                                   Checkpoint(
                                       Successor(
                                           previousLL2HistorySummary,
                                           prevInput,
                                           LL2NVRAM.hashBarrier))
                          BY <12>1 DEF Checkpoint
                        <12>3. QED
                          BY <12>2 DEF LL2HistorySummaryIsSuccessor
                      <11>5. QED
                        BY <11>2, <11>3, <11>4
                    (*******************************************************)
                    (* We have thus shown that the conditions for the      *)
                    (* HistorySummariesMatchRecursion predicate all hold.  *)
                    (*******************************************************)
                    <10>6. QED
                      BY <8>2, <8>3, <8>4, <10>2, <10>3, <10>4, <10>5 DEF HistorySummariesMatchRecursion
                  (*********************************************************)
                  (* Since the HistorySummariesMatch predicate equals the  *)
                  (* HistorySummariesMatchRecursion predicate, and since   *)
                  (* the HistorySummariesMatchRecursion predicate is       *)
                  (* satisfied, the HistorySummariesMatch predicate is     *)
                  (* satsisfied.                                           *)
                  (*********************************************************)
                  <9>3.QED
                    BY <9>1, <9>2
                (***********************************************************)
                (* Both the base case and the recursive case are           *)
                (* satisfied.                                              *)
                (***********************************************************)
                <8>7. QED
                  BY <8>5, <8>6
              (*************************************************************)
              (* The second conjunct in the HistorySummariesMatchRecursion *)
              (* predicate is that Hash(LL1RAM.historySummary, input) is   *)
              (* equal to itself.  We do not bother writing this.          *)
              (*                                                           *)
              (* The third conjunct in the HistorySummariesMatchRecursion  *)
              (* predicate is true by definition.                          *)
              (*************************************************************)
              <7>6. LL2HistorySummaryIsSuccessor(
                        ll2NewCheckpointedHistorySummary,
                        Checkpoint(LL2RAM.historySummary),
                        input,
                        LL2NVRAM.hashBarrier)
                BY DEF LL2HistorySummaryIsSuccessor,
                    ll2NewCheckpointedHistorySummary, ll2CheckpointedHistorySummary
              (*************************************************************)
              (* All conjuncts in the HistorySummariesMatchRecursion       *)
              (* predicate are satisfied.                                  *)
              (*************************************************************)
              <7>7. QED
                BY <7>1, <7>2, <7>3, <7>4, <7>6 DEF HistorySummariesMatchRecursion
              (*************************************************************)
              (* Since the HistorySummariesMatch predicate equals the      *)
              (* HistorySummariesMatchRecursion predicate, and the latter  *)
              (* predicate is satisfied, the former predicate is           *)
              (* satisfied.                                                *)
              (*************************************************************)
            <6>3. QED
              BY <6>1, <6>2
          (*****************************************************************)
          (* The third precondition is that in the `^\fastname^' spec, the *)
          (* logical history summary in the NVRAM and SPCR equals the      *)
          (* checkpointed new checkpointed history summary.  This follows  *)
          (* from a straightforward expansion of definitions, given that   *)
          (* an extension is not in progress when an LL2RepeatOperation    *)
          (* action is executed.                                           *)
          (*****************************************************************)
          <5>3. LL2NVRAMLogicalHistorySummary = ll2CheckpointedNewCheckpointedHistorySummary
            <6>1. ~LL2NVRAM.extensionInProgress
              BY <3>2
            <6>2. LL2NVRAMLogicalHistorySummary = ll2CurrentHistorySummary
              <7>1. LL2NVRAMLogicalHistorySummary = [
                        anchor |-> LL2NVRAM.historySummaryAnchor,
                        extension |-> BaseHashValue ]
                BY <6>1 DEF LL2NVRAMLogicalHistorySummary
              <7>2. ll2CurrentHistorySummary = [
                        anchor |-> LL2NVRAM.historySummaryAnchor,
                        extension |-> LL2SPCR ]
                BY DEF ll2CurrentHistorySummary
              <7>3. LL2SPCR = BaseHashValue
                BY <3>2, <6>1
              <7>4. QED
                BY <7>1, <7>2, <7>3
            <6>3. ll2CurrentHistorySummary = ll2CheckpointedNewCheckpointedHistorySummary
              BY <4>3
            <6>4. QED
              BY <6>2, <6>3
          (*****************************************************************)
          (* We use the HistorySummariesMatchAcrossCheckpointLemma to      *)
          (* prove that the fields are equal.  This requires proving some  *)
          (* types.                                                        *)
          (*****************************************************************)
          <5>4. QED
            <6>1. LL1NVRAM.historySummary \in HashType
              BY <2>1 DEF LL2Refinement, LL1TrustedStorageType
            <6>2. Hash(LL1RAM.historySummary, input) \in HashType
              BY <4>2
            <6>3. LL2NVRAMLogicalHistorySummary \in HistorySummaryType
              BY <2>1, LL2NVRAMLogicalHistorySummaryTypeSafe
            <6>4. ll2NewCheckpointedHistorySummary \in HistorySummaryType
              BY <3>3
            <6>5. LL2NVRAM.hashBarrier \in HashType
              BY <2>1, LL2SubtypeImplicationLemma DEF LL2SubtypeImplication
            <6>6. QED
              BY <5>1, <5>2, <5>3, <6>1, <6>2, <6>3, <6>4, <6>5,
                  HistorySummariesMatchAcrossCheckpointLemma
              DEF ll2CheckpointedNewCheckpointedHistorySummary
        (*******************************************************************)
        (* The two cases are exhaustive.                                   *)
        (*******************************************************************)
        <4>5. QED
          BY <4>1, <4>3, <4>4
      (*********************************************************************)
      (* The third conjunct in LL1RepeatOperation.  This conjunct asserts  *)
      (* a record equality, so we prove each field in the record           *)
      (* separately.                                                       *)
      (*********************************************************************)
      <3>15. LL1RAM' = [
                publicState |-> ll1sResult.newPublicState,
                privateStateEnc |-> ll1NewPrivateStateEnc,
                historySummary |-> LL1NVRAM.historySummary,
                authenticator |-> ll1NewAuthenticator ]
        (*******************************************************************)
        (* The public state field in the primed `^\basicname^' RAM equals  *)
        (* the public state in the result of the service, because the      *)
        (* primed RAM variables match across the two specs.                *)
        (*******************************************************************)
        <4>1. LL1RAM.publicState' = ll1sResult.newPublicState
          <5>1. LL1RAM.publicState' = LL2RAM.publicState'
            BY <2>1 DEF LL2Refinement
          <5>2. LL2RAM.publicState' = ll2SResult.newPublicState
            BY <3>2 DEF ll2SResult, ll2PrivateState
          <5>3. ll1sResult.newPublicState = ll2SResult.newPublicState
            BY <3>8
          <5>4. QED
            BY <5>1, <5>2, <5>3
        (*******************************************************************)
        (* The encrypted private state field in the primed `^\basicname^'  *)
        (* RAM equals the encrypted private state in the result of the     *)
        (* service, because the primed RAM variables match across the two  *)
        (* specs.                                                          *)
        (*******************************************************************)
        <4>2. LL1RAM.privateStateEnc' = ll1NewPrivateStateEnc
          <5>1. LL1RAM.privateStateEnc' = LL2RAM.privateStateEnc'
            BY <2>1 DEF LL2Refinement
          <5>2. LL2RAM.privateStateEnc' = ll2NewPrivateStateEnc
            BY <3>2 DEF ll2NewPrivateStateEnc, ll2SResult, ll2PrivateState
          <5>3. ll1NewPrivateStateEnc = ll2NewPrivateStateEnc
            BY <3>9
          <5>4. QED
            BY <5>1, <5>2, <5>3
        (*******************************************************************)
        (* In the `^\basicname^' spec, the history summary field in the    *)
        (* primed RAM equals the history summary in the unprimed NVRAM.    *)
        (* This is because (1) the history summaries in the RAM variables  *)
        (* match across the specs by refinement, (2) the history summary   *)
        (* in the primed `^\fastname^' RAM is set equal to the             *)
        (* ll2CurrentHistorySummary by the LL2RepeatOperation action, and  *)
        (* (3) the history summary in the `^\basicname^' NVRAM matches the *)
        (* ll2CurrentHistorySummary in the `^\fastname^' spec, as we       *)
        (* proved above.  We can thus use the                              *)
        (* HistorySummariesMatchUniqueLemma to prove the equality.         *)
        (*******************************************************************)
        <4>3. LL1RAM.historySummary' = LL1NVRAM.historySummary
          <5>1. HistorySummariesMatch(
                    LL1RAM.historySummary', LL2RAM.historySummary', LL2NVRAM.hashBarrier')
            BY <2>1 DEF LL2Refinement
          <5>2. LL2RAM.historySummary' = ll2CurrentHistorySummary
            BY <3>2 DEF ll2CurrentHistorySummary
          <5>3. HistorySummariesMatch(
                    LL1NVRAM.historySummary,
                    ll2CurrentHistorySummary,
                    LL2NVRAM.hashBarrier')
            BY <3>10
          (*****************************************************************)
          (* We use the HistorySummariesMatchUniqueLemma to prove that the *)
          (* fields are equal.  This requires proving some types.          *)
          (*****************************************************************)
          <5>4. QED
            <6>1. LL1RAM.historySummary' \in HashType
              BY <2>1 DEF LL2Refinement, LL1UntrustedStorageType
            <6>2. LL1NVRAM.historySummary \in HashType
              BY <2>1 DEF LL2Refinement, LL1TrustedStorageType
            <6>3. LL2RAM.historySummary' \in HistorySummaryType
              BY <2>1, LL2SubtypeImplicationLemma DEF LL2SubtypeImplication
            <6>4. LL2NVRAM.hashBarrier' \in HashType
              BY <2>1, LL2SubtypeImplicationLemma DEF LL2SubtypeImplication, LL2TypeInvariant, LL2TrustedStorageType
            <6>5. QED
              BY <5>1, <5>2, <5>3, <6>1, <6>2, <6>3, <6>4, HistorySummariesMatchUniqueLemma
        (*******************************************************************)
        (* The authenticator field in the primed `^\basicname^' RAM equals *)
        (* the new authenticator defined in the LL1RepeatOperation action, *)
        (* because the authenticators in the RAM variables match across    *)
        (* the specs by refinement.  We use the                            *)
        (* AuthenticatorsMatchUniqueLemma to prove the equality.           *)
        (*******************************************************************)
        <4>4. LL1RAM.authenticator' = ll1NewAuthenticator
          <5>1. PICK symmetricKey \in SymmetricKeyType :
                    AuthenticatorsMatch(
                        LL1RAM.authenticator',
                        LL2RAM.authenticator',
                        symmetricKey,
                        LL2NVRAM.hashBarrier')
            BY <2>1 DEF LL2Refinement
          <5>2. LL2RAM.authenticator' = ll2NewAuthenticator
            BY <3>2
            DEF ll2NewAuthenticator, ll2NewHistoryStateBinding, ll2CurrentHistorySummaryHash,
                ll2NewStateHash, ll2NewPrivateStateEnc, ll2SResult, ll2PrivateState,
                ll2NewHistorySummary, ll2CurrentHistorySummary
          <5>3. AuthenticatorsMatch(
                    ll1NewAuthenticator,
                    ll2NewAuthenticator,
                    LL2NVRAM.symmetricKey',
                    LL2NVRAM.hashBarrier')
            BY <3>12
          (*****************************************************************)
          (* We use the AuthenticatorsMatchUniqueLemma to prove that the   *)
          (* fields are equal.  This requires proving some types.          *)
          (*****************************************************************)
          <5>4. QED
            <6>1. LL1RAM.authenticator' \in MACType
              BY <2>1 DEF LL2Refinement, LL1UntrustedStorageType
            <6>2. ll1NewAuthenticator \in MACType
              BY <3>5
            <6>3. LL2RAM.authenticator' \in MACType
              BY <2>1, LL2SubtypeImplicationLemma DEF LL2SubtypeImplication
            <6>4. symmetricKey \in SymmetricKeyType
              BY <5>1
            <6>5. LL2NVRAM.symmetricKey' \in SymmetricKeyType
              BY <2>1, LL2SubtypeImplicationLemma DEF LL2SubtypeImplication
            <6>6. LL2NVRAM.hashBarrier' \in HashType
              BY <2>1, LL2SubtypeImplicationLemma DEF LL2SubtypeImplication, LL2TypeInvariant, LL2TrustedStorageType
            <6>7. QED
              BY <5>1, <5>2, <5>3, <6>1, <6>2, <6>3, <6>4, <6>5, <6>6,
                  AuthenticatorsMatchUniqueLemma
        (*******************************************************************)
        (* The refinement asserts that the RAM record has the appropriate  *)
        (* type.                                                           *)
        (*******************************************************************)
        <4>5. LL1RAM' \in LL1UntrustedStorageType
          BY <2>1 DEF LL2Refinement
        (*******************************************************************)
        (* We use the LL1RAMRecordCompositionLemma to unify the field      *)
        (* equalities into a record equality.                              *)
        (*******************************************************************)
        <4>6. QED
          BY <4>1, <4>2, <4>3, <4>4, <4>5, LL1RAMRecordCompositionLemma
      (*********************************************************************)
      (* The fourth conjunct in LL1RepeatOperation.  The set of observed   *)
      (* outputs is equal across the two specs.                            *)
      (*********************************************************************)
      <3>16. LL1ObservedOutputs' = LL1ObservedOutputs \union {ll1sResult.output}
        <4>1. LL1ObservedOutputs = LL2ObservedOutputs
          BY <2>1 DEF LL2Refinement
        <4>2. LL1ObservedOutputs' = LL2ObservedOutputs'
          BY <2>1 DEF LL2Refinement
        <4>3. ll1sResult.output = ll2SResult.output
          BY <3>8
        <4>4. LL2ObservedOutputs' = LL2ObservedOutputs \union {ll2SResult.output}
          BY <3>2 DEF ll2SResult, ll2PrivateState
        <4>5. QED
          BY <4>1, <4>2, <4>3, <4>4
      (*********************************************************************)
      (* The fifth conjunct in LL1RepeatOperation.  The NVRAM is unchanged *)
      (* by the UnchangedNVRAMLemma.                                       *)
      (*********************************************************************)
      <3>17. UNCHANGED LL1NVRAM
        BY <2>1, <3>2, UnchangedNVRAMLemma
      (*********************************************************************)
      (* The sixth conjunct in LL1RepeatOperation.  The disk is unchanged  *)
      (* by the UnchangedDiskLemma.                                        *)
      (*********************************************************************)
      <3>18. UNCHANGED LL1Disk
        BY <2>1, <3>2, UnchangedDiskLemma
      (*********************************************************************)
      (* The seventh conjunct in LL1RepeatOperation.  The set of available *)
      (* inputs is unchanged by the UnchangedAvailableInputsLemma.         *)
      (*********************************************************************)
      <3>19. UNCHANGED LL1AvailableInputs
        BY <2>1, <3>2, UnchangedAvailableInputsLemma
      (*********************************************************************)
      (* The eighth conjunct in LL1RepeatOperation.  We prove that the     *)
      (* primed set of observed authenticators matches across the specs,   *)
      (* that the unprimed set of observed authenticators matches across   *)
      (* the specs, and that the new authenticator matches across the      *)
      (* specs.  The AuthenticatorSetsMatchUniqueLemma then proves the     *)
      (* equality.                                                         *)
      (*********************************************************************)
      <3>20. LL1ObservedAuthenticators' =
                 LL1ObservedAuthenticators \union {ll1NewAuthenticator}
        (*******************************************************************)
        (* The primed set of observed authenticators matches across the    *)
        (* specs.  This follows directly from the refinement.              *)
        (*******************************************************************)
        <4>1. AuthenticatorSetsMatch(
            LL1ObservedAuthenticators',
            LL2ObservedAuthenticators',
            LL2NVRAM.symmetricKey',
            LL2NVRAM.hashBarrier')
          BY <2>1 DEF LL2Refinement
        (*******************************************************************)
        (* The union matches across the specs.  We prove this by proving   *)
        (* the matching of each constituent set.                           *)
        (*******************************************************************)
        <4>2. AuthenticatorSetsMatch(
            LL1ObservedAuthenticators \union {ll1NewAuthenticator},
            LL2ObservedAuthenticators \union {ll2NewAuthenticator},
            LL2NVRAM.symmetricKey',
            LL2NVRAM.hashBarrier')
          <5>1. UNCHANGED <<LL2NVRAM.symmetricKey, LL2NVRAM.hashBarrier>>
            BY <3>2
          (*****************************************************************)
          (* The unprimed set of observed authenticators matches across    *)
          (* the specs.  This follows directly from the refinement.        *)
          (*****************************************************************)
          <5>2. AuthenticatorSetsMatch(
                    LL1ObservedAuthenticators,
                    LL2ObservedAuthenticators,
                    LL2NVRAM.symmetricKey',
                    LL2NVRAM.hashBarrier')
              BY <2>1, <5>1 DEF LL2Refinement
          (*****************************************************************)
          (* The new authenticator matches across the specs.  This follows *)
          (* directly from the refinement.                                 *)
          (*****************************************************************)
          <5>3. AuthenticatorsMatch(
                    ll1NewAuthenticator,
                    ll2NewAuthenticator,
                    LL2NVRAM.symmetricKey',
                    LL2NVRAM.hashBarrier')
              BY <2>1, <5>1 DEF LL2Refinement
          <5>4. QED
            BY <5>2, <5>3 DEF AuthenticatorSetsMatch
        (*******************************************************************)
        (* In the `^\fastname^' spec, the primed set of observed           *)
        (* authenticators is formed from the union of the unprimed set of  *)
        (* observed authenticators and the new authenticator.              *)
        (*******************************************************************)
        <4>3. LL2ObservedAuthenticators' =
                    LL2ObservedAuthenticators \union {ll2NewAuthenticator}
          BY <3>2
          DEF ll2NewAuthenticator, ll2NewHistoryStateBinding, ll2CurrentHistorySummaryHash,
              ll2NewStateHash, ll2NewPrivateStateEnc, ll2SResult, ll2PrivateState,
              ll2NewHistorySummary, ll2CurrentHistorySummary
        <4>4. QED
          (*****************************************************************)
          (* We use the AuthenticatorSetsMatchUniqueLemma to prove that    *)
          (* the sets are equal.  This requires proving some types.        *)
          (*****************************************************************)
          <5>1. LL1ObservedAuthenticators' \in SUBSET MACType
            BY <2>1 DEF LL2Refinement, LL1UntrustedStorageType
          <5>2. LL1ObservedAuthenticators \union {ll1NewAuthenticator} \in
                    SUBSET MACType
            <6>1. LL1ObservedAuthenticators \in SUBSET MACType
              BY <2>1 DEF LL2Refinement, LL1UntrustedStorageType
            <6>2. ll1NewAuthenticator \in MACType
              BY <3>5
            <6>3. QED
              BY <6>1, <6>2
          <5>3. LL2ObservedAuthenticators' \in SUBSET MACType
            BY <2>1 DEF LL2TypeInvariant
          <5>4. LL2NVRAM.symmetricKey' \in SymmetricKeyType
            BY <2>1, LL2SubtypeImplicationLemma DEF LL2SubtypeImplication
          <5>5. LL2NVRAM.hashBarrier' \in HashType
            BY <2>1, LL2SubtypeImplicationLemma DEF LL2SubtypeImplication, LL2TypeInvariant, LL2TrustedStorageType
          <5>6. QED
            BY <4>1, <4>2, <4>3, <5>1, <5>2, <5>3, <5>4, <5>5,
                AuthenticatorSetsMatchUniqueLemma
      <3>21. QED
        BY <3>4, <3>13, <3>14, <3>15, <3>16, <3>17, <3>18, <3>19, <3>20
        DEF LL1RepeatOperation, ll1StateHash, ll1HistoryStateBinding, ll1PrivateState,
            ll1sResult, ll1NewPrivateStateEnc, ll1NewStateHash,
            ll1NewHistoryStateBinding, ll1NewAuthenticator
    (***********************************************************************)
    (* A `^\fastname^' LL2TakeCheckpoint action refines to a               *)
    (* `^\basicname^' stuttering step.  This is not at all obvious, and it *)
    (* is the essence of why the enhancement from `^\basicname^' to        *)
    (* `^\fastname^' spec continues to satisfy the implementation.         *)
    (***********************************************************************)
    <2>6. LL2TakeCheckpoint => UNCHANGED LL1Vars
      <3>1. HAVE LL2TakeCheckpoint
      (*********************************************************************)
      (* We re-state the definition from LL2TakeCheckpoint.                *)
      (*********************************************************************)
      <3> newHistorySummaryAnchor == Hash(LL2NVRAM.historySummaryAnchor, LL2SPCR)
      (*********************************************************************)
      (* We then hide the definition.                                      *)
      (*********************************************************************)
      <3> HIDE DEF newHistorySummaryAnchor
      (*********************************************************************)
      (* Two frequently useful facts are that the symmetric key and the    *)
      (* hash barrier in the `^\fastname^' NVRAM are unchanged.  This      *)
      (* follows directly from the definition of LL2TakeCheckpoint.        *)
      (*********************************************************************)
      <3>2. /\  UNCHANGED LL2NVRAM.symmetricKey
            /\  UNCHANGED LL2NVRAM.hashBarrier
        <4>1. LL2NVRAM' = [
                  historySummaryAnchor |-> newHistorySummaryAnchor,
                  symmetricKey |-> LL2NVRAM.symmetricKey,
                  hashBarrier |-> LL2NVRAM.hashBarrier,
                  extensionInProgress |-> FALSE ]
          BY <3>1 DEF LL2TakeCheckpoint, newHistorySummaryAnchor
        <4>2. QED
          BY <4>1 DEF LL2TrustedStorageType
      (*********************************************************************)
      (* We prove the UNCHANGED status for each `^\basicname^' variable in *)
      (* turn.  For LL1AvailableInputs, LL1ObservedOutputs,                *)
      (* LL1ObservedAuthenticators, LL1Disk, and LL1RAM, the UNCHANGED     *)
      (* status follows directly from the lemmas we have proven for this   *)
      (* purpose.                                                          *)
      (*********************************************************************)
      <3>3. UNCHANGED LL1AvailableInputs
        BY <2>1, <3>1, <3>2, UnchangedAvailableInputsLemma DEF LL2TakeCheckpoint
      <3>4. UNCHANGED LL1ObservedOutputs
        BY <2>1, <3>1, <3>2, UnchangedObservedOutputsLemma DEF LL2TakeCheckpoint
      <3>5. UNCHANGED LL1ObservedAuthenticators
        BY <2>1, <3>1, <3>2, UnchangedObservedAuthenticatorsLemma DEF LL2TakeCheckpoint
      <3>6. UNCHANGED LL1Disk
        BY <2>1, <3>1, <3>2, UnchangedDiskLemma DEF LL2TakeCheckpoint
      <3>7. UNCHANGED LL1RAM
        BY <2>1, <3>1, <3>2, UnchangedRAMLemma DEF LL2TakeCheckpoint
      (*********************************************************************)
      (* To prove the UNCHANGED status of the `^\basicname^' NVRAM value,  *)
      (* we prove each of the two fields separately.                       *)
      (*********************************************************************)
      <3>8. UNCHANGED LL1NVRAM
        (*******************************************************************)
        (* Proving the UNCHANGED status of the history summary in the      *)
        (* `^\basicname^' NVRAM is quite involved.  The top level is       *)
        (* straightforward.  We simply prove that both the unprimed and    *)
        (* primed history summaries in the `^\basicname^' NVRAM match the  *)
        (* primed logical history summary in the `^\fastname^' NVRAM and   *)
        (* SPCR.  Then, from the HistorySummariesMatchUniqueLemma, we      *)
        (* conclude that both unprimed and unprimed history summaries in   *)
        (* the `^\basicname^' NVRAM are equal.                             *)
        (*******************************************************************)
        <4>2. UNCHANGED LL1NVRAM.historySummary
          (*****************************************************************)
          (* The primed history summary in the `^\basicname^' NVRAM        *)
          (* matches the primed logical history summary in the             *)
          (* `^\fastname^' NVRAM and SPCR.  This follows directly from the *)
          (* refinement, given that the hash barrier in the `^\fastname^'  *)
          (* NVRAM has not changed.                                        *)
          (*****************************************************************)
          <5>1. HistorySummariesMatch(
                    LL1NVRAM.historySummary',
                    LL2NVRAMLogicalHistorySummary',
                    LL2NVRAM.hashBarrier)
            BY <3>2, <2>1 DEF LL2Refinement
          (*****************************************************************)
          (* The unprimed history summary in the `^\basicname^' NVRAM      *)
          (* matches the primed logical history summary in the             *)
          (* `^\fastname^' NVRAM and SPCR.  The proof involves three main  *)
          (* steps.                                                        *)
          (*****************************************************************)
          <5>2. HistorySummariesMatch(
                    LL1NVRAM.historySummary,
                    LL2NVRAMLogicalHistorySummary',
                    LL2NVRAM.hashBarrier)
            <6> ll2InitialHistorySummary == [ anchor |-> BaseHashValue, extension |-> BaseHashValue ]
            (***************************************************************)
            (* First, we pick a set of variables that satisfy the          *)
            (* existentials in the HistorySummariesMatchRecursion          *)
            (* operator, for the unprimed state.  This is more involved    *)
            (* than might seem necessary.                                  *)
            (***************************************************************)
            <6>1. PICK input \in InputType,
                       previousLL1HistorySummary \in HashType,
                       previousLL2HistorySummary \in HistorySummaryType :
                       HistorySummariesMatchRecursion(
                           LL1NVRAM.historySummary,
                           LL2NVRAMLogicalHistorySummary,
                           LL2NVRAM.hashBarrier)!(
                           input,
                           previousLL1HistorySummary,
                           previousLL2HistorySummary)
              (*************************************************************)
              (* First, we assert that the unprimed history summaries      *)
              (* match across the two specs, which follows from the        *)
              (* refinement.                                               *)
              (*************************************************************)
              <7>1. HistorySummariesMatch(
                        LL1NVRAM.historySummary,
                        LL2NVRAMLogicalHistorySummary,
                        LL2NVRAM.hashBarrier)
                BY <2>1 DEF LL2Refinement
              (*************************************************************)
              (* Then, we prove some types, to satisfy the universal       *)
              (* quantifiers in HistorySummariesMatchDefinition.           *)
              (*************************************************************)
              <7>2. LL1NVRAM.historySummary \in HashType
                BY <2>1 DEF LL2Refinement, LL1TrustedStorageType
              <7>3. LL2NVRAMLogicalHistorySummary \in HistorySummaryType
                BY <2>1, LL2NVRAMLogicalHistorySummaryTypeSafe
              <7>4. LL2NVRAM.hashBarrier \in HashType
                BY <2>1, LL2SubtypeImplicationLemma DEF LL2SubtypeImplication
              (*************************************************************)
              (* We prove that the `^\fastname^' logical history summary   *)
              (* does not equal the initial history summary.  This follows *)
              (* from an enablement condition in the LL2TakeCheckpoint     *)
              (* action, namely that LL2SPCR does not equal BaseHashValue. *)
              (*************************************************************)
              <7>5. LL2NVRAMLogicalHistorySummary # ll2InitialHistorySummary
                <8>1. LL2NVRAMLogicalHistorySummary.extension = LL2SPCR
                  <9>1. LL2NVRAMLogicalHistorySummary = [
                            anchor |-> LL2NVRAM.historySummaryAnchor,
                            extension |-> LL2SPCR ]
                    <10>1. LL2NVRAM.extensionInProgress <=> TRUE    (** changed = by <=> *)
                      BY <3>1 DEF LL2TakeCheckpoint
                    <10>2. LL2SPCR # BaseHashValue
                      BY <3>1, <10>1 DEF LL2TakeCheckpoint
                    <10>3. QED
                      BY <10>1, <10>2 DEF LL2NVRAMLogicalHistorySummary
                  <9>2. QED
                    BY <9>1
                <8>2. LL2SPCR # BaseHashValue
                  BY <3>1 DEF LL2TakeCheckpoint
                <8>3. ll2InitialHistorySummary.extension = BaseHashValue
                  BY DEF ll2InitialHistorySummary
                <8>4. QED
                  BY <8>1, <8>2, <8>3
              (*************************************************************)
              (* Finally from HistorySummariesMatchDefinition, we can      *)
              (* conclude that the quantified                              *)
              (* HistorySummariesMatchRecursion predicate is satisfied.    *)
              (*************************************************************)
              <7>6. QED
                BY <7>1, <7>2, <7>3, <7>4, <7>5, HistorySummariesMatchDefinition
                DEF HistorySummariesMatchRecursion
            (***************************************************************)
            (* Second, we prove that the quantified                        *)
            (* HistorySummariesMatchRecursion predicate is satisfied for   *)
            (* the unprimed history summary in the `^\basicname^' spec and *)
            (* the primed logical history summary in the `^\fastname^'     *)
            (* spec.                                                       *)
            (***************************************************************)
            <6>2. HistorySummariesMatchRecursion(
                      LL1NVRAM.historySummary,
                      LL2NVRAMLogicalHistorySummary',
                      LL2NVRAM.hashBarrier)!(
                      input,
                      previousLL1HistorySummary,
                      previousLL2HistorySummary)
              (*************************************************************)
              (* The quantified predicate is satisfied for the unprimed    *)
              (* variables, because we picked appropriate values for the   *)
              (* quantifiers above.                                        *)
              (*************************************************************)
              <7>1. HistorySummariesMatchRecursion(
                        LL1NVRAM.historySummary,
                        LL2NVRAMLogicalHistorySummary,
                        LL2NVRAM.hashBarrier)!(
                        input,
                        previousLL1HistorySummary,
                        previousLL2HistorySummary)
                BY <6>1
              (*************************************************************)
              (* We expand the definition of the quantified predicate into *)
              (* its three constituent conjuncts.  The first two conjuncts *)
              (* follow directly from the quantified                       *)
              (* HistorySummariesMatchRecursion predicate for the unprimed *)
              (* history summaries.                                        *)
              (*************************************************************)
              <7>2. HistorySummariesMatch(
                        previousLL1HistorySummary,
                        previousLL2HistorySummary,
                        LL2NVRAM.hashBarrier)
                BY <7>1
              <7>3. LL1NVRAM.historySummary = Hash(previousLL1HistorySummary, input)
                BY <7>1
              (*************************************************************)
              (* We expand the definition of the                           *)
              (* LL2HistorySummaryIsSuccessor predicate and ignore the     *)
              (* first disjunct, since we know that the history summary    *)
              (* has been checkpointed.                                    *)
              (*************************************************************)
              <7>4. LL2HistorySummaryIsSuccessor(
                        LL2NVRAMLogicalHistorySummary',
                        previousLL2HistorySummary,
                        input,
                        LL2NVRAM.hashBarrier)
                <8> successorHistorySummary ==
                        Successor(previousLL2HistorySummary, input, LL2NVRAM.hashBarrier)
                <8> checkpointedSuccessorHistorySummary == Checkpoint(successorHistorySummary)
                <8> HIDE DEF successorHistorySummary, checkpointedSuccessorHistorySummary
                (***********************************************************)
                (* All of the work is in proving the second disjunct of    *)
                (* the LL2HistorySummaryIsSuccessor predicate.             *)
                (***********************************************************)
                <8>1. LL2NVRAMLogicalHistorySummary' = checkpointedSuccessorHistorySummary
                  (*********************************************************)
                  (* First, we prove that the                              *)
                  (* LL2NVRAMLogicalHistorySummary equals a record with    *)
                  (* the fields of new history summary anchor and base     *)
                  (* hash value.  This follows fairly directly from the    *)
                  (* definitions of LL2NVRAMLogicalHistorySummary and      *)
                  (* LL2TakeCheckpoint.                                    *)
                  (*********************************************************)
                  <9>1. LL2NVRAMLogicalHistorySummary' = [
                            anchor |-> newHistorySummaryAnchor,
                            extension |-> BaseHashValue ]
                    <10>1. LL2NVRAM' = [
                              historySummaryAnchor |-> newHistorySummaryAnchor,
                              symmetricKey |-> LL2NVRAM.symmetricKey,
                              hashBarrier |-> LL2NVRAM.hashBarrier,
                              extensionInProgress |-> FALSE ]
                      BY <3>1 DEF LL2TakeCheckpoint, newHistorySummaryAnchor
                    <10>2. LL2NVRAMLogicalHistorySummary' = [
                              anchor |-> LL2NVRAM.historySummaryAnchor',
                              extension |-> BaseHashValue ]
                      <11>1. LL2NVRAM.extensionInProgress' <=> FALSE    (** changed = by <=> *)
                        BY <10>1
                      <11>2. QED
                        BY <11>1 DEF LL2NVRAMLogicalHistorySummary
                    <10>3. LL2NVRAM.historySummaryAnchor' = newHistorySummaryAnchor
                      BY <10>1
                    <10>4. QED
                      BY <10>2, <10>3
                  (*********************************************************)
                  (* Second, we prove that the                             *)
                  (* checkpointedSuccessorHistorySummary also equals a     *)
                  (* record with the fields of new history summary anchor  *)
                  (* and base hash value.                                  *)
                  (*********************************************************)
                  <9>2. checkpointedSuccessorHistorySummary = [
                            anchor |-> newHistorySummaryAnchor,
                            extension |-> BaseHashValue ]
                    (*******************************************************)
                    (* The main step is proving that the successor history *)
                    (* summary equals the logical history summary in the   *)
                    (* `^\fastname^' NVRAM and SPCR.                       *)
                    (*******************************************************)
                    <10>1. successorHistorySummary = LL2NVRAMLogicalHistorySummary
                      (*****************************************************)
                      (* From the LL2HistorySummaryIsSuccessor predicate,  *)
                      (* we know that the logical history summary must     *)
                      (* either equal the successor or the checkpointed    *)
                      (* successor.                                        *)
                      (*****************************************************)
                      <11>1. \/ LL2NVRAMLogicalHistorySummary = successorHistorySummary
                             \/ LL2NVRAMLogicalHistorySummary = checkpointedSuccessorHistorySummary
                        <12>1. LL2HistorySummaryIsSuccessor(
                                   LL2NVRAMLogicalHistorySummary,
                                   previousLL2HistorySummary,
                                   input,
                                   LL2NVRAM.hashBarrier)
                          BY <7>1
                        <12>2. QED
                          BY <12>1
                          DEF LL2HistorySummaryIsSuccessor, successorHistorySummary,
                              checkpointedSuccessorHistorySummary
                      (*****************************************************)
                      (* The logical history summary does not equal the    *)
                      (* checkpointed successor.                           *)
                      (*****************************************************)
                      <11>2. LL2NVRAMLogicalHistorySummary # checkpointedSuccessorHistorySummary
                        (***************************************************)
                        (* The extension field of the logical history      *)
                        (* summary does not equal the base hash value.     *)
                        (* This follows from an enablement condition of    *)
                        (* the TakeCheckpoint action.                      *)
                        (***************************************************)
                        <12>1. LL2NVRAMLogicalHistorySummary.extension # BaseHashValue
                          <13>1. LL2NVRAMLogicalHistorySummary.extension = LL2SPCR
                            <14>1. LL2NVRAMLogicalHistorySummary = [
                                      anchor |-> LL2NVRAM.historySummaryAnchor,
                                      extension |-> LL2SPCR ]
                              <15>1. LL2NVRAM.extensionInProgress <=> TRUE    (** changed = by <=> *)
                                BY <3>1 DEF LL2TakeCheckpoint
                              <15>2. LL2SPCR # BaseHashValue
                                BY <3>1, <15>1 DEF LL2TakeCheckpoint
                              <15>3. QED
                                BY <15>1, <15>2 DEF LL2NVRAMLogicalHistorySummary
                            <14>2. QED
                              BY <14>1
                          <13>2. LL2SPCR # BaseHashValue
                            BY <3>1 DEF LL2TakeCheckpoint
                          <13>3. QED
                            BY <13>1, <13>2
                        (***************************************************)
                        (* The extension field of the checkpointed         *)
                        (* successor does equal the base hash value.  This *)
                        (* follows from the                                *)
                        (* CheckpointHasBaseExtensionLemma.                *)
                        (***************************************************)
                        <12>2. checkpointedSuccessorHistorySummary.extension = BaseHashValue
                          <13>1. successorHistorySummary \in HistorySummaryType
                            <14>1. previousLL2HistorySummary \in HistorySummaryType
                              BY <6>1
                            <14>2. input \in InputType
                              BY <6>1
                            <14>3. LL2NVRAM.hashBarrier \in HashType
                              BY <2>1, LL2SubtypeImplicationLemma DEF LL2SubtypeImplication
                            <14>4. QED
                              BY <14>1, <14>2, <14>3, SuccessorTypeSafe DEF successorHistorySummary
                          <13>2. QED
                            BY <13>1, CheckpointHasBaseExtensionLemma
                            DEF checkpointedSuccessorHistorySummary
                        (***************************************************)
                        (* Since he extension field of the logical history *)
                        (* summary does not equal the base hash value, but *)
                        (* the extension field of the checkpointed         *)
                        (* successor does equal the base hash value, it    *)
                        (* follows that these two records are unequal.     *)
                        (***************************************************)
                        <12>3. QED
                          BY <12>1, <12>2
                      <11>3. QED
                        BY <11>1, <11>2
                    (*******************************************************)
                    (* Given that the successor history summary equals the *)
                    (* logical history summary in the `^\fastname^' NVRAM  *)
                    (* and SPCR, we can readily derive the specific field  *)
                    (* values for this record.                             *)
                    (*******************************************************)
                    <10>2. successorHistorySummary = [
                               anchor |-> LL2NVRAM.historySummaryAnchor,
                               extension |-> LL2SPCR ]
                      <11>1. LL2NVRAMLogicalHistorySummary = [
                                 anchor |-> LL2NVRAM.historySummaryAnchor,
                                 extension |-> LL2SPCR ]
                        <12>1. LL2NVRAM.extensionInProgress <=> TRUE    (** changed = by <=> *)
                          BY <3>1 DEF LL2TakeCheckpoint
                        <12>2. LL2SPCR # BaseHashValue
                          BY <3>1, <12>1 DEF LL2TakeCheckpoint
                        <12>3. QED
                          BY <12>1, <12>2 DEF LL2NVRAMLogicalHistorySummary
                      <11>2. QED
                        BY <10>1, <11>1
                    (*******************************************************)
                    (* Now, we merely need to show that the the extension  *)
                    (* field is not equal to the base hash value, so the   *)
                    (* definition of the Checkpoint operator yields the    *)
                    (* same value as the TakeCheckpoint action.            *)
                    (*******************************************************)
                    <10>3. QED
                      <11>1. successorHistorySummary.extension # BaseHashValue
                        <12>1. successorHistorySummary.extension = LL2SPCR
                          BY <10>2
                        <12>2. LL2SPCR # BaseHashValue
                          BY <3>1 DEF LL2TakeCheckpoint
                        <12>3. QED
                          BY <12>1, <12>2
                      <11>2. QED
                        BY <10>2, <11>1
                        DEF checkpointedSuccessorHistorySummary, Checkpoint, newHistorySummaryAnchor
                  (*********************************************************)
                  (* Since the LL2NVRAMLogicalHistorySummary and the       *)
                  (* checkpointedSuccessorHistorySummary each equal the    *)
                  (* same value, they equal each other.                    *)
                  (*********************************************************)
                  <9>3. QED
                    BY <9>1, <9>2

                <8>2. QED
                  BY <8>1
                  DEF LL2HistorySummaryIsSuccessor, checkpointedSuccessorHistorySummary,
                      successorHistorySummary
              (*************************************************************)
              (* The predicate is satisfied because each of its conjuncts  *)
              (* is satisfied.                                             *)
              (*************************************************************)
              <7>5. QED
                BY <7>2, <7>3, <7>4
            (***************************************************************)
            (* Third, we prove that for the unprimed `^\basicname^'        *)
            (* history summary and the primed `^\fastname^' logical        *)
            (* history summary, the HistorySummariesMatch predicate equals *)
            (* the quantified HistorySummariesMatchRecursion predicate.    *)
            (***************************************************************)
            <6>3. HistorySummariesMatch(
                      LL1NVRAM.historySummary,
                      LL2NVRAMLogicalHistorySummary',
                      LL2NVRAM.hashBarrier)
                <=> HistorySummariesMatchRecursion(     (* changed = by <=> *)
                      LL1NVRAM.historySummary,
                      LL2NVRAMLogicalHistorySummary',
                      LL2NVRAM.hashBarrier)
              (*************************************************************)
              (* We prove some types, to satisfy the universal quantifiers *)
              (* in HistorySummariesMatchDefinition.                       *)
              (*************************************************************)
              <7>1. LL1NVRAM.historySummary \in HashType
                BY <2>1 DEF LL2Refinement, LL1TrustedStorageType
              <7>2. LL2NVRAMLogicalHistorySummary' \in HistorySummaryType
                BY <2>1, LL2NVRAMLogicalHistorySummaryTypeSafe
              <7>3. LL2NVRAM.hashBarrier \in HashType
                BY <2>1, LL2SubtypeImplicationLemma DEF LL2SubtypeImplication
              (*************************************************************)
              (* We prove that the `^\fastname^' logical history summary   *)
              (* does not equal the initial history summary.  There are    *)
              (* three sub-steps.                                          *)
              (*************************************************************)
              <7>4. LL2NVRAMLogicalHistorySummary' # ll2InitialHistorySummary
                (***********************************************************)
                (* First, we prove that the primed value of the anchor     *)
                (* field in the `^\fastname^' logical history summary      *)
                (* equals the new history summary anchor defined in the    *)
                (* TakeCheckpoint action.  This follows fairly directly    *)
                (* from the definitions of the TakeCheckpoint action and   *)
                (* the LL2NVRAMLogicalHistorySummary operator.             *)
                (***********************************************************)
                <8>1. LL2NVRAMLogicalHistorySummary.anchor' = newHistorySummaryAnchor
                  <9>1. LL2NVRAM' = [
                            historySummaryAnchor |-> newHistorySummaryAnchor,
                            symmetricKey |-> LL2NVRAM.symmetricKey,
                            hashBarrier |-> LL2NVRAM.hashBarrier,
                            extensionInProgress |-> FALSE ]
                    BY <3>1 DEF LL2TakeCheckpoint, newHistorySummaryAnchor
                  <9>2. LL2NVRAMLogicalHistorySummary' = [
                            anchor |-> LL2NVRAM.historySummaryAnchor',
                            extension |-> BaseHashValue ]
                    <10>1. LL2NVRAM.extensionInProgress' <=> FALSE    (** changed = by <=> *)
                      BY <9>1
                    <10>2. QED
                      BY <10>1 DEF LL2NVRAMLogicalHistorySummary
                  <9>3. LL2NVRAM.historySummaryAnchor' = newHistorySummaryAnchor
                    BY <9>1
                  <9>4. QED
                    BY <9>2, <9>3
                (***********************************************************)
                (* Second, we prove that the new history summary anchor is *)
                (* not equal to the base hash value.  This follows because *)
                (* the new history summary anchor is generated as a hash   *)
                (* by the TakeCheckpoint action, and the                   *)
                (* BaseHashValueUnique property tells us that no hash      *)
                (* value generated by the Hash function can equal the base *)
                (* hash value.                                             *)
                (***********************************************************)
                <8>2. newHistorySummaryAnchor # BaseHashValue
                  <9>1. LL2NVRAM.historySummaryAnchor \in HashDomain
                    <10>1. LL2NVRAM.historySummaryAnchor \in HashType
                      BY <2>1, LL2SubtypeImplicationLemma DEF LL2SubtypeImplication
                    <10>2. QED
                      BY <10>1 DEF HashDomain
                  <9>2. LL2SPCR \in HashDomain
                    <10>1. LL2SPCR \in HashType
                      BY <2>1 DEF LL2TypeInvariant
                    <10>2. QED
                      BY <10>1 DEF HashDomain
                  <9>3. QED
                    BY <9>1, <9>2, BaseHashValueUnique DEF newHistorySummaryAnchor
                (***********************************************************)
                (* Third, the anchor field of the initial history summary  *)
                (* equals the base hash value, by definition.              *)
                (***********************************************************)
                <8>3. ll2InitialHistorySummary.anchor = BaseHashValue
                  BY DEF ll2InitialHistorySummary
                <8>4. QED
                  BY <8>1, <8>2, <8>3
              (*************************************************************)
              (* Finally, from HistorySummariesMatchDefinition, we can     *)
              (* conclude that the HistorySummariesMatch predicate equals  *)
              (* the quantified HistorySummariesMatchRecursion predicate.  *)
              (*************************************************************)
              <7>5. QED
                BY <7>1, <7>2, <7>3, <7>4, HistorySummariesMatchDefinition
                DEF ll2InitialHistorySummary
            (***************************************************************)
            (* We tie the above steps together by first deriving the       *)
            (* unquantified HistorySummariesMatchRecursion predicate from  *)
            (* the quantified predicate, and then proving the              *)
            (* straightforward equality.                                   *)
            (***************************************************************)
            <6>4. QED
              <7>1. HistorySummariesMatchRecursion(
                        LL1NVRAM.historySummary,
                        LL2NVRAMLogicalHistorySummary',
                        LL2NVRAM.hashBarrier)
                <8>1. /\  input \in InputType
                      /\  previousLL1HistorySummary \in HashType
                      /\  previousLL2HistorySummary \in HistorySummaryType
                  BY <6>1
                <8>2. QED
                  BY <6>2, <8>1 DEF HistorySummariesMatchRecursion
              <7>2. QED
                BY <6>2, <6>3, <7>1
          (*****************************************************************)
          (* We use the HistorySummariesMatchUniqueLemma to conclude that  *)
          (* both unprimed and unprimed history summaries in the           *)
          (* `^\basicname^' NVRAM are equal.                               *)
          (*****************************************************************)
          <5>3. QED
            <6>1. LL1NVRAM.historySummary' \in HashType
              BY <2>1 DEF LL2Refinement, LL1TrustedStorageType
            <6>2. LL1NVRAM.historySummary \in HashType
              BY <2>1 DEF LL2Refinement, LL1TrustedStorageType
            <6>3. LL2NVRAMLogicalHistorySummary' \in HistorySummaryType
              BY <2>1, LL2NVRAMLogicalHistorySummaryTypeSafe
            <6>4. LL2NVRAM.hashBarrier \in HashType
              BY <2>1, LL2SubtypeImplicationLemma DEF LL2SubtypeImplication
            <6>5. QED
              BY <5>1, <5>2, <6>1, <6>2, <6>3, <6>4, HistorySummariesMatchUniqueLemma
        (*******************************************************************)
        (* Proving the UNCHANGED status of the symmetric key in the        *)
        (* `^\basicname^' NVRAM is straightforward, using the lemma we     *)
        (* have proven for this purpose.                                   *)
        (*******************************************************************)
        <4>3. UNCHANGED LL1NVRAM.symmetricKey
          <5>1. UNCHANGED LL2NVRAM.symmetricKey
            BY <3>2
          <5>2. QED
            BY <2>1, <5>1, UnchangedNVRAMSymmetricKeyLemma
        (*******************************************************************)
        (* We can then use the LL1NVRAMRecordCompositionLemma directly,    *)
        (* once we prove some types.                                       *)
        (*******************************************************************)
        <4>8. QED
          <5>1. LL1NVRAM \in LL1TrustedStorageType
            BY <2>1 DEF LL2Refinement
          <5>2. LL1NVRAM' \in LL1TrustedStorageType
            BY <2>1 DEF LL2Refinement
          <5>3. QED
            BY <4>2, <4>3, <5>1, <5>2, LL1NVRAMRecordCompositionLemma
      (*********************************************************************)
      (* All of the `^\basicname^' variables are unchanged.                *)
      (*********************************************************************)
      <3>9. QED
        BY <3>3, <3>4, <3>5, <3>6, <3>7, <3>8 DEF LL1Vars
    (***********************************************************************)
    (* A `^\fastname^' LL2Restart action refines to one of two actions in  *)
    (* the `^\basicname^' spec.  If an extension is in progress at the     *)
    (* time the LL2Restart occurs, the loss of the SPCR value caused by    *)
    (* the LL2Restart is fatal, so this refines to a                       *)
    (* LL1RestrictedCorruption action in the `^\basicname^' spec, which in *)
    (* turn refines to an HLDie action in the high-level spec.  On the     *)
    (* other hand, if an extension is not in progress at the time the      *)
    (* LL2Restart occurs, the action refines to an LL1Restart action in    *)
    (* the `^\basicname^' spec.                                            *)
    (***********************************************************************)
    <2>7. LL2Restart =>
            IF LL2NVRAM.extensionInProgress
            THEN
                LL1RestrictedCorruption
            ELSE
                LL1Restart
      (*********************************************************************)
      (* We assume the antecedent.                                         *)
      (*********************************************************************)
      <3>1. HAVE LL2Restart
      (*********************************************************************)
      (* We pick a set of variables of the appropriate types that satisfy  *)
      (* the LL2CorruptRAM action.                                         *)
      (*********************************************************************)
      <3>2. PICK ll2UntrustedStorage \in LL2UntrustedStorageType,
                 ll2RandomSymmetricKey \in SymmetricKeyType \ {LL2NVRAM.symmetricKey},
                 ll2Hash \in HashType :
                 LL2Restart!(ll2UntrustedStorage, ll2RandomSymmetricKey, ll2Hash)
        BY <3>1 DEF LL2Restart
      (*********************************************************************)
      (* We first prove that the primed state of the `^\basicname^' RAM    *)
      (* has a value that satisfies a particular constraint that is        *)
      (* imposed by both the LL1RestrictedCorruption action and by the     *)
      (* LL1Restart action.                                                *)
      (*********************************************************************)
      <3>3. \E  ll1UntrustedStorage \in LL1UntrustedStorageType,
                ll1RandomSymmetricKey \in SymmetricKeyType \ {LL1NVRAM.symmetricKey},
                ll1Hash \in HashType :
                /\  ll1UntrustedStorage.authenticator =
                        GenerateMAC(ll1RandomSymmetricKey, ll1Hash)
                /\  LL1RAM' = ll1UntrustedStorage
        (*******************************************************************)
        (* We pick a symmetric key that satisfies the AuthenticatorsMatch  *)
        (* predicate for the primed states of the authenticators in the    *)
        (* RAM variables of the two specs.                                 *)
        (*******************************************************************)
        <4>3. PICK symmetricKey \in SymmetricKeyType :
                    AuthenticatorsMatch(
                        LL1RAM.authenticator',
                        LL2RAM.authenticator',
                        symmetricKey,
                        LL2NVRAM.hashBarrier')
          BY <2>1 DEF LL2Refinement
        (*******************************************************************)
        (* We pick a set of variables of the appropriate types that        *)
        (* satisfy the quantified AuthenticatorsMatch predicate.           *)
        (*******************************************************************)
        <4>4. PICK stateHash \in HashType,
                    ll1HistorySummary \in HashType,
                    ll2HistorySummary \in HistorySummaryType :
                    AuthenticatorsMatch(
                        LL1RAM.authenticator',
                        LL2RAM.authenticator',
                        symmetricKey,
                        LL2NVRAM.hashBarrier')!(
                        stateHash, ll1HistorySummary, ll2HistorySummary)!1
          BY <4>3 DEF AuthenticatorsMatch
        (*******************************************************************)
        (* We re-state the definitions from the LET in                     *)
        (* AuthenticatorsMatch.                                            *)
        (*******************************************************************)
        <4> ll1HistoryStateBinding == Hash(ll1HistorySummary, stateHash)
        <4> ll2HistorySummaryHash == Hash(ll2HistorySummary.anchor, ll2HistorySummary.extension)
        <4> ll2HistoryStateBinding == Hash(ll2HistorySummaryHash, stateHash)
        (*******************************************************************)
        (* We prove the types of the definitions, with help from the       *)
        (* AuthenticatorsMatchDefsTypeSafeLemma.                           *)
        (*******************************************************************)
        <4>5. /\  ll1HistoryStateBinding \in HashType
              /\  ll2HistorySummaryHash \in HashType
              /\  ll2HistoryStateBinding \in HashType
          BY <4>4, AuthenticatorsMatchDefsTypeSafeLemma
        (*******************************************************************)
        (* We hide the definitions.                                        *)
        (*******************************************************************)
        <4> HIDE DEF ll1HistoryStateBinding, ll2HistorySummaryHash, ll2HistoryStateBinding
        (*******************************************************************)
        (* To prove the constraints regarding the primed state of the      *)
        (* `^\basicname^' RAM, we first prove the types of the three       *)
        (* witnesses for the existentially quantified variables.           *)
        (*******************************************************************)
        <4>6. LL1RAM' \in LL1UntrustedStorageType
          BY <2>1 DEF LL2Refinement
        <4>7. ll2RandomSymmetricKey \in SymmetricKeyType \ {LL1NVRAM.symmetricKey}
          <5>1. ll2RandomSymmetricKey \in SymmetricKeyType \ {LL2NVRAM.symmetricKey}
            BY <3>2
          <5>2. LL1NVRAM.symmetricKey = LL2NVRAM.symmetricKey
            BY <2>1 DEF LL2Refinement
          <5>3. QED
            BY <5>1, <5>2
        <4>8. ll1HistoryStateBinding \in HashType
          BY <4>5
        (*******************************************************************)
        (* We prove that the authenticator in the `^\basicname^' spec is   *)
        (* generated as a MAC with a random symmetric key.  This follows   *)
        (* directly from the AuthenticatorGeneratedLemma, once we prove    *)
        (* the preconditions for the lemma.                                *)
        (*******************************************************************)
        <4>9. LL1RAM.authenticator' =
                  GenerateMAC(ll2RandomSymmetricKey, ll1HistoryStateBinding)
          (*****************************************************************)
          (* We need to prove some types.                                  *)
          (*****************************************************************)
          <5>1. stateHash \in HashType
            BY <4>4
          <5>2. ll1HistorySummary \in HashType
            BY <4>4
          <5>3. ll2HistorySummary \in HistorySummaryType
            BY <4>4
          <5>4. LL1RAM.authenticator' \in MACType
            <6>1. LL1RAM' \in LL1UntrustedStorageType
              BY <2>1 DEF LL2Refinement
            <6>2. QED
              BY <6>1 DEF LL1UntrustedStorageType
          <5>5. LL2RAM.authenticator' \in MACType
            BY <2>1, LL2SubtypeImplicationLemma DEF LL2SubtypeImplication
          <5>6. ll2RandomSymmetricKey \in SymmetricKeyType
            <6>1. ll2RandomSymmetricKey \in SymmetricKeyType \ {LL2NVRAM.symmetricKey}
              BY <3>2
            <6>2. QED
              BY <6>1
          <5>7. LL2NVRAM.hashBarrier' \in HashType
              BY <2>1, LL2SubtypeImplicationLemma DEF LL2SubtypeImplication, LL2TypeInvariant, LL2TrustedStorageType
          (*****************************************************************)
          (* Then we prove the three conjuncts in the antecedent of the    *)
          (* AuthenticatorGeneratedLemma.  The first conjunct follows from *)
          (* a conjunct in the refinement.                                 *)
          (*****************************************************************)
          <5>8. HistorySummariesMatch(ll1HistorySummary, ll2HistorySummary, LL2NVRAM.hashBarrier')
            BY <4>4
          (*****************************************************************)
          (* The second conjunct in the antecedent of the                  *)
          (* AuthenticatorGeneratedLemma mainly follows from the           *)
          (* refinement, but we also have to prove that the symmetric key  *)
          (* specified by an existential in the refinement matches the     *)
          (* random symmetric key specified in the LL2CorruptRAM action.   *)
          (* This follows from the MACUnforgeable property.                *)
          (*****************************************************************)
          <5>9. AuthenticatorsMatch(
                    LL1RAM.authenticator',
                    LL2RAM.authenticator',
                    ll2RandomSymmetricKey,
                    LL2NVRAM.hashBarrier')
            (***************************************************************)
            (* The main precondition for the MACUnforgeable property is    *)
            (* that the generated MAC is validated.  We first prove the    *)
            (* validation, which follows from the refinement.              *)
            (***************************************************************)
            <6>1. ValidateMAC(symmetricKey, ll2HistoryStateBinding, LL2RAM.authenticator')
              BY <4>4 DEF ll2HistoryStateBinding, ll2HistorySummaryHash
            (***************************************************************)
            (* We then prove the generation, which follows from the        *)
            (* LL2CorruptRAM action.                                       *)
            (***************************************************************)
            <6>2. LL2RAM.authenticator' = GenerateMAC(ll2RandomSymmetricKey, ll2Hash)
              <7>1. LL2RAM' = ll2UntrustedStorage
                BY <3>2
              <7>2. ll2UntrustedStorage.authenticator = GenerateMAC(ll2RandomSymmetricKey, ll2Hash)
                BY <4>2
              <7>3. QED
                BY <7>1, <7>2
            (***************************************************************)
            (* The remaining preconditions are types.                      *)
            (***************************************************************)
            <6>3. symmetricKey \in SymmetricKeyType
              BY <4>3
            <6>4. ll2RandomSymmetricKey \in SymmetricKeyType
              <7>1. ll2RandomSymmetricKey \in SymmetricKeyType \ {LL2NVRAM.symmetricKey}
                BY <3>2
              <7>2. QED
                BY <7>1
            <6>5. ll2HistoryStateBinding \in HashType
              BY <4>5
            <6>6. ll2Hash \in HashType
              BY <3>2
            (***************************************************************)
            (* The MACUnforgeable property tells us that the two keys are  *)
            (* equal.                                                      *)
            (***************************************************************)
            <6>7. symmetricKey = ll2RandomSymmetricKey
                BY <6>1, <6>2, <6>3, <6>4, <6>5, <6>6, MACUnforgeable
            <6>8. QED
              BY <4>3, <6>7
          (*****************************************************************)
          (* The third conjunct in the antecedent of the                   *)
          (* AuthenticatorGeneratedLemma mainly follows from the           *)
          (* LL2CorruptRAM action, but we also have to prove that the      *)
          (* history state binding specified in the refinement matches the *)
          (* arbitrary hash specified in the LL2CorruptRAM action.         *)
          (*****************************************************************)
          <5>10. LL2RAM.authenticator' =
                         GenerateMAC(ll2RandomSymmetricKey, ll2HistoryStateBinding)
            (***************************************************************)
            (* The state authenticaton in the primed `^\fastname^' RAM     *)
            (* equals the authenticator specified by the existential       *)
            (* ll2UntrustedStorage in the LL2CorruptRAM action.            *)
            (***************************************************************)
            <6>1. LL2RAM.authenticator' = ll2UntrustedStorage.authenticator
              <7>1. LL2RAM' = ll2UntrustedStorage
                BY <3>2
              <7>2. QED
                BY <7>1
            (***************************************************************)
            (* The authenticator specified by the existential              *)
            (* ll2UntrustedStorage in the LL2CorruptRAM action is          *)
            (* generated as a MAC of the history state binding from the    *)
            (* AuthenticatorsMatch predicate invoked by the refinement.    *)
            (* This follows from the MACCollisionResistant property.       *)
            (***************************************************************)
            <6>2. ll2UntrustedStorage.authenticator =
                      GenerateMAC(ll2RandomSymmetricKey, ll2HistoryStateBinding)
              (*************************************************************)
              (* The main precondition for the MACUnforgeable property is  *)
              (* that the generated MAC is validated.  We first prove the  *)
              (* validation, which follows from the refinement.            *)
              (*************************************************************)
              <7>1. ValidateMAC(symmetricKey, ll2HistoryStateBinding, LL2RAM.authenticator')
                BY <4>4 DEF ll2HistoryStateBinding, ll2HistorySummaryHash
              (*************************************************************)
              (* We then prove the generation, which follows from the      *)
              (* LL2CorruptRAM action.                                     *)
              (*************************************************************)
              <7>2. LL2RAM.authenticator' = GenerateMAC(ll2RandomSymmetricKey, ll2Hash)
                <8>1. LL2RAM' = ll2UntrustedStorage
                  BY <3>2
                <8>2. ll2UntrustedStorage.authenticator = GenerateMAC(ll2RandomSymmetricKey, ll2Hash)
                  BY <4>2
                <8>3. QED
                  BY <8>1, <8>2
              (*************************************************************)
              (* The remaining preconditions are types.                    *)
              (*************************************************************)
              <7>3. symmetricKey \in SymmetricKeyType
                BY <4>3
              <7>4. ll2RandomSymmetricKey \in SymmetricKeyType
                <8>1. ll2RandomSymmetricKey \in SymmetricKeyType \ {LL2NVRAM.symmetricKey}
                  BY <3>2
                <8>2. QED
                  BY <8>1
              <7>5. ll2HistoryStateBinding \in HashType
                BY <4>5
              <7>6. ll2Hash \in HashType
                BY <3>2
              (*************************************************************)
              (* The MACCollisionResistant property tells us that the two  *)
              (* hash values are equal.                                    *)
              (*************************************************************)
              <7>7. ll2Hash = ll2HistoryStateBinding
                BY <7>1, <7>2, <7>3, <7>4, <7>5, <7>6, MACCollisionResistant
              <7>8. QED
                BY <3>2, <7>7
            <6>3. QED
              BY <6>1, <6>2
          (*****************************************************************)
          (* We then invoke the AuthenticatorGeneratedLemma directly.      *)
          (*****************************************************************)
          <5>11. QED
            BY <5>1, <5>2, <5>3, <5>4, <5>5, <5>6, <5>7, <5>8, <5>9, <5>10,
            AuthenticatorGeneratedLemma
            DEF ll1HistoryStateBinding, ll2HistorySummaryHash, ll2HistoryStateBinding
        <4>10. QED
          BY <4>6, <4>7, <4>8, <4>9
      (*********************************************************************)
      (* Since the refinement is an IF-THEN-ELSE, we prove the THEN and    *)
      (* ELSE cases separately.  For the THEN case, we assume that an      *)
      (* extension is in progress and show that this refines to a          *)
      (* LL1RestrictedCorruption action.                                   *)
      (*********************************************************************)
      <3>4. ASSUME LL2NVRAM.extensionInProgress
            PROVE LL1RestrictedCorruption
        (*******************************************************************)
        (* One fact that will be useful in several places below is that    *)
        (* the extension field in the logical history summary of the       *)
        (* `^\fastname^' NVRAM and SPCR is equal to a crazy hash value,    *)
        (* because an extension is in progress but the SPCR (in the primed *)
        (* state) equals the base hash value.                              *)
        (*******************************************************************)
        <4>1. LL2NVRAMLogicalHistorySummary.extension' = CrazyHashValue
          <5>1. LL2NVRAMLogicalHistorySummary' = [
                    anchor |-> LL2NVRAM.historySummaryAnchor',
                    extension |-> CrazyHashValue ]
            <6>1. LL2NVRAM.extensionInProgress'
              <7>1. LL2NVRAM.extensionInProgress
                BY <3>3
              <7>2. UNCHANGED LL2NVRAM.extensionInProgress
                BY <3>2
              <7>3. QED
                BY <7>1, <7>2
            <6>2. LL2SPCR' = BaseHashValue
              BY <3>2
            <6>3. QED
              BY <6>1, <6>2 DEF LL2NVRAMLogicalHistorySummary
          <5>2. QED
            BY <5>1
        (*******************************************************************)
        (* We prove each conjunct of LL1RestrictedCorruption separately.   *)
        (* First, we prove the conjunct relating to the NVRAM.             *)
        (*******************************************************************)
        <4>2. LL1RestrictedCorruption!nvram
          (*****************************************************************)
          (* The primed value of the history summary field in the          *)
          (* `^\basicname^' NVRAM serves as our witness for the garbage    *)
          (* history summary.                                              *)
          (*****************************************************************)
          <5>1. LL1NVRAM.historySummary' \in HashType
            BY <2>1 DEF LL2Refinement, LL1TrustedStorageType
          (*****************************************************************)
          (* We prove that the constraint labeled `current' in the         *)
          (* LL1RestrictedCorruption action is satisfied.                  *)
          (*****************************************************************)
          <5>2. LL1RestrictedCorruption!nvram!current(LL1NVRAM.historySummary')
            (***************************************************************)
            (* To prove the universally quantified expression, we take a   *)
            (* set of variables of the appropriate types.                  *)
            (***************************************************************)
            <6>1. TAKE stateHash1 \in HashType,
                       ll1Authenticator \in LL1ObservedAuthenticators
            (***************************************************************)
            (* We re-state the definition from within the                  *)
            (* LL1RestrictedCorruption!nvram!current clause.               *)
            (***************************************************************)
            <6> ll1GarbageHistoryStateBinding == Hash(LL1NVRAM.historySummary', stateHash1)
            (***************************************************************)
            (* We hide the definition.                                     *)
            (***************************************************************)
            <6> HIDE DEF ll1GarbageHistoryStateBinding
            (***************************************************************)
            (* We need to prove the nvram::current conjunct, which asserts *)
            (* that the authenticator is not a valid MAC for the history   *)
            (* state binding formed from the history summary in the NVRAM  *)
            (* and any state hash.                                         *)
            (***************************************************************)
            <6>2. ~ValidateMAC(
                      LL1NVRAM.symmetricKey,
                      ll1GarbageHistoryStateBinding,
                      ll1Authenticator)
              (*************************************************************)
              (* We will use proof by contradiction.                       *)
              (*************************************************************)
              <7>1. SUFFICES
                    ASSUME
                        ValidateMAC(
                            LL1NVRAM.symmetricKey,
                            ll1GarbageHistoryStateBinding,
                            ll1Authenticator)
                    PROVE
                        FALSE
                OBVIOUS
              (*************************************************************)
              (* We first pick, from the set of `^\fastname^' observed     *)
              (* authenticators, a `^\fastname^' authenticator that        *)
              (* matches the `^\basicname^' authenticator.  We know that   *)
              (* such a authenticator exists, because the refinement       *)
              (* asserts that the sets of observed authenticators match    *)
              (* across the two specs.                                     *)
              (*************************************************************)
              <7>2. PICK ll2Authenticator \in LL2ObservedAuthenticators :
                        AuthenticatorsMatch(
                            ll1Authenticator,
                            ll2Authenticator,
                            LL2NVRAM.symmetricKey,
                            LL2NVRAM.hashBarrier)
                <8>1. ll1Authenticator \in LL1ObservedAuthenticators
                  BY <6>1
                <8>2. AuthenticatorSetsMatch(
                          LL1ObservedAuthenticators,
                          LL2ObservedAuthenticators,
                          LL2NVRAM.symmetricKey,
                          LL2NVRAM.hashBarrier)
                  BY <2>1 DEF LL2Refinement
                <8>3. QED
                  BY <8>1, <8>2 DEF AuthenticatorSetsMatch
              (*************************************************************)
              (* We pick a set of variables of the appropriate types that  *)
              (* satisfy the quantified AuthenticatorsMatch predicate.     *)
              (*************************************************************)
              <7>3. PICK stateHash2 \in HashType,
                        ll1HistorySummary \in HashType,
                        ll2HistorySummary \in HistorySummaryType :
                        AuthenticatorsMatch(
                            ll1Authenticator,
                            ll2Authenticator,
                            LL2NVRAM.symmetricKey,
                            LL2NVRAM.hashBarrier)!(
                            stateHash2, ll1HistorySummary, ll2HistorySummary)!1
                BY <7>2 DEF AuthenticatorsMatch
              (*************************************************************)
              (* We re-state the definitions from the LET in               *)
              (* AuthenticatorsMatch.                                      *)
              (*************************************************************)
              <7> ll1HistoryStateBinding == Hash(ll1HistorySummary, stateHash2)
              <7> ll2HistorySummaryHash == Hash(ll2HistorySummary.anchor, ll2HistorySummary.extension)
              <7> ll2HistoryStateBinding == Hash(ll2HistorySummaryHash, stateHash2)
              (*************************************************************)
              (* We prove the types of the definitions, with help from the *)
              (* AuthenticatorsMatchDefsTypeSafeLemma.                     *)
              (*************************************************************)
              <7>4. /\  ll1HistoryStateBinding \in HashType
                    /\  ll2HistorySummaryHash \in HashType
                    /\  ll2HistoryStateBinding \in HashType
                BY <7>3, AuthenticatorsMatchDefsTypeSafeLemma
              (*************************************************************)
              (* We hide the definitions.                                  *)
              (*************************************************************)
              <7> HIDE DEF ll1HistoryStateBinding, ll2HistorySummaryHash, ll2HistoryStateBinding
              (*************************************************************)
              (* We prove that the `^\basicname^'s history summary picked  *)
              (* to satisfy the AuthenticatorsMatch predicate equals the   *)
              (* history summary in the primed state of the `^\basicname^' *)
              (* NVRAM.                                                    *)
              (*************************************************************)
              <7>5. ll1HistorySummary = LL1NVRAM.historySummary'
                (***********************************************************)
                (* The first step is to show the equality of the history   *)
                (* state bindings that bind each of these history          *)
                (* summaries to their respective state hashes.             *)
                (***********************************************************)
                <8>1. ll1HistoryStateBinding = ll1GarbageHistoryStateBinding
                  (*********************************************************)
                  (* By hypothesis, the authenticator is a valid MAC for   *)
                  (* the garbage history state binding.                    *)
                  (*********************************************************)
                  <9>1. ValidateMAC(
                            LL2NVRAM.symmetricKey, ll1GarbageHistoryStateBinding, ll1Authenticator)
                    <10>1. LL1NVRAM.symmetricKey = LL2NVRAM.symmetricKey
                      BY <2>1 DEF LL2Refinement
                    <10>2. QED
                      BY <7>1, <10>1
                  (*********************************************************)
                  (* The definition of the AuthenticatorsMatch predicate   *)
                  (* tells us that the `^\basicname^' authenticator was    *)
                  (* generated as a MAC from the history state binding.    *)
                  (*********************************************************)
                  <9>2. ll1Authenticator = GenerateMAC(LL2NVRAM.symmetricKey, ll1HistoryStateBinding)
                    BY <7>3 DEF ll1HistoryStateBinding
                  (*********************************************************)
                  (* The remaining preconditions are types.                *)
                  (*********************************************************)
                  <9>3. LL2NVRAM.symmetricKey \in SymmetricKeyType
                    BY <2>1, LL2SubtypeImplicationLemma DEF LL2SubtypeImplication
                  <9>4. ll1HistoryStateBinding \in HashType
                    BY <7>4
                  <9>5. ll1GarbageHistoryStateBinding \in HashType
                    <10>1. LL1NVRAM.historySummary' \in HashDomain
                      BY <5>1 DEF HashDomain
                    <10>2. stateHash1 \in HashDomain
                      BY <5>1 DEF HashDomain
                    <10>3. QED
                      BY <10>1, <10>2, HashTypeSafe DEF ll1GarbageHistoryStateBinding
                  (*********************************************************)
                  (* The MACCollisionResistant property tells us that the  *)
                  (* two history state bindings are equal.                 *)
                  (*********************************************************)
                  <9>6. QED
                    BY <9>1, <9>2, <9>3, <9>4, <9>5, MACCollisionResistant
                (***********************************************************)
                (* By the collision resistance of the hash function, the   *)
                (* equality of the history state bindings implies the      *)
                (* equality of the history summaries.                      *)
                (***********************************************************)
                <8>2. QED
                  <9>1. LL1NVRAM.historySummary' \in HashDomain
                    <10>1. LL1NVRAM.historySummary' \in HashType
                      BY <2>1 DEF LL2Refinement, LL1TrustedStorageType
                    <10>2. QED
                      BY <10>1 DEF HashDomain
                  <9>2. ll1HistorySummary \in HashDomain
                    <10>1. ll1HistorySummary \in HashType
                      BY <7>3
                    <10>2. QED
                      BY <10>1 DEF HashDomain
                  <9>3. stateHash1 \in HashDomain
                    <10>1. stateHash1 \in HashType
                      BY <6>1
                    <10>2. QED
                      BY <10>1 DEF HashDomain
                  <9>4. stateHash2 \in HashDomain
                    <10>1. stateHash2 \in HashType
                      BY <7>3
                   <10>2. QED
                      BY <10>1 DEF HashDomain
                  <9>5. QED
                    BY <8>1, <9>2, <9>1, <9>4, <9>3, HashCollisionResistant
                    DEF ll1HistoryStateBinding, ll1GarbageHistoryStateBinding
              (*************************************************************)
              (* The `^\basicname^'s history summary picked to satisfy the *)
              (* AuthenticatorsMatch predicate matches the primed logical  *)
              (* history summary in the `^\fastname^' NVRAM and SPCR, by   *)
              (* the refinement and the above equality.                    *)
              (*************************************************************)
              <7>6. HistorySummariesMatch(
                        ll1HistorySummary,
                        LL2NVRAMLogicalHistorySummary',
                        LL2NVRAM.hashBarrier')
                <8>1. HistorySummariesMatch(
                          LL1NVRAM.historySummary',
                          LL2NVRAMLogicalHistorySummary',
                          LL2NVRAM.hashBarrier')
                  BY <2>1 DEF LL2Refinement
                <8>2. QED
                  BY <7>5, <8>1
              (*************************************************************)
              (* We prove that the HistorySummariesMatch predicate equals  *)
              (* the HistorySummariesMatchRecursion predicate in this      *)
              (* case.  We assert each condition required by the           *)
              (* definition of the predicate.                              *)
              (*************************************************************)
              <7>7. HistorySummariesMatch(
                        ll1HistorySummary, LL2NVRAMLogicalHistorySummary', LL2NVRAM.hashBarrier') =
                    HistorySummariesMatchRecursion(
                        ll1HistorySummary, LL2NVRAMLogicalHistorySummary', LL2NVRAM.hashBarrier')
                (***********************************************************)
                (* We prove some types, to satisfy the universal           *)
                (* quantifiers in HistorySummariesMatchDefinition.         *)
                (***********************************************************)
                <8>1. ll1HistorySummary \in HashType
                  BY <7>3
                <8>2. LL2NVRAMLogicalHistorySummary' \in HistorySummaryType
                  BY <2>1, LL2NVRAMLogicalHistorySummaryTypeSafe
                <8>3. LL2NVRAM.hashBarrier' \in HashType
                  BY <2>1, LL2SubtypeImplicationLemma DEF LL2SubtypeImplication
                (***********************************************************)
                (* We prove that the `^\fastname^' logical history summary *)
                (* does not equal the initial history summary.             *)
                (***********************************************************)
                <8> ll2InitialHistorySummary == [ anchor |-> BaseHashValue, extension |-> BaseHashValue ]
                <8>4. LL2NVRAMLogicalHistorySummary' # ll2InitialHistorySummary
                  <9>1. LL2NVRAMLogicalHistorySummary.extension' # BaseHashValue
                    BY <4>1, CrazyHashValueUnique
                  <9>2. ll2InitialHistorySummary.extension = BaseHashValue
                    BY DEF ll2InitialHistorySummary
                  <9>3. QED
                    BY <9>1, <9>2
                (***********************************************************)
                (* Finally, from HistorySummariesMatchDefinition, we can   *)
                (* conclude that the HistorySummariesMatch predicate       *)
                (* equals the quantified HistorySummariesMatchRecursion    *)
                (* predicate.                                              *)
                (***********************************************************)
                <8>5. QED
                  BY <8>1, <8>2, <8>3, <8>4, HistorySummariesMatchDefinition
                  DEF ll2InitialHistorySummary
              (*************************************************************)
              (* We pick values for the existential variables inside the   *)
              (* HistorySummariesMatchRecursion predicate that satisfy the *)
              (* predicate.  We know such variables exist, because the     *)
              (* predicate is satisfied by the two previous steps.         *)
              (*************************************************************)
              <7>8. PICK prevInput \in InputType,
                         previousLL1HistorySummary \in HashType,
                         previousLL2HistorySummary \in HistorySummaryType :
                         HistorySummariesMatchRecursion(
                             ll1HistorySummary,
                             LL2NVRAMLogicalHistorySummary',
                             LL2NVRAM.hashBarrier')!(
                             prevInput,
                             previousLL1HistorySummary,
                             previousLL2HistorySummary)
                (***********************************************************)
                (* We prove some types, to satisfy the universal           *)
                (* quantifiers in HistorySummariesMatchRecursion           *)
                (* predicate.                                              *)
                (***********************************************************)
                <8>1. ll1HistorySummary \in HashType
                  BY <7>3
                <8>2. LL2NVRAMLogicalHistorySummary' \in HistorySummaryType
                  BY <2>1, LL2NVRAMLogicalHistorySummaryTypeSafe
                <8>3. LL2NVRAM.hashBarrier' \in HashType
                  BY <2>1, LL2SubtypeImplicationLemma DEF LL2SubtypeImplication,LL2TypeInvariant, LL2TrustedStorageType
                <8>4. HistorySummariesMatchRecursion(
                          ll1HistorySummary, LL2NVRAMLogicalHistorySummary', LL2NVRAM.hashBarrier')
                  BY <7>6, <7>7
                <8>5. QED
                  BY <8>1, <8>2, <8>4 DEF HistorySummariesMatchRecursion
              (*************************************************************)
              (* One of the conjuncts in the definition of                 *)
              (* HistorySummariesMatchRecursion is that the `^\fastname^'  *)
              (* history summary is a successor of a previous history      *)
              (* summary.                                                  *)
              (*************************************************************)
              <7>9. LL2HistorySummaryIsSuccessor(
                        LL2NVRAMLogicalHistorySummary',
                        previousLL2HistorySummary,
                        prevInput,
                        LL2NVRAM.hashBarrier')
                BY <7>8 DEF HistorySummariesMatchRecursion
              (*************************************************************)
              (* We re-state the definitions from the LET in               *)
              (* LL2HistorySummaryIsSuccessor.                             *)
              (*************************************************************)
              <7> successorHistorySummary ==
                      Successor(previousLL2HistorySummary, prevInput, LL2NVRAM.hashBarrier')
              <7> checkpointedSuccessorHistorySummary == Checkpoint(successorHistorySummary)
              (*************************************************************)
              (* We hide the definitions.                                  *)
              (*************************************************************)
              <7> HIDE DEF successorHistorySummary, checkpointedSuccessorHistorySummary
              (*************************************************************)
              (* The definition of LL2HistorySummaryIsSuccessor tells us   *)
              (* that there are two ways that the logical history summary  *)
              (* could be a successor.  We will prove that neither of      *)
              (* these disjuncts is satisfiable.                           *)
              (*************************************************************)
              <7>10. \/  LL2NVRAMLogicalHistorySummary' = successorHistorySummary
                     \/  LL2NVRAMLogicalHistorySummary' = checkpointedSuccessorHistorySummary
                BY <7>9
                DEF LL2HistorySummaryIsSuccessor, successorHistorySummary,
                    checkpointedSuccessorHistorySummary
              (*************************************************************)
              (* First, we prove that the logical history summary cannot   *)
              (* be a successor.                                           *)
              (*************************************************************)
              <7>11. LL2NVRAMLogicalHistorySummary' # successorHistorySummary
                (***********************************************************)
                (* We re-state a definition from the LET in the Successor  *)
                (* operator.                                               *)
                (***********************************************************)
                <8> securedInput == Hash(LL2NVRAM.hashBarrier', prevInput)
                (***********************************************************)
                (* We hide the definition.                                 *)
                (***********************************************************)
                <8> HIDE DEF securedInput
                (***********************************************************)
                (* There is only one sub-step, which is proving that the   *)
                (* extension fields of these two records are unequal.      *)
                (***********************************************************)
                <8>1. LL2NVRAMLogicalHistorySummary.extension' # successorHistorySummary.extension
                  (*********************************************************)
                  (* If an extension is in progress but the SPCR equals    *)
                  (* the base hash value, the logical history summary      *)
                  (* equals a crazy hash value, as proven above.           *)
                  (*********************************************************)
                  <9>1. LL2NVRAMLogicalHistorySummary.extension' = CrazyHashValue
                    BY <4>1
                  (*********************************************************)
                  (* The extension field of the successor history summary  *)
                  (* is equal to a hash generated by the hash function.    *)
                  (*********************************************************)
                  <9>2. successorHistorySummary.extension =
                              Hash(previousLL2HistorySummary.extension, securedInput)
                    BY DEF successorHistorySummary, Successor, securedInput
                  (*********************************************************)
                  (* The arguments to the hash function are both in the    *)
                  (* hash domain.                                          *)
                  (*********************************************************)
                  <9>3. previousLL2HistorySummary.extension \in HashDomain
                    <10>1. previousLL2HistorySummary.extension \in HashType
                      <11>1. previousLL2HistorySummary \in HistorySummaryType
                        BY <7>8
                      <11>2. QED
                        BY <11>1 DEF HistorySummaryType
                    <10>2. QED
                      BY <10>1 DEF HashDomain
                  <9>4. securedInput \in HashDomain
                    <10>1. securedInput \in HashType
                      <11>1. LL2NVRAM.hashBarrier' \in HashDomain
                        <12>1. LL2NVRAM.hashBarrier' \in HashType
                          BY <2>1, LL2SubtypeImplicationLemma DEF LL2SubtypeImplication
                        <12>2. QED
                          BY <12>1 DEF HashDomain
                      <11>2. prevInput \in HashDomain
                        <12>1. prevInput \in InputType
                          BY <7>8
                        <12>2. QED
                          BY <12>1 DEF HashDomain
                      <11>3. QED
                        BY <11>1, <11>2, HashTypeSafe DEF securedInput
                    <10>2. QED
                      BY <10>1 DEF HashDomain
                  (*********************************************************)
                  (* The crazy hash value is not equal to any hash value   *)
                  (* that can be generated by the hash function when       *)
                  (* operating on arguments within its domain.             *)
                  (*********************************************************)
                  <9>5. QED
                      BY <9>1, <9>2, <9>3, <9>4, CrazyHashValueUnique
                <8>2. QED
                  BY <8>1
              (*************************************************************)
              (* Second, we prove that the logical history summary cannot  *)
              (* be a checkpoint.                                          *)
              (*************************************************************)
              <7>12. LL2NVRAMLogicalHistorySummary' # checkpointedSuccessorHistorySummary
                (***********************************************************)
                (* The extension field of the logical history summary does *)
                (* not equal the base hash value, as we proved above.      *)
                (***********************************************************)
                <8>1. LL2NVRAMLogicalHistorySummary.extension' # BaseHashValue
                  BY <4>1, CrazyHashValueUnique
                (***********************************************************)
                (* The extension field of the checkpointed successor does  *)
                (* equal the base hash value.  This follows from the       *)
                (* CheckpointHasBaseExtensionLemma.                        *)
                (***********************************************************)
                <8>2. checkpointedSuccessorHistorySummary.extension = BaseHashValue
                  <9>1. successorHistorySummary \in HistorySummaryType
                    <10>1. previousLL2HistorySummary \in HistorySummaryType
                      BY <7>8
                    <10>2. prevInput \in InputType
                      BY <7>8
                    <10>3. LL2NVRAM.hashBarrier' \in HashType
                      BY <2>1, LL2SubtypeImplicationLemma DEF LL2SubtypeImplication, LL2TypeInvariant, LL2TrustedStorageType
                    <10>4. QED
                      BY <10>1, <10>2, <10>3, SuccessorTypeSafe DEF successorHistorySummary
                  <9>2. QED
                    BY <9>1, CheckpointHasBaseExtensionLemma
                    DEF checkpointedSuccessorHistorySummary
                <8>3. QED
                  BY <8>1, <8>2
              (*************************************************************)
              (* We thus have a contradiction.                             *)
              (*************************************************************)
              <7>13. QED
                BY <7>10, <7>11, <7>12
            <6>3. QED
              BY <6>2 DEF ll1GarbageHistoryStateBinding
          (*****************************************************************)
          (* We prove that the constraint labeled `previous' in the        *)
          (* LL1RestrictedCorruption action is satisfied.                  *)
          (*****************************************************************)
          <5>3. LL1RestrictedCorruption!nvram!previous(LL1NVRAM.historySummary')
            (***************************************************************)
            (* To prove the universally quantified expression, we take a   *)
            (* set of variables of the appropriate types.                  *)
            (***************************************************************)
            <6>1. TAKE stateHash1 \in HashType,
                       ll1Authenticator \in LL1ObservedAuthenticators,
                       ll1SomeHistorySummary \in HashType,
                       someInput \in InputType
            (***************************************************************)
            (* We re-state the definition from within the                  *)
            (* LL1RestrictedCorruption!nvram!previous clause.              *)
            (***************************************************************)
            <6> ll1SomeHistoryStateBinding == Hash(ll1SomeHistorySummary, stateHash1)
            (***************************************************************)
            (* We hide the definition.                                     *)
            (***************************************************************)
            <6> HIDE DEF ll1SomeHistoryStateBinding
            (***************************************************************)
            (* We need to prove the nvram::previous conjunct, which        *)
            (* asserts an implication.  It suffices to assume the          *)
            (* antecedent and prove the consequent.                        *)
            (***************************************************************)
            <6>2. SUFFICES
                  ASSUME LL1NVRAM.historySummary' = Hash(ll1SomeHistorySummary, someInput)
                  PROVE ~ValidateMAC(
                      LL1NVRAM.symmetricKey,
                      ll1SomeHistoryStateBinding,
                      ll1Authenticator)
              BY DEF ll1SomeHistoryStateBinding
            (***************************************************************)
            (* The consequent of the nvram::previous conjunct asserts that *)
            (* the authenticator is not a valid MAC for the history state  *)
            (* binding formed from any predecessor of the history summary  *)
            (* in the NVRAM and any state hash.                            *)
            (***************************************************************)
            <6>3. ~ValidateMAC(
                      LL1NVRAM.symmetricKey,
                      ll1SomeHistoryStateBinding,
                      ll1Authenticator)
              (*************************************************************)
              (* We will use proof by contradiction.                       *)
              (*************************************************************)
              <7>1. SUFFICES
                    ASSUME
                        ValidateMAC(
                            LL1NVRAM.symmetricKey,
                            ll1SomeHistoryStateBinding,
                            ll1Authenticator)
                    PROVE
                        FALSE
                OBVIOUS
              (*************************************************************)
              (* We first pick, from the set of `^\fastname^' observed     *)
              (* authenticators, a `^\fastname^' authenticator that        *)
              (* matches the `^\basicname^' authenticator.  We know that   *)
              (* such a authenticator exists, because the refinement       *)
              (* asserts that the sets of observed authenticators match    *)
              (* across the two specs.                                     *)
              (*************************************************************)
              <7>2. PICK ll2Authenticator \in LL2ObservedAuthenticators :
                        AuthenticatorsMatch(
                            ll1Authenticator,
                            ll2Authenticator,
                            LL2NVRAM.symmetricKey,
                            LL2NVRAM.hashBarrier)
                <8>1. ll1Authenticator \in LL1ObservedAuthenticators
                  BY <6>1
                <8>2. AuthenticatorSetsMatch(
                          LL1ObservedAuthenticators,
                          LL2ObservedAuthenticators,
                          LL2NVRAM.symmetricKey,
                          LL2NVRAM.hashBarrier)
                  BY <2>1 DEF LL2Refinement
                <8>3. QED
                  BY <8>1, <8>2 DEF AuthenticatorSetsMatch
              (*************************************************************)
              (* We pick a set of variables of the appropriate types that  *)
              (* satisfy the quantified AuthenticatorsMatch predicate.     *)
              (*************************************************************)
              <7>3. PICK stateHash2 \in HashType,
                        ll1HistorySummary \in HashType,
                        ll2HistorySummary \in HistorySummaryType :
                        AuthenticatorsMatch(
                            ll1Authenticator,
                            ll2Authenticator,
                            LL2NVRAM.symmetricKey,
                            LL2NVRAM.hashBarrier)!(
                            stateHash2, ll1HistorySummary, ll2HistorySummary)!1
                BY <7>2 DEF AuthenticatorsMatch
              (*************************************************************)
              (* We re-state the definitions from the LET in               *)
              (* AuthenticatorsMatch.                                      *)
              (*************************************************************)
              <7> ll1HistoryStateBinding == Hash(ll1HistorySummary, stateHash2)
              <7> ll2HistorySummaryHash == Hash(ll2HistorySummary.anchor, ll2HistorySummary.extension)
              <7> ll2HistoryStateBinding == Hash(ll2HistorySummaryHash, stateHash2)
              (*************************************************************)
              (* We prove the types of the definitions, with help from the *)
              (* AuthenticatorsMatchDefsTypeSafeLemma.                     *)
              (*************************************************************)
              <7>4. /\  ll1HistoryStateBinding \in HashType
                    /\  ll2HistorySummaryHash \in HashType
                    /\  ll2HistoryStateBinding \in HashType
                BY <7>3, AuthenticatorsMatchDefsTypeSafeLemma
              (*************************************************************)
              (* We hide the definitions.                                  *)
              (*************************************************************)
              <7> HIDE DEF ll1HistoryStateBinding, ll2HistorySummaryHash, ll2HistoryStateBinding
              (*************************************************************)
              (* We prove that the `^\basicname^'s history summary picked  *)
              (* to satisfy the AuthenticatorsMatch predicate equals the   *)
              (* history summary in the primed state of the `^\basicname^' *)
              (* NVRAM.                                                    *)
              (*************************************************************)
              <7>5. ll1HistorySummary = ll1SomeHistorySummary
                (***********************************************************)
                (* The first step is to show the equality of the history   *)
                (* state bindings that bind each of these history          *)
                (* summaries to their respective state hashes.             *)
                (***********************************************************)
                <8>1. ll1HistoryStateBinding = ll1SomeHistoryStateBinding
                  (*********************************************************)
                  (* By hypothesis, the authenticator is a valid MAC for   *)
                  (* the garbage history state binding.                    *)
                  (*********************************************************)
                  <9>1. ValidateMAC(
                            LL2NVRAM.symmetricKey, ll1SomeHistoryStateBinding, ll1Authenticator)
                    <10>1. LL1NVRAM.symmetricKey = LL2NVRAM.symmetricKey
                      BY <2>1 DEF LL2Refinement
                    <10>2. QED
                      BY <7>1, <10>1
                  (*********************************************************)
                  (* The definition of the AuthenticatorsMatch predicate   *)
                  (* tells us that the `^\basicname^' authenticator was    *)
                  (* generated as a MAC from the history state binding.    *)
                  (*********************************************************)
                  <9>2. ll1Authenticator = GenerateMAC(LL2NVRAM.symmetricKey, ll1HistoryStateBinding)
                    BY <7>3 DEF ll1HistoryStateBinding
                  (*********************************************************)
                  (* The remaining preconditions are types.                *)
                  (*********************************************************)
                  <9>3. LL2NVRAM.symmetricKey \in SymmetricKeyType
                    BY <2>1, LL2SubtypeImplicationLemma DEF LL2SubtypeImplication
                  <9>4. ll1HistoryStateBinding \in HashType
                    BY <7>4
                  <9>5. ll1SomeHistoryStateBinding \in HashType
                    <10>1. ll1SomeHistorySummary \in HashDomain
                      <11>1. ll1SomeHistorySummary \in HashType
                        BY <6>1
                      <11>2. QED
                        BY <11>1 DEF HashDomain
                    <10>2. stateHash1 \in HashDomain
                      BY <5>1 DEF HashDomain
                    <10>3. QED
                      BY <10>1, <10>2, HashTypeSafe DEF ll1SomeHistoryStateBinding
                  (*********************************************************)
                  (* The MACCollisionResistant property tells us that the  *)
                  (* two history state bindings are equal.                 *)
                  (*********************************************************)
                  <9>6. QED
                    BY <9>1, <9>2, <9>3, <9>4, <9>5, MACCollisionResistant
                (***********************************************************)
                (* By the collision resistance of the hash function, the   *)
                (* equality of the history state bindings implies the      *)
                (* equality of the history summaries.                      *)
                (***********************************************************)
                <8>2. QED
                  <9>1. ll1SomeHistorySummary \in HashDomain
                    <10>1. ll1SomeHistorySummary \in HashType
                      BY <6>1
                    <10>2. QED
                      BY <10>1 DEF HashDomain
                  <9>2. ll1HistorySummary \in HashDomain
                    <10>1. ll1HistorySummary \in HashType
                      BY <7>3
                    <10>2. QED
                      BY <10>1 DEF HashDomain
                  <9>3. stateHash1 \in HashDomain
                    <10>1. stateHash1 \in HashType
                      BY <6>1
                    <10>2. QED
                      BY <10>1 DEF HashDomain
                  <9>4. stateHash2 \in HashDomain
                    <10>1. stateHash2 \in HashType
                      BY <7>3
                   <10>2. QED
                      BY <10>1 DEF HashDomain
                  <9>5. QED
                    BY <8>1, <9>2, <9>1, <9>4, <9>3, HashCollisionResistant
                    DEF ll1HistoryStateBinding, ll1SomeHistoryStateBinding
              (*************************************************************)
              (* We pick a value for the `^\fastname^'                     *)
              (* previous-inputs-summary existential variable inside the   *)
              (* HistorySummariesMatchRecursion predicate that satisfies   *)
              (* this predicate for (1) the `^\basicname^'s history        *)
              (* summary picked to satisfy the AuthenticatorsMatch         *)
              (* predicate and (2) the input taken from the universal      *)
              (* quantifier in the `previous' conjunct in the              *)
              (* LL1RestrictedCorruption action.                           *)
              (*************************************************************)
              <7>6. PICK previousLL2HistorySummary \in HistorySummaryType :
                         HistorySummariesMatchRecursion(
                             LL1NVRAM.historySummary',
                             LL2NVRAMLogicalHistorySummary',
                             LL2NVRAM.hashBarrier')!(
                             someInput,
                             ll1SomeHistorySummary,
                             previousLL2HistorySummary)
                (***********************************************************)
                (* The `^\basicname^'s history summary picked to satisfy   *)
                (* the AuthenticatorsMatch predicate matches the primed    *)
                (* logical history summary in the `^\fastname^' NVRAM and  *)
                (* SPCR, by the refinement and the above equality.         *)
                (***********************************************************)
                <8>1. HistorySummariesMatch(
                          LL1NVRAM.historySummary',
                          LL2NVRAMLogicalHistorySummary',
                          LL2NVRAM.hashBarrier')
                  <9>1. HistorySummariesMatch(
                            LL1NVRAM.historySummary',
                            LL2NVRAMLogicalHistorySummary',
                            LL2NVRAM.hashBarrier')
                    BY <2>1 DEF LL2Refinement
                  <9>2. QED
                    BY <9>1
                (***********************************************************)
                (* We prove that the HistorySummariesMatch predicate       *)
                (* equals the HistorySummariesMatchRecursion predicate in  *)
                (* this case.  We assert each condition required by the    *)
                (* definition of the predicate.                            *)
                (***********************************************************)
                <8>2. HistorySummariesMatch(
                          LL1NVRAM.historySummary',
                          LL2NVRAMLogicalHistorySummary',
                          LL2NVRAM.hashBarrier') <=>
                      HistorySummariesMatchRecursion(
                          LL1NVRAM.historySummary',
                          LL2NVRAMLogicalHistorySummary',
                          LL2NVRAM.hashBarrier')
                  (*********************************************************)
                  (* We prove some types, to satisfy the universal         *)
                  (* quantifiers in HistorySummariesMatchDefinition.       *)
                  (*********************************************************)
                  <9>1. LL1NVRAM.historySummary' \in HashType
                    BY <2>1 DEF LL2Refinement, LL1TrustedStorageType
                  <9>2. LL2NVRAMLogicalHistorySummary' \in HistorySummaryType
                    BY <2>1, LL2NVRAMLogicalHistorySummaryTypeSafe
                  <9>3. LL2NVRAM.hashBarrier' \in HashType
                    BY <2>1, LL2SubtypeImplicationLemma DEF LL2SubtypeImplication, LL2TypeInvariant, LL2TrustedStorageType
                  (*********************************************************)
                  (* We prove that the `^\fastname^' logical history       *)
                  (* summary does not equal the initial history summary.   *)
                  (*********************************************************)
                  <9> ll2InitialHistorySummary == [ anchor |-> BaseHashValue, extension |-> BaseHashValue ]
                  <9>4. LL2NVRAMLogicalHistorySummary' # ll2InitialHistorySummary
                    <10>1. LL2NVRAMLogicalHistorySummary.extension' # BaseHashValue
                      BY <4>1, CrazyHashValueUnique
                    <10>2. ll2InitialHistorySummary.extension = BaseHashValue
                      BY DEF ll2InitialHistorySummary
                    <10>3. QED
                      BY <10>1, <10>2
                  (*********************************************************)
                  (* Finally, from HistorySummariesMatchDefinition, we can *)
                  (* conclude that the HistorySummariesMatch predicate     *)
                  (* equals the quantified HistorySummariesMatchRecursion  *)
                  (* predicate.                                            *)
                  (*********************************************************)
                  <9>5. QED
                    BY <9>1, <9>2, <9>3, <9>4, HistorySummariesMatchDefinition
                    DEF ll2InitialHistorySummary
                (***********************************************************)
                (* We pick values for the remaining two existential        *)
                (* variables inside the HistorySummariesMatchRecursion     *)
                (* predicate that satisfy the predicate.  We know such     *)
                (* variables exist, because the predicate is satisfied by  *)
                (* the two previous steps.                                 *)
                (***********************************************************)
                <8>3. PICK prevInput \in InputType,
                           previousLL1HistorySummary \in HashType,
                           previousLL2HistorySummary \in HistorySummaryType :
                           HistorySummariesMatchRecursion(
                               LL1NVRAM.historySummary',
                               LL2NVRAMLogicalHistorySummary',
                               LL2NVRAM.hashBarrier')!(
                               prevInput,
                               previousLL1HistorySummary,
                               previousLL2HistorySummary)
                  (*********************************************************)
                  (* We prove some types, to satisfy the universal         *)
                  (* quantifiers in HistorySummariesMatchRecursion         *)
                  (* predicate.                                            *)
                  (*********************************************************)
                  <9>1. ll1HistorySummary \in HashType
                    BY <7>3
                  <9>2. LL2NVRAMLogicalHistorySummary' \in HistorySummaryType
                    BY <2>1, LL2NVRAMLogicalHistorySummaryTypeSafe
                  <9>3. LL2NVRAM.hashBarrier' \in HashType
                    BY <2>1, LL2SubtypeImplicationLemma DEF LL2SubtypeImplication, LL2TypeInvariant, LL2TrustedStorageType
                  <9>4. HistorySummariesMatchRecursion(
                            LL1NVRAM.historySummary',
                            LL2NVRAMLogicalHistorySummary',
                            LL2NVRAM.hashBarrier')
                    BY <8>1, <8>2
                  <9>5. QED
                    BY <9>1, <9>2, <9>4 DEF HistorySummariesMatchRecursion
                (***********************************************************)
                (* We prove that the existential variables for the         *)
                (* previous input and the previous history summary in the  *)
                (* above pick are equal to the input and history summary   *)
                (* taken from the universal quantifiers in the `previous'  *)
                (* conjunct in the LL1RestrictedCorruption action.  We use *)
                (* the HashCollisionResistant property.                    *)
                (***********************************************************)
                <8>4. /\  ll1SomeHistorySummary = previousLL1HistorySummary
                      /\  someInput = prevInput
                  (*********************************************************)
                  (* We prove the necessary types for the                  *)
                  (* HashCollisionResistant property.                      *)
                  (*********************************************************)
                  <9>1. ll1SomeHistorySummary \in HashDomain
                    <10>1. ll1SomeHistorySummary \in HashType
                      BY <6>1
                    <10>2. QED
                      BY <10>1 DEF HashDomain
                  <9>2. someInput \in HashDomain
                    <10>1. someInput \in InputType
                      BY <6>1
                    <10>2. QED
                      BY <10>1 DEF HashDomain
                  <9>3. previousLL1HistorySummary \in HashDomain
                    <10>1. previousLL1HistorySummary \in HashType
                      BY <8>3
                    <10>2. QED
                      BY <10>1 DEF HashDomain
                  <9>4. prevInput \in HashDomain
                    <10>1. prevInput \in InputType
                      BY <8>3
                    <10>2. QED
                      BY <10>1 DEF HashDomain
                  (*********************************************************)
                  (* The hashes are equal, because each is equal to the    *)
                  (* history summary in the primed `^\basicname^' NVRAM.   *)
                  (*********************************************************)
                  <9>5. Hash(ll1SomeHistorySummary, someInput) =
                            Hash(previousLL1HistorySummary, prevInput)
                    (*******************************************************)
                    (* The hash of the taken history summary and input are *)
                    (* equal to the history summary in the primed          *)
                    (* `^\basicname^' NVRAM by assumption of the           *)
                    (* antecedent in the `previous' conjunct in the        *)
                    (* LL1RestrictedCorruption action.                     *)
                    (*******************************************************)
                    <10>1. LL1NVRAM.historySummary' = Hash(ll1SomeHistorySummary, someInput)
                      BY <6>2
                    (*******************************************************)
                    (* The hash of the picked history summary and input    *)
                    (* are equal to the history summary in the primed      *)
                    (* `^\basicname^' NVRAM by the definition of the       *)
                    (* HistorySummariesMatchRecursion predicate.           *)
                    (*******************************************************)
                    <10>2. LL1NVRAM.historySummary' = Hash(previousLL1HistorySummary, prevInput)
                      BY <8>3
                    <10>3. QED
                      BY <10>1, <10>2
                  <9>6. QED
                    BY <9>1, <9>2, <9>3, <9>4, <9>5, HashCollisionResistant
                <8>5. QED
                  BY <8>3, <8>4
              (*************************************************************)
              (* One of the conjuncts in the definition of                 *)
              (* HistorySummariesMatchRecursion is that the `^\fastname^'  *)
              (* history summary is a successor of a previous history      *)
              (* summary.                                                  *)
              (*************************************************************)
              <7>7. LL2HistorySummaryIsSuccessor(
                        LL2NVRAMLogicalHistorySummary',
                        previousLL2HistorySummary,
                        someInput,
                        LL2NVRAM.hashBarrier')
                BY <7>6 DEF HistorySummariesMatchRecursion
              (*************************************************************)
              (* We re-state the definitions from the LET in               *)
              (* LL2HistorySummaryIsSuccessor.                             *)
              (*************************************************************)
              <7> successorHistorySummary ==
                      Successor(previousLL2HistorySummary, someInput, LL2NVRAM.hashBarrier')
              <7> checkpointedSuccessorHistorySummary == Checkpoint(successorHistorySummary)
              (*************************************************************)
              (* We hide the definitions.                                  *)
              (*************************************************************)
              <7> HIDE DEF successorHistorySummary, checkpointedSuccessorHistorySummary
              (*************************************************************)
              (* The definition of LL2HistorySummaryIsSuccessor tells us   *)
              (* that there are two ways that the logical history summary  *)
              (* could be a successor.  We will prove that neither of      *)
              (* these disjuncts is satisfiable.                           *)
              (*************************************************************)
              <7>8. \/  LL2NVRAMLogicalHistorySummary' = successorHistorySummary
                    \/  LL2NVRAMLogicalHistorySummary' = checkpointedSuccessorHistorySummary
                BY <7>7
                DEF LL2HistorySummaryIsSuccessor, successorHistorySummary,
                    checkpointedSuccessorHistorySummary
              (*************************************************************)
              (* First, we prove that the logical history summary cannot   *)
              (* be a successor.                                           *)
              (*************************************************************)
              <7>9. LL2NVRAMLogicalHistorySummary' # successorHistorySummary
                (***********************************************************)
                (* We re-state a definition from the LET in the Successor  *)
                (* operator.                                               *)
                (***********************************************************)
                <8> securedInput == Hash(LL2NVRAM.hashBarrier', someInput)
                (***********************************************************)
                (* We hide the definition.                                 *)
                (***********************************************************)
                <8> HIDE DEF securedInput
                (***********************************************************)
                (* There is only one sub-step, which is proving that the   *)
                (* extension fields of these two records are unequal.      *)
                (***********************************************************)
                <8>1. LL2NVRAMLogicalHistorySummary.extension' # successorHistorySummary.extension
                  (*********************************************************)
                  (* The extension field of the logical history summary is *)
                  (* a crazy hash value, as proven above.                  *)
                  (*********************************************************)
                  <9>1. LL2NVRAMLogicalHistorySummary.extension' = CrazyHashValue
                    BY <4>1
                  (*********************************************************)
                  (* Second, we prove that the extension field of the      *)
                  (* successor history summary from                        *)
                  (* LL2HistorySummaryIsSuccessor is computed from the     *)
                  (* hash function.                                        *)
                  (*********************************************************)
                  <9>2. successorHistorySummary.extension =
                            Hash(previousLL2HistorySummary.extension, securedInput)
                    BY DEF successorHistorySummary, Successor, securedInput
                  (*********************************************************)
                  (* Third, we prove that the two hashes are unequal.  We  *)
                  (* will use the CrazyHashValueUnique property.           *)
                  (*********************************************************)
                  <9>3. Hash(previousLL2HistorySummary.extension, securedInput) # CrazyHashValue
                    (*******************************************************)
                    (* To employ the CrazyHashValueUnique property, we     *)
                    (* need to prove some types.                           *)
                    (*******************************************************)
                    <10>1. previousLL2HistorySummary.extension \in HashDomain
                      <11>1. previousLL2HistorySummary.extension \in HashType
                        <12>1. previousLL2HistorySummary \in HistorySummaryType
                          BY <7>6
                        <12>2. QED
                          BY <12>1 DEF HistorySummaryType
                      <11>2. QED
                        BY <11>1 DEF HashDomain
                    <10>2. securedInput \in HashDomain
                      <11>1. securedInput \in HashType
                        <12>1. LL2NVRAM.hashBarrier' \in HashDomain
                          <13>1. LL2NVRAM.hashBarrier' \in HashType
                            BY <2>1, LL2SubtypeImplicationLemma DEF LL2SubtypeImplication
                          <13>2. QED
                            BY <13>1 DEF HashDomain
                        <12>2. someInput \in HashDomain
                          <13>1. someInput \in InputType
                            BY <6>1
                          <13>2. QED
                            BY <13>1 DEF HashDomain
                        <12>3. QED
                          BY <12>1, <12>2, HashTypeSafe DEF securedInput
                      <11>2. QED
                        BY <11>1 DEF HashDomain
                    <10>3. QED
                      BY <10>1, <10>2, CrazyHashValueUnique
                  <9>4. QED
                    BY <9>1, <9>2, <9>3
                <8>2. QED
                  BY <8>1
              (*************************************************************)
              (* Second, we prove that the logical history summary cannot  *)
              (* be a checkpoint.                                          *)
              (*************************************************************)
              <7>10. LL2NVRAMLogicalHistorySummary' # checkpointedSuccessorHistorySummary
                (***********************************************************)
                (* The extension field of the logical history summary does *)
                (* not equal the base hash value, as we proved above.      *)
                (***********************************************************)
                <8>1. LL2NVRAMLogicalHistorySummary.extension' # BaseHashValue
                  BY <4>1, CrazyHashValueUnique
                (***********************************************************)
                (* The extension field of the checkpointed successor does  *)
                (* equal the base hash value.  This follows from the       *)
                (* CheckpointHasBaseExtensionLemma.                        *)
                (***********************************************************)
                <8>2. checkpointedSuccessorHistorySummary.extension = BaseHashValue
                  <9>1. successorHistorySummary \in HistorySummaryType
                    <10>1. previousLL2HistorySummary \in HistorySummaryType
                      BY <7>6
                    <10>2. someInput \in InputType
                      BY <6>1
                    <10>3. LL2NVRAM.hashBarrier' \in HashType
                      BY <2>1, LL2SubtypeImplicationLemma DEF LL2SubtypeImplication, LL2TypeInvariant, LL2TrustedStorageType
                    <10>4. QED
                      BY <10>1, <10>2, <10>3, SuccessorTypeSafe DEF successorHistorySummary
                  <9>2. QED
                    BY <9>1, CheckpointHasBaseExtensionLemma
                    DEF checkpointedSuccessorHistorySummary
                <8>3. QED
                  BY <8>1, <8>2
              (*************************************************************)
              (* We thus have a contradiction.                             *)
              (*************************************************************)
              <7>11. QED
                BY <7>8, <7>9, <7>10
            <6>4. QED
              BY <6>3 DEF ll1SomeHistoryStateBinding
          (*****************************************************************)
          (* We prove the third conjunct within the `nvram' conjunct of    *)
          (* the LL1RestrictedCorruption action.                           *)
          (*****************************************************************)
          <5>4. LL1NVRAM' = [
                    historySummary |-> LL1NVRAM.historySummary',
                    symmetricKey |-> LL1NVRAM.symmetricKey ]
            <6>1. LL1NVRAM \in LL1TrustedStorageType
              BY <2>1 DEF LL2Refinement
            <6>2. LL1NVRAM' \in LL1TrustedStorageType
              BY <2>1 DEF LL2Refinement
            <6>3. UNCHANGED LL1NVRAM.symmetricKey
              <7>1. UNCHANGED LL2NVRAM.symmetricKey
                BY <3>2
              <7>2. QED
                BY <2>1, <7>1, UnchangedNVRAMSymmetricKeyLemma
            <6>4. QED
              BY <6>1, <6>2, <6>3, LL1NVRAMRecordCompositionLemma
          <5>5. QED
            BY <5>1, <5>2, <5>3, <5>4
        (*******************************************************************)
        (* The `ram' conjunct of the LL1RestrictedCorruption action is     *)
        (* satisfied because the `trashed' disjunct is satisfied.          *)
        (*******************************************************************)
        <4>3. LL1RestrictedCorruption!ram
        (*******************************************************************)
        (* The `trashed' disjunct of the LL1RestrictedCorruption action is *)
        (* satisfied by the proof above.                                   *)
        (*******************************************************************)
          <5>1. LL1RestrictedCorruption!ram!trashed 
            BY <3>3
          <5>2. QED
            BY <5>1
        (*******************************************************************)
        (* The disk is unchanged by the UnchangedDiskLemma.                *)
        (*******************************************************************)
        <4>4. UNCHANGED LL1Disk
          BY <2>1, <3>2, UnchangedDiskLemma
        (*******************************************************************)
        (* The set of available inputs is unchanged by the                 *)
        (* UnchangedAvailableInputsLemma.                                  *)
        (*******************************************************************)
        <4>5. UNCHANGED LL1AvailableInputs
          BY <2>1, <3>2, UnchangedAvailableInputsLemma
        (*******************************************************************)
        (* The set of observed outputs is unchanged by the                 *)
        (* UnchangedObservedOutputsLemma.                                  *)
        (*******************************************************************)
        <4>6. UNCHANGED LL1ObservedOutputs
          BY <2>1, <3>2, UnchangedObservedOutputsLemma
        (*******************************************************************)
        (* The set of observed authenticators is unchanged by the          *)
        (* UnchangedObservedAuthenticatorsLemma.                           *)
        (*******************************************************************)
        <4>7. UNCHANGED LL1ObservedAuthenticators
          BY <2>1, <3>2, UnchangedObservedAuthenticatorsLemma
        (*******************************************************************)
        (* Lastly, we tie together all of the required aspects of the      *)
        (* LL1RestrictedCorruption definition.                             *)
        (*******************************************************************)
        <4>8. QED
          BY <4>2, <4>3, <4>4, <4>5, <4>6, <4>7 DEF LL1RestrictedCorruption
      (*********************************************************************)
      (* For the ELSE case, we assume that an extension is not in progress *)
      (* and show that this refines to an LL1Restart action.               *)
      (*********************************************************************)
      <3>5. ASSUME ~LL2NVRAM.extensionInProgress
            PROVE LL1Restart
        (*******************************************************************)
        (* The extentials and the first two conjuncts in the LL1Restart    *)
        (* action are satisfied by a proof above.                          *)
        (*******************************************************************)
        <4>1. \E  ll1UntrustedStorage \in LL1UntrustedStorageType,
                  ll1RandomSymmetricKey \in SymmetricKeyType \ {LL1NVRAM.symmetricKey},
                  ll1Hash \in HashType :
                  /\  ll1UntrustedStorage.authenticator =
                          GenerateMAC(ll1RandomSymmetricKey, ll1Hash)
                  /\  LL1RAM' = ll1UntrustedStorage
          BY <3>3
        (*******************************************************************)
        (* The disk is unchanged by the UnchangedDiskLemma.                *)
        (*******************************************************************)
        <4>2. UNCHANGED LL1Disk
          BY <2>1, <3>2, UnchangedDiskLemma
        (*******************************************************************)
        (* The NVRAM is unchanged because the logical history summary is   *)
        (* unchanged and the symmetric key is unchanged.                   *)
        (*******************************************************************)
        <4>3. UNCHANGED LL1NVRAM
          <5>1. UNCHANGED LL1NVRAM.historySummary
            (***************************************************************)
            (* The logical history summary in the `^\fastname^' NVRAM and  *)
            (* SPCR is unchanged.                                          *)
            (***************************************************************)
            <6>1. UNCHANGED LL2NVRAMLogicalHistorySummary
              (*************************************************************)
              (* We reveal the definition of the unprimed logical history  *)
              (* summary in the `^\fastname^' NVRAM and SPCR, given that   *)
              (* there is no extension in progress.                        *)
              (*************************************************************)
              <7>1. LL2NVRAMLogicalHistorySummary = [
                        anchor |-> LL2NVRAM.historySummaryAnchor,
                        extension |-> BaseHashValue ]
                <8>1. ~LL2NVRAM.extensionInProgress
                  BY <3>5
                <8>2. QED
                  BY <8>1 DEF LL2NVRAMLogicalHistorySummary
              (*************************************************************)
              (* We reveal the definition of the primed logical history    *)
              (* summary in the `^\fastname^' NVRAM and SPCR, given that   *)
              (* there is no extension in progress.                        *)
              (*************************************************************)
              <7>2. LL2NVRAMLogicalHistorySummary' = [
                        anchor |-> LL2NVRAM.historySummaryAnchor',
                        extension |-> BaseHashValue ]
                <8>1. ~LL2NVRAM.extensionInProgress'
                  <9>1. ~LL2NVRAM.extensionInProgress
                    BY <3>5
                  <9>2. UNCHANGED LL2NVRAM.extensionInProgress
                    BY <3>2
                  <9>3. QED
                    BY <9>1, <9>2
                <8>2. QED
                  BY <8>1 DEF LL2NVRAMLogicalHistorySummary
              (*************************************************************)
              (* The history summary anchor in the `^\fastname^' NVRAM is  *)
              (* unchanged by a LL2CorruptSPCR action.                     *)
              (*************************************************************)
              <7>3. UNCHANGED LL2NVRAM.historySummaryAnchor
                BY <3>2
              (*************************************************************)
              (* Since the value of LL2NVRAMLogicalHistorySummary is       *)
              (* determined entirely by the history summary in the         *)
              (* `^\fastname^' NVRAM, and since this value is unchanged,   *)
              (* the value of LL2NVRAMLogicalHistorySummary is unchanged.  *)
              (*************************************************************)
              <7>4. QED
                BY <7>1, <7>2, <7>3
            <6>2. UNCHANGED LL2NVRAM.symmetricKey
              BY <3>2
            <6>3. UNCHANGED LL2NVRAM.hashBarrier
              BY <3>2
            <6>4. QED
              BY <2>1, <6>1, <6>2, <6>3, UnchangedNVRAMHistorySummaryLemma
          <5>2. UNCHANGED LL1NVRAM.symmetricKey
            <6>1. UNCHANGED LL2NVRAM.symmetricKey
              BY <3>2
            <6>2. QED
              BY <2>1, <6>1, UnchangedNVRAMSymmetricKeyLemma
          <5>3. LL1NVRAM \in LL1TrustedStorageType
            BY <2>1 DEF LL2Refinement
          <5>4. LL1NVRAM' \in LL1TrustedStorageType
            BY <2>1 DEF LL2Refinement
          <5>5. QED
            BY <5>1, <5>2, <5>3, <5>4, LL1NVRAMRecordCompositionLemma
        (*******************************************************************)
        (* The set of available inputs is unchanged by the                 *)
        (* UnchangedAvailableInputsLemma.                                  *)
        (*******************************************************************)
        <4>4. UNCHANGED LL1AvailableInputs
          BY <2>1, <3>2, UnchangedAvailableInputsLemma
        (*******************************************************************)
        (* The set of observed outputs is unchanged by the                 *)
        (* UnchangedObservedOutputsLemma.                                  *)
        (*******************************************************************)
        <4>5. UNCHANGED LL1ObservedOutputs
          BY <2>1, <3>2, UnchangedObservedOutputsLemma
        (*******************************************************************)
        (* The set of observed authenticators is unchanged by the          *)
        (* UnchangedObservedAuthenticatorsLemma.                           *)
        (*******************************************************************)
        <4>6. UNCHANGED LL1ObservedAuthenticators
          BY <2>1, <3>2, UnchangedObservedAuthenticatorsLemma
        (*******************************************************************)
        (* Lastly, we tie together all of the required aspects of the      *)
        (* LL1Restart definition.                                          *)
        (*******************************************************************)
        <4>7. QED
          BY <4>1, <4>2, <4>3, <4>4, <4>5, <4>6 DEF LL1Restart
      (*********************************************************************)
      (* Both THEN and ELSE cases are proven.                              *)
      (*********************************************************************)
      <3>6. QED
        BY <3>4, <3>5
    (***********************************************************************)
    (* A `^\fastname^' LL2ReadDisk action refines to a `^\basicname^'      *)
    (* LL1ReadDisk action.                                                 *)
    (***********************************************************************)
    <2>8. LL2ReadDisk => LL1ReadDisk
      <3>1. HAVE LL2ReadDisk
      (*********************************************************************)
      (* The primed state of the LL1RAM equals the unprimed state of the   *)
      (* LL1Disk.  The proof is somewhat tedious but completely            *)
      (* straightforward.                                                  *)
      (*********************************************************************)
      <3>2. LL1RAM' = LL1Disk
        <4>1. LL2RAM' = LL2Disk
          BY <3>1 DEF LL2ReadDisk
        <4>2. UNCHANGED LL2NVRAM.hashBarrier
          BY <3>1 DEF LL2ReadDisk
        (*******************************************************************)
        (* The primed public state field of the RAM equals the unprimed    *)
        (* public state field of the disk.                                 *)
        (*******************************************************************)
        <4>3. LL1RAM.publicState' = LL1Disk.publicState
          <5>1. LL2RAM.publicState' = LL2Disk.publicState
            BY <4>1
          <5>2. LL1Disk.publicState = LL2Disk.publicState
            BY <2>1 DEF LL2Refinement
          <5>3. LL1RAM.publicState' = LL2RAM.publicState'
            BY <2>1 DEF LL2Refinement
          <5>4. QED
            BY <5>1, <5>2, <5>3
        (*******************************************************************)
        (* The primed encrypted private state field of the RAM equals the  *)
        (* unprimed encrypted private state field of the disk.             *)
        (*******************************************************************)
        <4>4. LL1RAM.privateStateEnc' = LL1Disk.privateStateEnc
          <5>1. LL2RAM.privateStateEnc' = LL2Disk.privateStateEnc
            BY <4>1
          <5>2. LL1Disk.privateStateEnc = LL2Disk.privateStateEnc
            BY <2>1 DEF LL2Refinement
          <5>3. LL1RAM.privateStateEnc' = LL2RAM.privateStateEnc'
            BY <2>1 DEF LL2Refinement
          <5>4. QED
            BY <5>1, <5>2, <5>3
        (*******************************************************************)
        (* The primed history summary field of the RAM equals the unprimed *)
        (* history summary field of the disk.                              *)
        (*******************************************************************)
        <4>5. LL1RAM.historySummary' = LL1Disk.historySummary
          <5>1. LL2RAM.historySummary' = LL2Disk.historySummary
            BY <4>1
          <5>2. HistorySummariesMatch(
                    LL1Disk.historySummary,
                    LL2Disk.historySummary,
                    LL2NVRAM.hashBarrier)
            BY <2>1 DEF LL2Refinement
          <5>3. HistorySummariesMatch(
                    LL1RAM.historySummary',
                    LL2RAM.historySummary',
                    LL2NVRAM.hashBarrier')
            BY <2>1 DEF LL2Refinement
          <5>4. QED
            <6>1. LL1Disk.historySummary \in HashType
              BY <2>1 DEF LL2Refinement, LL1UntrustedStorageType
            <6>2. LL1RAM.historySummary' \in HashType
              BY <2>1 DEF LL2Refinement, LL1UntrustedStorageType
            <6>3. LL2Disk.historySummary \in HistorySummaryType
              BY <2>1, LL2SubtypeImplicationLemma DEF LL2SubtypeImplication
            <6>4. LL2NVRAM.hashBarrier \in HashType
              BY <2>1, LL2SubtypeImplicationLemma DEF LL2SubtypeImplication
            <6>5. QED
              BY <4>2, <5>1, <5>2, <5>3, <6>1, <6>2, <6>3, <6>4,
              HistorySummariesMatchUniqueLemma
        (*******************************************************************)
        (* The primed authenticator field of the RAM equals the unprimed   *)
        (* authenticator field of the disk.                                *)
        (*******************************************************************)
        <4>6. LL1RAM.authenticator' = LL1Disk.authenticator
          <5>1. LL2RAM.authenticator' = LL2Disk.authenticator
            BY <4>1
          <5>2. \E symmetricKey \in SymmetricKeyType :
                    AuthenticatorsMatch(
                        LL1Disk.authenticator,
                        LL2Disk.authenticator,
                        symmetricKey,
                        LL2NVRAM.hashBarrier)
            BY <2>1 DEF LL2Refinement
          <5>3. \E symmetricKey \in SymmetricKeyType :
                    AuthenticatorsMatch(
                        LL1RAM.authenticator',
                        LL2RAM.authenticator',
                        symmetricKey,
                        LL2NVRAM.hashBarrier')
            BY <2>1 DEF LL2Refinement
          <5>4. QED
            <6>1. LL1Disk.authenticator \in MACType
              BY <2>1 DEF LL2Refinement, LL1UntrustedStorageType
            <6>2. LL1RAM.authenticator' \in MACType
              BY <2>1 DEF LL2Refinement, LL1UntrustedStorageType
            <6>3. LL2Disk.authenticator \in MACType
              BY <2>1, LL2SubtypeImplicationLemma DEF LL2SubtypeImplication
            <6>4. LL2NVRAM.hashBarrier \in HashType
              BY <2>1, LL2SubtypeImplicationLemma DEF LL2SubtypeImplication
            <6>5. QED
              BY <4>2, <5>1, <5>2, <5>3, <6>1, <6>2, <6>3, <6>4,
                  AuthenticatorsMatchUniqueLemma
        <4>7. LL1RAM' \in LL1UntrustedStorageType
          BY <2>1 DEF LL2Refinement
        <4>8. LL1Disk \in LL1UntrustedStorageType
          BY <2>1 DEF LL2Refinement
        <4>9. QED
          BY <4>3, <4>4, <4>5, <4>6, <4>7, <4>8,
              LL1RAMRecordCompositionLemma, LL1DiskRecordCompositionLemma
      (*********************************************************************)
      (* All variables other than LL1RAM are unchanged, by virtue of the   *)
      (* corresponding lemmas.                                             *)
      (*********************************************************************)
      <3>3. UNCHANGED LL1Disk
        BY <2>1, <3>1, UnchangedDiskLemma DEF LL2ReadDisk
      <3>4. UNCHANGED LL1NVRAM
        BY <2>1, <3>1, UnchangedNVRAMLemma DEF LL2ReadDisk
      <3>5. UNCHANGED LL1AvailableInputs
        BY <2>1, <3>1, UnchangedAvailableInputsLemma DEF LL2ReadDisk
      <3>6. UNCHANGED LL1ObservedOutputs
        BY <2>1, <3>1, UnchangedObservedOutputsLemma DEF LL2ReadDisk
      <3>7. UNCHANGED LL1ObservedAuthenticators
        BY <2>1, <3>1, UnchangedObservedAuthenticatorsLemma DEF LL2ReadDisk
      <3>8. QED
        BY <3>2, <3>3, <3>4, <3>5, <3>6, <3>7 DEF LL1ReadDisk
    (***********************************************************************)
    (* A `^\fastname^' LL2WriteDisk action refines to a `^\basicname^'     *)
    (* LL1WriteDisk action.                                                *)
    (***********************************************************************)
    <2>9. LL2WriteDisk => LL1WriteDisk
      <3>1. HAVE LL2WriteDisk
      (*********************************************************************)
      (* The primed state of the LL1Disk equals the unprimed state of the  *)
      (* LL1RAM.  The proof is somewhat tedious but completely             *)
      (* straightforward.                                                  *)
      (*********************************************************************)
      <3>2. LL1Disk' = LL1RAM
        <4>1. LL2Disk' = LL2RAM
          BY <3>1 DEF LL2WriteDisk
        <4>2. /\  UNCHANGED LL2NVRAM.symmetricKey
              /\  UNCHANGED LL2NVRAM.hashBarrier
          BY <3>1 DEF LL2WriteDisk
        (*******************************************************************)
        (* The primed public state field of the disk equals the unprimed   *)
        (* public state field of the RAM.                                  *)
        (*******************************************************************)
        <4>3. LL1Disk.publicState' = LL1RAM.publicState
          <5>1. LL2Disk.publicState' = LL2RAM.publicState
            BY <4>1
          <5>2. LL1RAM.publicState = LL2RAM.publicState
            BY <2>1 DEF LL2Refinement
          <5>3. LL1Disk.publicState' = LL2Disk.publicState'
            BY <2>1 DEF LL2Refinement
          <5>4. QED
            BY <5>1, <5>2, <5>3
        (*******************************************************************)
        (* The primed encrypted private state field of the disk equals the *)
        (* unprimed encrypted private state field of the RAM.              *)
        (*******************************************************************)
        <4>4. LL1Disk.privateStateEnc' = LL1RAM.privateStateEnc
          <5>1. LL2Disk.privateStateEnc' = LL2RAM.privateStateEnc
            BY <4>1
          <5>2. LL1RAM.privateStateEnc = LL2RAM.privateStateEnc
            BY <2>1 DEF LL2Refinement
          <5>3. LL1Disk.privateStateEnc' = LL2Disk.privateStateEnc'
            BY <2>1 DEF LL2Refinement
          <5>4. QED
            BY <5>1, <5>2, <5>3
        (*******************************************************************)
        (* The primed history summary field of the disk equals the         *)
        (* unprimed history summary field of the RAM.                      *)
        (*******************************************************************)
        <4>5. LL1Disk.historySummary' = LL1RAM.historySummary
          <5>1. LL2Disk.historySummary' = LL2RAM.historySummary
            BY <4>1
          <5>2. HistorySummariesMatch(
                    LL1RAM.historySummary,
                    LL2RAM.historySummary,
                    LL2NVRAM.hashBarrier)
            BY <2>1 DEF LL2Refinement
          <5>3. HistorySummariesMatch(
                    LL1Disk.historySummary',
                    LL2Disk.historySummary',
                    LL2NVRAM.hashBarrier')
            BY <2>1 DEF LL2Refinement
          <5>4. QED
            <6>1. LL1RAM.historySummary \in HashType
              BY <2>1 DEF LL2Refinement, LL1UntrustedStorageType
            <6>2. LL1Disk.historySummary' \in HashType
              BY <2>1 DEF LL2Refinement, LL1UntrustedStorageType
            <6>3. LL2RAM.historySummary \in HistorySummaryType
              BY <2>1, LL2SubtypeImplicationLemma DEF LL2SubtypeImplication
            <6>4. LL2NVRAM.hashBarrier \in HashType
              BY <2>1, LL2SubtypeImplicationLemma DEF LL2SubtypeImplication
            <6>5. QED
              BY <4>2, <5>1, <5>2, <5>3, <6>1, <6>2, <6>3, <6>4,
              HistorySummariesMatchUniqueLemma
        (*******************************************************************)
        (* The primed authenticator field of the disk equals the unprimed  *)
        (* authenticator field of the RAM.                                 *)
        (*******************************************************************)
        <4>6. LL1Disk.authenticator' = LL1RAM.authenticator
          <5>1. LL2Disk.authenticator' = LL2RAM.authenticator
            BY <4>1
          <5>2. \E symmetricKey \in SymmetricKeyType :
                    AuthenticatorsMatch(
                        LL1RAM.authenticator,
                        LL2RAM.authenticator,
                        symmetricKey,
                        LL2NVRAM.hashBarrier)
            BY <2>1 DEF LL2Refinement
          <5>3. \E symmetricKey \in SymmetricKeyType :
                    AuthenticatorsMatch(
                        LL1Disk.authenticator',
                        LL2Disk.authenticator',
                        symmetricKey,
                        LL2NVRAM.hashBarrier')
            BY <2>1 DEF LL2Refinement
          <5>4. QED
            <6>1. LL1RAM.authenticator \in MACType
              BY <2>1 DEF LL2Refinement, LL1UntrustedStorageType
            <6>2. LL1Disk.authenticator' \in MACType
              BY <2>1 DEF LL2Refinement, LL1UntrustedStorageType
            <6>3. LL2RAM.authenticator \in MACType
              BY <2>1, LL2SubtypeImplicationLemma DEF LL2SubtypeImplication
            <6>4. LL2NVRAM.hashBarrier \in HashType
              BY <2>1, LL2SubtypeImplicationLemma DEF LL2SubtypeImplication
            <6>5. QED
              BY <4>2, <5>1, <5>2, <5>3, <6>1, <6>2, <6>3, <6>4,
                  AuthenticatorsMatchUniqueLemma
        <4>7. LL1Disk' \in LL1UntrustedStorageType
          BY <2>1 DEF LL2Refinement
        <4>8. LL1RAM \in LL1UntrustedStorageType
          BY <2>1 DEF LL2Refinement
        <4>9. QED
          BY <4>3, <4>4, <4>5, <4>6, <4>7, <4>8,
              LL1DiskRecordCompositionLemma, LL1RAMRecordCompositionLemma
      (*********************************************************************)
      (* All variables other than LL1Disk are unchanged, by virtue of the  *)
      (* corresponding lemmas.                                             *)
      (*********************************************************************)
      <3>3. UNCHANGED LL1RAM
        BY <2>1, <3>1, UnchangedRAMLemma DEF LL2WriteDisk
      <3>4. UNCHANGED LL1NVRAM
        BY <2>1, <3>1, UnchangedNVRAMLemma DEF LL2WriteDisk
      <3>5. UNCHANGED LL1AvailableInputs
        BY <2>1, <3>1, UnchangedAvailableInputsLemma DEF LL2WriteDisk
      <3>6. UNCHANGED LL1ObservedOutputs
        BY <2>1, <3>1, UnchangedObservedOutputsLemma DEF LL2WriteDisk
      <3>7. UNCHANGED LL1ObservedAuthenticators
        BY <2>1, <3>1, UnchangedObservedAuthenticatorsLemma DEF LL2WriteDisk
      <3>8. QED
        BY <3>2, <3>3, <3>4, <3>5, <3>6, <3>7 DEF LL1WriteDisk
    (***********************************************************************)
    (* A `^\fastname^' LL2CorruptRAM action refines to a `^\basicname^'    *)
    (* LL1CorruptRAM action.                                               *)
    (***********************************************************************)
    <2>10. LL2CorruptRAM => LL1CorruptRAM
      (*********************************************************************)
      (* We assume the antecedent.                                         *)
      (*********************************************************************)
      <3>1. HAVE LL2CorruptRAM
      (*********************************************************************)
      (* We pick a set of variables of the appropriate types that satisfy  *)
      (* the LL2CorruptRAM action.                                         *)
      (*********************************************************************)
      <3>2. PICK ll2UntrustedStorage \in LL2UntrustedStorageType,
                 ll2FakeSymmetricKey \in SymmetricKeyType \ {LL2NVRAM.symmetricKey},
                 ll2Hash \in HashType :
                 LL2CorruptRAM!(ll2UntrustedStorage, ll2FakeSymmetricKey, ll2Hash)
        BY <3>1 DEF LL2CorruptRAM
      (*********************************************************************)
      (* We pick a symmetric key that satisfies the AuthenticatorsMatch    *)
      (* predicate for the primed states of the authenticators in the RAM  *)
      (* variables of the two specs.                                       *)
      (*********************************************************************)
      <3>3. PICK symmetricKey \in SymmetricKeyType :
                AuthenticatorsMatch(
                    LL1RAM.authenticator',
                    LL2RAM.authenticator',
                    symmetricKey,
                    LL2NVRAM.hashBarrier')
        BY <2>1 DEF LL2Refinement
      (*********************************************************************)
      (* We pick a set of variables of the appropriate types that satisfy  *)
      (* the quantified AuthenticatorsMatch predicate.                     *)
      (*********************************************************************)
      <3>4. PICK stateHash \in HashType,
                ll1HistorySummary \in HashType,
                ll2HistorySummary \in HistorySummaryType :
                AuthenticatorsMatch(
                    LL1RAM.authenticator',
                    LL2RAM.authenticator',
                    symmetricKey,
                    LL2NVRAM.hashBarrier')!(
                    stateHash, ll1HistorySummary, ll2HistorySummary)!1
        BY <3>3 DEF AuthenticatorsMatch
      (*********************************************************************)
      (* We re-state the definitions from the LET in AuthenticatorsMatch.  *)
      (*********************************************************************)
      <3> ll1HistoryStateBinding == Hash(ll1HistorySummary, stateHash)
      <3> ll2HistorySummaryHash == Hash(ll2HistorySummary.anchor, ll2HistorySummary.extension)
      <3> ll2HistoryStateBinding == Hash(ll2HistorySummaryHash, stateHash)
      (*********************************************************************)
      (* We prove the types of the definitions, with help from the         *)
      (* AuthenticatorsMatchDefsTypeSafeLemma.                             *)
      (*********************************************************************)
      <3>5. /\  ll1HistoryStateBinding \in HashType
            /\  ll2HistorySummaryHash \in HashType
            /\  ll2HistoryStateBinding \in HashType
        BY <3>4, AuthenticatorsMatchDefsTypeSafeLemma
      (*********************************************************************)
      (* We hide the definitions.                                          *)
      (*********************************************************************)
      <3> HIDE DEF ll1HistoryStateBinding, ll2HistorySummaryHash, ll2HistoryStateBinding
      (*********************************************************************)
      (* We prove each aspect of LL1CorruptRAM separately.  First, we      *)
      (* prove the types of the three existentially quantified variables   *)
      (* in the definition of LL1CorruptRAM.                               *)
      (*********************************************************************)
      <3>6. LL1RAM' \in LL1UntrustedStorageType
        BY <2>1 DEF LL2Refinement
      <3>7. ll2FakeSymmetricKey \in SymmetricKeyType \ {LL1NVRAM.symmetricKey}
        <4>1. ll2FakeSymmetricKey \in SymmetricKeyType \ {LL2NVRAM.symmetricKey}
          BY <3>2
        <4>2. LL1NVRAM.symmetricKey = LL2NVRAM.symmetricKey
          BY <2>1 DEF LL2Refinement
        <4>3. QED
          BY <4>1, <4>2
      <3>8. ll1HistoryStateBinding \in HashType
        BY <3>5
      (*********************************************************************)
      (* We then prove the disjunction regarding the authenticator field   *)
      (* in the primed LL1RAM record.                                      *)
      (*********************************************************************)
      <3>9. \/  LL1RAM.authenticator' \in
                    LL1ObservedAuthenticators
            \/  LL1RAM.authenticator' =
                    GenerateMAC(ll2FakeSymmetricKey, ll1HistoryStateBinding)
        (*******************************************************************)
        (* We prove that when the authenticator in the `^\fastname^' spec  *)
        (* is in the set of observed authenticators, then the              *)
        (* authenticator in the `^\basicname^' spec is in the set of       *)
        (* observed authenticators.  This follows directly from the        *)
        (* AuthenticatorInSetLemma, once we prove the preconditions for    *)
        (* the lemma.                                                      *)
        (*******************************************************************)
        <4>1. ASSUME ll2UntrustedStorage.authenticator \in
                         LL2ObservedAuthenticators
              PROVE LL1RAM.authenticator' \in LL1ObservedAuthenticators
          (*****************************************************************)
          (* We need to prove some types.                                  *)
          (*****************************************************************)
          <5>1. LL1RAM.authenticator' \in MACType
            <6>1. LL1RAM' \in LL1UntrustedStorageType
              BY <2>1 DEF LL2Refinement
            <6>2. QED
              BY <6>1 DEF LL1UntrustedStorageType
          <5>2. LL2RAM.authenticator' \in MACType
            BY <2>1, LL2SubtypeImplicationLemma DEF LL2SubtypeImplication
          <5>3. LL1ObservedAuthenticators \in SUBSET MACType
            BY <2>1 DEF LL2Refinement
          <5>4. LL2ObservedAuthenticators \in SUBSET MACType
            BY <2>1 DEF LL2TypeInvariant
          <5>5. LL2NVRAM.symmetricKey \in SymmetricKeyType
              BY <2>1, LL2SubtypeImplicationLemma DEF LL2SubtypeImplication
          <5>6. LL2NVRAM.hashBarrier \in HashType
              BY <2>1, LL2SubtypeImplicationLemma DEF LL2SubtypeImplication
          (*****************************************************************)
          (* Then we prove the three conjuncts in the antecedent of the    *)
          (* AuthenticatorInSetLemma.  The first conjunct follows from the *)
          (* refinement.                                                   *)
          (*****************************************************************)
          <5>7. \E symmetricKey1 \in SymmetricKeyType :
                    AuthenticatorsMatch(
                        LL1RAM.authenticator',
                        LL2RAM.authenticator',
                        symmetricKey1,
                        LL2NVRAM.hashBarrier)
            <6>1. \E symmetricKey1 \in SymmetricKeyType :
                      AuthenticatorsMatch(
                          LL1RAM.authenticator',
                          LL2RAM.authenticator',
                          symmetricKey1,
                          LL2NVRAM.hashBarrier')
              BY <2>1 DEF LL2Refinement
            <6>2. UNCHANGED LL2NVRAM.hashBarrier
              BY <3>2
            <6>3. QED
              BY <6>1, <6>2
          (*****************************************************************)
          (* The second conjunct in the antecedent of the                  *)
          (* AuthenticatorInSetLemma also follows from the refinement.     *)
          (*****************************************************************)
          <5>8. AuthenticatorSetsMatch(
                    LL1ObservedAuthenticators,
                    LL2ObservedAuthenticators,
                    LL2NVRAM.symmetricKey,
                    LL2NVRAM.hashBarrier)
            BY <2>1 DEF LL2Refinement
          (*****************************************************************)
          (* The third conjunct in the antecedent of the                   *)
          (* AuthenticatorInSetLemma follows from the first disjunct in    *)
          (* the LL2CorruptRAM action.                                     *)
          (*****************************************************************)
          <5>9. LL2RAM.authenticator' \in LL2ObservedAuthenticators
            <6>1. LL2RAM.authenticator' = ll2UntrustedStorage.authenticator
              <7>1. LL2RAM' = ll2UntrustedStorage
                BY <3>2
              <7>2. QED
                BY <7>1
            <6>2. ll2UntrustedStorage.authenticator \in LL2ObservedAuthenticators
              BY <4>1
            <6>3. QED
              BY <6>1, <6>2
          (*****************************************************************)
          (* We then invoke the AuthenticatorInSetLemma directly.          *)
          (*****************************************************************)
          <5>10. QED
            BY <5>1, <5>2, <5>3, <5>4, <5>5, <5>6, <5>7, <5>8, <5>9, AuthenticatorInSetLemma
        (*******************************************************************)
        (* We prove that when the authenticator in the `^\fastname^' spec  *)
        (* is generated as a MAC with a fake symmetric key, then the       *)
        (* authenticator in the `^\basicname^' spec is generated as a MAC  *)
        (* with a fake symmetric key.  This follows directly from the      *)
        (* AuthenticatorGeneratedLemma, once we prove the preconditions    *)
        (* for the lemma.                                                  *)
        (*******************************************************************)
        <4>2. ASSUME ll2UntrustedStorage.authenticator =
                         GenerateMAC(ll2FakeSymmetricKey, ll2Hash)
              PROVE LL1RAM.authenticator' =
                        GenerateMAC(ll2FakeSymmetricKey, ll1HistoryStateBinding)
          (*****************************************************************)
          (* We need to prove some types.                                  *)
          (*****************************************************************)
          <5>1. stateHash \in HashType
            BY <3>4
          <5>2. ll1HistorySummary \in HashType
            BY <3>4
          <5>3. ll2HistorySummary \in HistorySummaryType
            BY <3>4
          <5>4. LL1RAM.authenticator' \in MACType
            <6>1. LL1RAM' \in LL1UntrustedStorageType
              BY <2>1 DEF LL2Refinement
            <6>2. QED
              BY <6>1 DEF LL1UntrustedStorageType
          <5>5. LL2RAM.authenticator' \in MACType
            BY <2>1, LL2SubtypeImplicationLemma DEF LL2SubtypeImplication
          <5>6. ll2FakeSymmetricKey \in SymmetricKeyType
            <6>1. ll2FakeSymmetricKey \in SymmetricKeyType \ {LL2NVRAM.symmetricKey}
              BY <3>2
            <6>2. QED
              BY <6>1
          <5>7. LL2NVRAM.hashBarrier' \in HashType
              BY <2>1, LL2SubtypeImplicationLemma DEF LL2SubtypeImplication, LL2TypeInvariant, LL2TrustedStorageType
          (*****************************************************************)
          (* Then we prove the three conjuncts in the antecedent of the    *)
          (* AuthenticatorGeneratedLemma.  The first conjunct follows from *)
          (* a conjunct in the refinement.                                 *)
          (*****************************************************************)
          <5>8. HistorySummariesMatch(ll1HistorySummary, ll2HistorySummary, LL2NVRAM.hashBarrier')
            BY <3>4
          (*****************************************************************)
          (* The second conjunct in the antecedent of the                  *)
          (* AuthenticatorGeneratedLemma mainly follows from the           *)
          (* refinement, but we also have to prove that the symmetric key  *)
          (* specified by an existential in the refinement matches the     *)
          (* fake symmetric key specified in the LL2CorruptRAM action.     *)
          (* This follows from the MACUnforgeable property.                *)
          (*****************************************************************)
          <5>9. AuthenticatorsMatch(
                    LL1RAM.authenticator',
                    LL2RAM.authenticator',
                    ll2FakeSymmetricKey,
                    LL2NVRAM.hashBarrier')
            (***************************************************************)
            (* The main precondition for the MACUnforgeable property is    *)
            (* that the generated MAC is validated.  We first prove the    *)
            (* validation, which follows from the refinement.              *)
            (***************************************************************)
            <6>1. ValidateMAC(symmetricKey, ll2HistoryStateBinding, LL2RAM.authenticator')
              BY <3>4 DEF ll2HistoryStateBinding, ll2HistorySummaryHash
            (***************************************************************)
            (* We then prove the generation, which follows from the        *)
            (* LL2CorruptRAM action.                                       *)
            (***************************************************************)
            <6>2. LL2RAM.authenticator' = GenerateMAC(ll2FakeSymmetricKey, ll2Hash)
              <7>1. LL2RAM' = ll2UntrustedStorage
                BY <3>2
              <7>2. ll2UntrustedStorage.authenticator = GenerateMAC(ll2FakeSymmetricKey, ll2Hash)
                BY <4>2
              <7>3. QED
                BY <7>1, <7>2
            (***************************************************************)
            (* The remaining preconditions are types.                      *)
            (***************************************************************)
            <6>3. symmetricKey \in SymmetricKeyType
              BY <3>3
            <6>4. ll2FakeSymmetricKey \in SymmetricKeyType
              <7>1. ll2FakeSymmetricKey \in SymmetricKeyType \ {LL2NVRAM.symmetricKey}
                BY <3>2
              <7>2. QED
                BY <7>1
            <6>5. ll2HistoryStateBinding \in HashType
              BY <3>5
            <6>6. ll2Hash \in HashType
              BY <3>2
            (***************************************************************)
            (* The MACUnforgeable property tells us that the two keys are  *)
            (* equal.                                                      *)
            (***************************************************************)
            <6>7. symmetricKey = ll2FakeSymmetricKey
                BY <6>1, <6>2, <6>3, <6>4, <6>5, <6>6, MACUnforgeable
            <6>8. QED
              BY <3>3, <6>7
          (*****************************************************************)
          (* The third conjunct in the antecedent of the                   *)
          (* AuthenticatorGeneratedLemma mainly follows from the           *)
          (* LL2CorruptRAM action, but we also have to prove that the      *)
          (* history state binding specified in the refinement matches the *)
          (* arbitrary hash specified in the LL2CorruptRAM action.         *)
          (*****************************************************************)
          <5>10. LL2RAM.authenticator' =
                         GenerateMAC(ll2FakeSymmetricKey, ll2HistoryStateBinding)
            (***************************************************************)
            (* The state authenticaton in the primed `^\fastname^' RAM     *)
            (* equals the authenticator specified by the existential       *)
            (* ll2UntrustedStorage in the LL2CorruptRAM action.            *)
            (***************************************************************)
            <6>1. LL2RAM.authenticator' = ll2UntrustedStorage.authenticator
              <7>1. LL2RAM' = ll2UntrustedStorage
                BY <3>2
              <7>2. QED
                BY <7>1
            (***************************************************************)
            (* The authenticator specified by the existential              *)
            (* ll2UntrustedStorage in the LL2CorruptRAM action is          *)
            (* generated as a MAC of the history state binding from the    *)
            (* AuthenticatorsMatch predicate invoked by the refinement.    *)
            (* This follows from the MACCollisionResistant property.       *)
            (***************************************************************)
            <6>2. ll2UntrustedStorage.authenticator =
                      GenerateMAC(ll2FakeSymmetricKey, ll2HistoryStateBinding)
              (*************************************************************)
              (* The main precondition for the MACUnforgeable property is  *)
              (* that the generated MAC is validated.  We first prove the  *)
              (* validation, which follows from the refinement.            *)
              (*************************************************************)
              <7>1. ValidateMAC(symmetricKey, ll2HistoryStateBinding, LL2RAM.authenticator')
                BY <3>4 DEF ll2HistoryStateBinding, ll2HistorySummaryHash
              (*************************************************************)
              (* We then prove the generation, which follows from the      *)
              (* LL2CorruptRAM action.                                     *)
              (*************************************************************)
              <7>2. LL2RAM.authenticator' = GenerateMAC(ll2FakeSymmetricKey, ll2Hash)
                <8>1. LL2RAM' = ll2UntrustedStorage
                  BY <3>2
                <8>2. ll2UntrustedStorage.authenticator = GenerateMAC(ll2FakeSymmetricKey, ll2Hash)
                  BY <4>2
                <8>3. QED
                  BY <8>1, <8>2
              (*************************************************************)
              (* The remaining preconditions are types.                    *)
              (*************************************************************)
              <7>3. symmetricKey \in SymmetricKeyType
                BY <3>3
              <7>4. ll2FakeSymmetricKey \in SymmetricKeyType
                <8>1. ll2FakeSymmetricKey \in SymmetricKeyType \ {LL2NVRAM.symmetricKey}
                  BY <3>2
                <8>2. QED
                  BY <8>1
              <7>5. ll2HistoryStateBinding \in HashType
                BY <3>5
              <7>6. ll2Hash \in HashType
                BY <3>2
              (*************************************************************)
              (* The MACCollisionResistant property tells us that the two  *)
              (* hash values are equal.                                    *)
              (*************************************************************)
              <7>7. ll2Hash = ll2HistoryStateBinding
                BY <7>1, <7>2, <7>3, <7>4, <7>5, <7>6, MACCollisionResistant
              <7>8. QED
                BY <4>2, <7>7
            <6>3. QED
              BY <6>1, <6>2
          (*****************************************************************)
          (* We then invoke the AuthenticatorGeneratedLemma directly.      *)
          (*****************************************************************)
          <5>11. QED
            BY <5>1, <5>2, <5>3, <5>4, <5>5, <5>6, <5>7, <5>8, <5>9, <5>10,
            AuthenticatorGeneratedLemma
            DEF ll1HistoryStateBinding, ll2HistorySummaryHash, ll2HistoryStateBinding
        (*******************************************************************)
        (* From the definition of LL2CorruptRAM, the above two cases are   *)
        (* exhaustive.                                                     *)
        (*******************************************************************)
        <4>3.  \/  ll2UntrustedStorage.authenticator \in
                       LL2ObservedAuthenticators
               \/  ll2UntrustedStorage.authenticator =
                       GenerateMAC(ll2FakeSymmetricKey, ll2Hash)
          BY <3>2
        <4>4. QED
          BY <4>1, <4>2, <4>3
      (*********************************************************************)
      (* All variables other than LL1RAM are unchanged, by virtue of the   *)
      (* corresponding lemmas.                                             *)
      (*********************************************************************)
      <3>10. UNCHANGED LL1Disk
        BY <2>1, <3>2, UnchangedDiskLemma DEF LL2CorruptRAM
      <3>11. UNCHANGED LL1NVRAM
        BY <2>1, <3>2, UnchangedNVRAMLemma DEF LL2CorruptRAM
      <3>12. UNCHANGED LL1AvailableInputs
        BY <2>1, <3>2, UnchangedAvailableInputsLemma DEF LL2CorruptRAM
      <3>13. UNCHANGED LL1ObservedOutputs
        BY <2>1, <3>2, UnchangedObservedOutputsLemma DEF LL2CorruptRAM
      <3>14. UNCHANGED LL1ObservedAuthenticators
        BY <2>1, <3>2, UnchangedObservedAuthenticatorsLemma DEF LL2CorruptRAM
      (*********************************************************************)
      (* Lastly, we tie together all of the required aspects of the        *)
      (* LL1CorruptRAM definition.                                         *)
      (*********************************************************************)
      <3>15. QED
        BY <3>6, <3>7, <3>8, <3>9, <3>10, <3>11, <3>12, <3>13, <3>14 DEF LL1CorruptRAM
    (***********************************************************************)
    (* A `^\fastname^' LL2CorruptSPCR action refines to one of two actions *)
    (* in the `^\basicname^' spec.  If an extension is in progress at the  *)
    (* time the LL2Restart occurs, the corruption of the SPCR value caused *)
    (* by the LL2CorruptSPCR is fatal, so this refines to a                *)
    (* LL1RestrictedCorruption action in the `^\basicname^' spec, which in *)
    (* turn refines to an HLDie action in the high-level spec.  On the     *)
    (* other hand, if an extension is not in progress at the time the      *)
    (* LL2Restart occurs, the action refines to a stuttering step in the   *)
    (* `^\basicname^' spec.                                                *)
    (***********************************************************************)
    <2>11. LL2CorruptSPCR =>
            IF LL2NVRAM.extensionInProgress
            THEN
                LL1RestrictedCorruption
            ELSE
                UNCHANGED LL1Vars
      (*********************************************************************)
      (* We assume the antecedent.                                         *)
      (*********************************************************************)
      <3>1. HAVE LL2CorruptSPCR
      (*********************************************************************)
      (* We pick a fake hash that satisfies the LL2CorruptSPCR action.     *)
      (*********************************************************************)
      <3>2. PICK fakeHash \in HashDomain : LL2CorruptSPCR!(fakeHash)
        BY <3>1 DEF LL2CorruptSPCR
      (*********************************************************************)
      (* We re-state the definition from LL2CorruptSPCR.                   *)
      (*********************************************************************)
      <3> newHistorySummaryExtension == Hash(LL2SPCR, fakeHash)
      (*********************************************************************)
      (* We then hide the definition.                                      *)
      (*********************************************************************)
      <3> HIDE DEF newHistorySummaryExtension
      (*********************************************************************)
      (* Since the refinement is an IF-THEN-ELSE, we prove the THEN and    *)
      (* ELSE cases separately.  For the THEN case, we assume that an      *)
      (* extension is in progress and show that this refines to a          *)
      (* LL1RestrictedCorruption action.                                   *)
      (*********************************************************************)
      <3>3. ASSUME LL2NVRAM.extensionInProgress
            PROVE LL1RestrictedCorruption
        (*******************************************************************)
        (* One fact that will be useful in several places below is that    *)
        (* the extension field in the logical history summary of the       *)
        (* `^\fastname^' NVRAM and SPCR is not equal to the base hash      *)
        (* value.  We'll separately prove the two cases of whether or not  *)
        (* the primed SPCR equals the base hash value.                     *)
        (*******************************************************************)
        <4>1. LL2NVRAMLogicalHistorySummary.extension' # BaseHashValue
          (*****************************************************************)
          (* One fact that will be useful for both cases is that an        *)
          (* extension is in progress in the primed state.                 *)
          (*****************************************************************)
          <5>1. LL2NVRAM.extensionInProgress'
            <6>1. LL2NVRAM.extensionInProgress
              BY <3>3
            <6>2. UNCHANGED LL2NVRAM.extensionInProgress
              BY <3>2
            <6>3. QED
              BY <6>1, <6>2
          (*****************************************************************)
          (* The first case is fairly simple.  When an extension is in     *)
          (* progress but the SPCR equals the base hash value, the logical *)
          (* history summary equals a crazy hash value, because this       *)
          (* sitution should never arise during normal operation.          *)
          (*****************************************************************)
          <5>2. CASE LL2SPCR' = BaseHashValue
            <6>1. LL2NVRAMLogicalHistorySummary.extension' = CrazyHashValue
              <7>1. LL2NVRAMLogicalHistorySummary' = [
                        anchor |-> LL2NVRAM.historySummaryAnchor',
                        extension |-> CrazyHashValue ]
                BY <5>1, <5>2 DEF LL2NVRAMLogicalHistorySummary
              <7>2. QED
                BY <7>1
            <6>2. CrazyHashValue # BaseHashValue
              BY CrazyHashValueUnique
            <6>3. QED
              BY <6>1, <6>2
          (*****************************************************************)
          (* The second case is slightly more involved.  We will prove     *)
          (* this in two steps.                                            *)
          (*****************************************************************)
          <5>3. CASE LL2SPCR' # BaseHashValue
            (***************************************************************)
            (* First, we prove that the primed value of the extension      *)
            (* field in the `^\fastname^' logical history summary equals   *)
            (* the new history summary extension defined in the            *)
            (* LL2CorruptSPCR action.  This follows fairly directly from   *)
            (* the definitions of the LL2CorruptSPCR action and the        *)
            (* LL2NVRAMLogicalHistorySummary operator.                     *)
            (***************************************************************)
            <6>1. LL2NVRAMLogicalHistorySummary.extension' = newHistorySummaryExtension
              <7>1. LL2SPCR' = newHistorySummaryExtension
                BY <3>2 DEF newHistorySummaryExtension
              <7>2. LL2NVRAMLogicalHistorySummary' = [
                           anchor |-> LL2NVRAM.historySummaryAnchor',
                           extension |-> LL2SPCR' ]
                BY <5>1, <5>3 DEF LL2NVRAMLogicalHistorySummary
              <7>3. QED
                BY <7>1, <7>2
            (***************************************************************)
            (* Second, we prove that the new history summary extension is  *)
            (* not equal to the base hash value.  This follows because the *)
            (* new history summary extension is generated as a hash by the *)
            (* LL2CorruptSPCR action, and the BaseHashValueUnique property *)
            (* tells us that no hash value generated by the Hash function  *)
            (* can equal the base hash value.                              *)
            (***************************************************************)
            <6>2. newHistorySummaryExtension # BaseHashValue
              <7>1. LL2SPCR \in HashDomain
                <8>1. LL2SPCR \in HashType
                  BY <2>1 DEF LL2TypeInvariant
                <8>2. QED
                  BY <8>1 DEF HashDomain
              <7>2. fakeHash \in HashDomain
                BY <3>2
              <7>3. QED
                BY <7>1, <7>2, BaseHashValueUnique DEF newHistorySummaryExtension
            <6>3. QED
              BY <6>1, <6>2
          (*****************************************************************)
          (* The two cases are exhaustive.                                 *)
          (*****************************************************************)
          <5>4. QED
            BY <5>2, <5>3
        (*******************************************************************)
        (* We prove each conjunct of LL1RestrictedCorruption separately.   *)
        (* First, we prove the conjunct relating to the NVRAM.             *)
        (*******************************************************************)
        <4>2. LL1RestrictedCorruption!nvram
          (*****************************************************************)
          (* The primed value of the history summary field in the          *)
          (* `^\basicname^' NVRAM serves as our witness for the garbage    *)
          (* history summary.                                              *)
          (*****************************************************************)
          <5>1. LL1NVRAM.historySummary' \in HashType
            BY <2>1 DEF LL2Refinement, LL1TrustedStorageType
          (*****************************************************************)
          (* We prove that the constraint labeled `current' in the         *)
          (* LL1RestrictedCorruption action is satisfied.                  *)
          (*****************************************************************)
          <5>2. LL1RestrictedCorruption!nvram!current(LL1NVRAM.historySummary')
            (***************************************************************)
            (* To prove the universally quantified expression, we take a   *)
            (* set of variables of the appropriate types.                  *)
            (***************************************************************)
            <6>1. TAKE stateHash1 \in HashType,
                       ll1Authenticator \in LL1ObservedAuthenticators
            (***************************************************************)
            (* We re-state the definition from within the                  *)
            (* LL1RestrictedCorruption!nvram!current clause.               *)
            (***************************************************************)
            <6> ll1GarbageHistoryStateBinding == Hash(LL1NVRAM.historySummary', stateHash1)
            (***************************************************************)
            (* We hide the definition.                                     *)
            (***************************************************************)
            <6> HIDE DEF ll1GarbageHistoryStateBinding
            (***************************************************************)
            (* We need to prove the nvram::current conjunct, which asserts *)
            (* that the authenticator is not a valid MAC for the history   *)
            (* state binding formed from the history summary in the NVRAM  *)
            (* and any state hash.                                         *)
            (***************************************************************)
            <6>2. ~ValidateMAC(
                      LL1NVRAM.symmetricKey,
                      ll1GarbageHistoryStateBinding,
                      ll1Authenticator)
              (*************************************************************)
              (* We will use proof by contradiction.                       *)
              (*************************************************************)
              <7>1. SUFFICES
                    ASSUME
                        ValidateMAC(
                            LL1NVRAM.symmetricKey,
                            ll1GarbageHistoryStateBinding,
                            ll1Authenticator)
                    PROVE
                        FALSE
                OBVIOUS
              (*************************************************************)
              (* We first pick, from the set of `^\fastname^' observed     *)
              (* authenticators, a `^\fastname^' authenticator that        *)
              (* matches the `^\basicname^' authenticator.  We know that   *)
              (* such a authenticator exists, because the refinement       *)
              (* asserts that the sets of observed authenticators match    *)
              (* across the two specs.                                     *)
              (*************************************************************)
              <7>2. PICK ll2Authenticator \in LL2ObservedAuthenticators :
                        AuthenticatorsMatch(
                            ll1Authenticator,
                            ll2Authenticator,
                            LL2NVRAM.symmetricKey,
                            LL2NVRAM.hashBarrier)
                <8>1. ll1Authenticator \in LL1ObservedAuthenticators
                  BY <6>1
                <8>2. AuthenticatorSetsMatch(
                          LL1ObservedAuthenticators,
                          LL2ObservedAuthenticators,
                          LL2NVRAM.symmetricKey,
                          LL2NVRAM.hashBarrier)
                  BY <2>1 DEF LL2Refinement
                <8>3. QED
                  BY <8>1, <8>2 DEF AuthenticatorSetsMatch
              (*************************************************************)
              (* We pick a set of variables of the appropriate types that  *)
              (* satisfy the quantified AuthenticatorsMatch predicate.     *)
              (*************************************************************)
              <7>3. PICK stateHash2 \in HashType,
                        ll1HistorySummary \in HashType,
                        ll2HistorySummary \in HistorySummaryType :
                        AuthenticatorsMatch(
                            ll1Authenticator,
                            ll2Authenticator,
                            LL2NVRAM.symmetricKey,
                            LL2NVRAM.hashBarrier)!(
                            stateHash2, ll1HistorySummary, ll2HistorySummary)!1
                BY <7>2 DEF AuthenticatorsMatch
              (*************************************************************)
              (* We re-state the definitions from the LET in               *)
              (* AuthenticatorsMatch.                                      *)
              (*************************************************************)
              <7> ll1HistoryStateBinding == Hash(ll1HistorySummary, stateHash2)
              <7> ll2HistorySummaryHash == Hash(ll2HistorySummary.anchor, ll2HistorySummary.extension)
              <7> ll2HistoryStateBinding == Hash(ll2HistorySummaryHash, stateHash2)
              (*************************************************************)
              (* We prove the types of the definitions, with help from the *)
              (* AuthenticatorsMatchDefsTypeSafeLemma.                     *)
              (*************************************************************)
              <7>4. /\  ll1HistoryStateBinding \in HashType
                    /\  ll2HistorySummaryHash \in HashType
                    /\  ll2HistoryStateBinding \in HashType
                BY <7>3, AuthenticatorsMatchDefsTypeSafeLemma
              (*************************************************************)
              (* We hide the definitions.                                  *)
              (*************************************************************)
              <7> HIDE DEF ll1HistoryStateBinding, ll2HistorySummaryHash, ll2HistoryStateBinding
              (*************************************************************)
              (* We prove that the `^\basicname^'s history summary picked  *)
              (* to satisfy the AuthenticatorsMatch predicate equals the   *)
              (* history summary in the primed state of the `^\basicname^' *)
              (* NVRAM.                                                    *)
              (*************************************************************)
              <7>5. ll1HistorySummary = LL1NVRAM.historySummary'
                (***********************************************************)
                (* The first step is to show the equality of the history   *)
                (* state bindings that bind each of these history          *)
                (* summaries to their respective state hashes.             *)
                (***********************************************************)
                <8>1. ll1HistoryStateBinding = ll1GarbageHistoryStateBinding
                  (*********************************************************)
                  (* By hypothesis, the authenticator is a valid MAC for   *)
                  (* the garbage history state binding.                    *)
                  (*********************************************************)
                  <9>1. ValidateMAC(
                            LL2NVRAM.symmetricKey, ll1GarbageHistoryStateBinding, ll1Authenticator)
                    <10>1. LL1NVRAM.symmetricKey = LL2NVRAM.symmetricKey
                      BY <2>1 DEF LL2Refinement
                    <10>2. QED
                      BY <7>1, <10>1
                  (*********************************************************)
                  (* The definition of the AuthenticatorsMatch predicate   *)
                  (* tells us that the `^\basicname^' authenticator was    *)
                  (* generated as a MAC from the history state binding.    *)
                  (*********************************************************)
                  <9>2. ll1Authenticator = GenerateMAC(LL2NVRAM.symmetricKey, ll1HistoryStateBinding)
                    BY <7>3 DEF ll1HistoryStateBinding
                  (*********************************************************)
                  (* The remaining preconditions are types.                *)
                  (*********************************************************)
                  <9>3. LL2NVRAM.symmetricKey \in SymmetricKeyType
                    BY <2>1, LL2SubtypeImplicationLemma DEF LL2SubtypeImplication
                  <9>4. ll1HistoryStateBinding \in HashType
                    BY <7>4
                  <9>5. ll1GarbageHistoryStateBinding \in HashType
                    <10>1. LL1NVRAM.historySummary' \in HashDomain
                      BY <5>1 DEF HashDomain
                    <10>2. stateHash1 \in HashDomain
                      BY <5>1 DEF HashDomain
                    <10>3. QED
                      BY <10>1, <10>2, HashTypeSafe DEF ll1GarbageHistoryStateBinding
                  (*********************************************************)
                  (* The MACCollisionResistant property tells us that the  *)
                  (* two history state bindings are equal.                 *)
                  (*********************************************************)
                  <9>6. QED
                    BY <9>1, <9>2, <9>3, <9>4, <9>5, MACCollisionResistant
                (***********************************************************)
                (* By the collision resistance of the hash function, the   *)
                (* equality of the history state bindings implies the      *)
                (* equality of the history summaries.                      *)
                (***********************************************************)
                <8>2. QED
                  <9>1. LL1NVRAM.historySummary' \in HashDomain
                    <10>1. LL1NVRAM.historySummary' \in HashType
                      BY <2>1 DEF LL2Refinement, LL1TrustedStorageType
                    <10>2. QED
                      BY <10>1 DEF HashDomain
                  <9>2. ll1HistorySummary \in HashDomain
                    <10>1. ll1HistorySummary \in HashType
                      BY <7>3
                    <10>2. QED
                      BY <10>1 DEF HashDomain
                  <9>3. stateHash1 \in HashDomain
                    <10>1. stateHash1 \in HashType
                      BY <6>1
                    <10>2. QED
                      BY <10>1 DEF HashDomain
                  <9>4. stateHash2 \in HashDomain
                    <10>1. stateHash2 \in HashType
                      BY <7>3
                   <10>2. QED
                      BY <10>1 DEF HashDomain
                  <9>5. QED
                    BY <8>1, <9>2, <9>1, <9>4, <9>3, HashCollisionResistant
                    DEF ll1HistoryStateBinding, ll1GarbageHistoryStateBinding
              (*************************************************************)
              (* The `^\basicname^'s history summary picked to satisfy the *)
              (* AuthenticatorsMatch predicate matches the primed logical  *)
              (* history summary in the `^\fastname^' NVRAM and SPCR, by   *)
              (* the refinement and the above equality.                    *)
              (*************************************************************)
              <7>6. HistorySummariesMatch(
                        ll1HistorySummary,
                        LL2NVRAMLogicalHistorySummary',
                        LL2NVRAM.hashBarrier')
                <8>1. HistorySummariesMatch(
                          LL1NVRAM.historySummary',
                          LL2NVRAMLogicalHistorySummary',
                          LL2NVRAM.hashBarrier')
                  BY <2>1 DEF LL2Refinement
                <8>2. QED
                  BY <7>5, <8>1
              (*************************************************************)
              (* We prove that the HistorySummariesMatch predicate equals  *)
              (* the HistorySummariesMatchRecursion predicate in this      *)
              (* case.  We assert each condition required by the           *)
              (* definition of the predicate.                              *)
              (*************************************************************)
              <7>7. HistorySummariesMatch(
                        ll1HistorySummary, LL2NVRAMLogicalHistorySummary', LL2NVRAM.hashBarrier') =
                    HistorySummariesMatchRecursion(
                        ll1HistorySummary, LL2NVRAMLogicalHistorySummary', LL2NVRAM.hashBarrier')
                (***********************************************************)
                (* We prove some types, to satisfy the universal           *)
                (* quantifiers in HistorySummariesMatchDefinition.         *)
                (***********************************************************)
                <8>1. ll1HistorySummary \in HashType
                  BY <7>3
                <8>2. LL2NVRAMLogicalHistorySummary' \in HistorySummaryType
                  BY <2>1, LL2NVRAMLogicalHistorySummaryTypeSafe
                <8>3. LL2NVRAM.hashBarrier' \in HashType
                  BY <2>1, LL2SubtypeImplicationLemma DEF LL2SubtypeImplication, LL2TypeInvariant, LL2TrustedStorageType
                (***********************************************************)
                (* We prove that the `^\fastname^' logical history summary *)
                (* does not equal the initial history summary.             *)
                (***********************************************************)
                <8> ll2InitialHistorySummary == [ anchor |-> BaseHashValue, extension |-> BaseHashValue ]
                <8>4. LL2NVRAMLogicalHistorySummary' # ll2InitialHistorySummary
                  <9>1. LL2NVRAMLogicalHistorySummary.extension' # BaseHashValue
                    BY <4>1
                  <9>2. ll2InitialHistorySummary.extension = BaseHashValue
                    BY DEF ll2InitialHistorySummary
                  <9>3. QED
                    BY <9>1, <9>2
                (***********************************************************)
                (* Finally, from HistorySummariesMatchDefinition, we can   *)
                (* conclude that the HistorySummariesMatch predicate       *)
                (* equals the quantified HistorySummariesMatchRecursion    *)
                (* predicate.                                              *)
                (***********************************************************)
                <8>5. QED
                  BY <8>1, <8>2, <8>3, <8>4, HistorySummariesMatchDefinition
                  DEF ll2InitialHistorySummary
              (*************************************************************)
              (* We pick values for the existential variables inside the   *)
              (* HistorySummariesMatchRecursion predicate that satisfy the *)
              (* predicate.  We know such variables exist, because the     *)
              (* predicate is satisfied by the two previous steps.         *)
              (*************************************************************)
              <7>8. PICK prevInput \in InputType,
                         previousLL1HistorySummary \in HashType,
                         previousLL2HistorySummary \in HistorySummaryType :
                         HistorySummariesMatchRecursion(
                             ll1HistorySummary,
                             LL2NVRAMLogicalHistorySummary',
                             LL2NVRAM.hashBarrier')!(
                             prevInput,
                             previousLL1HistorySummary,
                             previousLL2HistorySummary)
                (***********************************************************)
                (* We prove some types, to satisfy the universal           *)
                (* quantifiers in HistorySummariesMatchRecursion           *)
                (* predicate.                                              *)
                (***********************************************************)
                <8>1. ll1HistorySummary \in HashType
                  BY <7>3
                <8>2. LL2NVRAMLogicalHistorySummary' \in HistorySummaryType
                  BY <2>1, LL2NVRAMLogicalHistorySummaryTypeSafe
                <8>3. LL2NVRAM.hashBarrier' \in HashType
                  BY <2>1, LL2SubtypeImplicationLemma DEF LL2SubtypeImplication, LL2TypeInvariant, LL2TrustedStorageType
                <8>4. HistorySummariesMatchRecursion(
                          ll1HistorySummary, LL2NVRAMLogicalHistorySummary', LL2NVRAM.hashBarrier')
                  BY <7>6, <7>7
                <8>5. QED
                  BY <8>1, <8>2, <8>4 DEF HistorySummariesMatchRecursion
              (*************************************************************)
              (* One of the conjuncts in the definition of                 *)
              (* HistorySummariesMatchRecursion is that the `^\fastname^'  *)
              (* history summary is a successor of a previous history      *)
              (* summary.                                                  *)
              (*************************************************************)
              <7>9. LL2HistorySummaryIsSuccessor(
                        LL2NVRAMLogicalHistorySummary',
                        previousLL2HistorySummary,
                        prevInput,
                        LL2NVRAM.hashBarrier')
                BY <7>8 DEF HistorySummariesMatchRecursion
              (*************************************************************)
              (* We re-state the definitions from the LET in               *)
              (* LL2HistorySummaryIsSuccessor.                             *)
              (*************************************************************)
              <7> successorHistorySummary ==
                      Successor(previousLL2HistorySummary, prevInput, LL2NVRAM.hashBarrier')
              <7> checkpointedSuccessorHistorySummary == Checkpoint(successorHistorySummary)
              (*************************************************************)
              (* We hide the definitions.                                  *)
              (*************************************************************)
              <7> HIDE DEF successorHistorySummary, checkpointedSuccessorHistorySummary
              (*************************************************************)
              (* The definition of LL2HistorySummaryIsSuccessor tells us   *)
              (* that there are two ways that the logical history summary  *)
              (* could be a successor.  We will prove that neither of      *)
              (* these disjuncts is satisfiable.                           *)
              (*************************************************************)
              <7>10. \/  LL2NVRAMLogicalHistorySummary' = successorHistorySummary
                     \/  LL2NVRAMLogicalHistorySummary' = checkpointedSuccessorHistorySummary
                BY <7>9
                DEF LL2HistorySummaryIsSuccessor, successorHistorySummary,
                    checkpointedSuccessorHistorySummary
              (*************************************************************)
              (* First, we prove that the logical history summary cannot   *)
              (* be a successor.                                           *)
              (*************************************************************)
              <7>11. LL2NVRAMLogicalHistorySummary' # successorHistorySummary
                (***********************************************************)
                (* We re-state a definition from the LET in the Successor  *)
                (* operator.                                               *)
                (***********************************************************)
                <8> securedInput == Hash(LL2NVRAM.hashBarrier', prevInput)
                (***********************************************************)
                (* We hide the definition.                                 *)
                (***********************************************************)
                <8> HIDE DEF securedInput
                (***********************************************************)
                (* There is only one sub-step, which is proving that the   *)
                (* extension fields of these two records are unequal.      *)
                (* We'll separately prove the two cases of whether or not  *)
                (* the primed SPCR equals the base hash value.             *)
                (***********************************************************)
                <8>1. LL2NVRAMLogicalHistorySummary.extension' # successorHistorySummary.extension
                  (*********************************************************)
                  (* One fact that will be useful for both cases is that   *)
                  (* an extension is in progress in the primed state.      *)
                  (*********************************************************)
                  <9>1. LL2NVRAM.extensionInProgress'
                    <10>1. LL2NVRAM.extensionInProgress
                      BY <3>3
                    <10>2. UNCHANGED LL2NVRAM.extensionInProgress
                      BY <3>2
                    <10>3. QED
                      BY <10>1, <10>2
                  (*********************************************************)
                  (* The first case is when the SPCR equals the base hash  *)
                  (* value.                                                *)
                  (*********************************************************)
                  <9>2. CASE LL2SPCR' = BaseHashValue
                    (*******************************************************)
                    (* If an extension is in progress but the SPCR equals  *)
                    (* the base hash value, the logical history summary    *)
                    (* equals a crazy hash value, because this sitution    *)
                    (* should never arise during normal operation.         *)
                    (*******************************************************)
                    <10>1. LL2NVRAMLogicalHistorySummary.extension' = CrazyHashValue
                      <11>1. LL2NVRAMLogicalHistorySummary' = [
                                anchor |-> LL2NVRAM.historySummaryAnchor',
                                extension |-> CrazyHashValue ]
                        BY <9>1, <9>2 DEF LL2NVRAMLogicalHistorySummary
                      <11>2. QED
                        BY <11>1
                    (*******************************************************)
                    (* The extension field of the successor history        *)
                    (* summary is equal to a hash generated by the hash    *)
                    (* function.                                           *)
                    (*******************************************************)
                    <10>2. successorHistorySummary.extension =
                              Hash(previousLL2HistorySummary.extension, securedInput)
                      BY DEF successorHistorySummary, Successor, securedInput
                    (*******************************************************)
                    (* The arguments to the hash function are both in the  *)
                    (* hash domain.                                        *)
                    (*******************************************************)
                    <10>3. previousLL2HistorySummary.extension \in HashDomain
                      <11>1. previousLL2HistorySummary.extension \in HashType
                        <12>1. previousLL2HistorySummary \in HistorySummaryType
                          BY <7>8
                        <12>2. QED
                          BY <12>1 DEF HistorySummaryType
                      <11>2. QED
                        BY <11>1 DEF HashDomain
                    <10>4. securedInput \in HashDomain
                      <11>1. securedInput \in HashType
                        <12>1. LL2NVRAM.hashBarrier' \in HashDomain
                          <13>1. LL2NVRAM.hashBarrier' \in HashType
                            BY <2>1, LL2SubtypeImplicationLemma DEF LL2SubtypeImplication, LL2TypeInvariant, LL2TrustedStorageType
                          <13>2. QED
                            BY <13>1 DEF HashDomain
                        <12>2. prevInput \in HashDomain
                          <13>1. prevInput \in InputType
                            BY <7>8
                          <13>2. QED
                            BY <13>1 DEF HashDomain
                        <12>3. QED
                          BY <12>1, <12>2, HashTypeSafe DEF securedInput
                      <11>2. QED
                        BY <11>1 DEF HashDomain
                    (*******************************************************)
                    (* The crazy hash value is not equal to any hash value *)
                    (* that can be generated by the hash function when     *)
                    (* operating on arguments within its domain.           *)
                    (*******************************************************)
                    <10>5. QED
                      BY <10>1, <10>2, <10>3, <10>4, CrazyHashValueUnique
                  (*********************************************************)
                  (* We will prove the second case in two steps.           *)
                  (*********************************************************)
                  <9>3. CASE LL2SPCR' # BaseHashValue
                  (*********************************************************)
                  (* First, we prove that the extension field of the       *)
                  (* logical history summary is a hash whose second        *)
                  (* argument is the fakeHash from the LL2CorruptSPCR      *)
                  (* action.  This follows from the definitions of         *)
                  (* LL2CorruptSPCR and LL2NVRAMLogicalHistorySummary.     *)
                  (*********************************************************)
                    <10>1. LL2NVRAMLogicalHistorySummary.extension' = Hash(LL2SPCR, fakeHash)
                      <11>1. LL2NVRAMLogicalHistorySummary.extension' = LL2SPCR'
                        <12>1. LL2NVRAMLogicalHistorySummary' = [
                                   anchor |-> LL2NVRAM.historySummaryAnchor',
                                   extension |-> LL2SPCR' ]
                          BY <9>1, <9>3 DEF LL2NVRAMLogicalHistorySummary
                        <12>2. QED
                          BY <12>1
                      <11>2. LL2SPCR' = Hash(LL2SPCR, fakeHash)
                        BY <3>2
                      <11>3. QED
                        BY <11>1, <11>2
                    (*******************************************************)
                    (* Second, we prove that the extension field of the    *)
                    (* successor history summary from                      *)
                    (* LL2HistorySummaryIsSuccessor is a hash whose second *)
                    (* argument is the secured input from the Successor    *)
                    (* operator.  This follows directly from the           *)
                    (* definition of successorHistorySummary.              *)
                    (*******************************************************)
                    <10>2. successorHistorySummary.extension =
                               Hash(previousLL2HistorySummary.extension, securedInput)
                      BY DEF successorHistorySummary, Successor, securedInput
                    (*******************************************************)
                    (* Third, we prove that the two hashes are unequal.    *)
                    (* We will use the HashCollisionResistant property,    *)
                    (* along with the fact (which we will prove in a       *)
                    (* sub-step) that the second arguments to the two hash *)
                    (* functions are unequal.                              *)
                    (*******************************************************)
                    <10>3. Hash(LL2SPCR, fakeHash) #
                               Hash(previousLL2HistorySummary.extension, securedInput)
                      (*****************************************************)
                      (* To employ the HashCollisionResistant property, we *)
                      (* need to prove some types.                         *)
                      (*****************************************************)
                      <11>1. LL2SPCR \in HashDomain
                        <12>1. LL2SPCR \in HashType
                          BY <2>1 DEF LL2TypeInvariant
                        <12>2. QED
                          BY <12>1 DEF HashDomain
                      <11>2. fakeHash \in HashDomain
                        BY <3>2
                      <11>3. previousLL2HistorySummary.extension \in HashDomain
                        <12>1. previousLL2HistorySummary.extension \in HashType
                          <13>1. previousLL2HistorySummary \in HistorySummaryType
                            BY <7>8
                          <13>2. QED
                            BY <13>1 DEF HistorySummaryType
                        <12>2. QED
                          BY <12>1 DEF HashDomain
                      <11>4. securedInput \in HashDomain
                        <12>1. securedInput \in HashType
                          <13>1. LL2NVRAM.hashBarrier' \in HashDomain
                            <14>1. LL2NVRAM.hashBarrier' \in HashType
                              BY <2>1, LL2SubtypeImplicationLemma DEF LL2SubtypeImplication, LL2TypeInvariant, LL2TrustedStorageType
                            <14>2. QED
                              BY <14>1 DEF HashDomain
                          <13>2. prevInput \in HashDomain
                            <14>1. prevInput \in InputType
                              BY <7>8
                            <14>2. QED
                              BY <14>1 DEF HashDomain
                          <13>3. QED
                            BY <13>1, <13>2, HashTypeSafe DEF securedInput
                        <12>2. QED
                          BY <12>1 DEF HashDomain
                      (*****************************************************)
                      (* Then we need to prove that the second arguments   *)
                      (* to the hash functions are unequal.  We will       *)
                      (* employ the restriction on fake hash values from   *)
                      (* the definition of LL2CorruptSPCR.                 *)
                      (*****************************************************)
                      <11>5. fakeHash # securedInput
                        <12>1. \A fakeInput \in InputType :
                                   fakeHash # Hash(LL2NVRAM.hashBarrier', fakeInput)
                          <13>1. \A fakeInput \in InputType :
                                     fakeHash # Hash(LL2NVRAM.hashBarrier, fakeInput)
                            BY <3>2
                          <13>2. UNCHANGED LL2NVRAM.hashBarrier
                            BY <3>2
                          <13>3. QED
                            BY <13>1, <13>2
                        <12>2. prevInput \in InputType
                          BY <7>8
                        <12>3. QED
                          BY <12>1, <12>2 DEF securedInput
                      <11>6. QED
                        (***************************************************)
                        (* Ideally, this QED step should just read:        *)
                        (*                                                 *)
                        (*  BY <<11>>1, <<11>>2, <<11>>3, <<11>>4, <<11>>5, HashCollisionResistant *)
                        (*                                                 *)
                        (* However, the prover seems to get a little       *)
                        (* confused in this instance.  We make life easier *)
                        (* for the prover by definining some local         *)
                        (* variables and hiding their definitions before   *)
                        (* appealing to the HashCollisionResistant         *)
                        (* assumption.                                     *)
                        (***************************************************)
                        <12> h1a == LL2SPCR
                        <12> h2a == fakeHash
                        <12> h1b == previousLL2HistorySummary.extension
                        <12> h2b == securedInput
                        <12>1. h1a \in HashDomain
                          BY <11>1
                        <12>2. h2a \in HashDomain
                          BY <11>2
                        <12>3. h1b \in HashDomain
                          BY <11>3
                        <12>4. h2b \in HashDomain
                          BY <11>4
                        <12>5. h2a # h2b
                          BY <11>5
                        <12>6. Hash(h1a, h2a) # Hash(h1b, h2b)
                          <13> HIDE DEF h1a, h2a, h1b, h2b
                          <13>1. QED
                            BY <12>1, <12>2, <12>3, <12>4, <12>5, HashCollisionResistant
                        <12>7. QED
                          BY <12>6
                    <10>4. QED
                    BY <10>1, <10>2, <10>3
                  (*********************************************************)
                  (* The two cases are exhaustive.                         *)
                  (*********************************************************)
                  <9>4. QED
                    BY <9>2, <9>3
                <8>2. QED
                  BY <8>1
              (*************************************************************)
              (* Second, we prove that the logical history summary cannot  *)
              (* be a checkpoint.                                          *)
              (*************************************************************)
              <7>12. LL2NVRAMLogicalHistorySummary' # checkpointedSuccessorHistorySummary
                (***********************************************************)
                (* The extension field of the logical history summary does *)
                (* not equal the base hash value, as we proved above.      *)
                (***********************************************************)
                <8>1. LL2NVRAMLogicalHistorySummary.extension' # BaseHashValue
                  BY <4>1
                (***********************************************************)
                (* The extension field of the checkpointed successor does  *)
                (* equal the base hash value.  This follows from the       *)
                (* CheckpointHasBaseExtensionLemma.                        *)
                (***********************************************************)
                <8>2. checkpointedSuccessorHistorySummary.extension = BaseHashValue
                  <9>1. successorHistorySummary \in HistorySummaryType
                    <10>1. previousLL2HistorySummary \in HistorySummaryType
                      BY <7>8
                    <10>2. prevInput \in InputType
                      BY <7>8
                    <10>3. LL2NVRAM.hashBarrier' \in HashType
                      BY <2>1, LL2SubtypeImplicationLemma DEF LL2SubtypeImplication, LL2TypeInvariant, LL2TrustedStorageType
                    <10>4. QED
                      BY <10>1, <10>2, <10>3, SuccessorTypeSafe DEF successorHistorySummary
                  <9>2. QED
                    BY <9>1, CheckpointHasBaseExtensionLemma
                    DEF checkpointedSuccessorHistorySummary
                <8>3. QED
                  BY <8>1, <8>2
              (*************************************************************)
              (* We thus have a contradiction.                             *)
              (*************************************************************)
              <7>13. QED
                BY <7>10, <7>11, <7>12

            <6>3. QED
              BY <6>2 DEF ll1GarbageHistoryStateBinding
          (*****************************************************************)
          (* We prove that the constraint labeled `previous' in the        *)
          (* LL1RestrictedCorruption action is satisfied.                  *)
          (*****************************************************************)
          <5>3. LL1RestrictedCorruption!nvram!previous(LL1NVRAM.historySummary')
            (***************************************************************)
            (* To prove the universally quantified expression, we take a   *)
            (* set of variables of the appropriate types.                  *)
            (***************************************************************)
            <6>1. TAKE stateHash1 \in HashType,
                       ll1Authenticator \in LL1ObservedAuthenticators,
                       ll1SomeHistorySummary \in HashType,
                       someInput \in InputType
            (***************************************************************)
            (* We re-state the definition from within the                  *)
            (* LL1RestrictedCorruption!nvram!previous clause.              *)
            (***************************************************************)
            <6> ll1SomeHistoryStateBinding == Hash(ll1SomeHistorySummary, stateHash1)
            (***************************************************************)
            (* We hide the definition.                                     *)
            (***************************************************************)
            <6> HIDE DEF ll1SomeHistoryStateBinding
            (***************************************************************)
            (* We need to prove the nvram::previous conjunct, which        *)
            (* asserts an implication.  It suffices to assume the          *)
            (* antecedent and prove the consequent.                        *)
            (***************************************************************)
            <6>2. SUFFICES
                  ASSUME LL1NVRAM.historySummary' = Hash(ll1SomeHistorySummary, someInput)
                  PROVE ~ValidateMAC(
                      LL1NVRAM.symmetricKey,
                      ll1SomeHistoryStateBinding,
                      ll1Authenticator)
              BY DEF ll1SomeHistoryStateBinding
            (***************************************************************)
            (* The consequent of the nvram::previous conjunct asserts that *)
            (* the authenticator is not a valid MAC for the history state  *)
            (* binding formed from any predecessor of the history summary  *)
            (* in the NVRAM and any state hash.                            *)
            (***************************************************************)
            <6>3. ~ValidateMAC(
                      LL1NVRAM.symmetricKey,
                      ll1SomeHistoryStateBinding,
                      ll1Authenticator)
              (*************************************************************)
              (* We will use proof by contradiction.                       *)
              (*************************************************************)
              <7>1. SUFFICES
                    ASSUME
                        ValidateMAC(
                            LL1NVRAM.symmetricKey,
                            ll1SomeHistoryStateBinding,
                            ll1Authenticator)
                    PROVE
                        FALSE
                OBVIOUS
              (*************************************************************)
              (* We first pick, from the set of `^\fastname^' observed     *)
              (* authenticators, a `^\fastname^' authenticator that        *)
              (* matches the `^\basicname^' authenticator.  We know that   *)
              (* such a authenticator exists, because the refinement       *)
              (* asserts that the sets of observed authenticators match    *)
              (* across the two specs.                                     *)
              (*************************************************************)
              <7>2. PICK ll2Authenticator \in LL2ObservedAuthenticators :
                        AuthenticatorsMatch(
                            ll1Authenticator,
                            ll2Authenticator,
                            LL2NVRAM.symmetricKey,
                            LL2NVRAM.hashBarrier)
                <8>1. ll1Authenticator \in LL1ObservedAuthenticators
                  BY <6>1
                <8>2. AuthenticatorSetsMatch(
                          LL1ObservedAuthenticators,
                          LL2ObservedAuthenticators,
                          LL2NVRAM.symmetricKey,
                          LL2NVRAM.hashBarrier)
                  BY <2>1 DEF LL2Refinement
                <8>3. QED
                  BY <8>1, <8>2 DEF AuthenticatorSetsMatch
              (*************************************************************)
              (* We pick a set of variables of the appropriate types that  *)
              (* satisfy the quantified AuthenticatorsMatch predicate.     *)
              (*************************************************************)
              <7>3. PICK stateHash2 \in HashType,
                        ll1HistorySummary \in HashType,
                        ll2HistorySummary \in HistorySummaryType :
                        AuthenticatorsMatch(
                            ll1Authenticator,
                            ll2Authenticator,
                            LL2NVRAM.symmetricKey,
                            LL2NVRAM.hashBarrier)!(
                            stateHash2, ll1HistorySummary, ll2HistorySummary)!1
                BY <7>2 DEF AuthenticatorsMatch
              (*************************************************************)
              (* We re-state the definitions from the LET in               *)
              (* AuthenticatorsMatch.                                      *)
              (*************************************************************)
              <7> ll1HistoryStateBinding == Hash(ll1HistorySummary, stateHash2)
              <7> ll2HistorySummaryHash == Hash(ll2HistorySummary.anchor, ll2HistorySummary.extension)
              <7> ll2HistoryStateBinding == Hash(ll2HistorySummaryHash, stateHash2)
              (*************************************************************)
              (* We prove the types of the definitions, with help from the *)
              (* AuthenticatorsMatchDefsTypeSafeLemma.                     *)
              (*************************************************************)
              <7>4. /\  ll1HistoryStateBinding \in HashType
                    /\  ll2HistorySummaryHash \in HashType
                    /\  ll2HistoryStateBinding \in HashType
                BY <7>3, AuthenticatorsMatchDefsTypeSafeLemma
              (*************************************************************)
              (* We hide the definitions.                                  *)
              (*************************************************************)
              <7> HIDE DEF ll1HistoryStateBinding, ll2HistorySummaryHash, ll2HistoryStateBinding
              (*************************************************************)
              (* We prove that the `^\basicname^'s history summary picked  *)
              (* to satisfy the AuthenticatorsMatch predicate equals the   *)
              (* history summary in the primed state of the `^\basicname^' *)
              (* NVRAM.                                                    *)
              (*************************************************************)
              <7>5. ll1HistorySummary = ll1SomeHistorySummary
                (***********************************************************)
                (* The first step is to show the equality of the historyp  *)
                (* state bindings that bind each of these history          *)
                (* summaries to their respective state hashes.             *)
                (***********************************************************)
                <8>1. ll1HistoryStateBinding = ll1SomeHistoryStateBinding
                  (*********************************************************)
                  (* By hypothesis, the authenticator is a valid MAC for   *)
                  (* the garbage history state binding.                    *)
                  (*********************************************************)
                  <9>1. ValidateMAC(LL2NVRAM.symmetricKey, ll1SomeHistoryStateBinding, ll1Authenticator)
                    <10>1. LL1NVRAM.symmetricKey = LL2NVRAM.symmetricKey
                      BY <2>1 DEF LL2Refinement
                    <10>2. QED
                      BY <7>1, <10>1
                  (*********************************************************)
                  (* The definition of the AuthenticatorsMatch predicate   *)
                  (* tells us that the `^\basicname^' authenticator was    *)
                  (* generated as a MAC from the history state binding.    *)
                  (*********************************************************)
                  <9>2. ll1Authenticator = GenerateMAC(LL2NVRAM.symmetricKey, ll1HistoryStateBinding)
                    BY <7>3 DEF ll1HistoryStateBinding
                  (*********************************************************)
                  (* The remaining preconditions are types.                *)
                  (*********************************************************)
                  <9>3. LL2NVRAM.symmetricKey \in SymmetricKeyType
                    BY <2>1, LL2SubtypeImplicationLemma DEF LL2SubtypeImplication
                  <9>4. ll1HistoryStateBinding \in HashType
                    BY <7>4
                  <9>5. ll1SomeHistoryStateBinding \in HashType
                    <10>1. ll1SomeHistorySummary \in HashDomain
                      <11>1. ll1SomeHistorySummary \in HashType
                        BY <6>1
                      <11>2. QED
                        BY <11>1 DEF HashDomain
                    <10>2. stateHash1 \in HashDomain
                      BY <5>1 DEF HashDomain
                    <10>3. QED
                      BY <10>1, <10>2, HashTypeSafe DEF ll1SomeHistoryStateBinding
                  (*********************************************************)
                  (* The MACCollisionResistant property tells us that the  *)
                  (* two history state bindings are equal.                 *)
                  (*********************************************************)
                  <9>6. QED
                    BY <9>1, <9>2, <9>3, <9>4, <9>5, MACCollisionResistant
                (***********************************************************)
                (* By the collision resistance of the hash function, the   *)
                (* equality of the history state bindings implies the      *)
                (* equality of the history summaries.                      *)
                (***********************************************************)
                <8>2. QED
                  <9>1. ll1SomeHistorySummary \in HashDomain
                    <10>1. ll1SomeHistorySummary \in HashType
                      BY <6>1
                    <10>2. QED
                      BY <10>1 DEF HashDomain
                  <9>2. ll1HistorySummary \in HashDomain
                    <10>1. ll1HistorySummary \in HashType
                      BY <7>3
                    <10>2. QED
                      BY <10>1 DEF HashDomain
                  <9>3. stateHash1 \in HashDomain
                    <10>1. stateHash1 \in HashType
                      BY <6>1
                    <10>2. QED
                      BY <10>1 DEF HashDomain
                  <9>4. stateHash2 \in HashDomain
                    <10>1. stateHash2 \in HashType
                      BY <7>3
                   <10>2. QED
                      BY <10>1 DEF HashDomain
                  <9>5. QED
                    BY <8>1, <9>2, <9>1, <9>4, <9>3, HashCollisionResistant
                    DEF ll1HistoryStateBinding, ll1SomeHistoryStateBinding
              (*************************************************************)
              (* We pick a value for the `^\fastname^'                     *)
              (* previous-inputs-summary existential variable inside the   *)
              (* HistorySummariesMatchRecursion predicate that satisfies   *)
              (* this predicate for (1) the `^\basicname^'s history        *)
              (* summary picked to satisfy the AuthenticatorsMatch         *)
              (* predicate and (2) the input taken from the universal      *)
              (* quantifier in the `previous' conjunct in the              *)
              (* LL1RestrictedCorruption action.                           *)
              (*************************************************************)
              <7>6. PICK previousLL2HistorySummary \in HistorySummaryType :
                         HistorySummariesMatchRecursion(
                             LL1NVRAM.historySummary',
                             LL2NVRAMLogicalHistorySummary',
                             LL2NVRAM.hashBarrier')!(
                             someInput,
                             ll1SomeHistorySummary,
                             previousLL2HistorySummary)
                (***********************************************************)
                (* The `^\basicname^'s history summary picked to satisfy   *)
                (* the AuthenticatorsMatch predicate matches the primed    *)
                (* logical history summary in the `^\fastname^' NVRAM and  *)
                (* SPCR, by the refinement and the above equality.         *)
                (***********************************************************)
                <8>1. HistorySummariesMatch(
                          LL1NVRAM.historySummary',
                          LL2NVRAMLogicalHistorySummary',
                          LL2NVRAM.hashBarrier')
                  <9>1. HistorySummariesMatch(
                            LL1NVRAM.historySummary',
                            LL2NVRAMLogicalHistorySummary',
                            LL2NVRAM.hashBarrier')
                    BY <2>1 DEF LL2Refinement
                  <9>2. QED
                    BY <9>1
                (***********************************************************)
                (* We prove that the HistorySummariesMatch predicate       *)
                (* equals the HistorySummariesMatchRecursion predicate in  *)
                (* this case.  We assert each condition required by the    *)
                (* definition of the predicate.                            *)
                (***********************************************************)
                <8>2. HistorySummariesMatch(
                          LL1NVRAM.historySummary',
                          LL2NVRAMLogicalHistorySummary',
                          LL2NVRAM.hashBarrier') <=>
                      HistorySummariesMatchRecursion(
                          LL1NVRAM.historySummary',
                          LL2NVRAMLogicalHistorySummary',
                          LL2NVRAM.hashBarrier')
                  (*********************************************************)
                  (* We prove some types, to satisfy the universal         *)
                  (* quantifiers in HistorySummariesMatchDefinition.       *)
                  (*********************************************************)
                  <9>1. LL1NVRAM.historySummary' \in HashType
                    BY <2>1 DEF LL2Refinement, LL1TrustedStorageType
                  <9>2. LL2NVRAMLogicalHistorySummary' \in HistorySummaryType
                    BY <2>1, LL2NVRAMLogicalHistorySummaryTypeSafe
                  <9>3. LL2NVRAM.hashBarrier' \in HashType
                    BY <2>1, LL2SubtypeImplicationLemma DEF LL2SubtypeImplication
                  (*********************************************************)
                  (* We prove that the `^\fastname^' logical history       *)
                  (* summary does not equal the initial history summary.   *)
                  (*********************************************************)
                  <9> ll2InitialHistorySummary == [ anchor |-> BaseHashValue, extension |-> BaseHashValue ]
                  <9>4. LL2NVRAMLogicalHistorySummary' # ll2InitialHistorySummary
                    <10>1. LL2NVRAMLogicalHistorySummary.extension' # BaseHashValue
                      BY <4>1
                    <10>2. ll2InitialHistorySummary.extension = BaseHashValue
                      BY DEF ll2InitialHistorySummary
                    <10>3. QED
                      BY <10>1, <10>2
                  (*********************************************************)
                  (* Finally, from HistorySummariesMatchDefinition, we can *)
                  (* conclude that the HistorySummariesMatch predicate     *)
                  (* equals the quantified HistorySummariesMatchRecursion  *)
                  (* predicate.                                            *)
                  (*********************************************************)
                  <9>5. QED
                    BY <9>1, <9>2, <9>3, <9>4, HistorySummariesMatchDefinition
                    DEF ll2InitialHistorySummary
                (***********************************************************)
                (* We pick values for the remaining two existential        *)
                (* variables inside the HistorySummariesMatchRecursion     *)
                (* predicate that satisfy the predicate.  We know such     *)
                (* variables exist, because the predicate is satisfied by  *)
                (* the two previous steps.                                 *)
                (***********************************************************)
                <8>3. PICK prevInput \in InputType,
                           previousLL1HistorySummary \in HashType,
                           previousLL2HistorySummary \in HistorySummaryType :
                           HistorySummariesMatchRecursion(
                               LL1NVRAM.historySummary',
                               LL2NVRAMLogicalHistorySummary',
                               LL2NVRAM.hashBarrier')!(
                               prevInput,
                               previousLL1HistorySummary,
                               previousLL2HistorySummary)
                  (*********************************************************)
                  (* We prove some types, to satisfy the universal         *)
                  (* quantifiers in HistorySummariesMatchRecursion         *)
                  (* predicate.                                            *)
                  (*********************************************************)
                  <9>1. ll1HistorySummary \in HashType
                    BY <7>3
                  <9>2. LL2NVRAMLogicalHistorySummary' \in HistorySummaryType
                    BY <2>1, LL2NVRAMLogicalHistorySummaryTypeSafe
                  <9>3. LL2NVRAM.hashBarrier' \in HashType
                    BY <2>1, LL2SubtypeImplicationLemma DEF LL2SubtypeImplication
                  <9>4. HistorySummariesMatchRecursion(
                            LL1NVRAM.historySummary',
                            LL2NVRAMLogicalHistorySummary',
                            LL2NVRAM.hashBarrier')
                    BY <8>1, <8>2
                  <9>5. QED
                    BY <9>1, <9>2, <9>4 DEF HistorySummariesMatchRecursion
                (***********************************************************)
                (* We prove that the existential variables for the         *)
                (* previous input and the previous history summary in the  *)
                (* above pick are equal to the input and history summary   *)
                (* taken from the universal quantifiers in the `previous'  *)
                (* conjunct in the LL1RestrictedCorruption action.  We use *)
                (* the HashCollisionResistant property.                    *)
                (***********************************************************)
                <8>4. /\  ll1SomeHistorySummary = previousLL1HistorySummary
                      /\  someInput = prevInput
                  (*********************************************************)
                  (* We prove the necessary types for the                  *)
                  (* HashCollisionResistant property.                      *)
                  (*********************************************************)
                  <9>1. ll1SomeHistorySummary \in HashDomain
                    <10>1. ll1SomeHistorySummary \in HashType
                      BY <6>1
                    <10>2. QED
                      BY <10>1 DEF HashDomain
                  <9>2. someInput \in HashDomain
                    <10>1. someInput \in InputType
                      BY <6>1
                    <10>2. QED
                      BY <10>1 DEF HashDomain
                  <9>3. previousLL1HistorySummary \in HashDomain
                    <10>1. previousLL1HistorySummary \in HashType
                      BY <8>3
                    <10>2. QED
                      BY <10>1 DEF HashDomain
                  <9>4. prevInput \in HashDomain
                    <10>1. prevInput \in InputType
                      BY <8>3
                    <10>2. QED
                      BY <10>1 DEF HashDomain
                  (*********************************************************)
                  (* The hashes are equal, because each is equal to the    *)
                  (* history summary in the primed `^\basicname^' NVRAM.   *)
                  (*********************************************************)
                  <9>5. Hash(ll1SomeHistorySummary, someInput) =
                            Hash(previousLL1HistorySummary, prevInput)
                    (*******************************************************)
                    (* The hash of the taken history summary and input are *)
                    (* equal to the history summary in the primed          *)
                    (* `^\basicname^' NVRAM by assumption of the           *)
                    (* antecedent in the `previous' conjunct in the        *)
                    (* LL1RestrictedCorruption action.                     *)
                    (*******************************************************)
                    <10>1. LL1NVRAM.historySummary' = Hash(ll1SomeHistorySummary, someInput)
                      BY <6>2
                    (*******************************************************)
                    (* The hash of the picked history summary and input    *)
                    (* are equal to the history summary in the primed      *)
                    (* `^\basicname^' NVRAM by the definition of the       *)
                    (* HistorySummariesMatchRecursion predicate.           *)
                    (*******************************************************)
                    <10>2. LL1NVRAM.historySummary' = Hash(previousLL1HistorySummary, prevInput)
                      BY <8>3
                    <10>3. QED
                      BY <10>1, <10>2
                  <9>6. QED
                    BY <9>1, <9>2, <9>3, <9>4, <9>5, HashCollisionResistant
                <8>5. QED
                  BY <8>3, <8>4
              (*************************************************************)
              (* One of the conjuncts in the definition of                 *)
              (* HistorySummariesMatchRecursion is that the `^\fastname^'  *)
              (* history summary is a successor of a previous history      *)
              (* summary.                                                  *)
              (*************************************************************)
              <7>7. LL2HistorySummaryIsSuccessor(
                        LL2NVRAMLogicalHistorySummary',
                        previousLL2HistorySummary,
                        someInput,
                        LL2NVRAM.hashBarrier')
                BY <7>6 DEF HistorySummariesMatchRecursion
              (*************************************************************)
              (* We re-state the definitions from the LET in               *)
              (* LL2HistorySummaryIsSuccessor.                             *)
              (*************************************************************)
              <7> successorHistorySummary ==
                      Successor(previousLL2HistorySummary, someInput, LL2NVRAM.hashBarrier')
              <7> checkpointedSuccessorHistorySummary == Checkpoint(successorHistorySummary)
              (*************************************************************)
              (* We hide the definitions.                                  *)
              (*************************************************************)
              <7> HIDE DEF successorHistorySummary, checkpointedSuccessorHistorySummary
              (*************************************************************)
              (* The definition of LL2HistorySummaryIsSuccessor tells us   *)
              (* that there are two ways that the logical history summary  *)
              (* could be a successor.  We will prove that neither of      *)
              (* these disjuncts is satisfiable.                           *)
              (*************************************************************)
              <7>8. \/  LL2NVRAMLogicalHistorySummary' = successorHistorySummary
                     \/  LL2NVRAMLogicalHistorySummary' = checkpointedSuccessorHistorySummary
                BY <7>7
                DEF LL2HistorySummaryIsSuccessor, successorHistorySummary,
                    checkpointedSuccessorHistorySummary
              (*************************************************************)
              (* First, we prove that the logical history summary cannot   *)
              (* be a successor.                                           *)
              (*************************************************************)
              <7>9. LL2NVRAMLogicalHistorySummary' # successorHistorySummary
                (***********************************************************)
                (* We re-state a definition from the LET in the Successor  *)
                (* operator.                                               *)
                (***********************************************************)
                <8> securedInput == Hash(LL2NVRAM.hashBarrier', someInput)
                (***********************************************************)
                (* We hide the definition.                                 *)
                (***********************************************************)
                <8> HIDE DEF securedInput
                (***********************************************************)
                (* There is only one sub-step, which is proving that the   *)
                (* extension fields of these two records are unequal.      *)
                (***********************************************************)
                <8>1. LL2NVRAMLogicalHistorySummary.extension' # successorHistorySummary.extension
                  (*********************************************************)
                  (* First, we prove that the extension field of the       *)
                  (* logical history summary is a hash whose second        *)
                  (* argument is the fakeHash from the LL2CorruptSPCR      *)
                  (* action.  This follows from the definitions of         *)
                  (* LL2CorruptSPCR and LL2NVRAMLogicalHistorySummary.     *)
                  (*********************************************************)
                  <9>1. LL2NVRAMLogicalHistorySummary.extension' = Hash(LL2SPCR, fakeHash)
                    <10>1. LL2SPCR' = Hash(LL2SPCR, fakeHash)
                      BY <3>2
                    <10>2. LL2NVRAMLogicalHistorySummary.extension' = LL2SPCR'
                      <11>1. LL2NVRAMLogicalHistorySummary' = [
                                 anchor |-> LL2NVRAM.historySummaryAnchor',
                                 extension |-> LL2SPCR' ]
                        <12>1. LL2NVRAM.extensionInProgress'
                          <13>1. LL2NVRAM.extensionInProgress
                            BY <3>3
                          <13>2. UNCHANGED LL2NVRAM.extensionInProgress
                            BY <3>2
                          <13>3. QED
                            BY <13>1, <13>2
                        <12>2. LL2SPCR' # BaseHashValue
                          <13>1. LL2SPCR \in HashDomain
                            <14>1. LL2SPCR \in HashType
                              BY <2>1 DEF LL2TypeInvariant
                            <14>2. QED
                              BY <14>1 DEF HashDomain
                          <13>2. fakeHash \in HashDomain
                            BY <3>2
                          <13>3. QED
                            BY <10>1, <13>1, <13>2, BaseHashValueUnique
                        <12>3. QED
                          BY <12>1, <12>2 DEF LL2NVRAMLogicalHistorySummary
                      <11>2. QED
                        BY <11>1
                    <10>3. QED
                      BY <10>1, <10>2
                  (*********************************************************)
                  (* Second, we prove that the extension field of the      *)
                  (* successor history summary from                        *)
                  (* LL2HistorySummaryIsSuccessor is a hash whose second   *)
                  (* argument is the secured input from the Successor      *)
                  (* operator.  This follows directly from the definition  *)
                  (* of successorHistorySummary.                           *)
                  (*********************************************************)
                  <9>2. successorHistorySummary.extension =
                            Hash(previousLL2HistorySummary.extension, securedInput)
                    BY DEF successorHistorySummary, Successor, securedInput
                  (*********************************************************)
                  (* Third, we prove that the two hashes are unequal.  We  *)
                  (* will use the HashCollisionResistant property, along   *)
                  (* with the fact (which we will prove in a sub-step)     *)
                  (* that the second arguments to the two hash functions   *)
                  (* are unequal.                                          *)
                  (*********************************************************)
                  <9>3. Hash(LL2SPCR, fakeHash) #
                            Hash(previousLL2HistorySummary.extension, securedInput)
                    (*******************************************************)
                    (* To employ the HashCollisionResistant property, we   *)
                    (* need to prove some types.                           *)
                    (*******************************************************)
                    <10>1. LL2SPCR \in HashDomain
                      <11>1. LL2SPCR \in HashType
                        BY <2>1 DEF LL2TypeInvariant
                      <11>2. QED
                        BY <11>1 DEF HashDomain
                    <10>2. fakeHash \in HashDomain
                      BY <3>2
                    <10>3. previousLL2HistorySummary.extension \in HashDomain
                      <11>1. previousLL2HistorySummary.extension \in HashType
                        <12>1. previousLL2HistorySummary \in HistorySummaryType
                          BY <7>6
                        <12>2. QED
                          BY <12>1 DEF HistorySummaryType
                      <11>2. QED
                        BY <11>1 DEF HashDomain
                    <10>4. securedInput \in HashDomain
                      <11>1. securedInput \in HashType
                        <12>1. LL2NVRAM.hashBarrier' \in HashDomain
                          <13>1. LL2NVRAM.hashBarrier' \in HashType
                            BY <2>1, LL2SubtypeImplicationLemma DEF LL2SubtypeImplication, LL2TypeInvariant, LL2TrustedStorageType
                          <13>2. QED
                            BY <13>1 DEF HashDomain
                        <12>2. someInput \in HashDomain
                          <13>1. someInput \in InputType
                            BY <6>1
                          <13>2. QED
                            BY <13>1 DEF HashDomain
                        <12>3. QED
                          BY <12>1, <12>2, HashTypeSafe DEF securedInput
                      <11>2. QED
                        BY <11>1 DEF HashDomain
                    (*******************************************************)
                    (* Then we need to prove that the second arguments to  *)
                    (* the hash functions are unequal.  We will employ the *)
                    (* restriction on fake hash values from the definition *)
                    (* of LL2CorruptSPCR.                                  *)
                    (*******************************************************)
                    <10>5. fakeHash # securedInput
                      <11>1. \A fakeInput \in InputType :
                                 fakeHash # Hash(LL2NVRAM.hashBarrier', fakeInput)
                        <12>1. \A fakeInput \in InputType :
                                   fakeHash # Hash(LL2NVRAM.hashBarrier, fakeInput)
                          BY <3>2
                        <12>2. UNCHANGED LL2NVRAM.hashBarrier
                          BY <3>2
                        <12>3. QED
                          BY <12>1, <12>2
                      <11>2. someInput \in InputType
                        BY <6>1
                      <11>3. QED
                        BY <11>1, <11>2 DEF securedInput
                    <10>6. QED
                      (*****************************************************)
                      (* Ideally, this QED step should just read:          *)
                      (*                                                   *)
                      (*  BY <<10>>1, <<10>>2, <<10>>3, <<10>>4, <<10>>5, HashCollisionResistant *)
                      (*                                                   *)
                      (* However, the prover seems to get a little         *)
                      (* confused in this instance.  We make life easier   *)
                      (* for the prover by definining some local variables *)
                      (* and hiding their definitions before appealing to  *)
                      (* the HashCollisionResistant assumption.            *)
                      (*****************************************************)
                      <11> h1a == LL2SPCR
                      <11> h2a == fakeHash
                      <11> h1b == previousLL2HistorySummary.extension
                      <11> h2b == securedInput
                      <11>1. h1a \in HashDomain
                        BY <10>1
                      <11>2. h2a \in HashDomain
                        BY <10>2
                      <11>3. h1b \in HashDomain
                        BY <10>3
                      <11>4. h2b \in HashDomain
                        BY <10>4
                      <11>5. h2a # h2b
                        BY <10>5
                      <11>6. Hash(h1a, h2a) # Hash(h1b, h2b)
                        <12> HIDE DEF h1a, h2a, h1b, h2b
                        <12>1. QED
                          BY <11>1, <11>2, <11>3, <11>4, <11>5, HashCollisionResistant
                      <11>7. QED
                        BY <11>6
                  <9>4. QED
                    BY <9>1, <9>2, <9>3
                <8>2. QED
                  BY <8>1
              (*************************************************************)
              (* Second, we prove that the logical history summary cannot  *)
              (* be a checkpoint.                                          *)
              (*************************************************************)
              <7>10. LL2NVRAMLogicalHistorySummary' # checkpointedSuccessorHistorySummary
                (***********************************************************)
                (* The extension field of the logical history summary does *)
                (* not equal the base hash value, as we proved above.      *)
                (***********************************************************)
                <8>1. LL2NVRAMLogicalHistorySummary.extension' # BaseHashValue
                  BY <4>1
                (***********************************************************)
                (* The extension field of the checkpointed successor does  *)
                (* equal the base hash value.  This follows from the       *)
                (* CheckpointHasBaseExtensionLemma.                        *)
                (***********************************************************)
                <8>2. checkpointedSuccessorHistorySummary.extension = BaseHashValue
                  <9>1. successorHistorySummary \in HistorySummaryType
                    <10>1. previousLL2HistorySummary \in HistorySummaryType
                      BY <7>6
                    <10>2. someInput \in InputType
                      BY <6>1
                    <10>3. LL2NVRAM.hashBarrier' \in HashType
                      BY <2>1, LL2SubtypeImplicationLemma DEF LL2SubtypeImplication, LL2TypeInvariant, LL2TrustedStorageType
                    <10>4. QED
                      BY <10>1, <10>2, <10>3, SuccessorTypeSafe DEF successorHistorySummary
                  <9>2. QED
                    BY <9>1, CheckpointHasBaseExtensionLemma
                    DEF checkpointedSuccessorHistorySummary
                <8>3. QED
                  BY <8>1, <8>2
              (*************************************************************)
              (* We thus have a contradiction.                             *)
              (*************************************************************)
              <7>11. QED
                BY <7>8, <7>9, <7>10
            <6>4. QED
              BY <6>3 DEF ll1SomeHistoryStateBinding
          (*****************************************************************)
          (* We prove the third conjunct within the `nvram' conjunct of    *)
          (* the LL1RestrictedCorruption action.                           *)
          (*****************************************************************)
          <5>4. LL1NVRAM' = [
                    historySummary |-> LL1NVRAM.historySummary',
                    symmetricKey |-> LL1NVRAM.symmetricKey ]
            <6>1. LL1NVRAM \in LL1TrustedStorageType
              BY <2>1 DEF LL2Refinement
            <6>2. LL1NVRAM' \in LL1TrustedStorageType
              BY <2>1 DEF LL2Refinement
            <6>3. UNCHANGED LL1NVRAM.symmetricKey
              <7>1. UNCHANGED LL2NVRAM.symmetricKey
                BY <3>1 DEF LL2CorruptSPCR
              <7>2. QED
                BY <2>1, <7>1, UnchangedNVRAMSymmetricKeyLemma
            <6>4. QED
              BY <6>1, <6>2, <6>3, LL1NVRAMRecordCompositionLemma
          <5>5. QED
            BY <5>1, <5>2, <5>3, <5>4
        (*******************************************************************)
        (* Next, we prove the conjunct relating to the RAM.  For           *)
        (* LL2CorruptSPCR, the RAM is unchanged, so we can simply invoke   *)
        (* the UnchangedRAMLemma.                                          *)
        (*******************************************************************)
        <4>3. LL1RestrictedCorruption!ram
          <5>1. LL1RestrictedCorruption!ram!unchanged
            BY <2>1, <3>2, UnchangedRAMLemma DEF LL2CorruptSPCR
          <5>2. QED
            BY <5>1
        (*******************************************************************)
        (* The remaining variables are unchanged, so we can address each   *)
        (* with its appropriate lemma.                                     *)
        (*******************************************************************)
        <4>4. UNCHANGED LL1Disk
          BY <2>1, <3>2, UnchangedDiskLemma DEF LL2CorruptSPCR
        <4>5. UNCHANGED LL1AvailableInputs
          BY <2>1, <3>2, UnchangedAvailableInputsLemma DEF LL2CorruptSPCR
        <4>6. UNCHANGED LL1ObservedOutputs
          BY <2>1, <3>2, UnchangedObservedOutputsLemma DEF LL2CorruptSPCR
        <4>7. UNCHANGED LL1ObservedAuthenticators
          BY <2>1, <3>1, UnchangedObservedAuthenticatorsLemma DEF LL2CorruptSPCR
        <4>8. QED
          BY <4>2, <4>3, <4>4, <4>5, <4>6, <4>7 DEF LL1RestrictedCorruption
      (*********************************************************************)
      (* For the ELSE case, we assume that an extension is not in progress *)
      (* and show that this refines to a stuttering step.                  *)
      (*********************************************************************)
      <3>4. ASSUME ~LL2NVRAM.extensionInProgress
            PROVE UNCHANGED LL1Vars
        (*******************************************************************)
        (* We first prove that the `^\basicname^' NVRAM is unchanged.  We  *)
        (* do this one field at a time.                                    *)
        (*******************************************************************)
        <4>1. UNCHANGED LL1NVRAM
          (*****************************************************************)
          (* The history summary in the `^\basicname^' NVRAM is unchanged. *)
          (*****************************************************************)
          <5>1. UNCHANGED LL1NVRAM.historySummary
            (***************************************************************)
            (* The logical history summary in the `^\fastname^' NVRAM and  *)
            (* SPCR is unchanged.                                          *)
            (***************************************************************)
            <6>1. UNCHANGED LL2NVRAMLogicalHistorySummary
              (*************************************************************)
              (* We reveal the definition of the unprimed logical history  *)
              (* summary in the `^\fastname^' NVRAM and SPCR, given that   *)
              (* there is no extension in progress.                        *)
              (*************************************************************)
              <7>1. LL2NVRAMLogicalHistorySummary = [
                        anchor |-> LL2NVRAM.historySummaryAnchor,
                        extension |-> BaseHashValue ]
                <8>1. ~LL2NVRAM.extensionInProgress
                  BY <3>4
                <8>2. QED
                  BY <8>1 DEF LL2NVRAMLogicalHistorySummary
              (*************************************************************)
              (* We reveal the definition of the primed logical history    *)
              (* summary in the `^\fastname^' NVRAM and SPCR, given that   *)
              (* there is no extension in progress.                        *)
              (*************************************************************)
              <7>2. LL2NVRAMLogicalHistorySummary' = [
                        anchor |-> LL2NVRAM.historySummaryAnchor',
                        extension |-> BaseHashValue ]
                <8>1. ~LL2NVRAM.extensionInProgress'
                  <9>1. ~LL2NVRAM.extensionInProgress
                    BY <3>4
                  <9>2. UNCHANGED LL2NVRAM.extensionInProgress
                    BY <3>1 DEF LL2CorruptSPCR 
                  <9>3. QED
                    BY <9>1, <9>2
                <8>2. QED
                  BY <8>1 DEF LL2NVRAMLogicalHistorySummary
              (*************************************************************)
              (* The history summary anchor in the `^\fastname^' NVRAM is  *)
              (* unchanged by a LL2CorruptSPCR action.                     *)
              (*************************************************************)
              <7>3. UNCHANGED LL2NVRAM.historySummaryAnchor
                BY <3>1 DEF LL2CorruptSPCR
              (*************************************************************)
              (* Since the value of LL2NVRAMLogicalHistorySummary is       *)
              (* determined entirely by the history summary in the         *)
              (* `^\fastname^' NVRAM, and since this value is unchanged,   *)
              (* the value of LL2NVRAMLogicalHistorySummary is unchanged.  *)
              (*************************************************************)
              <7>4. QED
                BY <7>1, <7>2, <7>3
            <6>2. UNCHANGED LL2NVRAM.symmetricKey
              BY <3>1 DEF LL2CorruptSPCR
            <6>3. UNCHANGED LL2NVRAM.hashBarrier
              BY <3>1 DEF LL2CorruptSPCR
            <6>4. QED
              BY <2>1, <6>1, <6>2, <6>3, UnchangedNVRAMHistorySummaryLemma
          (*****************************************************************)
          (* The symmetric key in the `^\basicname^' NVRAM is unchanged.   *)
          (* This follows directly from the definition of the              *)
          (* LL2CorruptSPCR action.                                        *)
          (*****************************************************************)
          <5>2. UNCHANGED LL1NVRAM.symmetricKey
            <6>1. UNCHANGED LL2NVRAM.symmetricKey
              BY <3>1 DEF LL2CorruptSPCR
            <6>2. QED
              BY <2>1, <6>1, UnchangedNVRAMSymmetricKeyLemma
          (*****************************************************************)
          (* Since both fields are unchanged, we can use the               *)
          (* LL1NVRAMRecordCompositionLemma to show that the record is     *)
          (* unchanged.  We first prove the record types, and then we      *)
          (* invoke the LL1NVRAMRecordCompositionLemma directly.           *)
          (*****************************************************************)
          <5>3. LL1NVRAM \in LL1TrustedStorageType
            BY <2>1 DEF LL2Refinement
          <5>4. LL1NVRAM' \in LL1TrustedStorageType
            BY <2>1 DEF LL2Refinement
          <5>5. QED
            BY <5>1, <5>2, <5>3, <5>4, LL1NVRAMRecordCompositionLemma
        (*******************************************************************)
        (* The remaining variables are unchanged, so we can address each   *)
        (* with its appropriate lemma.                                     *)
        (*******************************************************************)
        <4>2. UNCHANGED LL1RAM
            BY <2>1, <3>2, UnchangedRAMLemma DEF LL2CorruptSPCR
        <4>3. UNCHANGED LL1Disk
          BY <2>1, <3>2, UnchangedDiskLemma DEF LL2CorruptSPCR
        <4>4. UNCHANGED LL1AvailableInputs
          BY <2>1, <3>2, UnchangedAvailableInputsLemma DEF LL2CorruptSPCR
        <4>5. UNCHANGED LL1ObservedOutputs
          BY <2>1, <3>2, UnchangedObservedOutputsLemma DEF LL2CorruptSPCR
        <4>6. UNCHANGED LL1ObservedAuthenticators
          BY <2>1, <3>1, UnchangedObservedAuthenticatorsLemma DEF LL2CorruptSPCR
        <4>7. QED
          BY <4>1, <4>2, <4>3, <4>4, <4>5, <4>6 DEF LL1Vars
      (*********************************************************************)
      (* Both THEN and ELSE cases are proven.                              *)
      (*********************************************************************)
      <3>5. QED
        BY <3>3, <3>4
    <2>12. QED
      BY <2>1, <2>2, <2>3, <2>4, <2>5, <2>6, <2>7, <2>8, <2>9, <2>10, <2>11 DEF LL1Next, LL2Next
  <1>4. QED
    (***********************************************************************)
    (* Using the StepSimulation proof rule, the base case and the          *)
    (* induction step together imply that the implication always holds.    *)
    (***********************************************************************)
    <2>1. [][LL2Next]_LL2Vars /\ []LL2Refinement /\ []LL2TypeInvariant => [][LL1Next]_LL1Vars
      BY <1>3, StepSimulation
    <2>2. QED
      BY <1>1, <1>2, <2>1 DEF LL2Spec, LL1Spec, LL2Refinement

====
